Package: mixtime
Title: Mixed Temporal Vectors and Operations
Version: 0.3.0
Authors@R: 
    c(person(given = "Mitchell",
             family = "O'Hara-Wild",
             email = "mail@mitchelloharawild.com",
             role = c("aut", "cre"),
             comment = c(ORCID = "0000-0001-6729-7695")))
Description: Flexible time classes for time series analysis and forecasting with
    mixed temporal granularities. Supports linear and cyclical time 
    representations in discrete and continuous forms, with timezone support, 
    across multiple calendar systems including Gregorian and ISO week date 
    calendars. Time points are stored numerically relative to a chronon; an 
    atomic time granule defined by time units of a calendar. Calendrical 
    arithmetic enables conversion between time granules (e.g. days to months) 
    and calendar systems. Multi-unit arithmetic allows for temporal analysis
    with other granules of common calendars (e.g. fortnights are 2-week units).
    Time vectors of different granularities (e.g. monthly and quarterly) can be
    combined in a single vector, making 'mixtime' ideal for data that changes
    observation frequency over time or requires temporal reconciliation across 
    scales. The package is extensible, allowing users to define custom calendars 
    that build upon civil and astronomical time systems.
License: MIT + file LICENSE
URL: https://pkg.mitchelloharawild.com/mixtime/,
        https://github.com/mitchelloharawild/mixtime
BugReports: https://github.com/mitchelloharawild/mixtime/issues
Encoding: UTF-8
Language: en-GB
Depends: R (>= 3.0.2)
Imports: lifecycle, vctrs, rlang, cli, S7, vecvec (>= 1.2.0), tzdb,
        methods
Suggests: stats, tsibble, testthat, pillar, knitr, rmarkdown, ggplot2,
        ggtime
LinkingTo: cpp11 (>= 0.5.2), tzdb (>= 0.5.0)
RdMacros: lifecycle
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-08-24 04:50:59 UTC; mitchell
Author: Mitchell O'Hara-Wild [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-6729-7695>)
Maintainer: Mitchell O'Hara-Wild <mail@mitchelloharawild.com>
Repository: CRAN
Date/Publication: 2026-08-24 05:30:02 UTC

# Generated by roxygen2: do not edit by hand

S3method("[<-","mixtime::mt_time")
S3method("[[<-","mixtime::mt_time")
S3method(Summary,"mixtime::mixtime")
S3method(all.equal,"mixtime::mt_time")
S3method(base::format,mixtime_interval)
S3method(base::format,mt_naive)
S3method(base::print,mixtime_interval)
S3method(base::print,mt_calendar)
S3method(seq,"mixtime::mixtime")
S3method(seq,"mixtime::mt_time")
S3method(vec_cast,mixtime.yearmonth)
S3method(vec_cast,mixtime.yearquarter)
S3method(vec_cast,mixtime.yearweek)
S3method(vec_proxy,mixtime_interval)
export(as_mixtime)
export(cal_gregorian)
export(cal_isoweek)
export(cal_sym454)
export(cal_time_civil)
export(cal_time_lunar)
export(cal_time_lunar_synodic)
export(cal_time_solar)
export(cal_time_solar_transit)
export(chronon_cardinality)
export(chronon_cardinality_fixed)
export(chronon_common)
export(chronon_divmod)
export(chronon_epoch)
export(chronon_format_attr)
export(chronon_format_cyclical)
export(chronon_format_duration)
export(chronon_format_linear)
export(chronon_parse_cyclical)
export(chronon_parse_linear)
export(circsum)
export(class_mixtime)
export(cyc)
export(cyclical_labels)
export(cyclical_labels_format)
export(cyclical_labels_parse)
export(cyclical_time)
export(date)
export(datetime)
export(day_of_month)
export(day_of_week)
export(day_of_year)
export(days)
export(duration)
export(hours)
export(is_mixtime)
export(is_time_cyclical)
export(is_time_duration)
export(is_time_linear)
export(label_scheme)
export(lin)
export(linear_labels)
export(linear_labels_format)
export(linear_labels_parse)
export(linear_time)
export(loc_altitude)
export(loc_latitude)
export(loc_longitude)
export(milliseconds)
export(minutes)
export(mixtime)
export(month_of_year)
export(months)
export(mt_cyclical)
export(mt_duration)
export(mt_linear)
export(mt_loc_unit)
export(mt_time)
export(mt_tz_unit)
export(mt_unit)
export(new_calendar)
export(new_cyclical_time_fn)
export(new_duration_fn)
export(new_linear_time_fn)
export(new_mixtime)
export(new_time)
export(new_time_unit)
export(parse_format)
export(quarters)
export(seconds)
export(time_calendar)
export(time_ceiling)
export(time_chronon)
export(time_components)
export(time_compose)
export(time_cycle)
export(time_floor)
export(time_is_complete_at)
export(time_is_cyclical)
export(time_is_determinate_at)
export(time_is_duration)
export(time_is_linear)
export(time_of_day)
export(time_parse)
export(time_round)
export(time_unit_abbr)
export(time_unit_full)
export(time_unit_plural)
export(tz_abbreviation)
export(tz_name)
export(tz_offset)
export(tz_transitions)
export(vocab_table)
export(week_of_year)
export(weeks)
export(year)
export(yearmonth)
export(yearquarter)
export(years)
export(yearweek)
import(lifecycle)
import(vctrs)
importFrom(S7,"S7_data<-")
importFrom(S7,"method<-")
importFrom(S7,"props<-")
importFrom(S7,S7_data)
importFrom(S7,S7_inherits)
importFrom(S7,class_any)
importFrom(S7,method)
importFrom(S7,new_class)
importFrom(S7,new_generic)
importFrom(S7,new_property)
importFrom(S7,props)
importFrom(S7,super)
importFrom(rlang,enexpr)
importFrom(rlang,enexprs)
importFrom(rlang,enquo)
importFrom(rlang,enquos)
importFrom(rlang,ensym)
importFrom(rlang,ensyms)
importFrom(rlang,eval_tidy)
importFrom(rlang,expr)
importFrom(rlang,exprs)
importFrom(rlang,list2)
importFrom(rlang,quo)
importFrom(rlang,quos)
importFrom(rlang,sym)
importFrom(rlang,syms)
importFrom(vctrs,vec_math)
importFrom(vctrs,vec_proxy_equal)
importFrom(vctrs,vec_proxy_order)
importFrom(vctrs,vec_restore)
importFrom(vecvec,class_vecvec)
importFrom(vecvec,unvecvec)
importFrom(vecvec,vecvec_apply)
useDynLib(mixtime, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# mixtime <a href="https://pkg.mitchelloharawild.com/mixtime/"><img src="man/figures/logo.svg" align="right" height="139" alt="mixtime website" /></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/mitchelloharawild/mixtime/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/mitchelloharawild/mixtime/actions/workflows/R-CMD-check.yaml)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![CRAN
status](https://www.r-pkg.org/badges/version/mixtime)](https://CRAN.R-project.org/package=mixtime)
<!-- badges: end -->

mixtime provides flexible time classes for time series analysis and
forecasting with mixed temporal granularities. It is designed for
handling temporal data at different frequencies, making it ideal for:

- Changing observation frequencies (e.g., historical quarterly records
  now reported monthly)
- Analysis involving data at different temporal resolutions (e.g.,
  quarterly tourism and monthly turnover)
- Forecasting with temporal reconciliation across multiple time scales

## Key Features

**📈 Linear Time** - Create linear time vectors with `linear_time()` or
with helpers:

- `year()`, `yearquarter()`, `yearmonth()`, `yearweek()`, `date()`,
  `datetime()`

**🔄 Cyclical Time** - Create cyclical time vectors with
`cyclical_time()` or with helpers:

- `month_of_year()`, `day_of_year()`, `day_of_month()`, `time_of_day()`,
  `day_of_week()`, `week_of_year()`

**⏳ Time durations** - Create time duration vectors with `duration()`
or with helpers:

- `years()`, `quarters()`, `months()`, `weeks()`, `days()`, `hours()`,
  `minutes()`, `seconds()`, `milliseconds()`

**🕰️ Temporal semantics**

- Discrete and continuous time models
- Timezone support for all chronons
- Custom granule sizes (e.g. `week(2L)` for fortnights)
- Mixed granularity vectors for different temporal resolution over time
  (or series)

**📅 Calendar Systems** - Support for several calendars:

- `cal_gregorian` - Gregorian dates (e.g. 2026-02-17)
- `cal_isoweek` - ISO week dates (e.g. 2026-W8-Tue)
- `cal_sym454` - Symmetry454 dates (e.g. 2026-Feb-W4-Tue)
- More calendars coming, including custom censored calendars

**🧮 Temporal Operations**

- Rounding: `time_floor()`, `time_round()`, `time_ceiling()`
- Sequencing: `seq()` for linear and cyclical time points

## Installation

The mixtime package can be installed from CRAN with:

``` r
install.packages("mixtime")
```

The development version can be installed from
[GitHub](https://github.com/mitchelloharawild/mixtime) with:

``` r
# install.packages("remotes")
remotes::install_github("mitchelloharawild/mixtime")
```

## Usage

``` r
library(mixtime)
#> 
#> Attaching package: 'mixtime'
#> The following objects are masked from 'package:base':
#> 
#>     date, months, quarters
demo_time <- as.POSIXct("2026-02-22 18:30:42", tz = "UTC")
demo_date <- as.Date("2026-02-22")
```

The mixtime package is designed around the time units of calendars.
These time units are used to create and manipulate time vectors.

### Calendars and time units

Calendars have a `cal_*` prefix, which contain a set of time units that
can be accessed with `<cal>$<unit>`. Using time units as functions
produces time granules.

``` r
# The Gregorian calendar
cal_gregorian
#> <cal_gregorian>
#> Time units:
#>   - year
#>   - quarter
#>   - month
#>   - day
#>   - ampm
#>   - hour
#>   - minute
#>   - second
#>   - millisecond
#>   - microsecond
#>   - nanosecond

# A 1-month time granule
cal_gregorian$month(1L) # (1L is integer 1)
#> <mixtime::tu_month>
#>  @ n : int 1
#>  @ tz: 'mt_naive' chr NA

# A 2-week time granule (fortnights)
cal_isoweek$week(2L)
#> <mixtime::tu_week>
#>  @ n : int 2
#>  @ tz: 'mt_naive' chr NA
```

### Linear Time

A linear time vector uses time granules to define the resolution of time
points, known as a `chronon` (the smallest granules of time). When the
input time has a different resolution than the `chronon`, it will be
automatically converted.

``` r
linear_time(demo_date, chronon = cal_gregorian$month(1L))
#> <mixtime[1]>
#> [1] 2026 Feb
```

Discrete time models (integer-based values) are used by default, however
continuous time models (double-based values) can be used with
`discrete = FALSE` to allow fractional chronons.

``` r
# February 22nd is 75% through the month (in non-leap years)
linear_time(demo_date, chronon = cal_gregorian$month(1L), discrete = FALSE)
#> <mixtime[1]>
#> [1] 2026 Feb 75.0%
```

Linear time helper functions are available to quickly create common time
points.

``` r
# Create time vectors at different granularities
yearquarter(demo_date) + 0:7
#> <mixtime[8]>
#> [1] 2026 Q1 2026 Q2 2026 Q3 2026 Q4 2027 Q1 2027 Q2 2027 Q3 2027 Q4
yearmonth(demo_date) + 0:11
#> <mixtime[12]>
#>  [1] 2026 Feb 2026 Mar 2026 Apr 2026 May 2026 Jun 2026 Jul 2026 Aug 2026 Sep
#>  [9] 2026 Oct 2026 Nov 2026 Dec 2027 Jan
yearweek(demo_date) + 0:10
#> <mixtime[11]>
#>  [1] 2026 W08 2026 W09 2026 W10 2026 W11 2026 W12 2026 W13 2026 W14 2026 W15
#>  [9] 2026 W16 2026 W17 2026 W18
date(demo_date) + 0:6
#> <mixtime[7]>
#> [1] 2026-02-22 2026-02-23 2026-02-24 2026-02-25 2026-02-26 2026-02-27 2026-02-28
```

The mixtime package allows time of different granularities to be
combined in a single vector.

``` r
c(
  year(demo_date), yearquarter(demo_date), 
  yearmonth(demo_date), yearweek(demo_date)
)
#> <mixtime[4]>
#> [1] 2026     2026 Q1  2026 Feb 2026 W08
```

### Cyclical Time

A cyclical time vector is defined by two time granules: a `chronon` (the
smaller granule of time) and a `cycle` (the larger granule that the
`chronon` loops over).

``` r
# The `calendar` argument provides a masking scope to `chronon` and `cycle`
cyclical_time(demo_date, chronon = day(1L), cycle = week(1L), calendar = cal_isoweek)
#> <mixtime[1]>
#> [1] Sun
```

There are several cyclical time helper functions for convenience.

``` r
# Extract cyclical components
month_of_year(demo_date)
#> <mixtime[1]>
#> [1] Feb
week_of_year(demo_date)
#> <mixtime[1]>
#> [1] W08
day_of_week(demo_date)
#> <mixtime[1]>
#> [1] Sun

# Continuous cyclical time shows progress through chronons
day_of_week(demo_time, discrete = FALSE)
#> <mixtime[1]>
#> [1] Sun 77.1%
```

### Timezones

All linear and cyclical time vectors support timezones via the `tz`
argument.

``` r
demo_time
#> [1] "2026-02-22 18:30:42 UTC"
# Same day (Sunday) in LA
date(demo_time, tz = "America/Los_Angeles")
#> <mixtime[1]>
#> [1] 2026-02-22 PST
date(demo_time, tz = "America/Los_Angeles", discrete = FALSE)
#> <mixtime[1]>
#> [1] 2026-02-22 43.8% PST
day_of_week(demo_time, tz = "America/Los_Angeles")
#> <mixtime[1]>
#> [1] Sun PST

# Next day (Monday) in Melbourne (23% through the 23rd)
date(demo_time, tz = "Australia/Melbourne")
#> <mixtime[1]>
#> [1] 2026-02-23 AEDT
date(demo_time, tz = "Australia/Melbourne", discrete = FALSE)
#> <mixtime[1]>
#> [1] 2026-02-23 23.0% AEDT
day_of_week(demo_time, tz = "Australia/Melbourne")
#> <mixtime[1]>
#> [1] Mon AEDT
```

### Temporal Manipulation

Linear time points can be adjusted to the floor, ceiling, or rounded to
a specified time granule.

``` r
# Round dates to different granularities
time_floor(demo_date, cal_gregorian$month(1L))
#> [1] "2026-02-01"
time_round(demo_date, cal_isoweek$week(1L))
#> [1] "2026-02-23"
time_ceiling(demo_date, cal_gregorian$month(1L))
#> [1] "2026-03-01"
```

### Time Sequences

The `seq()` function creates sequences of time points iterating by a
given time granule.

``` r
# Integer increments (advances by chronon's natural granule)
seq(yearmonth(demo_date), by = 1L, length.out = 10)
#> <mixtime[10]>
#>  [1] 2026 Feb 2026 Mar 2026 Apr 2026 May 2026 Jun 2026 Jul 2026 Aug 2026 Sep
#>  [9] 2026 Oct 2026 Nov

# Time can be sequenced by different granules than the chronon
seq(date(demo_date), by = cal_gregorian$month(1L), length.out = 8)
#> <mixtime[8]>
#> [1] 2026-02-22 2026-03-22 2026-04-22 2026-05-22 2026-06-22 2026-07-22 2026-08-22
#> [8] 2026-09-22
```

#' Friendly labels for cyclical relationships
#'
#' This S7 generic function provides the labels for cyclical relationships 
#' between time granules. These functions should return locale specific labels.
#'
#' @param granule A time granule object representing the granule (e.g., `month(1L)`)
#' @param cycle A time granule object representing the cycle (e.g., `year(1L)`)
#' @param ... Additional arguments for methods.
#'
#' @return Character vector of labels for the time point within the cycle.
#' 
#' @examples
#' # Labels for months in a year
#' with(cal_gregorian, cyclical_labels(month(1L), year(1L), 1:12))
#' 
#' # Labels for days in a week
#' with(cal_isoweek, cyclical_labels(day(1L), week(1L), 1:7))
#' 
#' # Labels for weeks in a year, defaulted from time_unit_abbr()
#' with(cal_isoweek, cyclical_labels(week(1L), year(1L), 1:52))
#' 
#' @export
cyclical_labels <- S7::new_generic("cyclical_labels", c("granule", "cycle"))

# #' Default friendly labels for cyclical relationships
# #' 
# #' The default method for [cyclical_labels()] simply pastes together the time unit 
# #' label and the position within the cycle. This is a fallback method that will
# #' be used if no more specific method is defined for the given time units.
# #' 
# #' @inheritParams cyclical_labels
# #' @param i Integer vector representing the position within the cycle.
# #' @param label If `TRUE`, return labels for the positions within the cycle. If `FALSE`, return the integer positions as character.
# #' @param abbreviate If `TRUE`, return abbreviated labels (e.g. "Mon" for "Monday"). If `FALSE`, return full labels (e.g. "Monday").
# #' 
# #' @seealso [cyclical_labels()]
# #' 
# #' @returns A character vector of labelled granules
# #' @rdname cyclical_labels
method(cyclical_labels, list(mt_unit, mt_unit)) <- function(granule, cycle, i, label = FALSE, abbreviate = TRUE, ...) {
  if (!label) return(as.character(i))
  # TODO - pluralise the full time unit
  paste0(
    if (abbreviate) time_unit_abbr(granule) else time_unit_full(granule), 
    i
  )
}

#' Base S7 class for creating new time units
#' 
#' This class is the primitive class for time units, and should
#' be extended from when creating new time units. A new class
#' is typically created with S7 using: 
#' `S7::new_class("tu_***", parent = mt_tz_unit)`
#' 
#' Time units are the building blocks of calendars in mixtime. Each unit
#' represents a specific temporal component (e.g., day, month, year) and
#' can be combined using [new_calendar()] to create a calendar system.
#' 
#' When creating custom calendars, define time unit classes that inherit
#' from either `mt_unit` (for standard units) or `mt_tz_unit` (for
#' timezone-aware units), then pass them as named arguments to
#' [new_calendar()]. The calendar will use these names to create
#' constructor functions accessible via `$` notation (e.g., `calendar$day(1L)`).
#' 
#' @section Calendar Algebra Methods:
#'
#' Time units enable calendar arithmetic through key generic methods
#' that should be implemented for custom time units:
#'
#' * `chronon_cardinality_fixed(x, y)` - Returns the number of unit `x`
#'   granules that fit within one unit `y` granule, for relationships that are
#'   a constant, context-independent number (e.g., 7 days per week, 24 hours
#'   per day). Prefer this over `chronon_cardinality()` whenever the
#'   relationship does not depend on `at`, since it is also used to determine
#'   which relationships are safe to use for [chronon_divmod()]'s graph
#'   traversal.
#'
#' * `chronon_cardinality(x, y, at)` - Returns the number of `x` granule
#'   that fit within one `y` granule. This is variable based on `at` (e.g.,
#'   28-31 days per month). Only implement this directly (rather than
#'   `chronon_cardinality_fixed()`) when the relationship genuinely depends on
#'   `at`, and pair it with a direct `chronon_divmod()` method so that the
#'   relationship remains reachable during graph traversal.
#'
#' * `chronon_divmod(x, from, to)` - Converts time point `x` from granules of
#'   `from` to granules of `to`, returning a list with `div` (the quotient)
#'   and `mod`. This enables conversions between granules that have
#'   variable cardinality (e.g., the date 2020-03-23 to the month 2020-03).
#'   All conversions should be based on chronons since epoch (1970-01-01),
#'   in the UTC time zone.
#' 
#' These methods work together to enable mixtime to perform calendar-aware
#' arithmetic, understanding that months have variable lengths and handling
#' timezone-aware conversions.
#' 
#' @param n The step size of time granule. For example, `n = 2L` is 2 time 
#'   units, and `cal_isoweek$week(2L)` would represent 2 weeks (a fortnight).
#' 
#' @return A time granule object of class `mt_unit`
#' 
#' @seealso [new_calendar()] for creating calendars from time units
#' 
#' @examples
#' # Create a timezone-aware unit class
#' 
#' # Use these units to create a calendar
#' my_calendar <- new_calendar(
#'   day = S7::new_class("tu_my_day", parent = mt_unit),
#'   month = S7::new_class("tu_my_month", parent = mt_tz_unit),
#'   class = "my_calendar"
#' )
#' 
#' # Access unit constructors from the calendar
#' my_calendar$day(1L)
#' my_calendar$month(3L, tz = "America/New_York")
#' 
#' @export
#' @rdname mt_unit
mt_unit <- S7::new_class(
  "mt_unit", 
  properties = list(
    n = S7::new_property(
      class = S7::class_numeric,
      default = 1L
    )
  ),
  # validator = function(self) {
  #   if (length(self@n) != 1L) {
  #     paste0("@n must be length 1 <numeric> value, not length ", length(self@n), ".")
  #   }
  # }
)

#' Time vector classes
#'
#' @description
#' The `mt_time` family are the S7 vector classes that store time points as a
#' numeric count of chronons. `mt_time` is the (internal) base class carrying 
#' the `chronon` property; the common modes of time are:
#'
#' * `mt_linear` - linear time points, typically produced with [linear_time()].
#' * `mt_cyclical` - cyclical time points with additional `cycle` granule, 
#'   typically produced with [cyclical_time()].
#' * `mt_duration` - time durations, typically produced with [duration()].
#'
#' The underlying data can be either **integer** (discrete time) or **double**
#' (continuous time). The `chronon` (and, for `mt_cyclical`, the `cycle`) are 
#' time granules, the result from a [mt_unit] object.
#'
#' @param .data A numeric vector of chronon counts (integer or double).
#' @param chronon,cycle Time granules ([mt_unit] objects) giving the unit of the
#'   chronon counts and, for `mt_cyclical`, the length of the cycle.
#'
#' @return An S7 class object (used for method dispatch), or a time vector when
#'   called as a constructor.
#'
#' @seealso [mt_linear()], [mt_duration()], and [mt_cyclical()] to construct
#'   time vectors, and [mt_unit] for the granule type stored in `chronon`/`cycle`.
#'
#' @name mt_time-class
NULL

# Lightweight S3 carrier for the mode of time (integer or double). It exists
# *only* so S7 accepts either an integer or double inputs (an S7 base-type
# class fixes a single base type).
mt_data <- S7::new_S3_class(
  "mt_time_data",
  constructor = function(.data = integer()) {
    structure(.data, class = "mt_time_data")
  },
  validator = function(self) {
    if (!is.numeric(self)) {
      cli::cli_abort("{.var self} must be an integer or double vector.", call. = FALSE)
    }
  }
)

granule_len1 <- function(value) {
  if (length(value@n) != 1L) {
    "must wrap a single time granule (its `@n` must be length 1)"
  }
}

#' @rdname mt_time-class
#' @usage NULL
#' @export
mt_time <- S7::new_class(
  "mt_time",
  parent = mt_data,
  properties = list(
    chronon = S7::new_property(
      class = mt_unit,
      default = quote(mt_unit(1L)),
      validator = granule_len1
    )
  )
)

#' @rdname mt_time-class
#' @export
mt_linear <- S7::new_class("mt_linear", parent = mt_time)

#' @rdname mt_time-class
#' @export
mt_duration <- S7::new_class("mt_duration", parent = mt_time)

#' @rdname mt_time-class
#' @export
mt_cyclical <- S7::new_class(
  "mt_cyclical",
  parent = mt_time,
  properties = list(
    cycle = S7::new_property(
      class = mt_unit,
      default = quote(mt_unit(1L)),
      validator = granule_len1
    )
  )
)

#' Base S7 class for mixtime vector objects
#'
#' `class_mixtime` is the base S7 class for all mixtime vector objects,
#' inheriting from [vecvec::class_vecvec]. While not intended to be used
#' directly, this S7 class is suitable to use when defining S7 methods for
#' mixtime vectors. S3 methods can be defined using the `mixtime::mixtime` 
#' class.
#' 
#' @param x A list of `"mt_time"` vectors, see [new_time()] for details.
#' @inheritParams vecvec::class_vecvec
#'
#' @return When used as a class definition (e.g., in `S7::method(generic,
#'   class_mixtime)`), an S7 class object representing the `mixtime` class,
#'   inheriting from [vecvec::class_vecvec]. When called as a constructor
#'   (`class_mixtime(list(...))`), a mixtime vector of S7 class `mixtime`
#'   (also inheriting the S3 class `"mixtime"`), containing the supplied list
#'   of time vectors as a [vecvec::class_vecvec] structure. End users should
#'   prefer [mixtime()] or [new_mixtime()] for construction.
#'
#' @seealso [mixtime()] for creating mixtime vectors, and [new_mixtime()] for
#' the low-level constructor function of this S7 class.
#'
#' @importFrom vecvec class_vecvec
#' @export
class_mixtime <- S7::new_class(
  "mixtime",
  parent = class_vecvec
)

# ----------------------------------------------------------------
# class_mixtime methods
# ----------------------------------------------------------------

#' @export
method(vec_ptype_full, class_mixtime) <- function(x, ...) "mixtime"

#' @export
method(vec_ptype_abbr, class_mixtime) <- function(x, ...) "mixtime"


time_valid <- function(x) {
  if (is_mixtime(x)) return(TRUE)
  !inherits(try(time_chronon(x), silent = TRUE), "try-error")
}

vec_ptype2_mixtime <- function(x, y, ...) {
  x_is_time <- time_valid(x)
  y_is_time <- time_valid(y)

  if (!(x_is_time && y_is_time) && !(is.numeric(x) || is.numeric(y))) {
    vctrs::stop_incompatible_type(x, y, x_arg = "", y_arg = "")
  }
  new_mixtime()
}

vec_cast_to_mixtime <- function(x, to, ...) {
  mixtime(x)
}

#' @importFrom vctrs vec_proxy_order
method(vec_proxy_order, class_mixtime) <- function(x, ...) {
  mode <- check_common_time_mode(x)
  if (!length(x@x)) return(vec_proxy_order(vecvec::unvecvec(x)))

  # The granules each part is reduced to, read before converting the parts below
  # replaces them with bare chronon counts carrying neither granule.
  chronon <- chronon_common_impl(lapply(x@x, function(v) attr(v, "chronon")))
  cycle <- if (identical(mode, "cyclical")) check_common_cycle(x)

  # Convert all time values to a common chronon, which a single part already is
  if (length(x@x) > 1L) {
    x@x <- lapply(x@x, function(v) {
      if (is.integer(v)) v <- v + 0.5
      chronon_convert(v, chronon)
    })
  }

  if (!is.null(cycle)) {
    # Cyclical values order by their position within the cycle, matching `==`
    # and `<` on `mt_cyclical` (see `mt_cyclical-compare`) rather than the
    # absolute chronon count stored underneath.
    x@x <- lapply(x@x, function(v) cyclical_position(vec_data(v), chronon, cycle))
  }

  vec_proxy_order(vecvec::unvecvec(x))
}

#' @importFrom vctrs vec_proxy_equal
method(vec_proxy_equal, class_mixtime) <- function(x, ...) {
  cycle <- if (identical(check_common_time_mode(x), "cyclical")) check_common_cycle(x)
  data_frame(
    # Cyclical values are equal when they share a position within the cycle,
    # matching `==` on `mt_cyclical` (see `mt_cyclical-compare`) rather than the
    # absolute chronon count stored underneath.
    x = if (is.null(cycle)) {
      as.numeric(x)
    } else {
      unvecvec(vecvec_apply(x, function(x) {
        cyclical_position(vec_data(x), attr(x, "chronon"), cycle)
      }))
    },
    g = unvecvec(vecvec_apply(x, function(x) rep(rlang::hash(attr(x, "chronon")), length(x))))
  )
}

# ------------------------------------------------------------------------------
# vec_cast / vec_ptype2 methods for fundamental modes of time
# ------------------------------------------------------------------------------

# --- casts from a time vector to a base type ---
mt_cast_to_character <- function(x, to, ...) time_format_impl(x)
mt_cast_to_integer <- function(x, to, ...) vec_cast(vec_data(x), integer())
mt_cast_to_double <- function(x, to, ...) vec_cast(vec_data(x), double())
mt_cast_to_Date <- function(x, to, ...) {
  # Convert to naive time zone dates
  vec_restore(chronon_convert(x, cal_gregorian$day(1L, tz = NA_character_)), to)
}
mt_cast_to_POSIXct <- function(x, to, ...) {
  vec_restore(chronon_convert(x, cal_gregorian$second(1L, tz = "UTC"), discrete = FALSE), to)
}

# --- casts from a base type to a time vector ---
mt_cast_from_numeric <- function(x, to, ...) {
  attributes(x) <- attributes(to)
  x
}

mt_cast_from_character <- function(x, to, ...) {
  cycle <- if (S7_inherits(to, mt_cyclical)) attr(to, "cycle") else NULL
  data <- vec_data(time_parse_impl(
    x,
    chronon = attr(to, "chronon"),
    cycle = cycle,
    discrete = is.integer(vec_data(to))
  ))
  attributes(data) <- attributes(to)
  data
}

# A cyclical `cycle` is a *modulus*, not a unit of measure: it names the
# equivalence relation the stored (absolute) chronon count is reduced by at
# format and comparison time. Unlike a chronon - where the finest common granule
# represents both operands losslessly - any granule common to two different
# cycles has *fewer* distinct positions than either operand, so
# `chronon_common_impl()` is not the applicable operation (week + year would give
# a one-day cycle, collapsing every value to the same position). Two cyclical
# vectors therefore combine only when they share a cycle, once naive properties
# (tz / location) have been reconciled between them.
cycle_common <- function(x, y) {
  x <- granule_inherit_props(x, y)
  y <- granule_inherit_props(y, x)
  if (!identical(x, y)) return(NULL)
  x
}

cycle_incompatible_details <- paste0(
  "Cyclical time vectors can only be combined when they share a cycle.\n",
  "Differing cycles can be held side by side in a `mixtime()` vector instead."
)

# --- time-to-time cast (same kind) ---
mt_cast_time_time <- function(x, to, ..., x_arg = "", to_arg = "") {
  if (S7_inherits(to, mt_duration)) {
    discrete <- is.integer(vec_data(to))
    x <- as.numeric(x) * chronon_cardinality(attr(to, "chronon"), attr(x, "chronon"))
    if (discrete) x <- as.integer(x)
    attributes(x) <- attributes(to)
    return(x)
  }
  if (S7_inherits(to, mt_cyclical) &&
      is.null(cycle_common(attr(x, "cycle"), attr(to, "cycle")))) {
    vctrs::stop_incompatible_cast(
      x, to, x_arg = x_arg, to_arg = to_arg,
      details = cycle_incompatible_details
    )
  }
  x <- chronon_convert(x, attr(to, "chronon"), discrete = is.integer(to))
  attributes(x) <- attributes(to)
  x
}

# --- prototype2 between two time vectors of the same kind ---
mt_ptype2_time_time <- function(x, y, ..., x_arg = "", y_arg = "") {
  x_dur <- S7_inherits(x, mt_duration)
  y_dur <- S7_inherits(y, mt_duration)
  x_cyc <- S7_inherits(x, mt_cyclical)
  y_cyc <- S7_inherits(y, mt_cyclical)
  if (x_dur != y_dur || x_cyc != y_cyc) {
    vctrs::stop_incompatible_type(x, y, x_arg = x_arg, y_arg = y_arg)
  }
  if (x_cyc) {
    # The chronon resolves to the finest common granule as usual, but the cycle
    # must match on both sides - see `cycle_common()`.
    cycle <- cycle_common(attr(x, "cycle"), attr(y, "cycle"))
    if (is.null(cycle)) {
      vctrs::stop_incompatible_type(
        x, y, x_arg = x_arg, y_arg = y_arg,
        details = cycle_incompatible_details
      )
    }
    return(mt_cyclical(
      chronon = chronon_common_impl(list(attr(x, "chronon"), attr(y, "chronon"))),
      cycle = cycle
    ))
  }
  data <- vec_ptype2(vec_data(x), vec_data(y))
  chronon <- chronon_common_impl(list(attr(x, "chronon"), attr(y, "chronon")))
  if (x_dur) {
    mt_duration(data, chronon = chronon)
  } else {
    mt_linear(data, chronon = chronon)
  }
}

# --- duration <-> numeric prototypes (only durations combine with bare numbers) ---
mt_ptype2_duration_x <- function(x, y, ...) x
mt_ptype2_y_duration <- function(x, y, ...) y

# Register the `vec_cast`/`vec_ptype2` methods above on the package-namespaced time
# classes. Called from `.onLoad()`.
register_mt_vctrs <- function() {
  types <- c("mixtime::mt_linear", "mixtime::mt_cyclical", "mixtime::mt_duration")
  for (cls in types) {
    # time -> base type
    register_s3_method("vctrs", "vec_cast.character", cls, mt_cast_to_character)
    register_s3_method("vctrs", "vec_cast.integer", cls, mt_cast_to_integer)
    register_s3_method("vctrs", "vec_cast.double", cls, mt_cast_to_double)
    # base type -> time
    register_s3_method("vctrs", "vec_cast", paste0(cls, ".integer"), mt_cast_from_numeric)
    register_s3_method("vctrs", "vec_cast", paste0(cls, ".double"), mt_cast_from_numeric)
    # time -> time (same kind) and prototype2 (same kind)
    register_s3_method("vctrs", "vec_cast", paste0(cls, ".", cls), mt_cast_time_time)
    register_s3_method("vctrs", "vec_ptype2", paste0(cls, ".", cls), mt_ptype2_time_time)
  }

  # A linear time vector can be cast to Date / POSIXct.
  register_s3_method("vctrs", "vec_cast.Date", "mixtime::mt_linear", mt_cast_to_Date)
  register_s3_method("vctrs", "vec_cast.POSIXct", "mixtime::mt_linear", mt_cast_to_POSIXct)

  register_s3_method("vctrs", "vec_cast", "mixtime::mt_linear.character", mt_cast_from_character)
  register_s3_method("vctrs", "vec_cast", "mixtime::mt_cyclical.character", mt_cast_from_character)

  # Durations combine with bare numbers.
  dur <- "mixtime::mt_duration"
  register_s3_method("vctrs", "vec_ptype2", paste0(dur, ".double"), mt_ptype2_duration_x)
  register_s3_method("vctrs", "vec_ptype2", paste0("double.", dur), mt_ptype2_y_duration)
  register_s3_method("vctrs", "vec_ptype2", paste0(dur, ".integer"), mt_ptype2_duration_x)
  register_s3_method("vctrs", "vec_ptype2", paste0("integer.", dur), mt_ptype2_y_duration)
}

# A cycle is a modulus rather than a unit of measure, so unlike a chronon there
# is no common cycle to reconcile two cyclical vectors to.
test_that("cyclical vectors of differing cycles can't be combined", {
  dow <- day_of_week(date("2020-01-15"))@x[[1L]]
  doy <- day_of_year(date("2020-01-15"))@x[[1L]]

  expect_error(vctrs::vec_ptype2(dow, doy), class = "vctrs_error_incompatible_type")
  expect_error(vctrs::vec_c(dow, doy), class = "vctrs_error_incompatible_type")
  expect_error(vctrs::vec_cast(dow, doy), class = "vctrs_error_cast")
})

test_that("cyclical vectors sharing a cycle combine on their common chronon", {
  doy <- day_of_year(as.Date("2020-01-15"))@x[[1L]]
  moy <- month_of_year(as.Date("2020-01-15"))@x[[1L]]

  ptype <- vctrs::vec_ptype2(doy, moy)
  expect_true(S7::S7_inherits(ptype, mt_cyclical))
  # finest common chronon, cycle carried through unchanged
  expect_equal(ptype@chronon, doy@chronon)
  expect_equal(ptype@cycle, doy@cycle)

  # the month casts to the first day of that month within the same cycle
  expect_equal(format(vctrs::vec_c(doy, moy)), c("D15", "D01"))
})

test_that("cycles differing only in inherited properties are reconciled", {
  naive <- day_of_week(date("2020-01-15"))@x[[1L]]
  tzed <- day_of_week(date("2020-01-22", tz = "Australia/Melbourne"))@x[[1L]]

  expect_equal(cycle_common(naive@cycle, tzed@cycle), tzed@cycle)
  expect_no_error(vctrs::vec_c(naive, tzed))
})

test_that("a mixtime still holds differing cycles side by side", {
  d <- as.Date("2020-01-15")
  x <- c(day_of_week(d), day_of_year(d))

  expect_length(x, 2L)
  expect_length(x@x, 2L)
  expect_equal(format(x), c("Wed", "D15"))
})

# Comparison reduces both operands to their position within the cycle, so a
# cyclical value means a weekday / a day-of-year, not a particular instant.
test_that("cyclical equality compares the position within the cycle", {
  # 2020-01-15 and 2020-01-22 are both a Wednesday
  expect_true(day_of_week(date("2020-01-15")) == day_of_week(date("2020-01-22")))
  expect_false(day_of_week(date("2020-01-15")) == day_of_week(date("2020-01-16")))
  expect_true(day_of_week(date("2020-01-15")) != day_of_week(date("2020-01-16")))

  # the same day of two different years
  expect_true(day_of_year(date("2020-01-15")) == day_of_year(date("2021-01-15")))
  expect_false(day_of_year(date("2020-01-15")) == day_of_year(date("2020-01-16")))
})

test_that("cyclical ordering follows the position within the cycle", {
  wed <- day_of_week(date("2020-01-15"))
  fri <- day_of_week(date("2021-01-15"))

  # Friday is later in its week than Wednesday, despite the earlier absolute date
  expect_true(wed < fri)
  expect_false(fri < wed)
  expect_true(fri > wed)
  expect_true(wed <= fri)
  expect_true(fri >= wed)
})

test_that("cyclical comparison of differing chronons uses interval endpoints", {
  jan <- month_of_year(date("2020-01-15"))
  d15 <- day_of_year(date("2020-01-15"))
  mar <- month_of_year(date("2020-03-01"))

  # D15 falls within January, so neither is wholly before the other
  expect_false(d15 == jan)
  expect_false(d15 < jan)
  expect_false(d15 > jan)

  # ... but March begins after D15 ends
  expect_true(d15 < mar)
  expect_true(mar > d15)
})

test_that("cyclical comparison of differing cycles is an error", {
  expect_error(day_of_week(date("2020-01-15")) == day_of_year(date("2020-01-15")))
  expect_error(day_of_week(date("2020-01-15")) < day_of_month(date("2020-01-15")))
})

test_that("continuous cyclical time compares fractional positions", {
  t <- as.POSIXct(c("2020-01-01 09:30:00", "2020-01-05 09:30:00"), tz = "UTC")
  same <- time_of_day(t, discrete = FALSE)
  expect_true(same[1] == same[2])

  later <- time_of_day(as.POSIXct("2020-01-02 09:45:00", tz = "UTC"), discrete = FALSE)
  expect_true(same[1] < later)
})

# Equality and ordering must agree with `==`/`<`, or `unique()` would keep two
# Wednesdays that `==` reports as equal.
test_that("unique(), duplicated() and sort() use the position within the cycle", {
  d <- as.Date("2020-01-13") + 0:13
  dow <- day_of_week(d)

  expect_length(unique(dow), 7L)
  expect_equal(format(unique(dow)), c("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"))
  expect_equal(duplicated(dow), rep(c(FALSE, TRUE), each = 7L))
  expect_equal(anyDuplicated(dow), 8L)
  expect_equal(
    format(sort(rev(dow))),
    rep(c("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"), each = 2L)
  )
  expect_equal(vctrs::vec_match(dow, unique(dow)), rep(1:7, times = 2L))
})

test_that("de-duplication spans the parts of a mixtime sharing a cycle", {
  x <- c(
    day_of_year(as.Date(c("2020-02-01", "2021-02-01"))),
    month_of_year(as.Date("2020-02-10"))
  )

  # the two 1 Februaries share a position; the month is a distinct chronon
  expect_equal(duplicated(x), c(FALSE, TRUE, FALSE))
  expect_equal(format(sort(x)), c("D32", "D32", "Feb"))
})

test_that("empty mixtime vectors and prototypes stay comparable", {
  expect_length(vctrs::vec_proxy_order(day_of_week(integer(0))), 0L)
  expect_length(vctrs::vec_proxy_order(vctrs::vec_ptype(day_of_week(0L))), 0L)
  expect_length(unique(vctrs::vec_ptype(day_of_week(0L))), 0L)
  expect_length(sort(day_of_week(integer(0))), 0L)
})

test_that("comparing a mixtime requires a single mode of time", {
  mixed <- c(yearmonth(as.Date("2020-01-15")), days(1L))
  expect_error(unique(mixed), "share one mode of time")
  expect_error(sort(mixed), "share one mode of time")

  cycles <- c(day_of_week(date("2020-01-15")), day_of_month(date("2020-01-15")))
  expect_error(unique(cycles), "differing cycles")
  expect_error(sort(cycles), "differing cycles")
})

# https://github.com/mitchelloharawild/mixtime/issues/81
test_that("vec_detect_complete() recycles hashed chronons correctly (#81)", {
  expect_true(all(vctrs::vec_detect_complete(yearmonth(0:3))))
})

test_that("vec_detect_complete() correctly identifies missing values", {
  expect_equal(vctrs::vec_detect_complete(yearmonth(c(0:3, NA))), c(TRUE, TRUE, TRUE, TRUE, FALSE))
})

test_that("vec_proxy_order() works for mixed-chronon mixtime vectors", {
  expect_no_error(vctrs::vec_proxy_order(c(yearmonth(360L), year(2000L))))
  expect_length(vctrs::vec_proxy_order(c(yearmonth(360L), year(2000L))), 2L)
})

test_that("vec_cast to character works for mt_duration", {
  expect_equal(vctrs::vec_cast(years(1L), character()), format(years(1L)))
  expect_equal(vctrs::vec_cast(months(3L), character()), format(months(3L)))
  expect_type(vctrs::vec_cast(years(1:3), character()), "character")
  expect_length(vctrs::vec_cast(years(1:3), character()), 3L)
})

test_that("inner time vectors are S7 objects of the right class", {
  lin <- yearmonth(0:3)@x[[1L]]
  cyc <- month_of_year(0:3)@x[[1L]]
  dur <- days(0:3)@x[[1L]]
  expect_true(S7::S7_inherits(lin, mt_linear))
  expect_true(S7::S7_inherits(cyc, mt_cyclical))
  expect_true(S7::S7_inherits(dur, mt_duration))
  # cyclical carries a typed `cycle` property; chronon is a typed `mt_unit`
  expect_true(S7::S7_inherits(cyc@cycle, mt_unit))
  expect_true(S7::S7_inherits(lin@chronon, mt_unit))
})

# vec_cast() from character (mt_cast_from_character() in vctrs.R) parses via
# time_parse_impl(), using the target's chronon/cycle to pick candidate
# formats. This is what lets comparison operators (see mt_linear-compare /
# mt_cyclical-compare in compare.R) accept a bare character operand.
test_that("vec_cast() parses character onto a linear time vector's own chronon", {
  to <- yearmonth(0L)@x[[1L]]
  cast <- vctrs::vec_cast("2020 Jan", to)
  expect_true(S7::S7_inherits(cast, mt_linear))
  expect_equal(cast, yearmonth("2020 Jan")@x[[1L]])
})

test_that("vec_cast() parses character onto a cyclical time vector's own chronon/cycle", {
  to <- month_of_year(0L)@x[[1L]]
  cast <- vctrs::vec_cast("Feb", to)
  expect_true(S7::S7_inherits(cast, mt_cyclical))
  expect_equal(cast, month_of_year(date("2020-02-15"))@x[[1L]])
})

test_that("vec_cast() from character fills in unset chronon properties (e.g. tz) from the target", {
  to <- datetime(0L, tz = "Australia/Melbourne")@x[[1L]]
  cast <- vctrs::vec_cast("2020-06-01 05:00:00", to)
  expect_equal(format(cast), "2020-06-01 05:00:00 AEST")
})

test_that("vec_cast() from character errors on unparseable text", {
  to <- yearmonth(0L)@x[[1L]]
  expect_error(vctrs::vec_cast("not a date", to), class = "mixtime_parse_no_match")
})

test_that("vec_cast() from character is not supported for durations (no meaningful text form)", {
  to <- days(1L)@x[[1L]]
  expect_error(vctrs::vec_cast("2 days", to), class = "vctrs_error_incompatible_type")
})

test_that("comparison operators accept a bare character operand, parsed onto the other side's chronon", {
  ym <- yearmonth("2020 Jan")

  expect_true(ym > "2019 Jan")
  expect_true("2019 Jan" < ym)
  expect_true(ym == "2020 Jan")
  expect_true(ym != "2019 Jan")
  expect_true(ym <= "2020 Jan")
  expect_true(ym >= "2020 Jan")

  expect_true(date("2020-01-15") > "2020-01-01")
  expect_true(month_of_year(date("2020-01-15")) == "Jan")
})

test_that("mode of time (integer/double) survives vec_slice/vec_c/vec_restore", {
  disc <- yearmonth(0:3)@x[[1L]]              # discrete -> integer data
  cont <- yearmonth(as.Date("2024-03-15") + 0:3, discrete = FALSE)@x[[1L]]  # continuous -> double data
  expect_type(vctrs::vec_data(disc), "integer")
  expect_type(vctrs::vec_data(cont), "double")
  expect_type(vctrs::vec_data(vctrs::vec_slice(disc, 2:3)), "integer")
  expect_type(vctrs::vec_data(vctrs::vec_c(disc, disc)), "integer")
  expect_type(vctrs::vec_data(vctrs::vec_c(cont, cont)), "double")
  # round-trip through the mixtime/vecvec container preserves class and type
  combined <- c(yearmonth(0:1), yearmonth(2:3))
  expect_s7_class(combined, class_mixtime)
  expect_type(vctrs::vec_data(combined@x[[1L]]), "integer")
})

scale_type <- S7::new_external_generic("ggplot2", "scale_type", "x")
method(scale_type, class_mixtime) <- function(x) {
  check_ggtime_attached("0.2.0.9000")
  "mixtime"
}

check_ggtime_attached <- function(version) {
  if ("package:ggtime" %in% search()) {
    return(invisible(TRUE))
  }
  cli::cli_abort(
    c(
      "Plotting {.pkg mixtime} vectors requires the {.pkg ggtime} package.",
      i = if (!requireNamespace("ggtime", quietly = TRUE)) {
        "Install it with {.run install.packages(\"ggtime\")}."
      } else if (package_version(getNamespaceVersion("ggtime")) < version) {
        "Update to {.pkg ggtime} {version} or later with {.run install.packages(\"ggtime\")}."
      } else {
        "Attach it with {.run library(ggtime)}."
      }
    ),
    call = NULL
  )
}

test_that("scale_type() selects the mixtime scale", {
  skip_if_not_installed("ggplot2")
  skip_if_not_installed("ggtime", minimum_version = "0.2.0.9000")
  library(ggtime)

  expect_identical(ggplot2::scale_type(yearmonth(36L + 0:11)), "mixtime")
})

test_that("scale_type() reports when ggtime is unavailable", {
  # search() is mocked, since the real search path cannot be manipulated
  # once ggtime has been attached by an earlier test in this session.
  local_mocked_bindings(search = function() c(".GlobalEnv", "package:ggtime"), .package = "base")
  expect_true(check_ggtime_attached("0.2.0.9000"))

  local_mocked_bindings(search = function() ".GlobalEnv", .package = "base")
  expect_error(check_ggtime_attached("0.2.0.9000"), "ggtime")
})

#' Convert between chronons of different time granules
#' 
#' This function converts between chronons measured in different time granules.
#' It is used internally for converting between different continuous time types,
#' and is particularly useful for efficiently converting between irregular time
#' granules. The default method uses `chronon_cardinality()` to cast between 
#' time granules, which is efficient for regular time granules.
#' 
#' @param from The time granule that `x` is measured in (e.g., `day(1L)`).
#' @param to The time granule to convert `x` into (e.g., `week(1L)`).
# #' @param x An integer vector of chronons measured in the `from` time granule.
#' @param ... Additional arguments for methods.
#' 
#' @return An list of two elements:
#' - `div`: integer vector of chronons measured in the `to` time granule.
#' - `mod`: integer vector of the remainder (in `from` time granule) after
#'  converting to the `to` time granule.
#' 
#' @examples
#' # Convert day 16 after epoch (1970-01-01) into weeks since epoch (and remainder days)
#' with(cal_isoweek, chronon_divmod(day(1L), week(1L), 16L))
#' 
#' # Convert week 4 after epoch (1970-W1) into days since epoch
#' with(cal_isoweek, chronon_divmod(week(1L), day(1L), 4L))
#'
#' @export
chronon_divmod <- S7::new_generic("chronon_divmod", c("from", "to"))

# #' @rdname chronon_divmod
S7::method(chronon_divmod, list(mt_unit, mt_unit)) <- function(from, to, x) {
  # No casting needed for identical time granules
  if (identical(S7_class_id(from), S7_class_id(to))) {
    # Shortcut to skip divmod for identical granule divmod ops
    if (identical(from@n, to@n)) {
      return(list(div = vec_data(x), mod = 0L))
    }
    return(fdivmod(vec_data(x), to@n / from@n))
  }

  # Apply graph dispatch to find shortest path between from and to using
  # divmod conversions between time granules (e.g. day -> month)
  path <- S7_graph_dispatch(chronon_divmod_graph(), from, to)

  path[[1]] <- from
  path[[length(path)]] <- to
  # Initialise intermediate classes with 1L and adjacent properties
  path[c(-1, -length(path))] <- lapply(path[c(-1, -length(path))], function(tu){
    # Ideally inherit from `to`, but if incomplete inherit from `from`
    granule_inherit_props(granule_inherit_props(tu(1L), to), from)
  })
  

  chronon <- vector("list", length(path))
  chronon[[1]] <- vec_data(x)
  remainder <- vector("list", length(path)-1L)
  # Forward convert chronons
  for (i in seq(2, length.out = length(path)-1)) {
    result <- chronon_divmod_dispatch(path[[i-1L]], path[[i]], chronon[[i-1L]])
    chronon[[i]] <- result$div
    remainder[[i-1L]] <- result$mod
  }

  # Backward convert remainder
  for (i in seq(length(remainder), by = -1L, length.out = length(remainder) - 1L)) {
    if (remainder[[i]] != 0) {
      remainder[[i-1L]] <- remainder[[i-1L]] + chronon_cardinality(path[[i-1L]], path[[i]], chronon[[i]])*remainder[[i]]
    }
  }

  list(
    div = chronon[[length(chronon)]],
    mod = remainder[[1L]]
  )

}

chronon_divmod_dispatch <- function(from, to, x) {
  if (is.null(chronon_divmod@methods[[S7_class_id(from)]][[S7_class_id(to)]])) {
    return(chronon_divmod_regular(from, to, x))
  } else {
    return(chronon_divmod(from, to, x))
  }
}

## Fallback to chronon_cardinality for regular time granules
chronon_divmod_regular <- function(from, to, x) {
  fdivmod(x, chronon_cardinality(from, to))
}

#' @keywords internal
"_PACKAGE"

# The following block is used by usethis to automatically manage
# roxygen namespace tags. Modify with care!
## usethis namespace: start
#' @useDynLib mixtime, .registration = TRUE
#' @importFrom S7 method method<-
#' @import lifecycle
#' @import vctrs
#' @importFrom rlang ensym ensyms enexpr enexprs enquo enquos sym syms expr exprs quo quos eval_tidy
#' @importFrom S7 method<- super new_generic new_class new_property class_any S7_inherits S7_data S7_data<-
#' @importFrom vecvec vecvec_apply unvecvec
#' @useDynLib mixtime, .registration = TRUE
## usethis namespace: end
NULL

#' Friendly labels for linear relationships
#'
#' This S7 generic function provides the labels for linear (non-repeating)
#' positions of a time granule. These functions should return locale specific
#' labels.
#'
#' @param granule A time granule object representing the granule (e.g., `year(1L)`)
# #' @param i Integer vector representing the position along the linear axis.
# #' @param label If `TRUE`, return labels for the positions along the linear axis. If `FALSE`, return the integer positions as character.
# #' @param abbreviate If `TRUE`, return abbreviated labels (e.g. "year" as "Y"). If `FALSE`, return full labels.
#' @param ... Additional arguments for methods.
#'
#' @return Character vector of labels for the time point.
#'
#' @examples
#' # Labels for years on a linear axis
#' with(cal_gregorian, linear_labels(year(1L), 2020:2025))
#'
#' @export
linear_labels <- S7::new_generic("linear_labels", "granule")

# #' @rdname linear_labels
method(linear_labels, mt_unit) <- function(granule, i, label = FALSE, abbreviate = TRUE, ...) {
  if (!label) return(as.character(i))
  # TODO - pluralise the full time unit
  paste0(i, if (abbreviate) time_unit_abbr(granule) else time_unit_full(granule))
}

#' Obtain the chronon of a time object
#' 
#' This S7 generic function extracts the chronon (the smallest time granule) from a
#' time object, such as continuous time or cyclical time representations.
#' 
#' @param x A time object (e.g., [base::Date], [base::POSIXct], [linear_time()], etc.)
#' @param ... Additional arguments for methods.
#' 
#' @return A time [duration()] vector representing the chronon of each
#'   value (e.g., `days(1L)`).
#' 
#' @examples
#' 
#' # The chronon of a Date object is 1 day
#' time_chronon(Sys.Date())
#' 
#' # The chronon of a POSIXct object is 1 second
#' time_chronon(Sys.time())
#' 
#' # The chronon of a continuous time year and month is 1 month
#' time_chronon(yearmonth(Sys.Date()))
#' 
#' # The common chronon of a mixed time object is the finest chronon
#' time_chronon(c(yearmonth(Sys.Date()), Sys.Date()))
#' 
#' @export
time_chronon <- S7::new_generic("time_chronon", "x")

S7::method(time_chronon, class_mixtime) <- function(x) {
  x@x <- lapply(x@x, function(x) time_chronon(x)@x[[1L]])
  x
}

S7::method(time_chronon, mt_time) <- function(x) {
  duration(rep(1L, length(x)), attr(x, "chronon"))
}

S7::method(time_chronon, class_any) <- function(x) {
  duration(rep(1L, length(x)), mt_unit(1L))
}

S7::method(time_chronon, S7::new_S3_class("Date")) <- function(x) {
  # Date is a 1-day naive chronon
  duration(
    rep(1L, length(x)), 
    chronon = cal_gregorian$day(1L)
  )
}

S7::method(time_chronon, S7::new_S3_class("POSIXt")) <- function(x) {
  # POSIXct is a 1-second chronon with time zone information
  duration(
    rep(1L, length(x)),
    chronon = cal_gregorian$second(1L, tz = attr(x, "tzone") %||% "UTC")
  )
}

# {tsibble} time classes
S7::method(time_chronon, S7::new_S3_class("yearquarter")) <- function(x) {
  duration(rep(1L, length(x)), chronon = cal_gregorian$quarter(1L))
}
S7::method(time_chronon, S7::new_S3_class("yearmonth")) <- function(x) {
  duration(rep(1L, length(x)), chronon = cal_gregorian$month(1L))
}
S7::method(time_chronon, S7::new_S3_class("yearweek")) <- function(x) {
  duration(rep(1L, length(x)), chronon = cal_isoweek$week(1L))
}

# {zoo} time classes
S7::method(time_chronon, S7::new_S3_class("yearqtr")) <- function(x) {
  duration(rep(1L, length(x)), chronon = cal_gregorian$quarter(1L))
}
S7::method(time_chronon, S7::new_S3_class("yearmon")) <- function(x) {
  duration(rep(1L, length(x)), chronon = cal_gregorian$month(1L))
}

# {hms} time class
S7::method(time_chronon, S7::new_S3_class("hms")) <- function(x) {
  duration(rep(1L, length(x)), chronon = cal_gregorian$second(1L))
}

# {lubridate::Period} time class
S7::method(time_chronon, new_S4_class("Period", package = "lubridate")) <- function(x) {
  components <- c(second = list(`attributes<-`(x, NULL)), attributes(unclass(x)))
  has_component <- vapply(components, function(x) sum(x) != 0L, logical(1L))
  if (sum(has_component) != 1L) {
    cli::cli_abort("Support for {.fn lubridate::period} data in mixtime is limited to single unit periods.")
  }
  chronon <- switch(
    names(components)[has_component],
    year = cal_gregorian$year(1L),
    month = cal_gregorian$month(1L),
    day = cal_gregorian$day(1L),
    hour = cal_gregorian$hour(1L),
    minute = cal_gregorian$minute(1L),
    second = cal_gregorian$second(1L)
  )
  duration(rep(1L, length(x)), chronon = chronon)
}

#' Symmetry454 time unit classes
#'
#' Time unit constructors for the Symmetry454 calendar system. These units can be
#' used with [linear_time()] to create custom time representations.
#'
#' @format A civil-based calendar containing Symmetry454 time units.
#'
#' @details
#' The Symmetry454 calendar (Sym454) is a perennial solar calendar proposed by
#' Dr. Irv Bromberg. It preserves the traditional 12-month structure and 7-day
#' week, with months arranged in a symmetrical 4-5-4 week pattern per quarter.
#' Every month starts on Monday and has a whole number of weeks, meaning no
#' month ever contains a partial week.
#' 
#' The following time units are available in the Symmetry454 calendar (`cal_sym454$`).
#'
#' - `year()`: Year unit
#' - `month()`: Month unit
#' - `week()`: Week unit
#' - `day()`: Day unit
#' - `hour()`: Hour unit
#' - `minute()`: Minute unit
#' - `second()`: Second unit
#' - `millisecond()`: Millisecond unit
#'
#' ## Leap years
#'
#' Rather than intercalary days, Symmetry454 uses a **leap week** appended to
#' December once every 5 or 6 years, making December a 5-week month in leap
#' years. A year is a leap year if `(52 * year + 146) %% 293 < 52`.
#'
#' This yields a mean year of 365 + 71/293 days (approx 365 days, 5 hours, 48
#' minutes, 56.5 seconds), intentionally slightly shorter than the mean
#' northward equinoctial year.
#'
#' @return An S3 list of class `c("cal_sym454", "mt_calendar")` containing
#'   the named time unit classes of the Symmetry454 calendar. Each unit is
#'   accessible via `$` notation and calling it with a step size produces a
#'   time granule (e.g., 1 week granule as `cal_sym454$week(1L)`).
#'
#' @seealso [linear_time()] for creating linear time points, and
#'   <https://en.wikipedia.org/wiki/Symmetry454> for more calendar details.
#'
#' @examples
#' # Create a custom time representation using Symmetry454 units
#' linear_time(
#'   Sys.time(),
#'   chronon = cal_sym454$week(1L)
#' )
#'
#' @name calendar_sym454
#' @export
cal_sym454 <- new_calendar(
  year = new_class("tu_sym454_year", parent = mt_tz_unit),
  month = new_class("tu_sym454_month", parent = mt_tz_unit),
  week = cal_isoweek$week,
  inherit = cal_time_civil,
  class = "cal_sym454"
)

# Time unit labels
method(time_unit_full, cal_sym454$year) <- function(x) "year{?/s}"
method(time_unit_abbr, cal_sym454$year) <- function(x) "Y"
method(time_unit_full, cal_sym454$month) <- function(x) "month{?/s}"
method(time_unit_abbr, cal_sym454$month) <- function(x) "M"

## Default formats
S7::method(
  chronon_format_linear,
  list(cal_sym454$year, S7::class_any)
) <- function(x, cal) "{lin(year(1L))}"
S7::method(
  chronon_format_linear,
  list(cal_sym454$month, S7::class_any)
) <- function(x, cal) "{lin(year(1L))}-{cyc(month(1L), year(1L), label=TRUE, abbreviate=TRUE)}"
S7::method(
  chronon_format_linear,
  list(cal_sym454$week, S7::new_S3_class("cal_sym454"))
) <- function(x, cal) "{lin(year(1L))}-{cyc(month(1L), year(1L), label=TRUE, abbreviate=TRUE)}-W{cyc(week(1L), month(1L))}"
S7::method(
  chronon_format_linear,
  list(cal_sym454$day, S7::new_S3_class("cal_sym454"))
) <- function(x, cal) "{lin(year(1L))}-{cyc(month(1L), year(1L), label=TRUE, abbreviate=TRUE)}-W{cyc(week(1L), month(1L))}-{cyc(day(1L), week(1L), label=TRUE, abbreviate=TRUE)}"

## Time labels
# Same "1BC"/"1"/"2"/... era scheme as `cal_gregorian$year` -- reuse its
# format/parse functions (they only ever touch `granule` generically via
# chronon_epoch(), so they work unmodified for the sym454 year class too).
method(linear_labels_format, cal_sym454$year) <- bc_ad_format
method(linear_labels_parse, cal_sym454$year) <- bc_ad_parse
method(cyclical_labels, list(cal_sym454$month, cal_sym454$year)) <- label_scheme(
  start = 1L,
  width = 2L,
  vocab = vocab_table(`en-GB` = list(wide = month.name, abbreviated = month.abb))
)
# Weeks count with 1-indexing
method(cyclical_labels, list(cal_sym454$week, cal_sym454$month)) <- label_scheme(start = 1L)

# Candidate parsing formats
method(chronon_parse_linear, list(cal_sym454$month, class_any)) <- function(x, cal) {
  parse_format(
    # e.g. "2024-01", "2024 Jan", "2024, January"
    "{lin(year)}[-/,\\s]+{cyc(month,year)}",
    # e.g. "01-2024", "Jan 2024", "January, 2024"
    "{cyc(month,year)}[-/,\\s]+{lin(year)}",
    regex = TRUE
  )
}
method(
  chronon_parse_linear,
  list(cal_sym454$day, S7::new_S3_class("cal_sym454"))
) <- function(x, cal) {
  parse_format(
    # e.g. "2024-Jan-W2-3", "2024 January w2 3rd"
    "{lin(year)}[-/,\\s]+{cyc(month,year)}[-/,\\s]+[Ww]?{cyc(week,month)}[-/,\\s]+{cyc(day,week)}(?:st|nd|rd|th)?",
    # e.g. "3rd-W2-Jan-2024", "3, w2, January 2024"
    "{cyc(day,week)}(?:st|nd|rd|th)?[-/,\\s]+[Ww]?{cyc(week,month)}[-/,\\s]+{cyc(month,year)}[-/,\\s]+{lin(year)}",
    regex = TRUE
  )
}
method(chronon_parse_cyclical, list(cal_sym454$month, cal_sym454$year)) <- function(x, y) {
  # e.g. "1", "Jan", "January"
  parse_format("{cyc(month,year)}")
}
method(chronon_parse_cyclical, list(cal_sym454$week, cal_sym454$month)) <- function(x, y) {
  parse_format(
    # e.g. "2", "W2", "w02"
    "[Ww]?{cyc(week,month)}",
    regex = TRUE
  )
}

# Calendar arithmetic
## Epochs
method(chronon_epoch, cal_sym454$year) <- function(x) 1970L

## Cardinality
S7::method(chronon_cardinality_fixed, list(cal_sym454$month, cal_sym454$year)) <-
  function(x, y) {
    12L
  }
S7::method(chronon_cardinality, list(cal_sym454$week, cal_sym454$month)) <-
  function(x, y, at = NULL) {
    # The number of weeks in each n-month period
    month_size <- y@n
    nweeks_cyc <- circsum(c(4L, 5L, 4L), month_size)

    # Find which n-month period we're in based on the "at" position (months since epoch)
    period <- at %% length(nweeks_cyc) + 1L

    nweeks <- nweeks_cyc[period]

    # Add the extra week to December for Symmetry454 leap years.
    # A year is a leap year when (52*year + 146) %% 293 < 52.
    m1 <- at * month_size
    contains_dec <- which((m1 %% 12L) >= (12L - month_size))
    year <- 1970L + m1[contains_dec] %/% 12L
    is_leap_year <- ((52 * year + 146L) %% 293L) < 52L

    nweeks[contains_dec[is_leap_year]] <- nweeks[contains_dec[is_leap_year]] + 1L

    # Scale by the number of weeks in the week time unit
    nweeks / x@n
  }

## Divmod
S7::method(chronon_divmod, list(cal_sym454$week, cal_sym454$month)) <-
  function(from, to, x) {
    # Most of this code works on 1-week units
    week_size <- from@n
    x <- x * week_size  # convert n-weeks to 1-weeks

    # 1. Account for leap weeks by regularising x to have a fixed 52 weeks per year

    # The symmetrical sub-cycles of the 293-year leap week cycle are:
    #   17+11+17 + 17+17+11+17+17 + 17+11+17 + 17+17+11+17+17 + 17+11+17
    #   = 45 + 79 + 45 + 79 + 45 = 293
    # Primary (length 17) sub-cycles have 3 leap years: 00100000100000100
    # Secondary (length 11) sub-cycles have 2 leap years: 00100000100
    leaps_cycle_17 <- function(w) (w >= 157L) + (w >= 470L) + (w >= 783L)
    leaps_cycle_11 <- function(w) (w >= 157L) + (w >= 470L)
    leaps_cycle_45 <- function(w) {
      ifelse(w < 887L,  leaps_cycle_17(w),
      ifelse(w < 1461L, 3L + leaps_cycle_11(w - 887L),
                        5L + leaps_cycle_17(w - 1461L)))
    }
    leaps_cycle_79 <- function(w) {
      ifelse(w < 887L,  leaps_cycle_17(w),
      ifelse(w < 1774L, 3L  + leaps_cycle_17(w - 887L),
      ifelse(w < 2348L, 6L  + leaps_cycle_11(w - 1774L),
      ifelse(w < 3235L, 8L  + leaps_cycle_17(w - 2348L),
                        11L + leaps_cycle_17(w - 3235L)))))
    }
    leaps_symmetry454 <- function(x) {
      # There are 15288 weeks in a full 293-year cycle (293*52 + 52 leap weeks)
      w <- x %% 15288L
      x %/% 15288L * 52L +
      ifelse(w < 2348L,  leaps_cycle_45(w),
      ifelse(w < 6470L,  8L  + leaps_cycle_79(w - 2348L),
      ifelse(w < 8818L,  22L + leaps_cycle_45(w - 6470L),
      ifelse(w < 12940L, 30L + leaps_cycle_79(w - 8818L),
                         44L + leaps_cycle_45(w - 12940L)))))
    }

    # Offset x to align with the nearest 293-year cycle boundary before the epoch.
    # There are 349 leap years between year 1-W1 and the 1970-W1 epoch, and the
    # nearest cycle start is (1969*52 + 349) %% (293*52 + 52) = 11009 weeks before epoch.
    x_cyc <- x + 11009L + week_size  # right align multi-week units
    n_leaps <- leaps_symmetry454(x_cyc)

    # Regularise x to have exactly 52 weeks per year by subtracting leap weeks.
    # (37 leap years occur in the cycle before the epoch, so we add 37 back.)
    x_reg <- x - n_leaps + 37L

    # 2. Use the 4-5-4 pattern to find the month (div) and week remainder (mod)
    ## The number of weeks in each n-month period
    month_size <- to@n
    weeks_len <- circsum(c(4L, 5L, 4L), month_size)

    ## The total weeks in a full n-month cycle
    weeks_tot <- sum(weeks_len)

    ## Find which n-month cycle we're in based on the regularised week count
    period_full <- x_reg %/% weeks_tot

    ## Find which part within the n-month cycle we're in
    weeks_seq  <- cumsum(weeks_len[-length(weeks_len)])
    period_part <- rowSums(outer(x_reg %% weeks_tot, weeks_seq, ">="))

    # div: total complete n-month cycles + complete n-months within the current cycle
    div <- period_full * length(weeks_len) + period_part
    # mod: remaining (regularised) weeks within the current n-month period
    mod <- x_reg %% weeks_tot - c(0L, weeks_seq)[period_part + 1L]


    # 3. Adjust the remainder to account for leap weeks that were removed during regularisation.

    # Identify leap weeks re-using cumulative in-cycle leap week counts: leaps_symmetry454()
    # If the week added a leap week from the previous, then it itself must be a leap week.
    # Applied only to regularised 52/53rd weeks of the year (only these weeks can be leap weeks)
    last_weeks  <- which(x_reg %% 52L >= 51L)
    leap_weeks  <- last_weeks[n_leaps[last_weeks] - leaps_symmetry454(x_cyc[last_weeks] - week_size) > 0L]
    mod[leap_weeks] <- mod[leap_weeks] + 1L  # restore leap week to remainder

    # Scale mod back to the original n-unit week size
    mod <- mod %/% week_size

    # Return the divmod result
    list(div = div, mod = mod)
  }

S7::method(chronon_divmod, list(cal_sym454$month, cal_sym454$week)) <-
  function(from, to, x) {

    # 1. Compute 4-5-4 cycle lengths for the given n-month period
    month_size <- from@n
    weeks_len <- circsum(c(4L, 5L, 4L), month_size)
    weeks_seq  <- cumsum(weeks_len)
    weeks_tot <- weeks_seq[length(weeks_seq)]
    
    # Modular arithmetic to find weeks from months (without leap weeks)
    period_full <- (x %/% length(weeks_len)) * weeks_tot
    period_part <- stats::approx(
      x = seq(0, length.out = length(weeks_len) + 1L), 
      y = c(0L, weeks_seq), 
      xout = x %% length(weeks_len), 
    )$y

    # 2. Compute the number of leap weeks to add back in
    # The symmetrical sub-cycles of the 293-year leap week cycle are:
    #   17+11+17 + 17+17+11+17+17 + 17+11+17 + 17+17+11+17+17 + 17+11+17
    #   = 45 + 79 + 45 + 79 + 45 = 293
    # Primary (length 17) sub-cycles have 3 leap years: 00100000100000100
    # Secondary (length 11) sub-cycles have 2 leap years: 00100000100
    leaps_cycle_17 <- function(m) (m >= 35L) + (m >= 107L) + (m >= 179L)
    leaps_cycle_11 <- function(m) (m >= 35L) + (m >= 107L)
    leaps_cycle_45 <- function(m) {
      ifelse(m < 204L,  leaps_cycle_17(m),
      ifelse(m < 336L,  3L + leaps_cycle_11(m - 204L),
                        5L + leaps_cycle_17(m - 336L)))
    }
    leaps_cycle_79 <- function(m) {
      ifelse(m < 204L,  leaps_cycle_17(m),
      ifelse(m < 408L,  3L  + leaps_cycle_17(m - 204L),
      ifelse(m < 540L,  6L  + leaps_cycle_11(m - 408L),
      ifelse(m < 744L,  8L  + leaps_cycle_17(m - 540L),
                        11L + leaps_cycle_17(m - 744L)))))
    }
    leaps_symmetry454 <- function(x) {
      # There are 3516 months in a full 293-year cycle (293*12 months)
      m <- x %% 3516L
      x %/% 3516L * 52L +
      ifelse(m < 540L,   leaps_cycle_45(m),
      ifelse(m < 1488L,  8L  + leaps_cycle_79(m - 540L),
      ifelse(m < 2028L,  22L + leaps_cycle_45(m - 1488L),
      ifelse(m < 2976L,  30L + leaps_cycle_79(m - 2028L),
                         44L + leaps_cycle_45(m - 2976L)))))
    }
    x_cyc <- x * month_size + 2531L
    n_leaps <- leaps_symmetry454(x_cyc) - 37L # number of leaps after 1970 Jan
    
    # 3. Combine complete cycles, partial cycles, and leap week adjustments 
    # to compute the final number of weeks (div) and remaining months (mod).
    week_size <- to@n
    weeks <- period_full + period_part + n_leaps
    list(div = weeks / week_size, mod = 0L)
  }

.onLoad <- function(...) {
  # Register S7 methods
  S7::methods_register()

  # Register vecvec methods
  vecvec::vecvec_register(
    class_mixtime,
    f_ptype2 = vec_ptype2_mixtime,
    f_cast_to = vec_cast_to_mixtime
  )

  # Register vec_cast / vec_ptype2 for the mt_time family on their package-namespaced
  # concrete classes (vctrs' custom double dispatch for these generics does not use
  # inheritance). Arithmetic is handled by native S7 operator methods (see arithmetic.R).
  register_mt_vctrs()

  # Initialise tzdb package
  tzdb::tzdb_initialize()

  invisible()
}

#' Create a new time unit class
#'
#' Define a new S7 class representing a time unit for use in a mixtime calendar.
#' Time units are the building blocks of calendars: each unit represents a
#' specific temporal component (e.g., day, month, year) and carries a step size
#' `n`. Units are combined via [new_calendar()] to form a complete calendar
#' system.
#'
#' Choose the parent class based on the type of time the unit represents:
#'
#' * [`mt_unit`] — abstract or calendar-only time (no timezone or location
#'   context; e.g., Gregorian year or ISO week).
#' * [`mt_tz_unit`] — civil time with a timezone (e.g., a clock hour that
#'   needs `tz` to resolve wall-clock ambiguity).
#' * [`mt_loc_unit`] — astronomical time with a geographic location (e.g., a
#'   solar day tied to observer longitude/latitude).
#'
#' @param name A string naming the new S7 class (e.g., `"tu_my_year"`).
#'   By convention, time unit class names are prefixed with `tu_`.
#' @param parent The parent S7 class. Should be one of [`mt_unit`],
#'   [`mt_tz_unit`], or [`mt_loc_unit`], or a class that itself inherits from
#'   one of these. Defaults to [`mt_unit`].
#' @param package A string giving the package name that owns the class.
#'   Defaults to the name of the calling namespace, so typically does not
#'   need to be set explicitly.
#' @param properties A named list of additional [`S7::new_property()`]
#'   definitions beyond those inherited from `parent`. Each entry becomes
#'   a slot on instances of the new class. Defaults to an empty list.
#' @param abstract Logical. If `TRUE` the class cannot be instantiated
#'   directly; it serves only as a base for further subclassing.
#' @param constructor A function to use as the class constructor, or `NULL`
#'   (default) to generate one automatically. The auto-generated constructor
#'   accepts `...` (forwarded to the parent constructor) plus one argument per
#'   entry in `properties`, with defaults taken from each property's
#'   `default` field.
#' @param validator A function of one argument (`self`) that returns `NULL`
#'   when the object is valid, or a character string describing the problem.
#'   See [S7::new_class()] for details.
#'
#' @return An S7 class object, as returned by [S7::new_class()].
#'
#' @seealso
#' * [`mt_unit`], [`mt_tz_unit`], [`mt_loc_unit`] for the base classes to
#'   inherit from.
#' * [new_calendar()] for assembling time units into a calendar.
#' * [S7::new_class()] for the underlying S7 class constructor.
#'
#' @examples
#' # An abstract unit with no properties
#' tu_my_seq <- new_time_unit("tu_my_seq", parent = mt_unit)
#'
#' # A civil-time unit with an extra property
#' tu_my_quarter <- new_time_unit(
#'   "tu_my_quarter",
#'   parent = mt_tz_unit,
#'   properties = list(
#'     fiscal = S7::new_property(S7::class_logical, default = FALSE)
#'   )
#' )
#'
#' @export
new_time_unit <- function(
  name,
  parent = mt_unit,
  package = topNamespaceName(parent.frame()),
  properties = list(),
  abstract = FALSE,
  constructor = NULL,
  validator = NULL
) {
  if (is.null(constructor)) {
    props <- syms(names(properties))
    names(props) <- names(properties)
    
    constructor <- rlang::new_function(
      args = rlang::pairlist2(... =, !!!lapply(properties, `[[`, "default")),
      body = expr(S7::new_object(parent(!!sym("...")), !!!props))
    )
  }

  S7::new_class(
    name = name,
    parent = parent,
    package = package,
    properties = properties,
    abstract = abstract,
    constructor = constructor,
    validator = validator
  )
}

#' Time units as a string
#' 
#' These S7 generic functions provide the full and abbreviated names for time
#' units. `time_unit_full()` is used in messages and durations (e.g., "2 months").
#' `time_unit_abbr()` is used for tsibble index interval displays (e.g., "1M").
#' `time_unit_plural()` pluralises the full unit name for a given quantity using
#' cli-style pluralisation (e.g., `"year{?/s}"` becomes `"year"` or `"years"`).
#' 
#' @param x A time granule object (e.g., `cal_gregorian$month(1L)`)
#' @param n Numeric quantity used to select singular or plural form.
#' @param ... Additional arguments for methods.
#' 
#' @return A string representing the time unit
#' 
#' @examples
#' time_unit_full(cal_gregorian$year(1L))
#' time_unit_plural(cal_gregorian$year(1L), 1L)
#' time_unit_plural(cal_gregorian$year(1L), 2L)
#' time_unit_abbr(cal_gregorian$year(1L))
#' 
#' @rdname time_unit_labels
#' @export
time_unit_full <- S7::new_generic("time_unit_full", "x")
method(time_unit_full, mt_unit) <- function(x) "unit{?/s}"

#' @rdname time_unit_labels
#' @export
time_unit_plural <- function(x, n = 2L) {
  str_plural(time_unit_full(x), as.numeric(n))
}

#' @rdname time_unit_labels
#' @export
time_unit_abbr <- S7::new_generic("time_unit_abbr", "x")
method(time_unit_abbr, mt_unit) <- function(x) ""

# A quantity of a time unit as text, e.g. "15 months". Defaults to describing
# the granule itself (`month(15L)`), since `mt_unit` objects are S7 scalars
# which neither paste() nor cli can interpolate directly. Vectorised over `n`.
time_granule_label <- function(x, n = x@n) {
  paste(n, time_unit_plural(x, n))
}

# ---------------------------------------------------------------------------
# Integer (discrete) durations
# ---------------------------------------------------------------------------

test_that("integer duration of 1 is singular", {
  expect_equal(format(minutes(1L)), "1 minute")
  expect_equal(format(seconds(1L)), "1 second")
  expect_equal(format(hours(1L)),   "1 hour")
  expect_equal(format(days(1L)),    "1 day")
  expect_equal(format(months(1L)),  "1 month")
  expect_equal(format(years(1L)),   "1 year")
})

test_that("integer duration > 1 is plural", {
  expect_equal(format(minutes(2L)),  "2 minutes")
  expect_equal(format(seconds(30L)), "30 seconds")
  expect_equal(format(hours(12L)),   "12 hours")
  expect_equal(format(days(7L)),     "7 days")
  expect_equal(format(months(6L)),   "6 months")
  expect_equal(format(years(3L)),    "3 years")
})

test_that("integer duration of 0 is plural", {
  expect_equal(format(minutes(0L)), "0 minutes")
})

# ---------------------------------------------------------------------------
# Double (fractional) durations
# ---------------------------------------------------------------------------

test_that("fractional duration 1.0 shows one decimal place and is plural", {
  expect_equal(format(minutes(1.0)), "1.0 minutes")
  expect_equal(format(seconds(1.0)), "1.0 seconds")
  expect_equal(format(hours(1.0)),   "1.0 hours")
})

test_that("fractional duration with existing decimal shows at least one decimal place", {
  expect_equal(format(minutes(1.1)),  "1.1 minutes")
  expect_equal(format(minutes(2.5)),  "2.5 minutes")
  expect_equal(format(minutes(0.5)),  "0.5 minutes")
})

test_that("fractional duration with more decimals is shown up to 3 decimal places by default", {
  expect_equal(format(minutes(1.1)),    "1.1 minutes")
  expect_equal(format(minutes(1.16)),   "1.16 minutes")
  expect_equal(format(minutes(1.123)),  "1.123 minutes")
  expect_equal(format(minutes(1.1234)), "1.1234 minutes")
})


test_that("fractional duration 0.0 is plural with one decimal", {
  expect_equal(format(minutes(0.0)), "0.0 minutes")
})

# ---------------------------------------------------------------------------
# NA values
# ---------------------------------------------------------------------------

test_that("NA integer duration formats as NA", {
  expect_equal(format(minutes(NA_integer_)), "NA")
})

test_that("NA double duration formats as NA", {
  expect_equal(format(minutes(NA_real_)), "NA")
})

# ---------------------------------------------------------------------------
# Vectors mixing values and NAs
# ---------------------------------------------------------------------------

test_that("vector of integer durations formats correctly", {
  expect_equal(
    format(minutes(c(1L, 2L, NA_integer_))),
    c("1 minute", "2 minutes", "NA")
  )
})

test_that("vector of fractional durations formats correctly", {
  expect_equal(
    format(minutes(c(1.0, 1.1, NA_real_))),
    c("1.0 minutes", "1.1 minutes", "NA")
  )
})

# ---------------------------------------------------------------------------
# Converting between duration chronons
# ---------------------------------------------------------------------------

test_that("converting a duration to a coarser chronon is continuous by default for doubles", {
  expect_equal(as.numeric(duration(days(4.0), cal_isoweek$week(1L))), 4/7)
})

test_that("converting a duration to a coarser chronon is discrete by default for integers", {
  expect_identical(as.integer(duration(days(4L), cal_isoweek$week(1L))), 0L)
})

test_that("discrete = FALSE returns a fractional duration", {
  expect_equal(as.numeric(duration(days(4L), cal_isoweek$week(1L), discrete = FALSE)), 4/7)
})

test_that("discrete = TRUE truncates to whole units", {
  expect_identical(as.integer(duration(days(4), cal_isoweek$week(1L), discrete = TRUE)), 0L)
  expect_identical(as.integer(duration(days(10), cal_isoweek$week(1L), discrete = TRUE)), 1L)
})

test_that("converting a duration to a finer chronon multiplies the magnitude", {
  expect_equal(as.numeric(duration(weeks(2L), cal_gregorian$day(1L))), 14)
})

test_that("Comparison - same chronon compares magnitudes directly", {
  expect_true(days(3L) == days(3L))
  expect_true(days(3L) < days(5L))
  expect_false(days(3L) > days(5L))
  expect_true(days(3L) != days(5L))
  expect_equal(days(1:3) < days(2L), c(TRUE, FALSE, FALSE))
})

test_that("Comparison - different chronons compare in a common chronon", {
  expect_true(weeks(1L) == days(7L))
  expect_true(weeks(1L) > days(6L))
  expect_true(days(8L) > weeks(1L))
  expect_false(weeks(1L) < days(7L))
  expect_true(weeks(1L) <= days(7L))
})

test_that("Comparison - a bare number is a duration in the other chronon", {
  expect_true(days(3L) == 3)
  expect_true(3 == days(3L))
  expect_true(days(3L) < 5)
})

# ---------------------------------------------------------------------------
# Converting a bare duration granule (pulled out of a mixtime's `@x`)
# ---------------------------------------------------------------------------

test_that("converting a bare duration granule agrees with converting its mixtime", {
  d <- days(1:3)
  expect_equal(
    duration(d@x[[1L]], chronon = cal_gregorian$hour(1L), discrete = FALSE),
    duration(d, chronon = cal_gregorian$hour(1L), discrete = FALSE)
  )
  expect_equal(
    as.numeric(duration(d@x[[1L]], chronon = cal_gregorian$hour(1L), discrete = FALSE)),
    c(24, 48, 72)
  )
})

test_that("converting a bare duration granule round-trips through another chronon", {
  d <- days(1:3)
  back <- duration(
    duration(d@x[[1L]], chronon = cal_gregorian$hour(1L), discrete = FALSE)@x[[1L]],
    chronon = cal_gregorian$day(1L), discrete = FALSE
  )
  expect_equal(as.numeric(back), as.numeric(d))
})

test_that("discrete = FALSE on a bare duration granule yields a double", {
  g <- days(1:3)@x[[1L]]
  res <- duration(g, chronon = cal_gregorian$hour(1L), discrete = FALSE)
  expect_false(is.integer(vctrs::vec_data(res@x[[1L]])))
})

test_that("duration() refuses a non-duration time granule", {
  tp <- mixtime(as.Date("2023-01-01") + 0:2)
  expect_error(
    duration(tp@x[[1L]], chronon = cal_gregorian$hour(1L)),
    "isn't a duration"
  )
})

test_that("calendar-dependent duration conversion errors with a duration-specific message", {
  expect_error(
    duration(days(60), chronon = cal_gregorian$month(1L), discrete = FALSE),
    "no anchor"
  )
})

test_that("zero-length durations can be built and converted", {
  # The element-wise duration predicate is `NA` for a zero-length input, so
  # zero-length magnitudes must be routed by class rather than by value.
  expect_length(seconds(integer()), 0L)
  expect_length(duration(numeric(), chronon = cal_gregorian$day(1L)), 0L)

  empty <- duration(
    seconds(integer())@x[[1L]],
    chronon = cal_gregorian$hour(1L),
    discrete = FALSE
  )
  expect_length(empty, 0L)
  expect_equal(time_unit_plural(attr(empty@x[[1L]], "chronon")), "hours")
})

#' Convert between chronons
#'
#' @param x A linear time object (of class `mt_linear`)
#' @param to The target chronon to convert to (a time granule object)
#' @param discrete If `TRUE`, the number of target chronons since Unix epoch
#' @param ... Additional arguments for methods.
#' that `x` falls into is returned as an integer. If `FALSE`, a fractional
#' number of target chronons is returned (analagous to time using a continuous
#' time model).
#'
#' @return A numeric or integer representing the time since Unix epoch in terms of the target
#' chronon's precision.
#'
#' @examples
#'
#' # Convert from months since epoch to years since epoch
#' chronon_convert(yearmonth(Sys.Date()), cal_gregorian$year(1L))
#'
#' # Convert from days since epoch to months since epoch
#' chronon_convert(Sys.Date(), cal_gregorian$month(1L))
#' chronon_convert(Sys.Date(), cal_gregorian$month(1L), discrete = TRUE)
#'
#' # Convert from seconds since epoch to hours since epoch
#' chronon_convert(Sys.time(), cal_gregorian$hour(1L))
#'
#' @noRd
#' @keywords internal
chronon_convert <- S7::new_generic("chronon_convert", "x")

chronon_convert_impl <- function(x, from, to, discrete, tz = NULL) {
  # Inherit granule properties from `from` if not given in `to` granule
  to <- granule_inherit_props(to, from)

  # Handle timezones
  # TODO - Generalise for converting to any different granule attribute
  # (e.g. locA -> locB)

  is_tz_conv <- S7_inherits(from, mt_tz_unit) && S7_inherits(to, mt_tz_unit)
  if (is_tz_conv) {
    if (is.na(tz_name(from)) && !is.na(tz_name(to))) {
      cli::cli_abort(
        "Cannot convert from timezone-naive chronon to timezone-aware chronon.",
        call = NULL
      )
    }
  }

  # Find path along convertable time units
  path <- S7_graph_dispatch(chronon_divmod_graph(), from, to)

  # Convert to target tz naive time for standard chronon boundaries
  # Skipped if:
  # * target tz is UTC (since that matches the internal representation)
  is_tz_utc <- identical(tz_name(to), "UTC") || identical(tz, "UTC")
  # * target is not a timezone-aware mt_tz_unit
  is_tz_aware <- is_tz_conv && !is.na(tz_name(to))

  if (!is_tz_utc && is_tz_aware) {
    x <- tz_to_local(x, from, tz_name(to))
  } else if (is_tz_conv && is.na(tz_name(to)) && !is.na(tz_name(from))) {
    # Converting from tz-aware to tz-naive - shift into the local wall-clock
    # domain the naive `to` chronon will read `x` as.
    x <- tz_to_local(x, from, tz_name(from))
  }

  # Convert between time points
  if (length(path) == 1L) {
    # Simple direct conversion by cardinality
    x <- vec_data(x) / chronon_cardinality(from, to, at = x)
  } else {
    # Convert along the path of time granules
    path[[1]] <- from
    prop_from <- S7::prop_names(from)
    path[[length(path)]] <- to
    prop_to <- S7::prop_names(to)
    # Initialise intermediate classes with 1L and adjacent properties
    path[c(-1, -length(path))] <- lapply(
      path[c(-1, -length(path))],
      function(tu) {
        # Ideally inherit from `to`, but if incomplete inherit from `from`
        granule_inherit_props(granule_inherit_props(tu(1L), to), from)
      }
    )

    # Convert chronons along the path
    not_na <- !is.na(x)
    for (i in seq(2, length.out = length(path) - 1)) {
      res <- chronon_divmod_dispatch(path[[i - 1L]], path[[i]], x[not_na])
      x[not_na] <- res$div
      nz_mod <- res$mod != 0
      part <- chronon_cardinality(
        path[[i - 1L]],
        path[[i]],
        floor(res$div[nz_mod])
      )
      x[not_na][nz_mod] <- x[not_na][nz_mod] + res$mod[nz_mod] / part
    }
  }

  # Undo the pre-conversion shift, for time zoned data.
  # Skipped if:
  # * target tz is NA (since that indicates a timezone-naive chronon)
  # * target tz is UTC (since that matches the internal representation)
  if (!is_tz_utc && is_tz_aware) {
    x <- tz_to_utc(x, to, tz_name(to), discrete = discrete, boundary = length(path) > 1L)
  } else if (discrete) {
    # Nudge by a few ULPs to avoid floating-point errors accumulated along
    # the path conversion landing just below an exact unit boundary.
    x <- as.integer(floor(x + 8 * .Machine$double.eps * pmax(abs(x), 1)))
  }

  x
}

method(chronon_convert, mt_linear) <- function(x, to, discrete = FALSE, ...) {
  chronon_convert_impl(vec_data(x), attr(x, "chronon"), to, discrete)
}

method(chronon_convert, class_mixtime) <- vecvec::vecvec_apply_fn(
  chronon_convert,
  SIMPLIFY = TRUE
)

method(chronon_convert, S7::class_any) <- function(
  x,
  to,
  discrete = FALSE,
  ...
) {
  chronon_convert_impl(as.numeric(x), chronon_common(x), to, discrete)
}

#' Cyclical time function factory
#' 
#' `new_cyclical_time_fn()` creates a cyclical time function for a specified
#' chronon and cycle. The cycle is the larger time granule that defines the time 
#' period over which the chronon loops (e.g., a week). The chronon is the smaller 
#' time granule that iterates within each cycle (e.g., a day). Combined, these two
#' granules form a cyclical time relationship (e.g., day of the week).
#' 
#' @param chronon A time granule object representing the chronon (e.g., `day(1L)`)
#' @param cycle A time granule object representing the cycle (e.g., `week(1L)`)
#' @param default_calendar A default calendar used to find the time units for 
#'   conversion if they don't exist in the calendar of the input data (e.g., `cal_isoweek`)
#' 
#' @return A function used to create cyclical time points with a specific chronon and cycle.
#' 
#' @examples
#' 
#' day_of_week <- new_cyclical_time_fn(day(1L), week(1L), default_calendar = cal_isoweek)
#' day_of_week(Sys.Date())
#' 
#' month_of_year <- new_cyclical_time_fn(month(1L), year(1L))
#' month_of_year(Sys.Date())
#' 
#' @export
new_cyclical_time_fn <- function(chronon, cycle, default_calendar = cal_gregorian) {
  chronon <- rlang::new_quosure(
    enexpr(chronon), 
    env = rlang::as_data_mask(default_calendar)
  )
  cycle <- rlang::new_quosure(
    enexpr(cycle), 
    env = rlang::as_data_mask(default_calendar)
  )

  function(
    data, discrete = TRUE, calendar = time_calendar(data), ...
  ) {
    # Add tz / loc to chronon
    chronon <- quo_add_dots(chronon, ...)

    cyclical_time(
      data, chronon = !!chronon, cycle = !!cycle, 
      discrete = discrete, calendar = calendar
    )
  }
}

#' Cyclical time points
#' 
#' `cyclical_time()` creates a vector of cyclical time points representing
#' positions within repeating cycles. This function is useful for creating custom
#' cyclical time representations that aren't covered by the convenience functions
#' like [day_of_week()] or [month_of_year()].
#' 
#' @param data Input data to convert to cyclical time. Can be:
#'   - Numeric values (interpreted as chronons, 1-indexed)
#'   - Character strings (parsed as dates/times)
#'   - Date or POSIXct objects
#'   - Other time objects
#' @param chronon A time granule representing the chronon (finest 
#'   indivisible time granule), evaluated in the context of `calendar`. Use 
#'   unquoted expressions like `day(1L)` or `month(1L)`. Chronons from a
#'   specific calendar can also be used (e.g. `cal_isoweek$day(1L)`).
#' @param cycle A time granule representing the cycle (coarser time granule
#'   that defines the period), evaluated in the context of `calendar`. Use
#'   unquoted expressions like `week(1L)` or `year(1L)`.
#' @param discrete Logical. If `TRUE` (default), returns integer positions within
#'   the cycle (discrete time model). If `FALSE`, returns fractional positions 
#'   allowing representation of fractional time chronons (continuous time model).
#' @param calendar Calendar system used to evaluate `chronon` and `cycle`.
#'   Defaults to `time_calendar(data)` for existing time objects. Common options
#'   include [cal_gregorian] and [cal_isoweek].
#' 
#' @return A `mixtime` time vector containing an `mt_cyclical` vector.
#' 
#' @seealso 
#' - [new_cyclical_time_fn()] for creating reusable cyclical time functions
#' - [day_of_week()], [day_of_month()], [day_of_year()] for common cyclical representations
#' - [month_of_year()], [week_of_year()] for other cyclical time helpers
#' - [cal_gregorian], [cal_isoweek] for calendar systems
#' 
#' @examples
#' # Day of week (1-7, Monday = 1)
#' cyclical_time(
#'   Sys.Date(),
#'   chronon = day(1L),
#'   cycle = week(1L),
#'   calendar = cal_isoweek
#' )
#' 
#' # Month of year (1-12)
#' cyclical_time(
#'   Sys.Date(),
#'   chronon = month(1L),
#'   cycle = year(1L)
#' )
#' 
#' # Continuous time (discrete = FALSE) for fractional month of year
#' cyclical_time(
#'   Sys.Date(),
#'   chronon = month(1L),
#'   cycle = year(1L),
#'   discrete = FALSE
#' )
#' 
#' # Day of month with Gregorian calendar
#' cyclical_time(
#'   Sys.Date(),
#'   chronon = day(1L),
#'   cycle = month(1L),
#'   calendar = cal_gregorian
#' )
#' 
#' # Hours, minutes, and seconds
#' cyclical_time(
#'   Sys.time(),
#'   chronon = second(1L),
#'   cycle = day(1L)
#' )
#' 
#' @export
cyclical_time <- function(
  data, chronon = time_chronon(data), cycle = time_cycle(data),
  discrete = TRUE, calendar = time_calendar(data)
) {
  # Evaluate chronon and cycle with a calendar mask
  quo_chronon <- enquo(chronon)
  quo_cycle <- enquo(cycle)
  tryCatch({
    chronon <- eval_tidy(quo_chronon, data = calendar, env = emptyenv())
    cycle <- eval_tidy(quo_cycle, data = time_calendar(chronon), env = emptyenv())
  }, error = function(e) {
    # Special hint for common error of 'week' unit not found in Gregorian calendar
    if (e$message == "could not find function \"week\"") {
      e$message <- c(
        e$message,
        "i" = "This error often occurs when trying to use a 'week' chronon without the ISO week calendar.\n",
        ">" = "Try specifying the calendar explicitly, e.g. `calendar = cal_isoweek`."
      )
    }
    cli::cli_abort(e$message, call = NULL)
  })

  # Make numeric data input 1-indexed
  # if (is.numeric(data)) data <- data - 1L
  # TODO - numeric data should default to origin-less cyclical time


  mixtime(data, chronon = chronon, cycle = cycle, discrete = discrete)

  # TODO - For origin-less cyclical time, reduce to cyclical time with divmod methods
  # data <- chronon_divmod(from = chronon, to = cycle, x = data)$mod
}

# #' @importFrom vctrs vec_arith
# #' @method vec_arith mt_cyclical
# #' @export
# vec_arith.mt_cyclical <- function(op, x, y, ...) {
#   UseMethod("vec_arith.mt_cyclical", y)
# }

# #' @importFrom vctrs vec_arith_base
# #' @method vec_arith.mt_cyclical integer
# #' @export
# vec_arith.mt_cyclical.integer <- function(op, x, y, ...) {
#   if (!op %in% c("+", "-")) {
#     stop("Only integer addition and subtraction supported for cyclical time", call. = FALSE)
#   }
#   period <- chronon_cardinality(attr(x, "chronon"), attr(x, "cycle"))
#   vec_restore((vec_arith_base(op, vec_data(x), y, ...) - 1L) %% period + 1L, x)
# }
# #' @importFrom vctrs vec_arith_base
# #' @method vec_arith.mt_cyclical double
# #' @export
# vec_arith.mt_cyclical.double <- vec_arith.mt_cyclical.integer

#' Cyclical time helpers
#' 
#' Helper functions for creating cyclical time representations. These functions
#' create time objects that repeat within a larger time cycle, useful for identifying
#' seasonal patterns or positions within a calendar period.
#' 
#' @param data Another object to be coerced into the specified cyclical time.
#' @param discrete If `TRUE`, the position within the cycle that `data` 
#' falls into is returned as an integer. If `FALSE`, a fractional 
#' position is returned (analagous to time using a continuous time model).
#' @param calendar A calendar object specifying the calendar system to use.
#' @param ... Additional arguments for [cyclical_time()], such as `tz` for timezones.
#' 
#' @section Cyclical time representations:
#' - `day_of_week()`: Represents the day position within a week (1-7) using
#'   the ISO 8601 standard where weeks start on Monday.
#' - `day_of_month()`: Represents the day position within a month (1-28, 1-29,
#'   1-30, or 1-31 depending on the month). The chronon is one day, cycling
#'   within a month.
#' - `day_of_year()`: Represents the day position within a year (1-365 or 1-366
#'   for leap years). The chronon is one day, cycling within a year.
#' - `week_of_year()`: Represents the week position within a year (1-52 or 1-53)
#'   using the ISO 8601 week numbering system.
#' - `month_of_year()`: Represents the month position within a year (1-12).
#'   The chronon is one month, cycling within a year.
#' 
#' @section Custom cyclical time representations:
#' You can create custom cyclical time representations using [cyclical_time()]
#' with any of the supported time units (see [calendar_gregorian] and [calendar_isoweek]).
#' 
#' For example, to create a representation for day of the month:
#' 
#' \preformatted{
#' day_of_month <- new_cyclical_time_fn(
#'   chronon = day(1L), cycle = month(1L),
#'   default_calendar = cal_gregorian
#' )
#' }
#'  
#' @return A `mixtime` time vector containing an `mt_cyclical` vector with chronon and cycle matching the function used.
#' 
#' @seealso [cyclical_time()] for creating cyclical time vectors,
#'   [new_cyclical_time_fn()] for creating cyclical time helper functions
#' 
#' @examples
#' 
#' month_of_year(Sys.Date())
#' day_of_year(Sys.Date())
#' day_of_week(Sys.Date())
#' day_of_week(as.Date("2025-12-15") + 0:6)
#' 
#' @name cyclical_time_helpers
#' @export
month_of_year <- new_cyclical_time_fn(
  chronon = month(1L),
  cycle = year(1L)
)

#' @rdname cyclical_time_helpers
#' @export
day_of_year <- new_cyclical_time_fn(
  chronon = day(1L),
  cycle = year(1L)
)

#' @rdname cyclical_time_helpers
#' @export
day_of_month <- new_cyclical_time_fn(
  chronon = day(1L),
  cycle = month(1L)
)

#' @rdname cyclical_time_helpers
#' @export
time_of_day <- new_cyclical_time_fn(
  chronon = second(1L),
  cycle = day(1L),
  default_calendar = cal_time_civil
)

#' @rdname cyclical_time_helpers
#' @export
day_of_week <- new_cyclical_time_fn(
  chronon = day(1L),
  cycle = week(1L),
  default_calendar = cal_isoweek
)

#' @rdname cyclical_time_helpers
#' @export
week_of_year <- new_cyclical_time_fn(
  chronon = week(1L),
  cycle = year(1L),
  default_calendar = cal_isoweek
)

test_that("Cyclical time handles text parsing with set timezones", {
  # Test that day_of_week preserves the correct day in the specified timezone
  result <- day_of_week(date("2020-01-01", tz = "America/Los_Angeles"))

  # 2020-01-01 is a Wednesday
  expect_equal(format(result), "Wed PST")

  # Test with UTC
  result_utc <- day_of_week(date("2020-01-01"))
  expect_equal(format(result_utc), "Wed")

  # Test with Europe/Berlin (same date, so same day of week)
  result_berlin <- day_of_week(date("2020-01-01", tz = "Europe/Berlin"))
  expect_equal(format(result_berlin), "Wed CET")
})

test_that("cyclical_time() handles timezone conversions showing position differences", {
  # Create a POSIXct representing 2020-01-01 09:00 in Australia/Melbourne
  melbourne_time <- as.POSIXct("2020-01-01 09:00:00", tz = "Australia/Melbourne")
  
  # In Melbourne, this is 2020-01-01 (Wednesday)
  melbourne_dow <- day_of_week(melbourne_time, tz = "Australia/Melbourne")
  expect_equal(format(melbourne_dow), "Wed AEDT")
  
  # Convert to America/Los_Angeles - should be 2019-12-31 (Tuesday)
  pacific_dow <- day_of_week(melbourne_time, tz = "America/Los_Angeles")
  expect_equal(format(pacific_dow), "Tue PST")
  
  # Convert to Europe/Berlin - should be 2019-12-31 (Tuesday)
  berlin_dow <- day_of_week(melbourne_time, tz = "Europe/Berlin")
  expect_equal(format(berlin_dow), "Tue CET")
})

test_that("cyclical_time() month_of_year handles timezone-induced date shifts", {
  # Create a POSIXct representing 2020-12-31 20:00 in America/Los_Angeles
  pacific_time <- as.POSIXct("2020-12-31 20:00:00", tz = "America/Los_Angeles")
  
  # In Pacific time, this is December (month 12)
  pacific_month <- month_of_year(pacific_time, tz = "America/Los_Angeles")
  expect_equal(format(pacific_month), "Dec PST")
  
  # Convert to Australia/Melbourne - should be next day (2021-01-01), January (month 1)
  melbourne_month <- month_of_year(pacific_time, tz = "Australia/Melbourne")
  expect_equal(format(melbourne_month), "Jan AEDT")
  
  # UTC should also be January 1st (month 1)
  utc_month <- month_of_year(pacific_time, tz = "UTC")
  expect_equal(format(utc_month), "Jan")
})

test_that("cyclical_time() day_of_month handles date boundaries across timezones", {
  # Create a POSIXct representing 2020-01-31 23:00 in UTC
  utc_time <- as.POSIXct("2020-01-31 23:00:00", tz = "UTC")
  
  # In UTC, this is the 31st
  utc_dom <- day_of_month(utc_time)
  expect_equal(as.character(format(utc_dom)), "D31")
  
  # In Australia/Sydney (UTC+11), this is February 1st (day 1)
  sydney_dom <- day_of_month(utc_time, tz = "Australia/Sydney")
  expect_equal(as.character(format(sydney_dom)), "D01 AEDT")
})

test_that("cyclical_time() day_of_year handles year boundary transitions", {
  # Create a POSIXct representing 2020-12-31 23:00 in America/Los_Angeles
  pacific_time <- as.POSIXct("2020-12-31 23:00:00", tz = "America/Los_Angeles")
  
  # In Pacific time, this is day 366 of 2020 (leap year)
  pacific_doy <- day_of_year(pacific_time, tz = "America/Los_Angeles")
  expect_equal(as.character(format(pacific_doy)), "D366 PST")
  
  # In UTC, this is 2021-01-01, so day 1 of 2021
  utc_doy <- day_of_year(pacific_time, tz = "UTC")
  expect_equal(as.character(format(utc_doy)), "D01")
})

test_that("cyclical_time() week_of_year handles timezone shifts across week boundaries", {
  # Create a POSIXct representing 2020-01-01 09:00 in America/Los_Angeles (Wednesday)
  pacific_time <- as.POSIXct("2020-01-01 09:00:00", tz = "America/Los_Angeles")
  
  # In Pacific time, 2020-01-01 is in week 1
  pacific_week <- week_of_year(pacific_time, tz = "America/Los_Angeles")
  expect_equal(format(pacific_week), "W01 PST")
  
  # Convert to Australia/Melbourne - becomes 2020-01-02, still in week 1
  melbourne_week <- week_of_year(pacific_time, tz = "Australia/Melbourne")
  expect_equal(format(melbourne_week), "W01 AEDT")
})

test_that("cyclical_time() preserves cycle position for same calendar day across timezones", {
  # Create a POSIXct representing a time on 2020-06-15 (Monday)
  utc_time <- as.POSIXct("2020-06-15 12:00:00", tz = "UTC")
  
  # Convert to different timezones - all should have same day of week
  # since the date 2020-06-15 is the same across timezones (just different times)
  pacific_dow <- day_of_week(utc_time, tz = "America/Los_Angeles")
  berlin_dow <- day_of_week(utc_time, tz = "Europe/Berlin")
  sydney_dow <- day_of_week(utc_time, tz = "Australia/Sydney")
  
  # All should show Monday
  expect_equal(format(pacific_dow), "Mon PDT")
  expect_equal(format(berlin_dow), "Mon CEST")
  expect_equal(format(sydney_dow), "Mon AEST")
})

test_that("cyclical_time() discrete vs continuous modes work with timezones", {
  # Create a POSIXct
  utc_time <- as.POSIXct("2020-06-15 12:30:45", tz = "UTC")
  
  # Discrete mode should return clean labels without percentage
  discrete_dow <- day_of_week(utc_time, discrete = TRUE)
  expect_equal(format(discrete_dow), "Mon")
  
  # Continuous mode should include fractional percentage
  continuous_dow <- day_of_week(utc_time, discrete = FALSE)
  expect_equal(format(continuous_dow), "Mon 52.1%")
})

test_that("cyclical_time() formats cycles of several of the chronon's own unit", {
  # The cycle is a multiple of the chronon's time unit, so the position within
  # the cycle counts the chronon up to the cycle's cardinality.
  expect_equal(
    format(cyclical_time(1:10, month(1L), month(15L))),
    paste0("M", 1:10)
  )

  # Positions wrap at the cardinality of the cycle (15 months)
  expect_equal(
    format(cyclical_time(13:17, month(1L), month(15L))),
    paste0("M", c(13L, 14L, 0L, 1L, 2L))
  )

  # Days within a cycle of 10 days, labelled by the day cyclical labels
  expect_equal(
    format(cyclical_time(0:10, day(1L), day(10L))),
    sprintf("D%02d", c(1:10, 1L))
  )
})

#' ISO 8601 time unit classes
#'
#' Time unit constructors for the ISO 8601 calendar system. These units can be
#' used with [linear_time()] to create custom time representations.
#'
#' @format A civil-based calendar containing ISO 8601 time units.
#' 
#' @details
#' The following time units are available in the ISO week date calendar:
#' 
#' - `year()`: ISO year unit (years start on the week containing the first Thursday)
#' - `week()`: Week unit (7-day periods)
#' - `day()`: Day unit
#' - `hour()`: Hour unit
#' - `minute()`: Minute unit
#' - `second()`: Second unit
#' - `millisecond()`: Millisecond unit
#' 
#' ISO 8601 weeks always start on Monday and the first week of a year is the week
#' containing the first Thursday of that year. This means that some days in early
#' January may belong to the last week of the previous ISO year, and some days in
#' late December may belong to the first week of the next ISO year.
#' 
#' @return An S3 list of class `c("cal_isoweek", "mt_calendar")` containing
#'   the named time unit classes of the ISO 8601 week calendar. Each unit is
#'   accessible via `$` notation and calling it with a step size produces a
#'   time granule (e.g., 1 week granule as `cal_isoweek$week(1L)`).
#'
#' @seealso [linear_time()] for creating custom time representations,
#'   [yearweek()] for a pre-defined ISO 8601 year-week representation
#' 
#' @name calendar_isoweek
#' @export
cal_isoweek <- new_calendar(
  year = S7::new_class("tu_isoyear", parent = mt_tz_unit),
  week = S7::new_class("tu_week", parent = mt_tz_unit),
  inherit = cal_time_civil,
  class = "cal_isoweek"
)

# Time unit labels
method(time_unit_full, cal_isoweek$year) <- function(x) "isoyear{?/s}"
method(time_unit_abbr, cal_isoweek$year) <- function(x) "IY"
method(time_unit_full, cal_isoweek$week) <- function(x) "week{?/s}"
method(time_unit_abbr, cal_isoweek$week) <- function(x) "W"

# Epoch for years
method(chronon_epoch, cal_isoweek$year) <- function(x) 1970L

# Default formats
method(chronon_format_linear, list(cal_isoweek$year, class_any)) <- function(x, cal) "{lin(year)}"
method(chronon_format_linear, list(cal_isoweek$week, class_any)) <- function(x, cal) "{lin(year)} W{cyc(week,year)}"
method(chronon_format_linear, list(cal_isoweek$day, S7::new_S3_class("cal_isoweek"))) <- function(x, cal) "{lin(year)}-W{cyc(week,year)}-{cyc(day,week,label=TRUE)}"
method(chronon_format_cyclical, list(cal_isoweek$day, cal_isoweek$week)) <- function(x, y) "{cyc(day,week,label=TRUE)}"
method(chronon_format_cyclical, list(cal_isoweek$week, cal_isoweek$year)) <- function(x, y) "W{cyc(week,year)}"

# Candidate parsing formats
method(chronon_parse_linear, list(cal_isoweek$week, class_any)) <- function(x, cal) {
  parse_format(
    # e.g. "2024-W01", "2024W01", "2024 w1"
    "{lin(year)}[-\\s]*[Ww]{cyc(week,year)}",
    # e.g. "W01-2024", "w1 2024"
    "[Ww]{cyc(week,year)}[-\\s]+{lin(year)}",
    regex = TRUE
  )
}
method(chronon_parse_linear, list(cal_isoweek$day, S7::new_S3_class("cal_isoweek"))) <- function(x, cal) {
  parse_format(
    # e.g. "2024-W01-3", "2024W01 Wed", "2024 w1 Wednesday"
    "{lin(year)}[-\\s]*[Ww]{cyc(week,year)}[-\\s]+{cyc(day,week)}",
    # e.g. "Wed W01-2024", "3 w1 2024"
    "{cyc(day,week)}[-\\s]+[Ww]{cyc(week,year)}[-\\s]+{lin(year)}",
    regex = TRUE
  )
}
method(chronon_parse_cyclical, list(cal_isoweek$day, cal_isoweek$week)) <- function(x, y) {
  # e.g. "3", "Wed", "Wednesday"
  parse_format("{cyc(day,week)}")
}
method(chronon_parse_cyclical, list(cal_isoweek$week, cal_isoweek$year)) <- function(x, y) {
  parse_format(
    # e.g. "1", "W1", "w01"
    "[Ww]?{cyc(week,year)}",
    regex = TRUE
  )
}

method(chronon_cardinality, list(cal_isoweek$week, cal_isoweek$year)) <- function(x, y, at = NULL) {
  if(y@n != 1L) {
    cli::cli_abort("The number of weeks in multi-year chronons is not yet supported.")
  }
  if(x@n != 1L) {
    cli::cli_abort("The number of multi-weeks in year chronons is not yet supported.")
  }
  year <- at + 1970L
  p <- (year + year %/% 4 - year %/% 100 + year %/% 400) %% 7
  ifelse(p == 4 | (p == 3 & is_leap_year(year)), 53L, 52L)  
}

method(chronon_cardinality_fixed, list(cal_isoweek$day, cal_isoweek$week)) <- function(x, y) {
  7L
}

method(chronon_divmod, list(cal_isoweek$day, cal_isoweek$week)) <- function(from, to, x) {
  # TODO: Add week start specification (e.g., week starts on Monday vs Sunday)
  divisor <- chronon_cardinality(from, to)
  list(
    div = (x + 3L) %/% divisor,
    mod = (x + 3L) %% divisor
  )
}

method(chronon_divmod, list(cal_isoweek$week, cal_isoweek$day)) <- function(from, to, x) {
  # TODO: Add week start specification (e.g., week starts on Monday vs Sunday)
  mult <- chronon_cardinality(to, from)

  list(
    div = x*mult - 3L,
    mod = 0L
  )
}

method(chronon_divmod, list(cal_isoweek$week, cal_isoweek$year)) <- function(from, to, x) {
  # Modulo arithmetic to convert from weeks to ISO years
  if (to@n != 1L) {
    stop("Converting to multi-year chronons from weeks is not yet supported", call. = FALSE)
  }
  if (from@n != 1L) {
    stop("Converting from multi-week chronons to years is not yet supported", call. = FALSE)
  }

  td   <- x * 7L
  y    <- as.integer(td %/% 365.2425 + 1970L)
  j    <- 365L * (y - 1970L) + (y - 1969L) %/% 4L - (y - 1901L) %/% 100L + (y - 1601L) %/% 400L
  leap <- ((y %% 4L == 0L) & (y %% 100L != 0L)) | (y %% 400L == 0L)
  jn   <- j + 365L + leap

  up   <- jn <= td
  dn   <- j > td
  y    <- y + up - dn
  j[up]  <- jn[up]
  j[dn]  <- 365L * (y[dn] - 1970L) + (y[dn] - 1969L) %/% 4L - (y[dn] - 1901L) %/% 100L + (y[dn] - 1601L) %/% 400L

  dow  <- (j + 3L) %% 7L + 1L
  w1   <- j + (4L - dow) %% 7L - 3L

  list(
    # year offset from epoch
    div = y - 1970L,  
    # 0-indexed week within ISO year [0, 51] or [0, 52]
    mod = (td - 3L - w1) %/% 7L
  )
}

method(chronon_divmod, list(cal_isoweek$year, cal_isoweek$week)) <- function(from, to, x) {
  if (to@n != 1L) {
    stop("Converting to multi-year chronons from weeks is not yet supported", call. = FALSE)
  }
  if (from@n != 1L) {
    stop("Converting from multi-week chronons to years is not yet supported", call. = FALSE)
  }
  
  # Split integer year offset and fractional (0-indexed week within year)
  y_off <- floor(x)
  frac  <- x - y_off
  y     <- as.integer(y_off) + 1970L

  j    <- 365L * (y - 1970L) + (y - 1969L) %/% 4L - (y - 1901L) %/% 100L + (y - 1601L) %/% 400L
  dow  <- (j + 3L) %% 7L + 1L
  w1   <- j + (4L - dow) %% 7L - 3L  # days since epoch of Monday of ISO week 1

  # weeks since epoch of week 1, plus fractional weeks within the year
  w1_week <- (w1 + 3L) %/% 7L        # since w1 = w*7 - 3, so w = (w1 + 3) / 7

  # frac is 0-indexed weeks within year (the mod from the forward direction)
  # total weeks count within year = 52 or 53
  leap  <- ((y %% 4L == 0L) & (y %% 100L != 0L)) | (y %% 400L == 0L)
  jn    <- j + 365L + leap
  dowN  <- (jn + 3L) %% 7L + 1L
  w1n   <- jn + (4L - dowN) %% 7L - 3L
  n_weeks <- (w1n - w1) %/% 7L       # 52 or 53

  list(
    div = w1_week + as.integer(floor(frac * n_weeks)),
    mod = 0L
  )
}

week.abb <- c("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun")
week.name <- c("Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday")

# TODO: Add offset for different week starting days
method(cyclical_labels, list(cal_isoweek$day, cal_isoweek$week)) <- label_scheme(
  start = 1L,
  vocab = vocab_table(`en-GB` = list(wide = week.name, abbreviated = week.abb))
)
# Weeks count with 1-indexing
method(cyclical_labels, list(cal_isoweek$week, S7::class_any)) <- label_scheme(start = 1L, width = 2L)

#' Lunar time unit classes
#'
#' Time unit constructors for the synodic lunar time system, where the boundary
#' of each month is at the new moon (lunar conjunction) and the boundary of
#' each phase is at a lunar octant (each eighth of the synodic cycle). Both
#' boundaries are location-independent, as the new moon is a geocentric
#' astronomical event. `cal_time_lunar` is an alias for
#' `cal_time_lunar_synodic`.
#'
#' @format A calendar containing synodic lunar time units.
#' 
#' @details
#' The following time units are available in the lunar calendar systems.
#' 
#' - `month()`: Synodic month unit
#' - `phase()`: Synodic phase unit
#' 
#' @return An S3 list of class `c("cal_time_lunar", "mt_calendar")` containing
#'   the named time unit classes of the synodic lunar calendar. Each unit is
#'   accessible via `$` notation and calling it with a step size produces a
#'   time granule (e.g., 1 synodic month granule as
#'   `cal_time_lunar$month(1L)`). Lunar month and phase boundaries are
#'   location-independent.
#'
#' @seealso [`cal_time_civil`], [`cal_time_solar`]
#' 
#' @examples
#' # Find the time of a new moon in the Gregorian calendar
#' t <- linear_time(Sys.Date(), cal_time_lunar$month(1L))
#' datetime(t, tz = "Australia/Melbourne")
#' 
#' 
#' @name calendar_time_lunar
#' @export
cal_time_lunar_synodic <- new_calendar(
  month = S7::new_class("tu_lunar_month", parent = mt_loc_unit),
  phase = S7::new_class("tu_lunar_phase", parent = mt_loc_unit)
)

#' @export
#' @rdname calendar_time_lunar
cal_time_lunar <- cal_time_lunar_synodic

# Time granule labels
method(time_unit_full, cal_time_lunar$month) <- function(x) "synodic month{?/s}"
method(time_unit_abbr, cal_time_lunar$month) <- function(x) "LM"
method(time_unit_full, cal_time_lunar$phase) <- function(x) "synodic phase{?/s}"
method(time_unit_abbr, cal_time_lunar$phase) <- function(x) "LP"

# Default formats
method(chronon_format_linear, list(cal_time_lunar$month, class_any)) <- function(x, cal) "L{lin(month)}"
method(chronon_format_linear, list(cal_time_lunar$phase, class_any)) <- function(x, cal) "L{lin(month)} {cyc(phase,month,label=TRUE)}"
method(chronon_format_cyclical, list(cal_time_lunar$phase, cal_time_lunar$month)) <- function(x, y) "{cyc(phase,month,label=TRUE)}"

# Epoch for lunations
method(chronon_epoch, cal_time_lunar$month) <- function(x) 953L

# The number of UTC seconds in a lunar (synodic) phase
method(chronon_cardinality, list(cal_time_civil$second, cal_time_lunar$phase)) <- function(x, y, at = NULL) {
  at <- approx_lunations_from_utc(at / 8)
  (approx_utc_from_lunations(at + y@n) - approx_utc_from_lunations(at)) / x@n
}

# The number of lunar phases in a lunar (synodic) month
method(chronon_cardinality_fixed, list(cal_time_lunar$phase, cal_time_lunar$month)) <- function(x, y) {
  8L
}

# The number of UTC seconds in a lunar (synodic) phase
method(chronon_divmod, list(cal_time_civil$second, cal_time_lunar$phase)) <- function(from, to, x) {
  list(
    div = approx_lunations_from_utc(as.double(x)) * 8,
    mod = 0
  )
}

method(chronon_divmod, list(cal_time_lunar$phase, cal_time_civil$second)) <- function(from, to, x) {
  list(
    div = approx_utc_from_lunations(as.double(x) / 8),
    mod = 0
  )
}

lunar_phase_abbr <- c("NM", "WxC", "FQ", "WxG", "FM", "WnG", "LQ", "WnC")
lunar_phase_full <- c("new moon", "waxing crescent", "first quarter", "waxing gibbous", "full moon", "waning gibbous", "last quarter", "waning crescent")
lunar_phase_emoji <- c("\U0001F311", "\U0001F312", "\U0001F313", "\U0001F314", "\U0001F315", "\U0001F316", "\U0001F317", "\U0001F318")

method(cyclical_labels, list(cal_time_lunar$phase, cal_time_lunar$month)) <- label_scheme(
  start = 1L,
  vocab = vocab_table(`en-GB` = list(
    wide = lunar_phase_full,
    abbreviated = lunar_phase_abbr,
    emoji = lunar_phase_emoji
  ))
)

naive_tz <- naive(NA_character_)

#' @param tz The timezone name for the unit (valid units can be found with `[tzdb::tzdb_names()]`)
#'
#' @rdname mt_unit
#' @export
mt_tz_unit <- new_class(
  "mt_tz_unit",
  parent = mt_unit,
  properties = list(tz = new_property(S7::class_character, default = naive_tz)),
  constructor = function(n = 1L, tz = naive_tz) {
    # Allow for logical NA for convenience
    if (is.na(tz) && !identical(tz, naive_tz)) {
      tz <- NA_character_
    }
    S7::new_object(mt_unit(n = n), tz = tz)
  },
  validator = function(self) {
    if (!is.na(self@tz)) {
      check_tz_name(self@tz)
    }
    NULL
  }
)

# Default formats
method(chronon_format_attr, mt_tz_unit) <- function(x) {
  if (!is.na(x@tz) && x@tz != "UTC") " {tz(.time)}" else ""
}

#' Civil time unit classes
#'
#' Time unit constructors for the civil time system where the boundary of each
#' day is at midnight on the 24 hour clock. This calendar is intended to be
#' built on by other calendars (e.g. `[cal_time_civil]` and `[cal_isoweek]`) to
#' add common time components.  These units can be
#' used with [linear_time()] to create custom time representations.
#'
#' @return A time granule object for the civil time system.
#'
#' @details
#' The following time units are available (`cal_time_civil$`).
#'
#' - `day()`: Day unit
#' - `hour()`: Hour unit
#' - `minute()`: Minute unit
#' - `second()`: Second unit
#' - `millisecond()`: Millisecond unit
#' - `microsecond()`: Microsecond unit
#' - `nanosecond()`: Nanosecond unit
#'
#' @seealso [`cal_time_civil`], [`cal_isoweek`]
#'
#' @examples
#' # Create a custom time representation using civil time granules
#' hms <- new_cyclical_time_fn(
#'   chronon = second(1L),
#'   cycle = hour(1L)
#' )
#'
#' @name calendar_time_civil
#' @export
cal_time_civil <- new_calendar(
  day = new_class("tu_day", parent = mt_tz_unit),
  ampm = new_class("tu_ampm", parent = mt_tz_unit),
  hour = new_class("tu_hour", parent = mt_tz_unit),
  minute = new_class("tu_minute", parent = mt_tz_unit),
  second = new_class("tu_second", parent = mt_tz_unit),
  millisecond = new_class("tu_millisecond", parent = mt_tz_unit),
  microsecond = new_class("tu_microsecond", parent = mt_tz_unit),
  nanosecond = new_class("tu_nanosecond", parent = mt_tz_unit),
  class = "cal_time_civil"
)

# Time unit labels
method(time_unit_full, cal_time_civil$day) <- function(x) "day{?/s}"
method(time_unit_abbr, cal_time_civil$day) <- function(x) "D"
method(time_unit_full, cal_time_civil$ampm) <- function(x) "halfday{?/s}"
method(time_unit_abbr, cal_time_civil$ampm) <- function(x) "hd"
method(time_unit_full, cal_time_civil$hour) <- function(x) "hour{?/s}"
method(time_unit_abbr, cal_time_civil$hour) <- function(x) "h"
method(time_unit_full, cal_time_civil$minute) <- function(x) "minute{?/s}"
method(time_unit_abbr, cal_time_civil$minute) <- function(x) "m"
method(time_unit_full, cal_time_civil$second) <- function(x) "second{?/s}"
method(time_unit_abbr, cal_time_civil$second) <- function(x) "s"
method(time_unit_full, cal_time_civil$millisecond) <- function(x) {
  "millisecond{?/s}"
}
method(time_unit_abbr, cal_time_civil$millisecond) <- function(x) "ms"
method(time_unit_full, cal_time_civil$microsecond) <- function(x) {
  "microsecond{?/s}"
}
method(time_unit_abbr, cal_time_civil$microsecond) <- function(x) "us"
method(time_unit_full, cal_time_civil$nanosecond) <- function(x) {
  "nanosecond{?/s}"
}
method(time_unit_abbr, cal_time_civil$nanosecond) <- function(x) "ns"

# Default formats
method(chronon_format_linear, list(cal_time_civil$day, class_any)) <- function(
  x,
  cal
) {
  # `year`/`month` are only in `cal` when the chronon's timezone is
  # resolvable (a Gregorian date is only defined relative to a timezone).
  # Without them, defer to the generic mt_unit fallback rather than
  # referencing calendar units that aren't actually in scope.
  if (is.null(cal$year) || is.null(cal$month)) {
    chronon_format_linear(S7::super(x, mt_unit), cal)
  } else {
    "{lin(year)}-{cyc(month,year)}-{cyc(day, month)}"
  }
}
method(chronon_format_linear, list(cal_time_civil$hour, class_any)) <- function(
  x,
  cal
) {
  paste(chronon_format_linear(cal$day(1L), cal), "{cyc(hour, day)}h")
}
method(
  chronon_format_linear,
  list(cal_time_civil$minute, class_any)
) <- function(x, cal) {
  paste(
    chronon_format_linear(cal$day(1L), cal),
    "{cyc(hour, day)}:{cyc(minute, hour)}"
  )
}
method(
  chronon_format_linear,
  list(cal_time_civil$second, class_any)
) <- function(x, cal) {
  paste(
    chronon_format_linear(cal$day(1L), cal),
    "{cyc(hour, day)}:{cyc(minute, hour)}:{cyc(second, minute)}"
  )
}
method(
  chronon_format_linear,
  list(cal_time_civil$millisecond, class_any)
) <- function(x, cal) {
  paste(
    chronon_format_linear(cal$day(1L), cal),
    "{cyc(hour, day)}:{cyc(minute, hour)}:{cyc(second, minute)}.{cyc(millisecond, second)}"
  )
}
method(
  chronon_format_cyclical,
  list(cal_time_civil$millisecond, cal_time_civil$day)
) <- function(x, y) {
  "{cyc(hour, day)}:{cyc(minute, hour)}:{cyc(second, minute)}.{cyc(millisecond, second)}"
}
method(
  chronon_format_cyclical,
  list(cal_time_civil$second, cal_time_civil$day)
) <- function(x, y) "{cyc(hour, day)}:{cyc(minute, hour)}:{cyc(second, minute)}"
method(
  chronon_format_cyclical,
  list(cal_time_civil$minute, cal_time_civil$day)
) <- function(x, y) "{cyc(hour, day)}:{cyc(minute, hour)}"

# Candidate parsing formats
method(chronon_parse_linear, list(cal_time_civil$hour, class_any)) <- function(
  x,
  cal
) {
  parse_format_time_of_day(
    cal,
    c(
      " {cyc(hour, day)}h",
      "T{cyc(hour, day)}h",
      " {cyc(hour, ampm)} {cyc(ampm, day, label = TRUE)}",
      "T{cyc(hour, ampm)} {cyc(ampm, day, label = TRUE)}"
    ),
    c(
      "[ T]{cyc(hour, day)}h",
      "[ T]{cyc(hour, ampm)} {cyc(ampm, day, label = TRUE)}"
    )
  )
}
method(
  chronon_parse_linear,
  list(cal_time_civil$minute, class_any)
) <- function(x, cal) {
  parse_format_time_of_day(
    cal,
    c(
      " {cyc(hour, day)}:{cyc(minute, hour)}",
      "T{cyc(hour, day)}:{cyc(minute, hour)}",
      " {cyc(hour, ampm)}:{cyc(minute, hour)} {cyc(ampm, day, label = TRUE)}",
      "T{cyc(hour, ampm)}:{cyc(minute, hour)} {cyc(ampm, day, label = TRUE)}"
    ),
    c(
      "[ T]{cyc(hour, day)}:{cyc(minute, hour)}",
      "[ T]{cyc(hour, ampm)}:{cyc(minute, hour)} {cyc(ampm, day, label = TRUE)}"
    )
  )
}
method(
  chronon_parse_linear,
  list(cal_time_civil$second, class_any)
) <- function(x, cal) {
  parse_format_time_of_day(
    cal,
    c(
      " {cyc(hour, day)}:{cyc(minute, hour)}:{cyc(second, minute)}",
      "T{cyc(hour, day)}:{cyc(minute, hour)}:{cyc(second, minute)}",
      " {cyc(hour, ampm)}:{cyc(minute, hour)}:{cyc(second, minute)} {cyc(ampm, day, label = TRUE)}",
      "T{cyc(hour, ampm)}:{cyc(minute, hour)}:{cyc(second, minute)} {cyc(ampm, day, label = TRUE)}"
    ),
    c(
      "[ T]{cyc(hour, day)}:{cyc(minute, hour)}:{cyc(second, minute)}",
      "[ T]{cyc(hour, ampm)}:{cyc(minute, hour)}:{cyc(second, minute)} {cyc(ampm, day, label = TRUE)}"
    )
  )
}
method(
  chronon_parse_linear,
  list(cal_time_civil$millisecond, class_any)
) <- function(x, cal) {
  parse_format_time_of_day(
    cal,
    c(
      " {cyc(hour, day)}:{cyc(minute, hour)}:{cyc(second, minute)}.{cyc(millisecond, second)}",
      "T{cyc(hour, day)}:{cyc(minute, hour)}:{cyc(second, minute)}.{cyc(millisecond, second)}",
      " {cyc(hour, ampm)}:{cyc(minute, hour)}:{cyc(second, minute)}.{cyc(millisecond, second)} {cyc(ampm, day, label = TRUE)}",
      "T{cyc(hour, ampm)}:{cyc(minute, hour)}:{cyc(second, minute)}.{cyc(millisecond, second)} {cyc(ampm, day, label = TRUE)}"
    ),
    c(
      "[ T]{cyc(hour, day)}:{cyc(minute, hour)}:{cyc(second, minute)}\\.{cyc(millisecond, second)}",
      "[ T]{cyc(hour, ampm)}:{cyc(minute, hour)}:{cyc(second, minute)}\\.{cyc(millisecond, second)} {cyc(ampm, day, label = TRUE)}"
    )
  )
}
method(
  chronon_parse_cyclical,
  list(cal_time_civil$hour, cal_time_civil$day)
) <- function(x, y) {
  parse_format(
    "{cyc(hour, day)}h",
    "{cyc(hour, ampm)} {cyc(ampm, day, label = TRUE)}"
  )
}
method(
  chronon_parse_cyclical,
  list(cal_time_civil$minute, cal_time_civil$day)
) <- function(x, y) {
  parse_format(
    "{cyc(hour, day)}:{cyc(minute, hour)}",
    "{cyc(hour, ampm)}:{cyc(minute, hour)} {cyc(ampm, day, label = TRUE)}"
  )
}
method(
  chronon_parse_cyclical,
  list(cal_time_civil$second, cal_time_civil$day)
) <- function(x, y) {
  parse_format(
    "{cyc(hour, day)}:{cyc(minute, hour)}:{cyc(second, minute)}",
    "{cyc(hour, ampm)}:{cyc(minute, hour)}:{cyc(second, minute)} {cyc(ampm, day, label = TRUE)}"
  )
}
method(
  chronon_parse_cyclical,
  list(cal_time_civil$millisecond, cal_time_civil$day)
) <- function(x, y) {
  parse_format(
    "{cyc(hour, day)}:{cyc(minute, hour)}:{cyc(second, minute)}.{cyc(millisecond, second)}",
    "{cyc(hour, ampm)}:{cyc(minute, hour)}:{cyc(second, minute)}.{cyc(millisecond, second)} {cyc(ampm, day, label = TRUE)}"
  )
}

## HOURs in DAYs
method(
  chronon_cardinality_fixed,
  list(cal_time_civil$hour, cal_time_civil$day)
) <- function(x, y) {
  24L
}

## AMPMs in DAYs
method(
  chronon_cardinality_fixed,
  list(cal_time_civil$ampm, cal_time_civil$day)
) <- function(x, y) {
  2L
}

## HOURs in AMPMs
method(
  chronon_cardinality_fixed,
  list(cal_time_civil$hour, cal_time_civil$ampm)
) <- function(x, y) {
  12L
}

## MINUTEs in HOURs
method(
  chronon_cardinality_fixed,
  list(cal_time_civil$minute, cal_time_civil$hour)
) <- function(x, y) {
  60L
}

## SECONDs in MINUTEs
method(
  chronon_cardinality_fixed,
  list(cal_time_civil$second, cal_time_civil$minute)
) <- function(x, y) {
  60L
}

## MILLISECONDs in SECONDs
method(
  chronon_cardinality_fixed,
  list(cal_time_civil$millisecond, cal_time_civil$second)
) <- function(x, y) {
  1000L
}

## MICROSECONDs in MILLISECONDs
method(
  chronon_cardinality_fixed,
  list(cal_time_civil$microsecond, cal_time_civil$millisecond)
) <- function(x, y) {
  1000L
}

## NANOSECONDs in MICROSECONDs
method(
  chronon_cardinality_fixed,
  list(cal_time_civil$nanosecond, cal_time_civil$microsecond)
) <- function(x, y) {
  1000L
}


# Cyclical labels
method(
  cyclical_labels,
  list(cal_time_civil$day, S7::class_any)
) <- label_scheme(start = 1L, width = 2L)

method(
  cyclical_labels,
  list(cal_time_civil$ampm, S7::class_any)
) <- label_scheme(
  start = 0L,
  vocab = vocab_table(
    `en-GB` = list(wide = c("AM", "PM"), abbreviated = c("AM", "PM"))
  ),
  transform = list(
    encode = function(i, at = NULL) i + 1L,
    decode = function(d, at = NULL) d - 1L
  )
)

# Hours need custom methods for the 12-hour clock's "12,1,2,...,11" wraparound
method(
  cyclical_labels_format,
  list(cal_time_civil$hour, S7::class_any)
) <- function(granule, cycle, i, at = NULL, ...) {
  if (S7_inherits(cycle, cal_time_civil$ampm)) {
    # 12 hours count with 12,1,2,...,11
    sprintf("%02d", (i - 1L) %% 12L + 1L)
  } else {
    # 24 hours count with 0-indexing
    sprintf("%02d", i)
  }
}
method(
  cyclical_labels_parse,
  list(cal_time_civil$hour, S7::class_any)
) <- function(granule, cycle, ...) {
  list(
    pattern = "\\d+",
    decode = function(text, at = NULL) {
      d <- as.integer(text)
      if (S7_inherits(cycle, cal_time_civil$ampm)) d %% 12L else d
    }
  )
}

# Minutes/seconds/milliseconds count with 0-indexing -- plain numeric (tier 1).
method(
  cyclical_labels,
  list(cal_time_civil$minute, S7::class_any)
) <- label_scheme(start = 0L, width = 2L)
method(
  cyclical_labels,
  list(cal_time_civil$second, S7::class_any)
) <- label_scheme(start = 0L, width = 2L)
method(
  cyclical_labels,
  list(cal_time_civil$millisecond, S7::class_any)
) <- label_scheme(start = 0L, width = 3L)

# called internally by mt_glue_fmt
mt_unit_display <- function(x, units, parts, ...) {
  if (is_mt <- inherits(x, "S7_class")) {
    # Match based on class only, e.g. {year}-{month}-{day}
    xi <- which(vapply(units, S7_inherits, logical(1L), x))
    if (length(xi) == 0L) {
      # This doesn't match one of the expected units, maybe it's not a mixtime S7 class?
      is_mt <- FALSE
    } else if (length(xi) > 1L) {
      # This matches multiple expected units, more precision is needed.
      cal <- time_calendar(units[[length(units)]])
      bad_tu <- names(cal)[match(S7_class_id(x), vapply(cal, S7_class_id, character(1L)))]
      cli::cli_abort(
        c(
          "Multiple units match {.code {paste0('{', bad_tu, '}')}} in the mixtime format string.",
          i =  "Specify chronon size precisely using e.g. {.code {paste0('{', bad_tu, '(1L)}')}}"
        ),
        call = NULL
      )
    }
  } else if (is_mt <- S7_inherits(x, mt_unit)) {
    xi <- vec_match(
      data.frame(x = x@n, tu = S7_class_id(x)),
      data.frame(x = vapply(units, function(u) u@n, numeric(1L)), tu = vapply(units, S7_class_id, character(1L)))
    )
    # No matches found
    if (is.na(xi)) {
      cli::cli_abort(
        c(
          "The calendar time unit of {time_granule_label(x)} could not be found."
        ),
        call = NULL
      )
    }
  }
  if (is_mt) {
    # Return cyclical labels for cyclical parts
    if (xi > 1) return(cyclical_labels_format(units[[xi]], units[[xi-1L]], parts[[xi]], at = NULL, ...))
    # Return linear labels for coarsest part
    return(as.character(parts[[xi]]))
  } else {
    as.character(x)
  }
}

# Decodes one mt_glue_fmt() `{lin(...)}`/`{cyc(...)}` token with attributes
format_token_spec <- function(tok) {
  list(
    chronon = tok[[1L]],
    cycle = if (length(tok) == 2L) tok[[2L]] else NULL,
    attrs = attributes(tok) %||% list()
  )
}

time_format_default <- function(x, attr = TRUE) {
  chronon <- base::attr(x, "chronon")

  if (S7_inherits(x, mt_duration)) {
    return(chronon_format_duration(chronon))
  }

  cycle <- base::attr(x, "cycle")
  fmt <- if (is.null(cycle)) {
    chronon_format_linear(chronon)
  } else {
    chronon_format_cyclical(chronon, cycle)
  }

  # Add fractional
  if (is.double(x)) fmt <- paste(fmt, "{frac(.time)}")

  # Add format attributes (e.g. tz or location)
  if (attr) fmt <- paste0(fmt, chronon_format_attr(chronon))

  fmt
}

time_format_impl <- function(x, format = time_format_default(x, attr = attr), ..., attr = TRUE) {
  # Obtain core time information
  chronon <- attr(x, "chronon")

  # Evaluation calendar: the chronon's own, extended with the cycle's units so
  # cyclical granules (e.g. {cyc(month, year)}) can be named.
  cycle <- attr(x, "cycle")
  calendar <- if (is.null(cycle)) time_calendar(chronon) else time_calendar(cycle)

  x_special <- is.na(x) | is.infinite(vec_data(x))

  # Create the glue evaluation environment by layering format-only helpers on top
  # of the shared lin()/cyc() component mask: the attribute helpers (tz / loc /
  # frac) and .time are added to the same calendar + lin/cyc vocabulary that
  # time_components() evaluates against.
  env <- rlang::new_environment(
    data = c(
      component_mask(calendar, chronon),
      list(
        # Attribute helper functions
        tz = tz_abbreviation,
        loc = function(x) {
          chronon <- attr(x, "chronon")
          if (!S7_inherits(mt_loc_unit)) return("")
          lat <- chronon@lat
          lon <- chronon@lon
          alt <- chronon@alt
          lat_str <- sprintf("%.2f%s", abs(lat), if (lat >= 0) "N" else "S")
          lon_str <- sprintf("%.2f%s", abs(lon), if (lon >= 0) "E" else "W")
          if (alt != 0) {
            paste0(lat_str, " ", lon_str, " ", sprintf("%.0fm", alt))
          } else {
            paste0(lat_str, " ", lon_str)
          }
        },
        frac = function(x) {
          # Apply time zone offset to x, with truncation for discrete time models.
          chronon <- attr(x, "chronon")
          x <- vec_data(x)
          x_tz <- tz_offset_impl(x, chronon)
          if(is.integer(x)) x_tz <- trunc(x_tz)
          x <- x + x_tz
          sprintf("%.1f%%", (x - floor(x))*100)
        },

        # Attach .time for specialised usage (e.g. tz_abbreviation(.time))
        .time = x
      )
    ),
    parent = rlang::caller_env()
  )

  fmt <- mt_glue_fmt(format, env = env)
  fmt_parts <- vapply(fmt, is.list, logical(1L))

  # Resolve bare S7 class tokens (e.g. from {year}, {month}, {day}) into
  # lin/cyc lists using the chronon_cardinality graph to order fine->coarse.
  s7_class_lgl <- vapply(fmt, inherits, logical(1L), "S7_class")
  if (any(s7_class_lgl)) {
    bare_units <- vapply(fmt[s7_class_lgl], S7_class_id, character(1L))
    cal <- time_calendar(x)

    tu_i <- match(
      bare_units,
      vapply(cal, S7_class_id, character(1L))
    )

    cli::cli_abort(
      c(
        "Bare time granule tokens are not currently supported in format strings.",
        i = "Use {.fn lin} for linear (coarsest) units, e.g. {.code {paste0('{lin(', names(cal)[tu_i[1]], ')}')}}}",
        i = "Use {.fn cyc} for cyclical units, e.g. {.code {paste0('{cyc(', names(cal)[tu_i[1]], ', <coarser_unit>)}')}}"
      ),
      call = NULL
    )

    # class_idx <- which(s7_class_lgl)
    # class_num <- length(class_idx)

    # # Initialise as time units of size 1L
    # fmt[class_idx] <- lapply(fmt[class_idx], function(f) f(1L))

    # # Find order of granularity from fine -> coarse
    # ordered_idx <- class_idx[S7_order_granules(fmt[class_idx])]

    # # Replace string parts such that:
    # # * finer time units are cyclical with the next coarser
    # for (i in seq_len(class_num - 1L)) {
    #   fmt[[ordered_idx[i]]] <- list(fmt[[ordered_idx[i]]], fmt[[ordered_idx[i+1L]]])
    # }
    # # * coarsest time unit is linear
    # fmt[[ordered_idx[class_num]]] <- list(fmt[[ordered_idx[class_num]]])
  }

  # Compute the numeric parts for display
  res_split <- split(fmt[fmt_parts], lengths(fmt[fmt_parts]))
  parts <- chronon_parts(
    x        = x[!x_special],
    linear   = unlist(res_split[["1"]], recursive = FALSE),
    cyclical = res_split[["2"]]
  )

  # Apply labels
  parts$linear <- .mapply(
    function(tu, x) {
      spec <- format_token_spec(tu)
      rlang::exec(linear_labels_format, spec$chronon, x, !!!spec$attrs)
    },
    dots = list(res_split[["1"]], parts$linear),
    MoreArgs = NULL
  )
  parts$cyclical <- .mapply(
    # TODO floor(x) shouldn't be necessary, fix chronon_parts()?
    function(tu, x, at) {
      spec <- format_token_spec(tu)
      rlang::exec(cyclical_labels_format, spec$chronon, spec$cycle, floor(x), at = at, !!!spec$attrs)
    },
    dots = list(res_split[["2"]], parts$cyclical, parts$cyclical_at),
    MoreArgs = NULL
  )

  # Insert time labels into format string parts
  if (any(fmt_parts)) {
    fmt[fmt_parts] <- unsplit(Filter(length, parts), lengths(fmt[fmt_parts]))
  }

  # Handle format parts which include special values (e.g. timezones)
  if (any(x_special) && any(fmt_inc_special <- lengths(fmt) == length(x))) {
    fmt[fmt_inc_special] <- lapply(fmt[fmt_inc_special], function(x) x[!x_special])
  }

  out <- character(length(x))
  out[!x_special] <- trimws(rlang::exec(paste0, !!!fmt))
  out[x_special] <- format(vec_data(x)[x_special])
  out
}

#' Format mixtime vectors
#'
#' Formats a mixtime vector as a character vector, using a glue-style format
#' string of `{lin(...)}`/`{cyc(...)}` tokens tailored to the vector's chronon
#' and cycle. If `format` is omitted, a sensible default is derived
#' automatically. See `vignette("time-format-strings")` for the format string
#' syntax.
#'
#' @param x A mixtime vector.
#' @param ... Additional arguments for methods, including `format`: a
#'   glue-style format string, defaulting to one derived automatically for `x`.
#' @param attr If `TRUE` (default), append attribute information (e.g.
#'   timezone) to the default format.
#'
#' @return A character vector the same length as `x`.
#'
#' @usage
#' \S4method{format}{mt_time}(x, ..., attr = TRUE)
#' @aliases format,mt_time-method
#' @name format.mt_time
#' @export
S7::method(format, mt_time) <- function(x, ..., attr = TRUE) {
  time_format_impl(x, ..., attr = attr)
}

#' @export
S7::method(print, mt_time) <- function(x, ...) {
  cat(paste0("<", class(x)[[1L]], "[", length(x), "]>"), "\n", sep = "")
  print(format(x), quote = FALSE)
  invisible(x)
}

#' @export
S7::method(str, mt_linear) <- function(object, ..., vec.len = getOption("str")$vec.len %||% 4L) {
  cat("linear<", format(time_chronon(object)[1L]), "> [1:", length(object), "] ", .mt_str_values(object, vec.len), "\n", sep = "")
}

#' @export
S7::method(str, mt_cyclical) <- function(object, ..., vec.len = getOption("str")$vec.len %||% 4L) {
  cat("cyclical<", format(time_chronon(object)[1L]), "/", format(time_cycle(object)[1L]), "> [1:", length(object), "] ", .mt_str_values(object, vec.len), "\n", sep = "")
}

#' @export
S7::method(str, mt_duration) <- function(object, ..., vec.len = getOption("str")$vec.len %||% 4L) {
  cat("duration<", format(time_chronon(object)[1L]), "> [1:", length(object), "] ", .mt_str_values(object, vec.len), "\n", sep = "")
}

.mt_str_values <- function(x, vec_len) {
  vals <- format(x)
  if (length(vals) > vec_len) {
    paste0(paste0(vals[seq_len(vec_len)], collapse = " "), " ...")
  } else {
    paste0(vals, collapse = " ")
  }
}

test_that("format() handles NA in linear time", {
  expect_equal(format(yearmonth(NA_real_)), "NA")
  expect_equal(format(date(NA_character_)), "NA")
})

test_that("format() handles NaN in linear time", {
  expect_equal(format(yearmonth(NaN)), "NaN")
})

test_that("format() handles Inf and -Inf in linear time", {
  expect_equal(format(yearmonth(Inf)),  "Inf")
  expect_equal(format(yearmonth(-Inf)), "-Inf")
})

test_that("format() handles mixed regular, NA, NaN and Inf values", {
  x <- yearmonth(c(0, NA, NaN, Inf, -Inf))
  result <- format(x)
  expect_equal(result[[1]], "1970 Jan 0.0%")
  expect_equal(result[[2]], "  NA")
  expect_equal(result[[3]], " NaN")
  expect_equal(result[[4]], " Inf")
  expect_equal(result[[5]], "-Inf")
})

test_that("format() handles NA in durations", {
  expect_equal(format(months(NA_integer_)), "NA")
  expect_equal(format(days(NA_real_)), "NA")
})

test_that("format() handles NaN in durations", {
  expect_equal(format(days(NaN)), "NaN")
})

test_that("format() handles Inf and -Inf in durations", {
  expect_equal(format(months(Inf)),  "Inf")
  expect_equal(format(months(-Inf)), "-Inf")
})

test_that("format() handles mixed special values in durations", {
  x <- months(c(1L, NA_integer_, 2L))
  result <- format(x)
  expect_equal(result[[1]], "1 month")
  expect_equal(result[[2]], "NA")
  expect_equal(result[[3]], "2 months")

  x_frac <- days(c(1.5, NaN, Inf, -Inf))
  result_frac <- format(x_frac)
  expect_equal(result_frac[[2]], " NaN")
  expect_equal(result_frac[[3]], " Inf")
  expect_equal(result_frac[[4]], "-Inf")
})

test_that("format() handles special values in cyclical time", {
  x <- yearmonth(c(0, NA, Inf))
  result <- format(x)
  expect_length(result, 3L)
  expect_equal(result[[2]], " NA")
  expect_equal(result[[3]], "Inf")
})

test_that("format() reports an unmatched granule in the format string", {
  # Granules are S7 scalars, so their size comes from @n rather than vec_data()
  expect_error(
    mt_unit_display(
      cal_gregorian$day(2L),
      units = list(cal_gregorian$day(1L)),
      parts = list(1L)
    ),
    "2 days"
  )
})

test_that("format() falls back to the generic mt_unit template for a naive-tz civil chronon", {
  # A common chronon between an untimezoned Date and a zoned datetime is
  # genuinely tz-naive; its calendar has no year/month, so the default
  # templates must not reference them (see tz-merge.md). Instead they defer
  # (via S7::super()) to the same generic mt_unit fallback used for any other
  # chronon without a bespoke format template.
  combined <- c(date(Sys.Date()), datetime(as.POSIXct("2015-01-01", tz = "UTC")))
  ch <- chronon_common(combined)
  expect_true(is.na(tz_name(ch)))
  expect_null(time_calendar(ch)$year)

  naive_second <- vecvec::unvecvec(mixtime(as.numeric(combined), chronon = ch, discrete = FALSE))
  expect_no_error(out <- format(naive_second))
  expect_match(out, "^D[0-9]")
})

# A naive time granule property inherits from specified properties
naive <- function(x = NA) structure(x, class = "mt_naive")

#' @exportS3Method base::format mt_naive
format.mt_naive <- function(x, ...) {
  paste0("naive [", NextMethod(), "]")
}

is.naive <- function(x) {
  vapply(x, inherits, logical(1L), "mt_naive")
}

# ---------------------------------------------------------------------------
# loc_latitude / loc_longitude / loc_altitude
# ---------------------------------------------------------------------------

lat <- -37.8136
lon <- 144.9631
alt <- 100

# ---------------------------------------------------------------------------
# mt_loc_unit directly
# ---------------------------------------------------------------------------
test_that("loc_* accessors work on mt_loc_unit", {
  u <- mt_loc_unit(1L, lat = lat, lon = lon, alt = alt)
  expect_equal(loc_latitude(u),  lat)
  expect_equal(loc_longitude(u), lon)
  expect_equal(loc_altitude(u),  alt)
})

test_that("loc_altitude defaults to 0 when not supplied", {
  u <- mt_loc_unit(1L, lat = lat, lon = lon)
  # The default alt is naive(0); loc_altitude should return 0 (possibly naive)
  expect_equal(as.numeric(loc_altitude(u)), 0)
})

# ---------------------------------------------------------------------------
# mt_time (linear_time with a location-aware chronon)
# ---------------------------------------------------------------------------
test_that("loc_* accessors work on mt_time via solar day chronon", {
  t <- linear_time(1:3, cal_time_solar$day(1L, lat = lat, lon = lon))
  expect_equal(loc_latitude(t),  rep(lat, 3))
  expect_equal(loc_longitude(t), rep(lon, 3))
  expect_equal(loc_altitude(t),  rep(0, 3))   # default altitude
})

test_that("loc_* accessors work on mt_time via solar ampm chronon", {
  t <- linear_time(1:4, cal_time_solar$ampm(1L, lat = lat, lon = lon, alt = alt))
  expect_equal(loc_latitude(t),  rep(lat, 4))
  expect_equal(loc_longitude(t), rep(lon, 4))
  expect_equal(loc_altitude(t),  rep(alt, 4))
})

test_that("loc_* accessors return correct length matching input vector", {
  n <- 5L
  t <- linear_time(seq_len(n), cal_time_solar$day(1L, lat = lat, lon = lon))
  expect_length(loc_latitude(t),  n)
  expect_length(loc_longitude(t), n)
  expect_length(loc_altitude(t),  n)
})

# ---------------------------------------------------------------------------
# class_mixtime (mixed-chronon vector)
# ---------------------------------------------------------------------------
test_that("loc_* accessors work on class_mixtime", {
  t1 <- linear_time(1:2, cal_time_solar$day(1L, lat = lat,    lon = lon))
  t2 <- linear_time(1:2, cal_time_solar$day(1L, lat = 48.133, lon = 11.583))
  mt <- c(t1, t2)
  expect_equal(loc_latitude(mt),  c(lat, lat, 48.133, 48.133))
  expect_equal(loc_longitude(mt), c(lon, lon, 11.583, 11.583))
})

# ---------------------------------------------------------------------------
# class_any fallback — non-location objects return NA_real_
# ---------------------------------------------------------------------------
test_that("loc_* accessors return NA_real_ for non-location objects", {
  expect_identical(loc_latitude(42),       NA_real_)
  expect_identical(loc_longitude("hello"), NA_real_)
  expect_identical(loc_altitude(NULL),     NA_real_)
})

test_that("loc_* accessors return NA_real_ for non-astronomical mt_time", {
  t <- linear_time(1:3, cal_gregorian$day(1L))
  # No location info: each element maps to NA_real_ (length matches input)
  expect_equal(loc_latitude(t),  rep(NA_real_, 3))
  expect_equal(loc_longitude(t), rep(NA_real_, 3))
  expect_equal(loc_altitude(t),  rep(NA_real_, 3))
})

# ---------------------------------------------------------------------------
# Numeric correctness — west longitude / southern hemisphere / high altitude
# ---------------------------------------------------------------------------
test_that("loc_* return correct values for west longitude and high altitude", {
  hi_lat <- 21.3069; hi_lon <- -157.8583; hi_alt <- 350
  t <- linear_time(1:2, cal_time_solar$day(1L, lat = hi_lat, lon = hi_lon, alt = hi_alt))
  expect_equal(loc_latitude(t),  rep(hi_lat, 2))
  expect_equal(loc_longitude(t), rep(hi_lon, 2))
  expect_equal(loc_altitude(t),  rep(hi_alt, 2))
})

#' Compose a linear time vector from linear and cyclical components
#'
#' `time_compose()` is the inverse of [time_components()]: given a set of
#' [lin()]/[cyc()] components it reconstructs the corresponding time points.
#' Each component is either a two-sided formula pairing a spec with its value,
#' or an already-tagged linear/cyclical time vector (e.g. produced by
#' [linear_time()], [cyclical_time()], or a [time_components()] column).
#'
#' A `lin()` component (the **anchor**), when supplied, fixes the absolute
#' position at some granule (e.g. the year). Every other component must be
#' `cyc()`, chaining without gaps or branches from the anchor down to the
#' target chronon: each cycle must equal another component's chronon exactly.
#'
#' With no `lin()` anchor, every component must be `cyc()`, chained the same
#' way but rooted at whichever component's cycle isn't itself another
#' component's chronon. The result is cyclical time tagged with that root's
#' cycle: `cyc(month, year) ~ 3` alone matches `month_of_year()` for any
#' March; chaining `cyc(day, month) ~ 15` onto it collapses to one (day,
#' year) pair, day-of-year 74, matching `day_of_year()` for 15 March.
#'
#' Values of linear and cyclical components are specified on the right-hand-side
#' of the formula. A `lin()` value is the real-world count (e.g. the literal
#' year 1980); a `cyc()` value is the 1-indexed position within the cycle
#' (e.g. `cyc(month, year) ~ 3` is the 3rd month, March), matching everyday
#' counting rather than the raw 0-indexed position [time_components()] uses
#' internally.
#'
#' @param ... Components used to build the time point. Each element is either:
#'   * a two-sided formula, `lin(<granule>) ~ <value>` or
#'   `cyc(<granule>, <cycle>) ~ <value>` (see [lin()]/[cyc()]), or
#'   * an existing linear or cyclical `mixtime` vector.
#' @param calendar Calendar used to resolve bare granule names in `lin()`/
#'   `cyc()` formulas. Defaults to [cal_gregorian].
#' @inheritParams linear_time
#'
#' @return A `mixtime` time vector, at the finest chronon reached by the
#'   chain (or the root's own chronon, if only one component is given).
#'   Linear with a `lin()` anchor, cyclical otherwise.
#'
#' @seealso [time_components()] for the inverse operation, [lin()]/[cyc()] for
#'   the component vocabulary shared with [time_components()] and [format()].
#'
#' @examples
#' # cyc() values are 1-indexed positions: month 3 is March, day 15 is the 15th
#' time_compose(lin(year) ~ 1980, cyc(month, year) ~ 3, cyc(day, month) ~ 15)
#'
#' # A lin() anchor alone is a valid (coarser) time point
#' time_compose(lin(year) ~ 1980)
#'
#' # No lin() anchor: composes cyclical time
#' time_compose(cyc(month, year) ~ 3)
#'
#' # Chaining collapses to one (chronon, cycle) pair: day 15 of month 3
#' # becomes day-of-year 74
#' time_compose(cyc(day, month) ~ 15, cyc(month, year) ~ 3)
#'
#' # Round-tripping through time_components()
#' parts <- time_components(as.Date("2024-03-15"), yr = lin(year), mth = cyc(month, year))
#' with(parts, time_compose(yr, mth))
#'
#' # Multi-unit (self-referencing) cycles: the 3rd month (1-indexed) of the
#' # 4th 3-month block since epoch (block 3 = months 9-11 -> December 1970)
#' time_compose(lin(month(3L)) ~ 3, cyc(month(1L), month(3L)) ~ 3)
#'
#' @export
time_compose <- function(..., discrete = TRUE, calendar = cal_gregorian) {
  dots <- list(...)
  if (length(dots) == 0L) {
    cli::cli_abort(
      c(
        "Must supply at least one component to compose a time point.",
        i = "Use {.fn lin} for the anchor, e.g. {.code lin(year) ~ 1980}.",
        i = "Use {.fn cyc} for the rest, e.g. {.code cyc(month, year) ~ 3}."
      ),
      call = NULL
    )
  }

  comps <- lapply(dots, compose_tag, calendar = calendar)

  sizes <- vctrs::vec_size_common(!!!lapply(comps, `[[`, "value"))
  comps <- lapply(comps, function(x) {
    x$value <- vctrs::vec_recycle(x$value, sizes)
    x
  })

  chain <- compose_chain(comps)
  new_mixtime(compose_recompose(chain$chain, discrete = discrete, cycle = chain$cycle))
}

# Turn one `...` element into a normalized component: a plain
# list(chronon, cycle, value), where `cycle` is NULL for a linear (anchor)
# component, and `value` is always the *raw, 0-indexed within-cycle offset*
# for a cyclical component (never the absolute epoch-relative count
# mt_cyclical itself stores, and never the formula's 1-indexed input) -- this
# is the one convention chain-walking and the already-tagged and
# formula-derived inputs all agree on.
#
# Already-tagged mixtime input passes through (reduced to this convention);
# a `spec ~ value` formula is evaluated and tagged.
compose_tag <- function(x, calendar) {
  if (S7_inherits(x, class_mixtime)) {
    if (length(x@x) != 1L) {
      cli::cli_abort(
        c(
          "Each {.fn time_compose} component must have a single, consistent chronon.",
          i = "This one mixes {length(x@x)} different chronons; split it first."
        ),
        call = NULL
      )
    }
    x <- x@x[[1L]]
  }
  if (S7_inherits(x, mt_linear)) {
    return(list(
      chronon = attr(x, "chronon"),
      cycle = NULL,
      value = vec_data(x)
    ))
  }
  if (S7_inherits(x, mt_cyclical)) {
    chronon <- attr(x, "chronon")
    cycle <- attr(x, "cycle")
    # A real mt_cyclical stores the same absolute, epoch-relative count as the
    # equivalent mt_linear (see time_components()); reduce it to the raw
    # within-cycle offset the same way chronon_parts() does, so it matches a
    # formula-derived component's convention.
    offset <- chronon_divmod(from = chronon, to = cycle, x = vec_data(x))$mod
    return(list(chronon = chronon, cycle = cycle, value = offset))
  }

  if (!inherits(x, "formula") || length(x) != 3L) {
    cli::cli_abort(
      c(
        "Each {.fn time_compose} component must be a {.fn lin}/{.fn cyc} formula or an existing linear/cyclical time vector.",
        i = "e.g. {.code lin(year) ~ 1980} or {.code cyc(month, year) ~ 3}."
      ),
      call = NULL
    )
  }

  spec <- eval_tidy(
    rlang::f_lhs(x),
    data = component_mask(calendar),
    env = rlang::f_env(x)
  )
  value <- eval(rlang::f_rhs(x), rlang::f_env(x))

  if (length(spec) == 1L) {
    # The value is the real-world count (e.g. the literal year 1980);
    # chronon_epoch() gives the same origin offset mixtime() itself applies
    # to reach the raw, epoch-relative count.
    list(
      chronon = spec[[1L]],
      cycle = NULL,
      value = value - chronon_epoch(spec[[1L]])
    )
  } else {
    # The value is the 1-indexed position within the cycle (e.g. 3 is the
    # 3rd month, matching everyday counting); shift to the raw 0-indexed
    # offset chronon_parts()/time_components() use internally.
    list(chronon = spec[[1L]], cycle = spec[[2L]], value = value - 1L)
  }
}

# Returns list(chain, cycle): chain is the ordered root-to-finest component
# list; cycle is NULL for a lin() anchor, or the root's cycle for cyclical
# time (no lin() given).
#
# A lin() anchor, if supplied, must be the only one; every cyc() then chains
# from it via exact chronon equality, without gaps or branches. With no
# lin(), every component must be cyc(), rooted at whichever one's cycle
# isn't supplied as another component's chronon; its own within-cycle
# offset seeds the chain as an absolute count (mt_cyclical's convention,
# see compose_tag()).
compose_chain <- function(comps) {
  is_lin <- vapply(comps, function(x) is.null(x$cycle), logical(1L))
  n_lin <- sum(is_lin)

  if (n_lin > 1L) {
    labels <- vapply(
      comps[is_lin],
      function(x) time_granule_label(x$chronon, 1),
      character(1L)
    )
    cli::cli_abort(
      c(
        "{.fn time_compose} needs exactly one {.fn lin} anchor, but {n_lin} were supplied ({paste(labels, collapse = ', ')}).",
        i = "Keep one as the anchor and express the rest as {.fn cyc} components relative to it."
      ),
      call = NULL
    )
  }

  if (n_lin == 1L) {
    is_root <- is_lin
    root_cycle <- NULL
  } else {
    supplied_chronons <- lapply(comps, `[[`, "chronon")
    is_root <- vapply(
      comps,
      function(x) !any(vapply(supplied_chronons, identical, logical(1L), y = x$cycle)),
      logical(1L)
    )
    n_root <- sum(is_root)

    if (n_root == 0L) {
      cli::cli_abort(
        c(
          "{.fn time_compose} needs a {.fn lin} anchor, or a {.fn cyc} chain with a coarsest link, but every {.fn cyc} component's cycle is itself another's chronon.",
          i = "Add a {.fn lin} anchor, or drop one of the components."
        ),
        call = NULL
      )
    }
    if (n_root > 1L) {
      labels <- vapply(
        comps[is_root],
        function(x) paste0(time_granule_label(x$chronon, 1), "/", time_granule_label(x$cycle, 1)),
        character(1L)
      )
      cli::cli_abort(
        c(
          "Without a {.fn lin} anchor, {.fn time_compose} needs a single connected {.fn cyc} chain, but {n_root} disconnected chain{?s} were supplied ({paste(labels, collapse = ', ')}).",
          i = "Add a {.fn lin} anchor to combine them, or keep only one chain."
        ),
        call = NULL
      )
    }

    root_cycle <- comps[[which(is_root)]]$cycle
  }

  anchor <- comps[[which(is_root)]]
  links <- comps[!is_root]
  used <- logical(length(links))

  chain <- list(anchor)
  frontier <- anchor$chronon

  repeat {
    matches <- which(
      !used &
        vapply(links, function(l) identical(l$cycle, frontier), logical(1L))
    )
    if (length(matches) == 0L) {
      break
    }
    if (length(matches) > 1L) {
      labels <- vapply(
        links[matches],
        function(x) time_granule_label(x$chronon, 1),
        character(1L)
      )
      cli::cli_abort(
        c(
          "Multiple {.fn cyc} components build from the same {time_granule_label(frontier, 1)}: {paste(labels, collapse = ', ')}.",
          i = "Supply only one path from the anchor to the target granularity."
        ),
        call = NULL
      )
    }
    used[matches] <- TRUE
    chain[[length(chain) + 1L]] <- links[[matches]]
    frontier <- links[[matches]]$chronon
  }

  if (!all(used)) {
    labels <- vapply(
      links[!used],
      function(x) {
        paste0(
          time_granule_label(x$chronon, 1),
          "/",
          time_granule_label(x$cycle, 1)
        )
      },
      character(1L)
    )
    cli::cli_abort(
      c(
        "The following {.fn cyc} component{?s} do{?es/} not connect to the other supplied components: {paste(labels, collapse = ', ')}.",
        i = "Each {.fn cyc}'s cycle must exactly match another component's chronon (same unit and size)."
      ),
      call = NULL
    )
  }

  list(chain = chain, cycle = root_cycle)
}

# Walk the validated chain coarse -> fine, reconstructing the absolute count
# at each link from the base of its coarser period (chronon_divmod()'s
# coarse -> fine `div`, which is always an exact base with no remainder) plus
# the link's own within-cycle offset.
#
# `chain[[1L]]` seeds `running_count`/`running_chronon` directly, whether it
# is a lin() anchor (`cycle` NULL) or, for cyclical time, the chain's root
# cyc() component (see compose_chain()). `cycle` then tags the result
# mt_cyclical instead of mt_linear.
#
# `strict = FALSE` produces NA for invalid rows rather than aborting
compose_recompose <- function(chain, discrete, cycle = NULL, strict = TRUE) {
  anchor <- chain[[1L]]
  running_chronon <- anchor$chronon
  running_count <- anchor$value
  keep <- !(is.na(running_count) | is.infinite(running_count))

  for (link in chain[-1L]) {
    chronon_i <- link$chronon

    base_i <- chronon_divmod(
      from = running_chronon,
      to = chronon_i,
      x = running_count[keep]
    )$div

    # `link$value` is usually a plain offset vector (from a `spec ~ value`
    # formula, or an already-tagged mixtime component). time_parse() instead
    # supplies a closure `function(at) -> offset` for a decoded cyc() token,
    # since decoding it may depend on `at` -- the coarser instance's resolved
    # position -- which has only just become available above. `at` is built
    # full-length (NA outside `keep`) so the closure sees the same shape a
    # plain offset vector would, and everything below stays unchanged.
    offset_i <- if (is.function(link$value)) {
      at <- rep(NA_real_, length(running_count))
      at[keep] <- base_i
      link$value(at)
    } else {
      link$value
    }

    # `at` is the raw, epoch-relative count in terms of the coarser
    # (running) chronon -- the same convention chronon_cardinality() methods
    # use everywhere else (e.g. compare.R, predicates.R), so no mixtime()
    # round-trip is needed here.
    cardinality <- chronon_cardinality(
      chronon_i,
      running_chronon,
      at = running_count[keep]
    )
    bad <- offset_i[keep] < 0 | offset_i[keep] >= cardinality
    if (any(bad, na.rm = TRUE)) {
      if (!strict) {
        # Drop the invalid rows out of the chain quietly as NA
        bad[is.na(bad)] <- TRUE
        idx <- which(keep)

        new_count <- running_count
        new_count[idx[!bad]] <- base_i[!bad] + offset_i[keep][!bad]
        new_count[idx[bad]] <- NA_real_
        running_count <- new_count
        keep[idx[bad]] <- FALSE
        running_chronon <- chronon_i
        next
      }
      bad_i <- which(bad)[1L]
      cli::cli_abort(
        c(
          "{.val {offset_i[keep][bad_i] + 1L}} is not a valid {time_granule_label(chronon_i, 1)} within that {time_granule_label(running_chronon, 1)}.",
          i = "Valid values are {.val {1}} to {.val {cardinality[bad_i]}}."
        ),
        call = NULL
      )
    }

    new_count <- running_count
    new_count[keep] <- base_i + offset_i[keep]
    running_count <- new_count
    running_chronon <- chronon_i
  }

  if (discrete) {
    running_count <- round(running_count)
    # Integer storage can't hold +/-Inf (unlike double), so only narrow when
    # nothing special remains -- matches how e.g. year(c(2020, Inf)) itself
    # keeps a special-containing discrete vector as double.
    if (all(is.na(running_count) | is.finite(running_count))) {
      running_count <- as.integer(running_count)
    }
  }
  if (is.null(cycle)) {
    mt_linear(running_count, chronon = running_chronon)
  } else {
    mt_cyclical(running_count, chronon = running_chronon, cycle = cycle)
  }
}

#' Constructor for mixtime time vectors
#'
#' @description
#'
#' `r lifecycle::badge("deprecated")`
#'
#' `new_time()` was the low-level constructor for `mt_time` vectors. It has been
#' deprecated in favour of calling the concrete time class constructors directly:
#' [mt_linear()] for linear time, [mt_duration()] for durations, and
#' [mt_cyclical()] for cyclical time.
#'
#' Creates a `mixtime` time vector at a specific time point, with a specified
#' chronon and optional cycle. The `chronon` defines the smallest indivisible
#' time granule for the time vector, while the `cycle` allows for cyclical time
#' representations (e.g. day-of-week, month-of-year).
#' 
#' @param x A numeric vector of time points, integers for discrete time or 
#'   doubles for continuous time. 
#' @param chronon A time granule object representing the smallest indivisible time
#'   granule (chronon) for the time vector (e.g. `cal_gregorian$day(1L)`). 
#' @param cycle An optional time granule object representing the cycle for cyclical
#'   time (e.g. `cal_gregorian$week(1L)` for day-of-week). If not provided, the 
#'   time vector will be treated as linear time.
#' @param class An optional character vector of additional S3 classes to assign 
#'   to the resulting time vector. This allows for further subclassing of 
#'   `mt_time` for specific time types (e.g. linear, cyclical, durations, etc.).
#' 
#' @return A `mt_time` vector representing the time points in `x` according to 
#'   the specified `chronon` and `cycle`.
#' 
#' @examples
#' # Create a continuous mixtime time vector for today
#' new_time(
#'   as.double(Sys.Date()),
#'   chronon = cal_gregorian$day(1L, tz = Sys.timezone()),
#'   class = "mt_linear"
#' )
#'
#' # Create a discrete mixtime time vector for the current date and time
#' new_time(
#'   as.integer(Sys.time()),
#'   chronon = cal_gregorian$second(1L, tz = Sys.timezone()),
#'   class = "mt_linear"
#' )
#' 
#' # Create a discrete mixtime time vector for the time of day (cyclical time)
#' new_time(
#'   as.integer(Sys.time()), 
#'   chronon = cal_gregorian$second(1L, tz = Sys.timezone()), 
#'   cycle = cal_gregorian$day(1L, tz = Sys.timezone()),
#'   class = "mt_cyclical"
#' )
#' 
#' @export
new_time <- function(x = integer(), chronon = mt_unit(1L), cycle = NULL, class = NULL) {
  # Route to the concrete S7 time class. The `chronon`/`cycle` length is validated
  # by the typed properties of these classes (see `mt_time-class`).
  class <- class %||% if (is.null(cycle)) "mt_linear" else "mt_cyclical"
  # Validate `class` before warning so the deprecation message can point at a
  # valid replacement constructor.
  if (!class %in% c("mt_linear", "mt_duration", "mt_cyclical")) {
    cli::cli_abort("Unknown time class {.val {class}}.", call. = FALSE)
  }
  lifecycle::deprecate_warn(
    "0.2.0",
    "new_time()",
    with = paste0(class, "()")
  )
  switch(
    class,
    mt_linear = mt_linear(x, chronon = chronon),
    mt_duration = mt_duration(x, chronon = chronon),
    mt_cyclical = {
      if (is.null(cycle)) {
        cli::cli_abort("{.var cycle} is required for cyclical time.", call. = FALSE)
      }
      mt_cyclical(x, chronon = chronon, cycle = cycle)
    }
  )
}

# mt_time vector compatibility methods -----------------------------------------

#' @importFrom vctrs vec_restore
#' @export
S7::method(vec_restore, mt_time) <- function(x, to, ...) {
  if (!is.numeric(x)) {
    cli::cli_abort("{.var x} must be a numeric vector.", call. = FALSE)
  }
  attributes(x) <- attributes(to)
  x
}

#' @export
S7::method(`[`, mt_time) <- function(x, i, ...) {
  S7::S7_data(x) <- S7::S7_data(x)[i]
  x
}

#' @export
S7::method(`[[`, mt_time) <- function(x, i, ...) {
  S7::S7_data(x) <- S7::S7_data(x)[[i]]
  x
}

#' @method [<- mixtime::mt_time
#' @export
`[<-.mixtime::mt_time` <- function(x, i, value) {
  data <- S7::S7_data(x)
  data[i] <- S7::S7_data(value)
  S7::S7_data(x) <- data
  x
}

#' @method [[<- mixtime::mt_time
#' @export
`[[<-.mixtime::mt_time` <- function(x, i, value) {
  data <- S7::S7_data(x)
  data[[i]] <- S7::S7_data(value)
  S7::S7_data(x) <- data
  x
}

#' @export
S7::method(rep, mt_time) <- function(x, ...) {
  S7::S7_data(x) <- rep(S7::S7_data(x), ...)
  x
}

#' @export
S7::method(c, mt_time) <- function(...) {
  dots <- list(...)
  x <- dots[[1L]]
  S7::S7_data(x) <- do.call(base::c, lapply(dots, S7::S7_data))
  x
}

#' Duration vectors
#' 
#' `duration()` creates a vector of durations with a specified chronon.
#' Durations represent a fixed span of time measured in a given time granule
#' (e.g., 3 months, 5 days), without reference to a specific point in time.
#' 
#' @param data A time vector of duration magnitudes, or an existing
#'  [duration()] vector to convert to `chronon` granules.
#' @param chronon A time granule expression representing the chronon, evaluated
#'  in the context of `calendar`. Use unquoted expressions like `month(1L)` or
#'  `day(1L)`. Chronons from a specific calendar can also be used (e.g.
#'  `cal_gregorian$month(1L)`). Defaults to the time chronon of the input
#'  `data` (`time_chronon(data)`).
#' @param discrete Logical. If `TRUE` (default), returns integer durations 
#'   always rounding down (discrete time model). If `FALSE`, returns fractional 
#'   durations (continuous time model).
#' @param calendar Calendar system used to evaluate `chronon`. Defaults to
#'   `time_calendar(data)` for existing time objects. Common options include
#'   [cal_gregorian] and [cal_isoweek].
#'
#' @return A `mixtime` vector containing an `mt_duration` vector.
#'
#' @seealso
#' - [new_duration_fn()] for creating reusable duration functions
#' - [cal_gregorian], [cal_isoweek] for calendar systems
#'
#' @examples
#' # A duration of 3 months
#' duration(3L, cal_gregorian$month(1L))
#'
#' # A vector of durations in days
#' duration(1:7, cal_gregorian$day(1L))
#'
#' # Convert a duration of 4 days into weeks
#' duration(days(4), cal_isoweek$week(1L), discrete = FALSE)
#' duration(days(4), cal_isoweek$week(1L), discrete = TRUE)
#'
#' @export
duration <- function(
  data, chronon = time_chronon(data), discrete = NULL,
  calendar = time_calendar(data)
) {
  # Evaluate chronon and cycle with a calendar mask
  quo_chronon <- enquo(chronon)
  chronon <- eval_tidy(quo_chronon, data = calendar, env = emptyenv())

  if (!inherits(chronon, "mixtime::mt_unit")) {
    cli::cli_abort("{.var chronon} must be a time granule object.", call. = FALSE)
  }

  # Convert an existing duration (mixtime or a bare duration vector) into
  # the requested chronon.
  if (time_is_duration(data[1L])) {
    if (is_mixtime(data)) {
      if (length(data@x) > 1L) {
        cli::cli_abort(
          c(
            "{.fn duration} currently only supports converting durations with a single chronon.",
            i = "To combine several chronons, combine several duration vectors with {.fun c}."
          ),
          call = NULL
        )
      }
      raw <- data@x[[1L]]
    } else {
      raw <- data
    }
    from_chronon <- attr(raw, "chronon")
    if (is.null(discrete)) discrete <- is.integer(vec_data(raw))
    cardinality <- tryCatch(
      chronon_cardinality(chronon, from_chronon),
      error = function(e) {
        cli::cli_abort(
          c(
            "Can't convert a duration from {.val {time_unit_plural(from_chronon)}} to {.val {time_unit_plural(chronon)}}.",
            i = "Their ratio depends on where in the calendar the duration falls, but a duration has no anchor to resolve that against.",
            i = "Convert a {.cls mixtime} time point instead, which carries the calendar context this conversion needs."
          ),
          call = NULL,
          parent = e
        )
      }
    )
    x <- as.numeric(raw) * cardinality
    if (discrete) x <- as.integer(floor(x))
    return(new_mixtime(mt_duration(x, chronon = chronon)))
  }

  if (S7_inherits(data, mt_time)) {
    cli::cli_abort(
      c(
        "{.var data} already has chronon {.val {time_unit_plural(attr(data, 'chronon'))}}, but it isn't a duration.",
        i = "{.fn duration} converts duration vectors; use {.fn mixtime} to convert time points."
      ),
      call = NULL
    )
  }

  if (!is.numeric(data)) {
    cli::cli_abort("{.var data} must be a numeric vector.", call. = FALSE)
  }
  new_mixtime(mt_duration(data, chronon = chronon))
}

#' Duration function factory
#' 
#' `new_duration_fn()` creates a duration function for a specified chronon. A
#' chronon is the smallest indivisible time unit (e.g., days, months) that
#' defines what the numeric magnitudes in the resulting duration vector
#' represent.
#' 
#' @param chronon A bare call for a time unit object representing the chronon
#'   (e.g., `month(1L)`, `day(1L)`).
#' @param default_calendar A default calendar used to resolve the time units
#'   if they don't exist in the calendar of the input data (e.g.,
#'   `cal_gregorian`).
#' 
#' @return A function used to create duration vectors with a specific chronon.
#'   The returned function accepts:
#'   \describe{
#'     \item{`data`}{A numeric vector of duration magnitudes.}
#'     \item{`calendar`}{A calendar system used to evaluate `chronon`. Defaults
#'       to `time_calendar(data)`.}
#'     \item{`...`}{Additional arguments passed to the chronon (e.g., `tz` for
#'       timezones).}
#'   }
#' 
#' @seealso 
#' - [duration()] for creating duration vectors directly
#' - [cal_gregorian], [cal_isoweek] for calendar systems
#' 
#' @examples
#' # Create a months duration function
#' months <- new_duration_fn(month(1L), default_calendar = cal_gregorian)
#' months(1:6)
#' 
#' # Create a days duration function
#' days <- new_duration_fn(day(1L), default_calendar = cal_gregorian)
#' days(1:7)
#' 
#' @export
new_duration_fn <- function(chronon, default_calendar = cal_gregorian) {
  chronon <- rlang::new_quosure(
    enexpr(chronon), 
    env = rlang::as_data_mask(default_calendar)
  )
  function(
    data, calendar = time_calendar(data), ...
  ) {
    # Add tz / loc to chronon
    chronon <- quo_add_dots(chronon, ...)

    duration(data, chronon = !!chronon, calendar = calendar)
  }
}

#' Duration helper functions
#'
#' Convenience functions for creating duration vectors of common time units.
#' Each function wraps [new_duration_fn()] for its respective chronon.
#'
#' @inheritParams duration
#' @param ... Additional arguments passed to the chronon (e.g. `tz` for
#'   timezones).
#'
#' @return A `mixtime` vector containing an `mt_duration` vector.
#'
#' @seealso
#' - [new_duration_fn()] for creating custom duration functions
#' - [duration()] for creating duration vectors directly
#' - [cal_gregorian], [cal_isoweek] for calendar systems
#'
#' @examples
#' years(3L)
#' quarters(2L)
#' months(6L)
#' weeks(4L)
#' days(7L)
#' hours(12L)
#' minutes(30L)
#' seconds(45L)
#' milliseconds(500L)
#'
#' @name duration_helpers
NULL

#' @rdname duration_helpers
#' @export
years <- new_duration_fn(year(1L))

#' @rdname duration_helpers
#' @export
quarters <- new_duration_fn(quarter(1L))

#' @rdname duration_helpers
#' @export
months <- new_duration_fn(month(1L))

#' @rdname duration_helpers
#' @export
weeks <- new_duration_fn(week(1L), default_calendar = cal_isoweek)

#' @rdname duration_helpers
#' @export
days <- new_duration_fn(day(1L))

#' @rdname duration_helpers
#' @export
hours <- new_duration_fn(hour(1L))

#' @rdname duration_helpers
#' @export
minutes <- new_duration_fn(minute(1L))

#' @rdname duration_helpers
#' @export
seconds <- new_duration_fn(second(1L))

#' @rdname duration_helpers
#' @export
milliseconds <- new_duration_fn(millisecond(1L))

test_that("seq.mixtime works with linear time using integer by", {
  # yearmonth sequences
  result <- seq(yearmonth("2020 Jan"), yearmonth("2020 Dec"))
  expect_s7_class(result, class_mixtime)
  expect_length(result, 12)
  expect_equal(format(result[[1]]), format(yearmonth("2020 Jan")))
  expect_equal(format(result[[12]]), format(yearmonth("2020 Dec")))

  # yearquarter with length.out and integer by
  result <- seq(yearquarter("2020 Q1"), length.out = 5, by = 1L)
  expect_s7_class(result, class_mixtime)
  expect_length(result, 5)
  expect_equal(format(result[[1]]), format(yearquarter("2020 Q1")))
  expect_equal(format(result[[5]]), format(yearquarter("2021 Q1")))

  # date sequences
  result <- seq(date("2020-01-01"), date("2020-01-10"))
  expect_s7_class(result, class_mixtime)
  expect_length(result, 10)
  expect_equal(format(result[[1]]), format(date("2020-01-01")))
  expect_equal(format(result[[10]]), format(date("2020-01-10")))
})

test_that("seq.mixtime works with linear time using string intervals", {
  # Month intervals
  result <- seq(date("2020-01-01"), date("2020-12-31"), by = "1 month")
  expect_s7_class(result, class_mixtime)
  expect_length(result, 12)
  expect_equal(format(result[[1]]), format(date("2020-01-01")))
  expect_equal(format(result[[12]]), format(date("2020-12-01")))

  # Year intervals
  result <- seq(yearmonth("2020 Jan"), yearmonth("2025 Jan"), by = "1 year")
  expect_s7_class(result, class_mixtime)
  expect_length(result, 6)
  expect_equal(format(result[[1]]), format(yearmonth("2020 Jan")))
  expect_equal(format(result[[6]]), format(yearmonth("2025 Jan")))

  # Week intervals with length.out
  result <- seq(date("2020-01-01"), length.out = 10, by = "2 weeks")
  expect_s7_class(result, class_mixtime)
  expect_length(result, 10)
  expect_equal(format(result[[1]]), format(date("2020-01-01")))
  expect_equal(
    format(result[[10]]),
    format(date(as.Date("2020-01-01") + 2 * 7 * 9))
  )
})

test_that("seq.mixtime works with linear time using time unit objects", {
  # Month time units
  result <- seq(
    yearmonth("2020 Jan"),
    yearmonth("2020 Dec"),
    by = cal_gregorian$month(2L)
  )
  expect_s7_class(result, class_mixtime)
  expect_length(result, 6)
  expect_equal(format(result[[1]]), format(yearmonth("2020 Jan")))
  expect_equal(format(result[[6]]), format(yearmonth("2020 Nov")))

  # Year time units with length.out
  result <- seq(date("2020-01-01"), length.out = 5, by = cal_gregorian$year(1L))
  expect_s7_class(result, class_mixtime)
  expect_length(result, 5)
  expect_equal(format(result[[1]]), format(date("2020-01-01")))
  expect_equal(format(result[[5]]), format(date("2024-01-01")))

  # Day time units
  result <- seq(
    date("2020-01-01"),
    date("2020-01-31"),
    by = cal_gregorian$day(7L)
  )
  expect_s7_class(result, class_mixtime)
  expect_length(result, 5)
  expect_equal(format(result[[1]]), format(date("2020-01-01")))
  expect_equal(format(result[[5]]), format(date("2020-01-29")))
})

test_that("seq.mixtime works with cyclical time", {
  # month_of_year sequence (full cycle)
  result <- seq(month_of_year(0L), month_of_year(11L))
  expect_s7_class(result, class_mixtime)
  expect_length(result, 12)
  expect_equal(format(result[[1]]), format(month_of_year(0L)))
  expect_equal(format(result[[12]]), format(month_of_year(11L)))

  # month_of_year with wrap-around and time unit by
  # result <- seq(month_of_year(5L), month_of_year(3L), by = cal_gregorian$month(2L))
  # expect_s7_class(result, class_mixtime)
  # Should wrap around: Jun -> Aug -> Oct -> Dec -> Feb -> Apr
  # expect_length(result, 6)
  # expect_equal(format(result[[1]]), format(month_of_year(5L)))
  # expect_equal(format(result[[6]]), format(month_of_year(3L)))

  # day_of_week sequence with integer by
  result <- seq(day_of_week(0L), day_of_week(6L), by = 1L)
  expect_s7_class(result, class_mixtime)
  expect_length(result, 7)
  expect_equal(format(result[[1]]), format(day_of_week(0L)))
  expect_equal(format(result[[7]]), format(day_of_week(6L)))
})

test_that("seq.mixtime handles length.out correctly", {
  # With linear time
  result <- seq(yearmonth("2020 Jan"), length.out = 24)
  expect_s7_class(result, class_mixtime)
  expect_length(result, 24)
  expect_equal(format(result[[1]]), format(yearmonth("2020 Jan")))
  expect_equal(format(result[[24]]), format(yearmonth("2021 Dec")))

  # With cyclical time
  result <- seq(month_of_year(0L), length.out = 15)
  expect_s7_class(result, class_mixtime)
  expect_length(result, 15)
  expect_equal(format(result[[1]]), format(month_of_year(0L)))
})

test_that("seq.mixtime handles along.with correctly", {
  reference_vec <- 1:10

  result <- seq(yearmonth("2020 Jan"), along.with = reference_vec)
  expect_s7_class(result, class_mixtime)
  expect_length(result, 10)
  expect_equal(format(result[[1]]), format(yearmonth("2020 Jan")))
  expect_equal(format(result[[10]]), format(yearmonth("2020 Oct")))
})

test_that("seq.mixtime handles backward sequences", {
  # Linear time going backward
  result <- seq(yearmonth("2020 Dec"), yearmonth("2020 Jan"), by = -1L)
  expect_s7_class(result, class_mixtime)
  expect_length(result, 12)
  expect_equal(format(result[[1]]), format(yearmonth("2020 Dec")))
  expect_equal(format(result[[12]]), format(yearmonth("2020 Jan")))

  # With negative time units
  result <- seq(
    date("2020-12-31"),
    date("2020-12-01"),
    by = cal_gregorian$day(-5L)
  )
  expect_s7_class(result, class_mixtime)
  expect_length(result, 7)
  expect_equal(format(result[[1]]), format(date("2020-12-31")))
  expect_equal(format(result[[7]]), format(date("2020-12-01")))
})

test_that("seq.mixtime handles edge cases", {
  # Single element sequence
  result <- seq(yearmonth("2020 Jan"), yearmonth("2020 Jan"))
  expect_s7_class(result, class_mixtime)
  expect_length(result, 1)
  expect_equal(format(result[[1]]), format(yearmonth("2020 Jan")))

  # Length.out = 1
  result <- seq(yearmonth("2020 Jan"), length.out = 1)
  expect_s7_class(result, class_mixtime)
  expect_length(result, 1)

  # Empty along.with (though unusual)
  result <- seq(yearmonth("2020 Jan"), along.with = integer(0))
  expect_s7_class(result, class_mixtime)
  expect_length(result, 0)
})

test_that("seq.mixtime works with Date objects in mixtime", {
  # Dates should work through mixtime
  result <- seq(mixtime(as.Date("2020-01-01")), mixtime(as.Date("2020-01-10")))
  expect_s7_class(result, class_mixtime)
  expect_length(result, 10)
})

test_that("seq.mixtime works with different by specifications", {
  # Integer by
  result1 <- seq(yearmonth("2020 Jan"), yearmonth("2020 Dec"), by = 2L)
  expect_length(result1, 6)

  # String by
  result2 <- seq(date("2020-01-01"), date("2020-02-01"), by = "1 week")
  expect_s7_class(result2, class_mixtime)

  # Time unit by
  result3 <- seq(
    yearmonth("2020 Jan"),
    yearmonth("2020 Dec"),
    by = cal_gregorian$quarter(1L)
  )
  expect_s7_class(result3, class_mixtime)
  expect_length(result3, 4)
})


test_that("seq.mixtime with fractional linear time", {
  # Fractional yearmonth with day-based by
  result <- seq(
    yearmonth("2020 Jan", discrete = FALSE),
    yearmonth("2025 Jan"),
    by = "30 days"
  )
  expect_s7_class(result, class_mixtime)
  expect_length(result, 61)
  expect_equal(
    format(result[[1]]),
    format(yearmonth("2020 Jan", discrete = FALSE))
  )
  # Last element should be close to but not exceed the end
  expect_match(format(result[[61]]), "2024 Dec ")

  # Fractional date with week-based by
  result <- seq(
    date("2020-01-01", discrete = FALSE),
    date("2020-02-01"),
    by = "1 week"
  )
  expect_s7_class(result, class_mixtime)
  expect_gt(length(result), 4) # More than 4 weeks
  expect_equal(
    format(result[[1]]),
    format(date("2020-01-01", discrete = FALSE))
  )

  # Fractional yearmonth with month-based by (should work cleanly)
  result <- seq(
    yearmonth("2020 Jan", discrete = FALSE),
    yearmonth("2020 Jun"),
    by = "1 month"
  )
  expect_s7_class(result, class_mixtime)
  expect_length(result, 6)
  expect_equal(
    format(result[[1]]),
    format(yearmonth("2020 Jan", discrete = FALSE))
  )
  expect_equal(
    format(result[[6]]),
    format(yearmonth("2020 Jun", discrete = FALSE))
  )

  # Fractional yearquarter with day-based by
  result <- seq(
    yearquarter("2020 Q1", discrete = FALSE),
    length.out = 10,
    by = "15 days"
  )
  expect_s7_class(result, class_mixtime)
  expect_length(result, 10)
  # Should show fractional percentages within quarters
  expect_equal(
    format(result[[1]]),
    format(yearquarter("2020 Q1", discrete = FALSE))
  )

  # Fractional with length.out
  result <- seq(
    yearmonth("2020 Jan", discrete = FALSE),
    length.out = 13,
    by = "25 days"
  )
  expect_s7_class(result, class_mixtime)
  expect_length(result, 13)
  expect_equal(
    format(result[[13]]),
    # 2020 Oct, 83.9% through the month (day 27 of 31) - no string format
    # captures a mid-month fraction, so this stays numeric
    format(yearmonth(609 + 26 / 31, discrete = FALSE))
  )

  # Backward fractional sequence
  result <- seq(
    yearmonth("2020 Dec", discrete = FALSE),
    yearmonth("2020 Jan"),
    by = "-30 days"
  )
  expect_s7_class(result, class_mixtime)
  expect_gt(length(result), 10)
  expect_equal(
    format(result[[1]]),
    format(yearmonth("2020 Dec", discrete = FALSE))
  )
  expect_equal(
    format(result[[12]]),
    format(yearmonth(600 + 5 / 31, discrete = FALSE))
  )
})

test_that("seq.mixtime with duration by", {
  # Using a duration for by should work
  result <- seq(
    date("2020-01-01"),
    date("2020-01-10"),
    by = duration(2L, day(1L))
  )
  expect_s7_class(result, class_mixtime)
  expect_length(result, 5)
  expect_equal(format(result[[1]]), format(date("2020-01-01")))
  expect_equal(format(result[[5]]), format(date("2020-01-09")))
})
test_that("seq.mixtime by a coarser granule of the same unit works (day(7L) etc.)", {
  # chronon and by share the same time unit (day), just different `n` - not a calendar-field
  # shift, so no clamping is involved.
  result <- seq(
    date("2020-01-01"),
    date("2020-01-31"),
    by = cal_gregorian$day(7L)
  )
  expect_equal(
    format(result),
    c("2020-01-01", "2020-01-08", "2020-01-15", "2020-01-22", "2020-01-29")
  )
})

test_that("seq.mixtime by = '1 month' clamps day-of-month but preserves time-of-day", {
  # Regression test: clamping used to flatten day-of-month and time-of-day into one
  # remainder, so overflowing into a shorter month also destroyed the time-of-day.
  result <- suppressWarnings(seq(
    datetime("2026-01-31 14:30:00"),
    length.out = 2,
    by = "1 month"
  ))
  expect_equal(format(result[[2]]), "2026-02-28 14:30:00")
})

test_that("seq.mixtime on_invalid = 'overflow' lets the day-of-month overflow into the next month", {
  result <- seq(
    date("2020-01-31"),
    length.out = 3,
    by = "1 month",
    on_invalid = "overflow"
  )
  expect_equal(format(result), c("2020-01-31", "2020-03-02", "2020-03-31"))
})

test_that("seq.mixtime messages describe the granules involved", {
  # Granules are S7 scalars, so their size comes from @n rather than vec_data()
  expect_error(
    seq(date("2020-01-01"), yearmonth("2020 Jun")),
    "1 day"
  )
  expect_warning(
    seq(date("2020-01-31"), by = months(1L), length.out = 3L),
    "31 days"
  )
})

test_that("seq.mixtime with negative-magnitude `by` clamps and steps backwards", {
  expect_warning(
    result <- seq(date("2020-03-31"), by = "-1 month", length.out = 3),
    "31 days"
  )
  expect_equal(format(result), c("2020-03-31", "2020-02-29", "2020-01-31"))

  result2 <- suppressWarnings(seq(
    date("2020-03-31"),
    by = months(-1L),
    length.out = 3
  ))
  expect_equal(format(result2), format(result))
})

# Tests for time arithmetic: addition, subtraction, and duration operations.
#
# Cross-chronon arithmetic (e.g. date + months, date + years) uses calendar-field
# arithmetic: the duration's own chronon (e.g. month) is advanced by its magnitude, and the
# finer remainder (e.g. day-of-month) is preserved exactly, clamped to the shifted period's
# valid range if it would otherwise overflow (e.g. 2020-01-31 + months(1) -> 2020-02-29,
# not 2020-03-02).

# ---------------------------------------------------------------------------
# date + duration (same chronon: days)
# ---------------------------------------------------------------------------

test_that("date() + days() advances by the correct number of days", {
  d <- date(as.Date("2020-01-02"))

  expect_equal(format(d + days(1L)), "2020-01-03")
  expect_equal(format(d + days(7L)), "2020-01-09")
  expect_equal(format(d + days(30L)), "2020-02-01")
  expect_equal(format(d + days(0L)), "2020-01-02")
})

test_that("date() - days() retreats by the correct number of days", {
  d <- date(as.Date("2020-03-01"))

  expect_equal(format(d - days(1L)), "2020-02-29") # 2020 is a leap year
  expect_equal(format(d - days(29L)), "2020-02-01")
})

# ---------------------------------------------------------------------------
# date + duration (cross-chronon: days <-> months)
#
# The day-of-month is preserved exactly, clamped to the target month's length
# if it would otherwise overflow. Subtraction clamps against the earlier
# month's length in the same way.
# ---------------------------------------------------------------------------

test_that("date() + months(1L) advances to the same day next month", {
  expect_equal(format(date(as.Date("2020-01-02")) + months(1L)), "2020-02-02")
  expect_equal(format(date(as.Date("2019-12-15")) + months(1L)), "2020-01-15")
})

test_that("date() + months(n) advances by n calendar months, preserving day-of-month", {
  expect_equal(format(date(as.Date("2020-01-02")) + months(3L)), "2020-04-02")
  expect_equal(format(date(as.Date("2019-12-15")) + months(2L)), "2020-02-15")
})

test_that("date() + months() clamps an overflowing day-of-month to the target month's length", {
  # January 31 has no equivalent in February -> clamps to the last day
  expect_equal(format(date(as.Date("2020-01-31")) + months(1L)), "2020-02-29") # 2020 is a leap year
  expect_equal(format(date(as.Date("2021-01-31")) + months(1L)), "2021-02-28")
  # Clamping applies per intervening month, not just the destination
  expect_equal(format(date(as.Date("2020-01-31")) + months(2L)), "2020-03-31")
})

test_that("months() + date() is commutative", {
  expect_equal(format(months(1L) + date(as.Date("2020-01-02"))), "2020-02-02")
})

test_that("date() - months() retreats to the same day the given number of months earlier", {
  d <- date(as.Date("2020-03-15"))

  expect_equal(format(d - months(1L)), "2020-02-15")
  expect_equal(format(d - months(3L)), "2019-12-15")
})

test_that("date() - months() clamps an overflowing day-of-month to the earlier month's length", {
  # March 31 has no equivalent in February -> clamps to the last day
  expect_equal(format(date(as.Date("2020-03-31")) - months(1L)), "2020-02-29") # 2020 is a leap year
  expect_equal(format(date(as.Date("2021-03-31")) - months(1L)), "2021-02-28")
})

# ---------------------------------------------------------------------------
# date + duration (cross-chronon: days <-> years)
#
# days <-> years also has a direct (context-dependent) chronon_divmod() method - a flat
# day-of-year encoding, with no notion of "month" - but calendar-field arithmetic
# deliberately does *not* use it: it decomposes along the *longest* available path
# (day -> month -> year, not the shortcut edge) so that day-of-month is the component
# that gets clamped, exactly as it is for days <-> months. This is what keeps
# `+ years(1L)` and `+ months(12L)` in agreement.
# ---------------------------------------------------------------------------

test_that("date() + years(1L) advances to the same day next year", {
  expect_equal(format(date(as.Date("2020-01-02")) + years(1L)), "2021-01-02")
  expect_equal(format(date(as.Date("2021-01-02")) + years(1L)), "2022-01-02")
})

test_that("date() + years(1L) clamps a leap day to the last day of February in a common year", {
  expect_equal(format(date(as.Date("2024-02-29")) + years(1L)), "2025-02-28")
  # Agrees with the equivalent months() shift
  expect_equal(
    format(date(as.Date("2024-02-29")) + years(1L)),
    format(date(as.Date("2024-02-29")) + months(12L))
  )
})

test_that("date() - years(1L) retreats to the same day the given number of years earlier", {
  expect_equal(format(date(as.Date("2020-01-02")) - years(1L)), "2019-01-02")
  expect_equal(format(date(as.Date("2021-01-02")) - years(1L)), "2020-01-02")
})

# ---------------------------------------------------------------------------
# yearmonth + duration (same chronon: months) -- exact by definition
# ---------------------------------------------------------------------------

test_that("yearmonth() + months() advances by the correct number of months", {
  ym <- yearmonth(as.Date("2020-01-01"))

  expect_equal(format(ym + months(1L)), "2020 Feb")
  expect_equal(format(ym + months(12L)), "2021 Jan")
  expect_equal(format(ym + months(0L)), "2020 Jan")
})

test_that("yearmonth() - months() retreats by the correct number of months", {
  ym <- yearmonth(as.Date("2020-03-01"))

  expect_equal(format(ym - months(1L)), "2020 Feb")
  expect_equal(format(ym - months(3L)), "2019 Dec")
})

# ---------------------------------------------------------------------------
# yearmonth + duration (cross-chronon: months <-> years) -- exact (12 months/yr)
# ---------------------------------------------------------------------------

test_that("yearmonth() + years() advances by 12 months per year", {
  ym <- yearmonth(as.Date("2020-06-01"))

  expect_equal(format(ym + years(1L)), "2021 Jun")
  expect_equal(format(ym + years(2L)), "2022 Jun")
})

test_that("yearmonth() - years() retreats by 12 months per year", {
  ym <- yearmonth(as.Date("2023-03-01"))

  expect_equal(format(ym - years(1L)), "2022 Mar")
})

# ---------------------------------------------------------------------------
# time - time => duration
# ---------------------------------------------------------------------------

test_that("date() - date() returns the difference in days", {
  d1 <- date(as.Date("2020-02-01"))
  d2 <- date(as.Date("2020-01-01"))

  result <- d1 - d2
  expect_s7_class(result, class_mixtime)
  expect_equal(as.integer(result), 31L) # January has 31 days
})

test_that("date() - date() handles leap-year February correctly", {
  d1 <- date(as.Date("2020-03-01"))
  d2 <- date(as.Date("2020-02-01"))

  expect_equal(as.integer(d1 - d2), 29L) # 2020 is a leap year
})

test_that("date() - date() is zero for identical times", {
  d <- date(as.Date("2023-07-04"))
  expect_equal(as.integer(d - d), 0L)
})

test_that("yearmonth() - yearmonth() returns the difference in months", {
  ym1 <- yearmonth(as.Date("2021-04-01"))
  ym2 <- yearmonth(as.Date("2020-01-01"))

  expect_equal(as.integer(ym1 - ym2), 15L) # 12 + 3
})

test_that("yearmonth() - yearmonth() is negative when ym1 < ym2", {
  ym1 <- yearmonth(as.Date("2020-01-01"))
  ym2 <- yearmonth(as.Date("2020-04-01"))

  expect_equal(as.integer(ym1 - ym2), -3L)
})

# ---------------------------------------------------------------------------
# date +/- integer (direct numeric shift along the chronon)
# ---------------------------------------------------------------------------

test_that("date() + integer shifts by raw day units", {
  d <- date(as.Date("2020-01-01"))

  expect_equal(format(d + 1L), "2020-01-02")
  expect_equal(format(d + 31L), "2020-02-01")
})

test_that("yearmonth() + integer shifts by raw month units", {
  ym <- yearmonth(as.Date("2020-01-01"))

  expect_equal(format(ym + 1L), "2020 Feb")
  expect_equal(format(ym + 12L), "2021 Jan")
})

# ---------------------------------------------------------------------------
# Duration arithmetic (duration +/- duration, duration +/- integer)
# ---------------------------------------------------------------------------

test_that("duration + duration with the same chronon sums correctly", {
  expect_equal(as.integer(months(3L) + months(2L)), 5L)
  expect_equal(as.integer(days(10L) + days(5L)), 15L)
  expect_equal(as.integer(years(1L) + years(2L)), 3L)
})

test_that("duration - duration with the same chronon differences correctly", {
  expect_equal(as.integer(months(5L) - months(2L)), 3L)
  expect_equal(as.integer(days(10L) - days(3L)), 7L)
})

test_that("duration + integer shifts the duration magnitude", {
  expect_equal(as.integer(months(3L) + 2L), 5L)
  expect_equal(as.integer(days(7L) - 3L), 4L)
})

test_that("integer + duration shifts the duration magnitude", {
  expect_equal(as.integer(2L + months(3L)), 5L)
})

test_that("duration * numeric scales the duration magnitude", {
  expect_equal(as.integer(days(3L) * 3), 9L)
  expect_equal(as.integer(months(2L) * 6), 12L)
  expect_equal(as.integer(days(5L) * 0), 0L)
  expect_equal(as.integer(days(5L) * 1), 5L)
  expect_equal(as.integer(days(5L) * -1), -5L)
})

test_that("numeric * duration is commutative", {
  expect_equal(as.integer(3 * days(3L)), 9L)
  expect_equal(as.integer(0.5 * days(6L)), 3L)
})

test_that("duration / numeric scales the duration magnitude", {
  expect_equal(as.integer(days(6L) / 2), 3L)
  expect_equal(as.integer(months(9L) / 3), 3L)
  expect_equal(as.double(days(5L) / 2), 2.5)
})

test_that("duration / duration returns a dimensionless ratio", {
  expect_equal(as.double(days(6L) / days(2L)), 3)
  expect_equal(as.double(days(6L) / days(6L)), 1)
  expect_equal(as.double(months(4L) / months(2L)), 2)
})

test_that("duration / duration converts to a common chronon before dividing (same calendar)", {
  # months and years share cal_gregorian; years(1L) == 12 months
  expect_equal(as.double(years(2L) / months(1L)), 24)
  expect_equal(as.double(months(6L) / years(1L)), 0.5)
})

test_that("duration / duration converts to a common chronon before dividing (cross-calendar)", {
  # weeks (cal_isoweek) and days (cal_gregorian) share a common day chronon
  expect_equal(as.double(weeks(1L) / days(1L)), 7)
  expect_equal(as.double(days(14L) / weeks(1L)), 2)
})

# ---------------------------------------------------------------------------
# Vectorised operations
# ---------------------------------------------------------------------------

test_that("vectorised date() + days() works element-wise", {
  d <- date(as.Date("2020-01-01") + 0:2)

  result <- d + days(1L)
  expect_equal(format(result), c("2020-01-02", "2020-01-03", "2020-01-04"))
})

test_that("vectorised date() + months() works element-wise", {
  d <- date(as.Date(c("2020-01-15", "2020-02-15", "2020-03-15")))

  # Jan=31, Feb=29 (2020 leap), Mar=31 -> each advances by its own month's length
  result <- d + months(1L)
  expect_equal(format(result), c("2020-02-15", "2020-03-15", "2020-04-15"))
})

test_that("vectorised date() - date() returns element-wise differences in days", {
  d1 <- date(as.Date("2020-12-31") + 0:2)
  d2 <- date(as.Date("2020-12-31"))

  result <- d1 - d2
  expect_equal(as.integer(result), 0:2)
})

# ---------------------------------------------------------------------------
# NA propagation
# ---------------------------------------------------------------------------

test_that("NA propagates through date() + months()", {
  d <- date(as.Date(c("2020-01-01", NA, "2020-03-01")))

  result <- d + months(1L)
  expect_equal(is.na(result), c(FALSE, TRUE, FALSE))
})

test_that("NA propagates through date() - date()", {
  d1 <- date(as.Date(c("2020-02-01", NA)))
  d2 <- date(as.Date("2020-01-01"))

  result <- d1 - d2
  expect_equal(is.na(result), c(FALSE, TRUE))
})

# ---------------------------------------------------------------------------
# Unsupported operations error gracefully
# ---------------------------------------------------------------------------

test_that("multiplying two time points errors", {
  d <- date(as.Date("2020-01-01"))
  expect_error(d * d)
})

test_that("adding two time points errors", {
  d <- date(as.Date("2020-01-01"))
  expect_error(d + d)
})

# ---------------------------------------------------------------------------
# Timezone-aware time - date subtraction (issue #97)
# ---------------------------------------------------------------------------

test_that("datetime() - date() uses the date's timezone for its day boundary", {
  # 2026-06-16 09:30:00 AEST = 2026-06-15 23:30:00 UTC = 2026-06-16 01:30:00 CEST
  t <- datetime("2026-06-16 09:30:00", tz = "Australia/Melbourne")

  # Melbourne date: 2026-06-16 AEST; time-of-day in Melbourne = 9h30m = 34200s
  result_mel <- t - date(t, tz = "Australia/Melbourne", discrete = TRUE)
  expect_equal(as.double(result_mel), 34200)

  # Berlin date: 2026-06-16 CEST; time-of-day in Berlin = 1h30m = 5400s
  result_ber <- t - date(t, tz = "Europe/Berlin", discrete = TRUE)
  expect_equal(as.double(result_ber), 5400)

  # UTC date: 2026-06-15 UTC; time-of-day in UTC = 23h30m = 84600s
  result_utc <- t - date(t, tz = "UTC", discrete = TRUE)
  expect_equal(as.double(result_utc), 84600)
})

test_that("datetime() with a timezone can be shifted by durations with a regular cardinality", {
  d <- datetime("2026-06-03 02:30:00", tz = "Australia/Melbourne")

  expect_equal(format(d + hours(1L)), "2026-06-03 03:30:00 AEST")
  expect_equal(format(d + days(1L)), "2026-06-04 02:30:00 AEST")
  expect_equal(format(d - hours(1L)), "2026-06-03 01:30:00 AEST")
  expect_equal(format(hours(1L) + d), "2026-06-03 03:30:00 AEST")
})

#' Epoch offset for chronons
#'
#' Returns the epoch offset for a given chronon (time unit). The epoch defines
#' the starting point of the chronon's linear numbering, used when converting
#' from internal representations to common displays (e.g. applying an epoch of
#' 1970-01-01).
#'
#' @param x A chronon (time unit) object.
#' @param ... Additional arguments for methods.
#'
#' @return A numeric value representing the epoch for the chronon.
#'
#' @export
#' @examples
#' # The epoch for year linear time displays is 1970
#' chronon_epoch(cal_gregorian$year(1L))
chronon_epoch <- S7::new_generic("chronon_epoch", c("x"))

# No epoch specified, default 0
method(chronon_epoch, mt_unit) <- function(x) 0L

#' Generate sequences of mixtime values
#'
#' Create regular sequences of time values. This method handles both linear
#' time sequences (dates, date-times) and cyclical time sequences (day of week,
#' month of year).
#'
#' @param from Starting value of the sequence.
#' @param to End value of the sequence (if provided).
#' @param by Increment of the sequence. Can be:
#'   * A numeric for the number of time chronons
#'   * A character string specifying the interval (e.g., "1 day", "2 weeks",
#'     "1 month", "1 year")
#'   * A time granule object created with time unit functions (e.g., `cal_gregorian$year(1L)`,
#'     `cal_gregorian$month(1L)`, `cal_gregorian$day(1L)`)
#'   * A time [duration()] object (e.g., `years(1L)`, `months(1L)`, `days(1L)`)
#' @param length.out Desired length of the sequence (alternative to `to`).
#' @param along.with Take the length from this argument (alternative to `length.out`).
#' @param on_invalid How to handle time points that overflow the cycle when
#'   using a `by` argument with different time granule than the sequence chronon.
#'   Options are:
#'   * `"nearest"` (default): Adjust overflowing time points to the nearest
#'     valid time point within the cycle
#'   * `"overflow"`: Allow time points to overflow into the next cycle
#'
#'   This is relevant when the starting time point has an offset that doesn't
#'   exist in all cycles. For example, starting on day 31 with `by = "1 month"`
#'   will overflow in months with fewer than 31 days (e.g., February). With
#'   `"nearest"`, these will be adjusted to the last day of the month (e.g.,
#'   Feb 28/29). With `"overflow"`, the extra days carry into the next month.
#'
#'   If not explicitly specified and overflow occurs, a warning is issued with
#'   the default `"nearest"` behavior applied.
#' @param ... Additional arguments passed to the underlying sequence method.
#'
#' @return A mixtime vector containing the sequence.
#'
#' @details
#' For linear time types (Date, POSIXct, yearmonth, etc.), sequences progress
#' forward or backward in time. For cyclical time types (month_of_year,
#' day_of_week, etc.), sequences wrap around cyclically.
#' 
#' @examples
#' # Linear time sequences with integer by
#' seq(yearmonth("2020 Jan"), yearmonth("2020 Dec"))
#' seq(yearquarter("2020 Q1"), length.out = 5, by = 3)
#'
#' # Linear time sequences with string intervals
#' seq(date("2020-01-01"), date("2020-12-31"), by = "1 month")
#' seq(yearmonth("2020 Jan"), yearmonth("2025 Jan"), by = "1 year")
#' seq(date("2020-01-01"), length.out = 10, by = "2 weeks")
#'
#' # Linear time sequences incrementing by time granules
#' seq(yearmonth("2020 Jan"), yearmonth(("2020 Dec")), by = cal_gregorian$month(2L))
#' seq(date("2020-01-01"), length.out = 5, by = cal_gregorian$year(1L))
#' seq(date("2020-01-01"), date("2020-01-31"), by = cal_gregorian$day(7L))
#'
#' # Handling invalid dates with on_invalid
#' seq(date("2020-01-31"), length.out = 3, by = "1 month")  # warns, uses nearest
#' seq(date("2020-01-31"), length.out = 3, by = "1 month", on_invalid = "nearest")
#' seq(date("2020-01-31"), length.out = 3, by = "1 month", on_invalid = "overflow")
#'
#' # Cyclical time sequences
#' seq(month_of_year(0L), month_of_year(11L))
#' seq(day_of_week(0L), day_of_week(6L), by = 1)
#'
#' @rdname seq.mixtime
#' @method seq mixtime::mixtime
#' @export
`seq.mixtime::mixtime` <- function(...) {
  # Strip mixtime vecvec wrapper
  arg <- lapply(
    rlang::list2(...),
    function(x) if (is_mixtime(x)) vecvec::unvecvec(x) else x
  )

  # Call seq method with bare vectors
  new_mixtime(rlang::exec(seq, !!!arg))
}

#' @rdname seq.mixtime
#' @method seq mixtime::mt_time
#' @export
`seq.mixtime::mt_time` <- function(
  from,
  to,
  by,
  length.out = NULL,
  along.with = NULL,
  on_invalid = c("nearest", "overflow"),
  ...
) {
  if (!is.null(along.with)) {
    length.out <- length(along.with)
  } else if (!is.null(length.out)) {
    if (length(length.out) != 1L) {
      stop(sprintf("'%s' must be of length 1", length.out))
    }
    length.out <- ceiling(length.out)
  }

  missing_from <- missing(from)
  missing_to <- missing(to)
  missing_by <- missing(by)

  # Check chronon compatibility
  if (!missing_from && !missing_to) {
    from_chronon <- chronon_common(from)
    to_chronon <- chronon_common(to)

    # Check that from and to have the same time classes
    if (!identical(from_chronon, to_chronon)) {
      cli::cli_abort(c(
        "Incompatible time chronones in {.fn seq}.",
        "i" = "{.arg from} has {.cls {time_granule_label(from_chronon)}} chronons.",
        "i" = "{.arg to} has {.cls {time_granule_label(to_chronon)}} chronons."
      ))
    }
  }
  ptype <- if (missing_from) to else from

  if (!missing_by) {
    # If by is provided, exactly two of from/to/length.out must be specified
    n_provided <- sum(!missing_from, !missing_to, !is.null(length.out))
    if (n_provided != 2L) {
      cli::cli_abort(
        "When {.arg by} is provided, exactly two of {.arg from}, {.arg to}, or {.arg length.out} must be specified."
      )
    }

    # Allow duration input for by, but convert to granule
    if (length(by) != 1L) {
      cli::cli_abort("{.var by} must be a single time duration", call. = FALSE)
    }
    if (time_is_duration(by)) {
      by_size <- as.numeric(by)
      by <- attr(by, "chronon")
      by@n <- by@n * by_size
    }

    # Sequence chronon
    chronon <- chronon_common(ptype)

    # Parse `by` argument
    if (is.character(by)) {
      by <- parse_time_unit(by)
    }
    # Numeric `by` uses chronon
    if (is.numeric(by) && !S7_inherits(by, mt_unit)) {
      by <- S7::S7_class(chronon)(by)
    }

    # Inherit granule properties from `by` if not given in `chronon` granule
    by <- granule_inherit_props(by, chronon)

    # Calendar-field clamping applies when `by` has an irregular cardinality
    # with the `chronon`; the irregularity potentially allows for out-of-range
    # time points (e.g. 30 Jan -> 30 Feb is invalid, because day<->month is
    # irregular). Rounding to the nearest valid time point impacts other time
    # components (e.g. the time of the day), which is usually undesirable.
    path <- if (chronon_needs_clamping(chronon, by)) {
      chronon_longest_path(chronon, by)
    }

    # Convert `from`/`to` time chronons to `by` time chronons. `align_value` records whichever
    # of `from`/`to` ends up determining `seq_part` (the sequence's within-`by` remainder) -
    # it's what the reconstruction below decomposes for the clamped path.
    seq_part <- NULL
    align_value <- NULL
    # Increment by 1 since the time point chronon is rebased with `by`
    arg <- list(by = 1L)
    # `chronon_divmod()`/`chronon_decompose()` work on a UTC(-shaped) numeric
    # value, but for the clamped path a `by` boundary (e.g. "start of month")
    # is defined on the local wall clock, so shift into local time first for
    # that case - the same convention `chronon_parts()` uses. Regular
    # (fixed-cardinality) `by` granules like hour/day are stepped as raw
    # durations instead (matching e.g. `seq.POSIXt()`), so they're left alone.
    to_by_domain <- if (is.null(path)) {
      function(x) x
    } else {
      function(x) tz_to_local(x, chronon, tz_name(by))
    }
    if (!missing_from) {
      # Convert `from` into `by` chronons
      from_local <- to_by_domain(as.numeric(from))
      divmod <- chronon_divmod(chronon, by, from_local)
      arg$from <- divmod$div

      # Left aligned sequencing
      seq_part <- divmod$mod
      align_value <- from_local
    }
    if (!missing_to) {
      # Shift `to` left to account for `from` alignment
      if (!is.null(seq_part)) {
        to <- to - seq_part
      }

      # Convert `to` into `by` chronons
      to_local <- to_by_domain(as.numeric(to))
      divmod <- chronon_divmod(chronon, by, to_local)
      arg$to <- divmod$div

      # Right aligned sequencing if from isn't provided
      if (is.null(seq_part)) {
        seq_part <- divmod$mod
        align_value <- to_local
      }
    }
    if (!is.null(length.out)) {
      arg$length.out <- length.out
    }

    res <- rlang::exec(seq.int, !!!arg)

    missing_on_invalid <- is.character(on_invalid) && length(on_invalid) > 1L
    on_invalid <- match.arg(on_invalid)

    # Reconstruct `chronon` units from `by`-unit counts (from seq.int())
    #
    # If sequenced by an irregular cardinality (`path` is non-NULL), use
    # clamping to return component-by-component valid time points.
    #
    # Otherwise, regular cardinality can be directly converted and always
    # produce valid time points.
    if (is.null(path)) {
      res <- chronon_convert_impl(
        res,
        by,
        chronon,
        discrete = is.integer(ptype)
      ) +
        seq_part
    } else {
      seq_mod <- chronon_decompose(align_value, path)$mod
      recomposed <- chronon_recompose(
        res,
        seq_mod,
        path,
        clamp = on_invalid == "nearest"
      )

      if (missing_on_invalid && any(recomposed$clamped, na.rm = TRUE)) {
        cli::cli_warn(c(
          "The cycle offset ({time_granule_label(chronon, seq_part + 1L)}) has produced time points that overflow the {time_unit_plural(by, 1L)} cycle.",
          "!" = "Using the nearest valid time points in the cycle, {.code on_invalid = \"nearest\"} (the default).",
          "i" = "Specify {.arg on_invalid} explicitly to suppress this warning."
        ))
      }
      # `chronon_recompose()` stays in the local-shifted domain used by
      # `align_value` above - undo the shift to recover the true `chronon`
      # value.
      res <- tz_to_utc(recomposed$value, chronon, tz_name(by))
    }

    # Restore integer type for discrete time input
    if (is.integer(ptype)) res <- as.integer(res)
  } else {
    # If by is missing, at least two of from/to/length.out must be specified
    n_provided <- sum(!missing_from, !missing_to, !is.null(length.out))
    if (n_provided < 2L) {
      cli::cli_abort(
        "When {.arg by} is missing, at least two of {.arg from}, {.arg to}, or {.arg length.out} must be specified."
      )
    }

    arg <- list(length.out = length.out)
    if (!missing_from) {
      arg$from <- as.numeric(from)
    }
    if (!missing_to) {
      arg$to <- as.numeric(to)
    }

    res <- rlang::exec(seq.int, !!!arg)
  }

  vec_restore(res, ptype)
}

# #' @rdname seq.mixtime
# #' @export
# seq.mt_cyclical <- function(from, to, by, length.out = NULL, along.with = NULL, ...) {
#   if (!is.null(along.with)) {
#     length.out <- length(along.with)
#   }
#   else if (!is.null(length.out)) {
#     if (length(length.out) != 1L)
#       stop(sprintf("'%s' must be of length 1", length.out))
#     length.out <- ceiling(length.out)
#   }

#   # Capture extra arguments
#   arg <- rlang::list2(...)
#   if (!missing(by)) arg$by <- by
#   if (!is.null(length.out)) arg$length.out <- length.out
#   # NB: Passing along.with = NULL to seq.int causes an overflow in R
#   if (!is.null(along.with)) arg$along.with <- along.with

#   # Parse `by` argument
#   if (is.character(arg$by)) arg$by <- parse_time_unit(arg$by)

#   # Convert mt_cyclical to numeric for seq() method
#   if (!missing(from) && inherits(from, "mt_cyclical")) {
#     ptype <- from
#     arg$from <- from@n
#   }
#   if (!missing(to) && inherits(to, "mt_cyclical")) {
#     arg$to <- vctrs::vec_data(to)
#     if (!is.null(arg$from)) {
#       # Require compatible cyclical time objects for from:to
#       vec_assert(to, from)
#     } else {
#       ptype <- to
#     }
#   }

#   # Cyclical period
#   period <- chronon_cardinality(attr(ptype, "chronon"), attr(ptype, "cycles")[[1L]])

#   # Adjust from:to for looping around cycles
#   if (!is.null(arg$to) && !is.null(arg$from)) {
#     if (arg$by%||%1 > 0 && to < from) {
#       arg$to <- arg$to + period
#     } else if (to > from && arg$by%||%1 < 0) {
#       arg$to <- arg$to - period
#     }
#   }

#   # Convert `by` to match `ptype` units
#   if (!is.null(arg$by) && S7_inherits(arg$by, mt_unit)) {
#     arg$by <- chronon_cardinality(time_chronon(ptype), arg$by)
#   }

#   # Generate linear sequence
#   res <- rlang::inject(seq.int(!!!arg))

#   # Compute cyclical component
#   res <- res %% period

#   # Restore integer type for discrete time input
#   if (is.integer(ptype)) res <- as.integer(res)

#   # Restore cyclical time attributes
#   vec_restore(res, ptype)
# }

naive_loc <- naive(NA_real_)

# #' Location unit class with geographic coordinates
# #'
# #' A time unit class that extends mt_unit to include geographic location
# #' information for solar-based time calculations.
# #'
#' @param lat Numeric. Latitude in decimal degrees. Range: -90 to 90.
#'   Default: naive location (astronomical calculations require `lat`).
#' @param lon Numeric. Longitude in decimal degrees. Range: -180 to 180.
#'   Default: naive location (astronomical calculations require `lon`).
#' @param alt Numeric. Altitude in meters above sea level.
#'   Default: 0 (sea level).
#'
#' @rdname mt_unit
#' @export
mt_loc_unit <- S7::new_class(
  "mt_loc_unit", 
  parent = mt_unit,
  properties = list(
    lat = S7::new_property(S7::class_numeric, default = naive_loc),
    lon = S7::new_property(S7::class_numeric, default = naive_loc),
    alt = S7::new_property(S7::class_numeric, default = naive(0))
  ),
  constructor = function (n = 1L, lat = naive_loc, lon = naive_loc, alt = naive(0)) {
    # Allow for logical NA for convenience
    if (is.na(lat) && !identical(lat, naive_loc)) lat <- NA_real_
    if (is.na(lon) && !identical(lon, naive_loc)) lon <- NA_real_
    
    S7::new_object(mt_unit(n = n), lat = lat, lon = lon, alt = alt)
  },
  validator = function(self) {
    NULL
  }
)

# Default formats
method(chronon_format_attr, mt_loc_unit) <- function(x) {
  if (!is.naive(x@lat) && !is.naive(x@lon)) " [{loc(.time)}]" else ""
}

# tzdb::tzdb_names() re-derives its result on every call; the timezone
# database only changes via an explicit tzdb::tzdb_initialize() (done once,
# in .onLoad()), so the name list is cached lazily on first use.
tzdb_names_cached <- local({
  names <- NULL
  function() {
    if (is.null(names)) {
      names <<- tzdb::tzdb_names()
    }
    names
  }
})

check_tz_name <- function(zone) {
  # Naive time zone
  if (is.na(zone)) {
    return(invisible(TRUE))
  }

  # Specified time zone
  if (!zone %in% tzdb_names_cached()) {
    cli::cli_abort(
      c(
        "Timezone {.val {zone}} not found in timezone database.",
        "i" = "Valid timezone names can be found with {.fn tzdb::tzdb_names}."
      ),
      call = NULL
    )
  }
  invisible(TRUE)
}

time_mode_label <- function(x) {
  if (S7_inherits(x, mt_cyclical)) {
    "cyclical"
  } else if (S7_inherits(x, mt_duration)) {
    "duration"
  } else {
    "linear"
  }
}

# Operations that reduce a mixtime to mutually comparable values - comparison,
# ordering, de-duplication - are only meaningful when every element shares one
# mode of time: an absolute position (`mt_linear`), a position within a cycle
# (`mt_cyclical`) and a magnitude (`mt_duration`) cannot be compared with one
# another.
#
# Returns the shared mode, or `character(0)` for an empty mixtime.
check_common_time_mode <- function(x) {
  modes <- unique(vapply(x@x, time_mode_label, character(1L)))
  if (length(modes) > 1L) {
    cli::cli_abort(
      c(
        "Can't compare {.field {modes}} time in a single {.cls mixtime} vector.",
        "i" = "Comparing, sorting and de-duplicating a {.cls mixtime} requires every element to share one mode of time."
      ),
      call = NULL
    )
  }
  modes
}

# Cyclical time compares on its position within a cycle, so those positions are
# comparable only when every element shares one cycle. A cycle is a modulus
# rather than a unit of measure, so unlike a chronon there is no common value
# two differing cycles could be reconciled to (see `cycle_common()`).
#
# Returns the shared cycle granule.
check_common_cycle <- function(x) {
  cycles <- lapply(x@x, function(x) attr(x, "cycle"))
  cycle <- Reduce(
    function(x, y) if (is.null(x)) NULL else cycle_common(x, y),
    cycles
  )
  if (is.null(cycle)) {
    cli::cli_abort(
      c(
        "Can't compare cyclical time of differing cycles.",
        "i" = "Found {.field {unique(vapply(cycles, time_granule_label, character(1L)))}}."
      ),
      call = NULL
    )
  }
  cycle
}

#' Compute circular rolling sums
#'
#' Calculates rolling sums of length `k` for all contiguous subsequences
#' around a circular vector. Returns sums for each valid k-element window
#' that wraps around the vector as if arranged in a circle.
#'
#' @param x A numeric vector to compute circular sums over.
#' @param size Integer; the window size (number of consecutive elements to
#'   sum). A negative size anchors the window at its end, with the window
#'   counting backwards.
#' @param step Integer; the step size (the increment in starting index for
#'   each sum). A negative step walks the windows backwards around the circle.
#'
#' @return A numeric vector containing the sum of each contiguous subsequence
#'   around the circle. The length of the resulting vector is the number of
#'   combinations until the pattern between `x` and `step` repeats
#'
#' @examples
#' # Simple circular sum with window of 2
#' circsum(c(1, 2, 3, 4), 2)
#' # Returns: 3 7 (1+2, 3+4)
#'
#' # Window of 3 elements
#' circsum(c(1, 2, 3, 4, 5), 3)
#' # Returns: 6 10 9 8 12 (1+2+3, 4+5+1, 2+3+4, 5+1+2, 3+4+5)
#'
#' # Negative step walks the same windows backwards
#' circsum(c(1, 2, 3, 4, 5), 1, -1)
#' # Returns: 1 5 4 3 2
#'
#' # Negative size anchors each window at its end instead of its start - a
#' # trailing rather than leading rolling sum
#' circsum(c(1, 2, 3, 4, 5), -2, 1)
#' # Returns: 6 3 5 7 9 (5+1, 1+2, 2+3, 3+4, 4+5)
#'
#' @export
circsum <- function(x, size, step = size) {
  n <- length(x)
  size_abs <- abs(size)
  step_abs <- abs(step)
  if (n == 0L || size_abs <= 0L || step_abs <= 0L) {
    return(numeric(0))
  }

  # Offsets from `start` covered by one window: forward (0, 1, ..., k - 1)
  # for a positive size, or backward (-(k - 1), ..., -1, 0) - ending at
  # `start` - for a negative one. A single-element window (k == 1) is the
  # same set either way, so the shortcut below doesn't need this.
  offsets <- seq_len(size_abs) - 1L
  if (size < 0L) {
    offsets <- -rev(offsets)
  }

  if (size_abs == 1L) {
    if (step == 1L) {
      return(x)
    }
    num_windows <- n / gcd(n, step_abs)
    indices <- ((seq_len(num_windows) - 1L) * step) %% n + 1L
    return(x[indices])
  }

  # Number of unique windows before pattern repeats
  num_windows <- n / gcd(n, step_abs)

  # Generate all window sums
  vapply(
    seq_len(num_windows),
    function(i) {
      start <- ((i - 1L) * step) %% n
      indices <- (start + offsets) %% n + 1L
      sum(x[indices])
    },
    numeric(1)
  )
}

gcd <- function(a, b) {
  while (b != 0L) {
    temp <- b
    b <- a %% b
    a <- temp
  }
  a
}

# Floor division of doubles, and the matching remainder.
#
# R's `%/%` and `%%` put every double element through a correction step guarding
# against a quotient landing on the wrong side of an integer (~75x slower).
# A directly rounded quotient is already exact and needs no correction.
fdiv <- function(x, y) {
  floor(x / y)
}

fdivmod <- function(x, y) {
  div <- floor(x / y)
  list(div = div, mod = x - div * y)
}

# Evaluate time units in a calendar context
eval_cal <- function(expr, cal) {
  eval_tidy({{ expr }}, data = cal)
}

quo_add_dots <- function(quo, ...) {
  dots <- enquos(...)
  expr <- rlang::quo_get_expr(quo)
  rlang::quo_set_expr(
    quo,
    as.call(c(as.list(expr), list2(...)))
  )
}

# Similar to S7::new_S3_class to define S4 methods without depending on packages
new_S4_class <- function(className, package) {
  methods::newClassRepresentation(
    className = className,
    package = package
  )
}

# From S7:::topNamespaceName, required for new_time_unit wrapper
# Wrapper is required to patch https://github.com/RConsortium/S7/issues/609
topNamespaceName <- function(env = parent.frame()) {
  env <- topenv(env)
  if (!isNamespace(env)) {
    return()
  }
  as.character(getNamespaceName(env))
}

# Resolve a cli-style {?singular/plural} or {?zero/singular/plural} tokens
str_plural <- function(template, n) {
  m <- regexpr("\\{\\?[^}]*\\}", template, perl = TRUE)
  if (m[[1L]] == -1L) {
    return(rep_len(template, length(n)))
  }

  ml <- attr(m, "match.length")
  parts <- strsplit(substr(template, m + 2L, m + ml - 2L), "/", fixed = TRUE)[[
    1L
  ]]
  prefix <- substr(template, 1L, m - 1L)
  suffix <- substr(template, m + ml, nchar(template))

  # +s suffix if no plural form is provided
  if (length(parts) == 1L) {
    parts <- c(parts, paste0(parts, "s"))
  }

  idx <- if (length(parts) == 3L) {
    ifelse(n == 0, 1L, ifelse(n == 1L, 2L, 3L))
  } else {
    ifelse(n == 1L, 1L, 2L)
  }

  paste0(prefix, parts[idx], suffix)
}

test_that("year() converts dates correctly", {
  # Basic conversions
  expect_equal(format(year(as.Date("1970-01-01"))), "1970")
  expect_equal(format(year(as.Date("2025-12-16"))), "2025")
  
  # Edge case: last day of year
  expect_equal(format(year(as.Date("1970-12-31"))), "1970")
  expect_equal(format(year(as.Date("2025-12-31"))), "2025")
  
  # Edge case: first day of year
  expect_equal(format(year(as.Date("1971-01-01"))), "1971")
  expect_equal(format(year(as.Date("2026-01-01"))), "2026")
  
  # Vector consistency check over large range
  dates <- as.Date(0:25000, origin = "1970-01-01")
  diff <- format(year(dates)) == format(dates, "%Y")
  expect_true(all(diff))
  
  # Fractional years from dates
  y_start <- as.numeric(year(as.Date("2025-01-01"), discrete = FALSE))
  y_mid <- as.numeric(year(as.Date("2025-07-01"), discrete = FALSE))
  y_end <- as.numeric(year(as.Date("2025-12-31"), discrete = FALSE))
  
  expect_equal(y_start, 55)
  expect_equal(y_mid, 55.49589, tolerance = 1e-5)
  expect_equal(y_end, 55.99726, tolerance = 1e-5)
})

test_that("year() handles leap years correctly", {
  # Leap year boundaries
  expect_equal(format(year(as.Date("1972-02-29"))), "1972")
  expect_equal(format(year(as.Date("2000-02-29"))), "2000")
  expect_equal(format(year(as.Date("2024-02-29"))), "2024")
  
  # Century years (1900 not leap, 2000 is leap)
  expect_equal(format(year(as.Date("1900-02-28"))), "1900")
  expect_equal(format(year(as.Date("2000-02-29"))), "2000")
  expect_equal(format(year(as.Date("2100-02-28"))), "2100")
  
  # Day after leap day
  expect_equal(format(year(as.Date("1972-03-01"))), "1972")
  expect_equal(format(year(as.Date("2000-03-01"))), "2000")
  
  # Fractional years for leap year dates
  y_leap_start <- as.numeric(year(as.Date("2024-01-01"), discrete = FALSE))
  y_leap_mid <- as.numeric(year(as.Date("2024-07-01"), discrete = FALSE))
  
  # Leap year has 366 days, so midpoint fraction should be different
  expect_equal(y_leap_start, 54)
  expect_equal(y_leap_mid, 54.49727, tolerance = 1e-5)
})

test_that("year() handles discrete vs continuous time", {
  # Discrete should return integer chronons
  y_discrete <- year(Sys.time(), discrete = TRUE)
  expect_true(is.integer(as.integer(y_discrete)))
  
  # Continuous should return fractional chronons
  y_continuous <- year(Sys.time(), discrete = FALSE)
  y_numeric <- as.numeric(y_continuous)
  expect_true(y_numeric - floor(y_numeric) > 0)
  
  # Continuous should progress through the year
  t1 <- as.POSIXct("2025-01-01 00:00:00", tz = "UTC")
  t2 <- as.POSIXct("2025-07-01 00:00:00", tz = "UTC")
  y1 <- as.numeric(year(t1, discrete = FALSE))
  y2 <- as.numeric(year(t2, discrete = FALSE))
  expect_true(y2 > y1)
})

test_that("yearmonth() converts dates correctly", {
  # Basic conversions
  expect_equal(format(yearmonth(as.Date("1970-01-01"))), "1970 Jan")
  expect_equal(format(yearmonth(as.Date("1970-02-01"))), "1970 Feb")
  
  # Edge case: end of month
  expect_equal(format(yearmonth(as.Date("1970-01-31"))), "1970 Jan")
  expect_equal(format(yearmonth(as.Date("1970-02-28"))), "1970 Feb")
  
  # Edge case: start of next month
  expect_equal(format(yearmonth(as.Date("1970-02-01"))), "1970 Feb")
  
  # Vector consistency check
  dates <- as.Date(0:150000, origin = "1970-01-01")
  diff <- format(yearmonth(dates)) == format(dates, "%Y %b")
  expect_true(all(diff))
})

test_that("yearmonth() handles leap year months correctly", {
  # February in leap years
  expect_equal(format(yearmonth(as.Date("1972-02-29"))), "1972 Feb")
  expect_equal(format(yearmonth(as.Date("2000-02-29"))), "2000 Feb")
  expect_equal(format(yearmonth(as.Date("2024-02-29"))), "2024 Feb")
  
  # March after leap day
  expect_equal(format(yearmonth(as.Date("1972-03-01"))), "1972 Mar")
  expect_equal(format(yearmonth(as.Date("2000-03-01"))), "2000 Mar")
})

test_that("yearquarter() converts correctly", {
  # All quarters
  expect_equal(format(yearquarter(0L)), "1970 Q1")
  expect_equal(format(yearquarter(1L)), "1970 Q2")
  expect_equal(format(yearquarter(2L)), "1970 Q3")
  expect_equal(format(yearquarter(3L)), "1970 Q4")
  
  # Next year
  expect_equal(format(yearquarter(as.Date("1971-02-04"))), "1971 Q1")
  
  # Vector
  expect_length(yearquarter(0:7), 8)
})

test_that("chronon_divmod between days and months handles edge cases", {
  # Algorithm anchor date (0000-03-01 in proleptic Gregorian)
  # Days since epoch for key dates
  
  # Start of months
  expect_equal(format(yearmonth(as.Date("1970-01-01"))), "1970 Jan")
  expect_equal(format(yearmonth(as.Date("1970-12-01"))), "1970 Dec")
  expect_equal(format(yearmonth(as.Date("1971-01-01"))), "1971 Jan")
  
  # End of months (should still be in that month)
  expect_equal(format(yearmonth(as.Date("1970-01-31"))), "1970 Jan")
  expect_equal(format(yearmonth(as.Date("1970-04-30"))), "1970 Apr")
  expect_equal(format(yearmonth(as.Date("1970-12-31"))), "1970 Dec")
  
  # Leap year February
  expect_equal(format(yearmonth(as.Date("1972-02-29"))), "1972 Feb")
  expect_equal(format(yearmonth(as.Date("1972-03-01"))), "1972 Mar")
})

test_that("chronon_divmod between days and years handles edge cases", {
  # Year boundaries
  expect_equal(format(year(as.Date("1970-12-31"))), "1970")
  expect_equal(format(year(as.Date("1971-01-01"))), "1971")
  
  # Leap year transitions
  expect_equal(format(year(as.Date("1972-02-29"))), "1972")
  expect_equal(format(year(as.Date("1972-12-31"))), "1972")
  expect_equal(format(year(as.Date("1973-01-01"))), "1973")
  
  # Century transitions
  expect_equal(format(year(as.Date("1999-12-31"))), "1999")
  expect_equal(format(year(as.Date("2000-01-01"))), "2000")
  expect_equal(format(year(as.Date("2000-12-31"))), "2000")
  expect_equal(format(year(as.Date("2001-01-01"))), "2001")
})

test_that("reverse conversion from year to Date works correctly", {
  # Converting back to Date should give Jan 1
  years_seq <- year(0:500)
  dates_back <- as.Date(years_seq)
  
  # All should be January 1st
  days <- format(dates_back, "%m-%d")
  expect_true(all(days == "01-01"))
  
  # Years should match
  years_check <- as.integer(format(dates_back, "%Y"))
  expect_equal(years_check, 0:500)
})

test_that("chronon_cardinality handles months with variable days", {
  # This requires the `at` context parameter
  # Test that error is thrown when at is NULL
  expect_error(
    chronon_cardinality(cal_gregorian$day(1L), cal_gregorian$month(1L), at = NULL),
    "time context"
  )
  
  # Test that error is thrown for years without at
  expect_error(
    chronon_cardinality(cal_gregorian$day(1L), cal_gregorian$year(1L), at = NULL),
    "time context"
  )
})

test_that("chronon_cardinality(day, month) handles a negative-magnitude month chronon", {
  # Regression test: `y@n < 0` (e.g. from a descending `by = months(-1L)`
  # sequence step) used to pass a negative window size straight into
  # circsum(), which returns numeric(0) for size <= 0 and cascades to NA.
  # `at` is a block index in `y` units, so `at * y@n` recovers the correct
  # absolute month regardless of `y@n`'s sign.
  march_2020 <- (2020L - 1970L) * 12L + 3L - 1L # absolute month index, 0-based
  expect_equal(
    chronon_cardinality(cal_gregorian$day(1L), cal_gregorian$month(-1L), at = -march_2020),
    31 # March
  )
  expect_equal(
    chronon_cardinality(cal_gregorian$day(1L), cal_gregorian$month(-1L), at = -(march_2020 - 1L)),
    29 # February 2020 (leap year)
  )

  # Same absolute month, positive vs negative `y@n`, should agree
  expect_equal(
    chronon_cardinality(cal_gregorian$day(1L), cal_gregorian$month(1L), at = march_2020),
    chronon_cardinality(cal_gregorian$day(1L), cal_gregorian$month(-1L), at = -march_2020)
  )
})

test_that("chronon_divmod(day, month) round-trips through multi-month `to` chronons", {
  # Regression test: a stray `res <- chronon_cardinality(...)` reassignment
  # (an accidental rebase regression, originally a discarded validation-only
  # call) corrupted `div` for any `to@n != 1`, e.g. `by = "2 months"`.
  expect_equal(
    format(seq(date("2020-01-01"), length.out = 6, by = "2 months")),
    c("2020-01-01", "2020-03-01", "2020-05-01", "2020-07-01", "2020-09-01", "2020-11-01")
  )
})

test_that("chronon_divmod(day, month)'s `mod` accounts for whole months already elapsed in a wide window", {
  # Regression test: `mod` used to be `day + x_frac` unconditionally - the
  # offset into the *calendar* month containing `x`, which only coincides
  # with the offset into the `to@n`-month *window* when `to@n == 1`. Starting
  # mid-window (here, the 2nd month of a 2-month window) exposed the gap: the
  # whole first month of the window was silently dropped from `mod`, so every
  # subsequent step landed a full window early.
  expect_equal(
    format(seq(date("2020-02-15"), length.out = 4, by = "2 months")),
    c("2020-02-15", "2020-04-15", "2020-06-15", "2020-08-15")
  )
})

test_that("chronon_divmod(day, month)'s `div` is correct for a negative, multi-month `to`", {
  # Regression test: `div` used to be `fdiv(res, res_scale)` unconditionally,
  # which only agrees with chronon_cardinality()'s `at * n_months` window
  # anchor when `res_scale > 0`. For `res_scale < 0` and `abs(res_scale) > 1`
  # it picked a window running *backwards* from `res` instead of forwards
  # from `div * res_scale`, disagreeing with chronon_cardinality() (and thus
  # with clamping, which calls chronon_cardinality() using `div` as `at`).
  feb_1970 <- as.numeric(date("1970-02-01"))
  expect_equal(
    with(cal_gregorian, chronon_divmod(day(1L), month(-2L), feb_1970)),
    list(div = 0L, mod = 31)
  )
  # div/mod must always reconstruct `x` via chronon_cardinality()'s own
  # window convention: x = month_start_days(div * n) + mod.
  for (n in c(-13L, -7L, -2L, -1L, 1L, 2L, 7L, 13L)) {
    dm <- with(cal_gregorian, chronon_divmod(day(1L), month(n), feb_1970))
    expect_equal(
      with(cal_gregorian, chronon_divmod(month(n), day(1L), dm$div))$div + dm$mod,
      feb_1970,
      info = paste("n =", n)
    )
  }
})

test_that("chronon_cardinality/chronon_divmod(day, month) support month chronons >= 12", {
  # Previously hard-errored ("Month chronons >= 12 are not yet supported"):
  # circsum()'s per-window cost scaled with `abs(n_months)` and its leap-day
  # correction only ever accounted for a single February per window, both of
  # which broke down for a >= 12-month window. Replaced by a difference of
  # two O(1) `month_start_days()` calls, which has no such limit.

  # `at` is a block index in `month(13L)` units, so its window is the 13
  # calendar months [at * 13, at * 13 + 13). `at = 1` lands the window on
  # Feb 1971 - Feb 1972 inclusive, which contains *two* Februaries (one
  # non-leap, one leap - 1972 is a leap year), exercising the case the old
  # single-February assumption couldn't.
  expect_equal(
    chronon_cardinality(cal_gregorian$day(1L), cal_gregorian$month(13L), at = 1L),
    as.numeric(date("1972-03-01") - date("1971-02-01"))
  )
  # `at = 0`'s window (Jan 1970 - Jan 1971 inclusive) contains only one.
  expect_equal(
    chronon_cardinality(cal_gregorian$day(1L), cal_gregorian$month(13L), at = 0L),
    as.numeric(date("1971-02-01") - date("1970-01-01"))
  )

  # Round-trips through seq().
  expect_equal(
    format(seq(date("1970-01-01"), length.out = 3, by = months(13L))),
    c("1970-01-01", "1971-02-01", "1972-03-01")
  )
})

test_that("is_leap_year helper function works correctly", {
  # Standard leap years (divisible by 4)
  expect_true(is_leap_year(1972))
  expect_true(is_leap_year(2024))
  
  # Non-leap years
  expect_false(is_leap_year(1970))
  expect_false(is_leap_year(2025))
  
  # Century years (not leap unless divisible by 400)
  expect_false(is_leap_year(1900))
  expect_true(is_leap_year(2000))
  expect_false(is_leap_year(2100))
  
  # Edge case: year 2400
  expect_true(is_leap_year(2400))
})

test_that("conversion handles dates before Unix epoch", {
  # Negative days since epoch
  expect_equal(format(year(as.Date("1969-12-31"))), "1969")
  expect_equal(format(year(as.Date("1969-01-01"))), "1969")
  expect_equal(format(year(as.Date("1900-01-01"))), "1900")
  
  expect_equal(format(yearmonth(as.Date("1969-12-31"))), "1969 Dec")
  expect_equal(format(yearmonth(as.Date("1969-06-15"))), "1969 Jun")
})

test_that("cyclical_labels_format for months work correctly", {
  # The method should return abbreviated month names
  labels <- cyclical_labels_format(cal_gregorian$month(1L), cal_gregorian$year(1L), 0:11, label = TRUE, abbreviate = TRUE)
  expect_equal(labels, month.abb)
})

test_that("time_unit methods return correct strings", {
  expect_equal(time_unit_full(cal_gregorian$year(1L)), "year{?/s}")
  expect_equal(time_unit_abbr(cal_gregorian$year(1L)), "Y")
  
  expect_equal(time_unit_full(cal_gregorian$month(1L)), "month{?/s}")
  expect_equal(time_unit_abbr(cal_gregorian$month(1L)), "M")
  
  expect_equal(time_unit_full(cal_gregorian$day(1L)), "day{?/s}")
  expect_equal(time_unit_abbr(cal_gregorian$day(1L)), "D")
  
  expect_equal(time_unit_full(cal_gregorian$hour(1L)), "hour{?/s}")
  expect_equal(time_unit_abbr(cal_gregorian$hour(1L)), "h")
})

test_that("time_unit_plural produces correct singular and plural forms", {
  expect_equal(time_unit_plural(cal_gregorian$year(1L), 1L), "year")
  expect_equal(time_unit_plural(cal_gregorian$year(1L), 2L), "years")
  expect_equal(time_unit_plural(cal_gregorian$month(1L), 1L), "month")
  expect_equal(time_unit_plural(cal_gregorian$month(1L), 3L), "months")
  expect_equal(time_unit_plural(cal_gregorian$day(1L), 1L), "day")
  expect_equal(time_unit_plural(cal_gregorian$day(1L), 7L), "days")
})

test_that("formatting of fractional/continuous dates (discrete = FALSE) around March 31st", {
  t <- date(.POSIXct(1774949379), discrete = FALSE)
  expect_equal(format(t), "2026-03-31 39.6%")
})

test_that("chronon_divmod day->year is correct for sub-day times on Dec 31", {
  # Regression test: fractional day values on Dec 31 previously corrupted the
  # month-prime calculation (mp), causing year to roll over to the next year at
  # 14:24 instead of midnight.  All times on Dec 31 must return the correct year.

  # The old boundary: 14:24:00 on Dec 31 2007 would (wrongly) return 2008
  expect_equal(format(year(datetime("2007-12-31 14:23:59"))), "2007")
  expect_equal(format(year(datetime("2007-12-31 14:24:00"))), "2007")
  expect_equal(format(year(datetime("2007-12-31 14:24:01"))), "2007")

  # Full day of Dec 31 in a non-leap year
  expect_equal(format(year(datetime("2007-12-31 00:00:00"))), "2007")
  expect_equal(format(year(datetime("2007-12-31 23:59:59"))), "2007")

  # Full day of Dec 31 in a leap year
  expect_equal(format(year(datetime("2008-12-31 14:23:59"))), "2008")
  expect_equal(format(year(datetime("2008-12-31 14:24:00"))), "2008")
  expect_equal(format(year(datetime("2008-12-31 14:24:01"))), "2008")
  expect_equal(format(year(datetime("2008-12-31 23:59:59"))), "2008")

  # Midnight tick into the new year must return the new year
  expect_equal(format(year(datetime("2008-01-01 00:00:00"))), "2008")
  expect_equal(format(year(datetime("2025-01-01 00:00:00"))), "2025")

  # Vector: every hour of 2007-12-31 should be 2007
  hours <- seq(
    as.POSIXct("2007-12-31 00:00:00", tz = "UTC"),
    as.POSIXct("2007-12-31 23:00:00", tz = "UTC"),
    by = "hour"
  )
  expect_true(all(format(year(hours)) == "2007"))
})

test_that("chronon_divmod day->month is correct for sub-day times on Dec 31", {
  # Companion regression test for the day->month path, which shares the same
  # fractional-day bug.

  expect_equal(format(yearmonth(datetime("2007-12-31 14:23:59"))), "2007 Dec")
  expect_equal(format(yearmonth(datetime("2007-12-31 14:24:00"))), "2007 Dec")
  expect_equal(format(yearmonth(datetime("2007-12-31 14:24:01"))), "2007 Dec")
  expect_equal(format(yearmonth(datetime("2007-12-31 23:59:59"))), "2007 Dec")

  # Same checks for a leap-year December
  expect_equal(format(yearmonth(datetime("2008-12-31 14:24:00"))), "2008 Dec")
  expect_equal(format(yearmonth(datetime("2008-12-31 23:59:59"))), "2008 Dec")

  # Midnight tick into January
  expect_equal(format(yearmonth(datetime("2008-01-01 00:00:00"))), "2008 Jan")

  # Vector: every hour of 2007-12-31 should be 2007 Dec
  hours <- seq(
    as.POSIXct("2007-12-31 00:00:00", tz = "UTC"),
    as.POSIXct("2007-12-31 23:00:00", tz = "UTC"),
    by = "hour"
  )
  expect_true(all(format(yearmonth(hours)) == "2007 Dec"))
})

test_that("circsum works with basic examples", {
  # Example from documentation
  expect_equal(circsum(c(1, 2, 3, 4), 2), c(3, 7))
  expect_equal(circsum(c(1, 2, 3, 4, 5), 3), c(6, 10, 9, 8, 12))
  
  # Size equals step (default behavior)
  expect_equal(circsum(c(1, 2, 3, 4), 2, 2), c(3, 7))
  
  # Different step sizes
  expect_equal(circsum(c(1, 2, 3, 4), 2, 1), c(3, 5, 7, 5))
})

test_that("circsum works with month days", {
  monthdays <- c(31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)
  
  expect_equal(circsum(monthdays, 1L), monthdays)
  expect_equal(circsum(monthdays, 2L), c(59, 61, 61, 62, 61, 61))
  expect_equal(circsum(monthdays, 3L), c(90, 91, 92, 92))
  expect_equal(circsum(monthdays, 4L), c(120, 123, 122))
  expect_equal(circsum(monthdays, 9L), c(273, 273, 274, 275))
  expect_equal(circsum(monthdays, 13L), c(396, 393, 396, 395, 396, 395, 396, 396, 395, 396, 395, 396))
})

test_that("circsum handles edge cases", {
  # Empty vector
  expect_equal(circsum(numeric(0), 2), numeric(0))
  
  # Size of 1
  expect_equal(circsum(c(1, 2, 3), 1), c(1, 2, 3))
  
  # Size equals length
  expect_equal(circsum(c(1, 2, 3, 4), 4), c(10))
  
  # Size greater than length (wraps around)
  expect_equal(circsum(c(1, 2, 3), 5), c(9, 10, 11))
})

test_that("circsum handles invalid inputs", {
  expect_equal(circsum(c(1, 2, 3), 0), numeric(0))
  expect_equal(circsum(c(1, 2, 3), 2, 0), numeric(0))
})

test_that("circsum treats a negative step as reversed direction", {
  expect_equal(circsum(c(1, 2, 3, 4, 5), 1, -1), c(1, 5, 4, 3, 2))
  expect_equal(circsum(c(1, 2, 3), 1, -1), c(1, 3, 2))
  expect_equal(circsum(c(1, 2, 3), 2, -1), c(3, 4, 5))
})

test_that("circsum treats a negative size as a trailing (end-anchored) window", {
  # size = -k at position i sums x[i - k + 1 .. i] instead of x[i .. i + k - 1]
  # - the same window widths and sums as a positive size, just anchored at
  # the other end. With `step = 1`, this is exactly a trailing rolling sum.
  expect_equal(circsum(c(1, 2, 3, 4, 5), -2, 1), c(6, 3, 5, 7, 9))
  expect_equal(circsum(c(1, 2, 3, 4, 5), -2, 1), rev(circsum(rev(c(1, 2, 3, 4, 5)), 2, 1)))

  # size = 1 is its own start and end - sign of size doesn't matter
  expect_equal(circsum(c(1, 2, 3), -1, 1), circsum(c(1, 2, 3), 1, 1))
  expect_equal(circsum(c(1, 2, 3), -1, -1), circsum(c(1, 2, 3), 1, -1))

  # Negative size and step (the default) together: windows both trail their
  # position *and* are visited backwards.
  expect_equal(circsum(c(1, 2, 3, 4, 5), -2), circsum(c(1, 2, 3, 4, 5), -2, -2))
})

test_that("circsum step parameter works correctly", {
  # Step larger than size
  expect_equal(circsum(c(1, 2, 3, 4, 5), 2, 3), c(3, 9, 5, 6, 7))
  
  # Step of 1 with size 2
  expect_equal(circsum(c(1, 2, 3, 4), 2, 1), c(3, 5, 7, 5))
})

test_that("circsum shortcut: size=1, step=1 returns input vector", {
  # This tests the first shortcut path: returns x directly
  x <- c(1, 2, 3, 4, 5)
  expect_identical(circsum(x, 1, 1), x)
  
  # Single element
  expect_identical(circsum(10, 1, 1), 10)
  
  # Larger vector
  x <- 1:100
  expect_identical(circsum(x, 1, 1), x)
  
  # Non-integer values
  x <- c(1.5, 2.7, 3.2)
  expect_identical(circsum(x, 1, 1), x)
})

test_that("circsum shortcut: size=1, step>1 uses GCD-based indexing", {
  # This tests the second shortcut path
  
  # Step of 2 with vector length 4 (GCD=2, num_windows=2)
  expect_equal(circsum(c(1, 2, 3, 4), 1, 2), c(1, 3))
  
  # Step of 2 with vector length 5 (GCD=1, num_windows=5)
  expect_equal(circsum(c(1, 2, 3, 4, 5), 1, 2), c(1, 3, 5, 2, 4))
  
  # Step of 3 with vector length 6 (GCD=3, num_windows=2)
  expect_equal(circsum(c(10, 20, 30, 40, 50, 60), 1, 3), c(10, 40))
  
  # Step of 3 with vector length 7 (GCD=1, num_windows=7)
  expect_equal(circsum(c(1, 2, 3, 4, 5, 6, 7), 1, 3), c(1, 4, 7, 3, 6, 2, 5))
  
  # Step equals length (wraps back to start)
  expect_equal(circsum(c(1, 2, 3, 4), 1, 4), c(1))
  
  # Step of 5 with vector length 10 (GCD=5, num_windows=2)
  expect_equal(circsum(1:10, 1, 5), c(1, 6))
})

test_that("circsum shortcut: size=1 with various step sizes", {
  # Step larger than length
  expect_equal(circsum(c(1, 2, 3), 1, 5), c(1, 3, 2))
  
  # Prime number step with prime length
  expect_equal(circsum(c(10, 20, 30, 40, 50), 1, 3), c(10, 40, 20, 50, 30))
  
  # Step coprime with length
  expect_equal(circsum(c(1, 2, 3, 4, 5, 6), 1, 5), c(1, 6, 5, 4, 3, 2))
})

test_that("fdiv() and fdivmod() agree with %/% and %%", {
  x <- as.numeric(c(-1000:1000, -2^40, 2^40, -719468L, 719468L))

  for (y in c(1, 2, 5L, 7L, 12L, 60L, 146097L, 0.5, 1.5)) {
    expect_equal(fdiv(x, y), x %/% y, info = paste("divisor", y))

    dm <- fdivmod(x, y)
    expect_equal(dm$div, x %/% y, info = paste("divisor", y))
    expect_equal(dm$mod, x %% y, info = paste("divisor", y))
  }

  # Vectorised divisors (e.g. a context-dependent cardinality per element)
  y <- rep_len(c(28, 29, 30, 31), length(x))
  expect_equal(fdiv(x, y), x %/% y)
  expect_equal(fdivmod(x, y)$mod, x %% y)
})

test_that("coarse-to-fine conversion takes a monotone path", {
  quarter <- cal_gregorian$quarter(1L)
  day <- cal_gregorian$day(1L)

  # Routing quarters to days via years would make a quarter an even 1/4 of the
  # year (91.5 days in 2020) rather than the 91 days of January to March.
  path <- S7_graph_dispatch(chronon_divmod_graph(), quarter, day)
  expect_equal(
    vapply(path, function(unit) class(unit(1L))[[1L]], character(1L)),
    c("mixtime::tu_quarter", "mixtime::tu_month", "mixtime::tu_day")
  )

  # Quarter boundaries land exactly on the day they begin.
  starts <- as.Date(c("2020-01-01", "2020-04-01", "2020-07-01", "2020-10-01"))
  expect_equal(
    chronon_convert_impl(200:203, quarter, day, discrete = FALSE),
    as.numeric(starts)
  )
  # Interpolating within the granule is not the same as splitting the interval
  # it spans: half a quarter is 1.5 months, which is only the middle of the
  # quarter when its months are of equal length. Q3 2020 spans 92 days, so its
  # midpoint is 46 days in, while mid-August is a day-and-a-half later. This is
  # why the ordering proxy takes the mean of the two bounds instead.
  expect_equal(chronon_convert_impl(202.5, quarter, day, discrete = FALSE), 18490.5)
  expect_equal(mean(chronon_convert_impl(202:203, quarter, day, discrete = FALSE)), 18490)
})

test_that("units with no monotone path still convert", {
  # A week is neither finer nor coarser than a month, so the unrestricted
  # search is still needed to route between them (via days).
  expect_no_error(
    chronon_convert_impl(2611L, cal_isoweek$week(1L), cal_gregorian$month(1L), discrete = TRUE)
  )
})

#' Comparison operators for linear time (`mt_linear`)
#'
#' @description
#' Discrete linear time values represent a *closed interval* spanning their
#' chronon (e.g. `year(2020)` spans every instant from the start of 2020 to
#' the end of 2020), while continuous linear time values represent a single
#' instant. Comparing two `mt_linear` vectors therefore compares the
#' start/end instants of the (possibly zero-width) interval each value
#' represents:
#'
#' - `a == b` iff `start(a) == start(b)` and `end(a) == end(b)`
#' - `a < b` iff `end(a) < start(b)`
#' - `a > b` iff `start(a) > end(b)`
#' - `a <= b` iff `end(a) <= end(b)` (right-bound comparison)
#' - `a >= b` iff `start(a) >= start(b)` (left-bound comparison)
#'
#' This is **not** a total order: `<=`/`>=` are endpoint comparisons, not
#' shorthand for `(< or ==)`/`(> or ==)`, so it is possible for none of
#' `==`, `<`, `>` to hold between two values.
#'
#' If both operands are continuous (fractional chronons), or share an
#' identical chronon, the comparison simplifies to a direct numeric
#' comparison, since there is no interval to consider.
#'
#' @param e1,e2 `mt_linear` vectors (or values castable to one, such as
#'   plain numeric vectors sharing the other operand's chronon).
#'
#' @return A logical vector.
#'
#' @usage
#' \S4method{==}{mt_linear}(e1, e2)
#' \S4method{!=}{mt_linear}(e1, e2)
#' \S4method{<}{mt_linear}(e1, e2)
#' \S4method{<=}{mt_linear}(e1, e2)
#' \S4method{>}{mt_linear}(e1, e2)
#' \S4method{>=}{mt_linear}(e1, e2)
#'
#' @aliases ==,mt_linear-method !=,mt_linear-method <,mt_linear-method <=,mt_linear-method >,mt_linear-method >=,mt_linear-method
#' @name mt_linear-compare
NULL

linear_compare <- function(op, e1, e2) {
  if (!S7_inherits(e1, mt_linear)) e1 <- vec_cast(e1, e2)
  if (!S7_inherits(e2, mt_linear)) e2 <- vec_cast(e2, e1)

  x_chronon <- e1@chronon
  y_chronon <- e2@chronon
  xv <- S7_data(e1)
  yv <- S7_data(e2)

  if (identical(x_chronon, y_chronon)) {
    # Identical granules: no span to consider, compare the raw data directly
    x_start <- x_end <- xv
    y_start <- y_end <- yv
  } else {
    # Different granules: express both values as closed intervals
    # ([start, end]) in their common (finest shared) chronon, then compare
    # endpoints. Continuous values are zero-width intervals (start == end).
    common <- chronon_common_impl(list(x_chronon, y_chronon))
    x_start <- chronon_convert_impl(xv, x_chronon, common, discrete = is.integer(xv))
    y_start <- chronon_convert_impl(yv, y_chronon, common, discrete = is.integer(yv))

    x_end <- x_start
    if (is.integer(xv)) {
      x_end <- x_start + chronon_cardinality(common, x_chronon, at = xv) - 1
    }
    y_end <- y_start
    if (is.integer(yv)) {
      y_end <- y_start + chronon_cardinality(common, y_chronon, at = yv) - 1
    }
  }

  switch(
    op,
    "==" = x_start == y_start & x_end == y_end,
    "!=" = !(x_start == y_start & x_end == y_end),
    "<"  = x_end < y_start,
    ">"  = x_start > y_end,
    "<=" = x_end <= y_end,
    ">=" = x_start >= y_start
  )
}

method(`==`, list(mt_linear, class_any)) <- function(e1, e2) linear_compare("==", e1, e2)
method(`==`, list(class_any, mt_linear)) <- function(e1, e2) linear_compare("==", e1, e2)

method(`!=`, list(mt_linear, class_any)) <- function(e1, e2) linear_compare("!=", e1, e2)
method(`!=`, list(class_any, mt_linear)) <- function(e1, e2) linear_compare("!=", e1, e2)

method(`<`, list(mt_linear, class_any)) <- function(e1, e2) linear_compare("<", e1, e2)
method(`<`, list(class_any, mt_linear)) <- function(e1, e2) linear_compare("<", e1, e2)

method(`<=`, list(mt_linear, class_any)) <- function(e1, e2) linear_compare("<=", e1, e2)
method(`<=`, list(class_any, mt_linear)) <- function(e1, e2) linear_compare("<=", e1, e2)

method(`>`, list(mt_linear, class_any)) <- function(e1, e2) linear_compare(">", e1, e2)
method(`>`, list(class_any, mt_linear)) <- function(e1, e2) linear_compare(">", e1, e2)

method(`>=`, list(mt_linear, class_any)) <- function(e1, e2) linear_compare(">=", e1, e2)
method(`>=`, list(class_any, mt_linear)) <- function(e1, e2) linear_compare(">=", e1, e2)

#' Comparison operators for durations (`mt_duration`)
#'
#' @description
#' A duration is a scalar magnitude of time measured in a given chronon (e.g.
#' `days(3)`), with no reference to a point in time. Comparing two durations is
#' therefore a plain magnitude comparison, once both operands have been
#' expressed in a common chronon:
#'
#' - `a == b` iff the two magnitudes are equal in their common chronon
#' - `a < b`, `a <= b`, `a > b`, `a >= b` compare the magnitudes directly
#'
#' Unlike `mt_linear` comparison there is no interval/span to consider, so this
#' *is* a total order. When both operands already share a chronon the magnitudes
#' are compared as-is; otherwise both are scaled to their finest common chronon
#' (the same scaling used when combining durations arithmetically, see
#' `duration_combine()`).
#'
#' @param e1,e2 `mt_duration` vectors (or values castable to one, such as plain
#'   numeric vectors interpreted in the other operand's chronon).
#'
#' @return A logical vector.
#'
#' @usage
#' \S4method{==}{mt_duration}(e1, e2)
#' \S4method{!=}{mt_duration}(e1, e2)
#' \S4method{<}{mt_duration}(e1, e2)
#' \S4method{<=}{mt_duration}(e1, e2)
#' \S4method{>}{mt_duration}(e1, e2)
#' \S4method{>=}{mt_duration}(e1, e2)
#'
#' @aliases ==,mt_duration-method !=,mt_duration-method <,mt_duration-method <=,mt_duration-method >,mt_duration-method >=,mt_duration-method
#' @name mt_duration-compare
NULL

duration_compare <- function(op, e1, e2) {
  if (!S7_inherits(e1, mt_duration)) e1 <- vec_cast(e1, e2)
  if (!S7_inherits(e2, mt_duration)) e2 <- vec_cast(e2, e1)

  x_chronon <- e1@chronon
  y_chronon <- e2@chronon
  xd <- S7_data(e1)
  yd <- S7_data(e2)

  if (!identical(x_chronon, y_chronon)) {
    # Different chronons: scale both magnitudes to their finest common chronon
    # before comparing (mirrors `duration_combine()` in arithmetic.R).
    tu <- chronon_common_impl(list(x_chronon, y_chronon))
    xd <- xd * chronon_cardinality(tu, x_chronon)
    yd <- yd * chronon_cardinality(tu, y_chronon)
  }

  switch(
    op,
    "==" = xd == yd,
    "!=" = xd != yd,
    "<"  = xd < yd,
    ">"  = xd > yd,
    "<=" = xd <= yd,
    ">=" = xd >= yd
  )
}

method(`==`, list(mt_duration, class_any)) <- function(e1, e2) duration_compare("==", e1, e2)
method(`==`, list(class_any, mt_duration)) <- function(e1, e2) duration_compare("==", e1, e2)

method(`!=`, list(mt_duration, class_any)) <- function(e1, e2) duration_compare("!=", e1, e2)
method(`!=`, list(class_any, mt_duration)) <- function(e1, e2) duration_compare("!=", e1, e2)

method(`<`, list(mt_duration, class_any)) <- function(e1, e2) duration_compare("<", e1, e2)
method(`<`, list(class_any, mt_duration)) <- function(e1, e2) duration_compare("<", e1, e2)

method(`<=`, list(mt_duration, class_any)) <- function(e1, e2) duration_compare("<=", e1, e2)
method(`<=`, list(class_any, mt_duration)) <- function(e1, e2) duration_compare("<=", e1, e2)

method(`>`, list(mt_duration, class_any)) <- function(e1, e2) duration_compare(">", e1, e2)
method(`>`, list(class_any, mt_duration)) <- function(e1, e2) duration_compare(">", e1, e2)

method(`>=`, list(mt_duration, class_any)) <- function(e1, e2) duration_compare(">=", e1, e2)
method(`>=`, list(class_any, mt_duration)) <- function(e1, e2) duration_compare(">=", e1, e2)

#' Comparison operators for cyclical time (`mt_cyclical`)
#'
#' @description
#' A cyclical time value stores an absolute chronon count but *means* a position
#' within its cycle (e.g. `day_of_week()` means a weekday, not a particular
#' Wednesday). Comparison therefore reduces both operands to their position
#' within the cycle - the same reduction [format()] displays - and compares
#' those:
#'
#' - `day_of_week(date("2020-01-15")) == day_of_week(date("2020-01-22"))` is
#'   `TRUE`, since both are a Wednesday.
#' - `day_of_year(date("2020-01-15")) == day_of_year(date("2021-01-15"))` is
#'   `TRUE`, since both are the 15th day of their year.
#'
#' Both operands must share a `cycle`: a cycle is a modulus rather than a unit,
#' so there is no meaningful common cycle between (say) a weekday and a
#' day-of-year, and comparing them is an error. Differing *chronons* within a
#' shared cycle are reconciled as they are for [mt_linear][mt_linear-compare]:
#' both positions are expressed in the finest common chronon, and a discrete
#' value spans the closed interval of its chronon, so
#'
#' - `a == b` iff `start(a) == start(b)` and `end(a) == end(b)`
#' - `a < b` iff `end(a) < start(b)`, `a > b` iff `start(a) > end(b)`
#' - `a <= b` iff `end(a) <= end(b)`, `a >= b` iff `start(a) >= start(b)`
#'
#' As for `mt_linear`, this is not a total order when chronons differ. Ordering
#' follows the position within the cycle (so `Mon < Wed`); the cycle's wrap-around
#' is not treated as circular.
#'
#' @param e1,e2 `mt_cyclical` vectors sharing a cycle (or values castable to one,
#'   such as plain numeric vectors sharing the other operand's chronon).
#'
#' @return A logical vector.
#'
#' @usage
#' \S4method{==}{mt_cyclical}(e1, e2)
#' \S4method{!=}{mt_cyclical}(e1, e2)
#' \S4method{<}{mt_cyclical}(e1, e2)
#' \S4method{<=}{mt_cyclical}(e1, e2)
#' \S4method{>}{mt_cyclical}(e1, e2)
#' \S4method{>=}{mt_cyclical}(e1, e2)
#'
#' @aliases ==,mt_cyclical-method !=,mt_cyclical-method <,mt_cyclical-method <=,mt_cyclical-method >,mt_cyclical-method >=,mt_cyclical-method
#' @name mt_cyclical-compare
NULL

# Reduce absolute chronon counts to their position within `cycle`, matching the
# reduction format() displays: the integer position from chronon_parts(), plus
# the fraction carried by a continuous (double) time model, which chronon_parts()
# floors away for display but which distinguishes values here.
cyclical_position <- function(x, chronon, cycle) {
  x <- mt_cyclical(x, chronon = chronon, cycle = cycle)
  pos <- chronon_parts(x, cyclical = list(list(chronon, cycle)))$cyclical[[1L]]
  xd <- S7_data(x)
  if (is.double(xd)) {
    xd <- xd + tz_offset_impl(xd, attr(x, "chronon"))
    pos <- pos + (xd - floor(xd))
  }
  pos
}

cyclical_compare <- function(op, e1, e2) {
  if (!S7_inherits(e1, mt_cyclical)) e1 <- vec_cast(e1, e2)
  if (!S7_inherits(e2, mt_cyclical)) e2 <- vec_cast(e2, e1)

  # A cycle is a modulus, not a unit: there is no common cycle to reconcile to
  # (see `cycle_common()`), so differing cycles cannot be compared.
  cycle <- cycle_common(e1@cycle, e2@cycle)
  if (is.null(cycle)) {
    vctrs::stop_incompatible_op(op, e1, e2, details = cycle_incompatible_details)
  }

  x_chronon <- e1@chronon
  y_chronon <- e2@chronon
  xv <- S7_data(e1)
  yv <- S7_data(e2)

  if (identical(x_chronon, y_chronon)) {
    # Identical granules: no span to consider, compare the positions directly
    x_start <- x_end <- cyclical_position(xv, x_chronon, cycle)
    y_start <- y_end <- cyclical_position(yv, y_chronon, cycle)
  } else {
    # Different granules: express both positions as closed intervals ([start,
    # end]) in their common (finest shared) chronon, then compare endpoints,
    # mirroring `linear_compare()`.
    common <- chronon_common_impl(list(x_chronon, y_chronon))
    x_start <- cyclical_position(
      chronon_convert_impl(xv, x_chronon, common, discrete = is.integer(xv)),
      common, cycle
    )
    y_start <- cyclical_position(
      chronon_convert_impl(yv, y_chronon, common, discrete = is.integer(yv)),
      common, cycle
    )

    x_end <- x_start
    if (is.integer(xv)) {
      x_end <- x_start + chronon_cardinality(common, x_chronon, at = xv) - 1
    }
    y_end <- y_start
    if (is.integer(yv)) {
      y_end <- y_start + chronon_cardinality(common, y_chronon, at = yv) - 1
    }
  }

  switch(
    op,
    "==" = x_start == y_start & x_end == y_end,
    "!=" = !(x_start == y_start & x_end == y_end),
    "<"  = x_end < y_start,
    ">"  = x_start > y_end,
    "<=" = x_end <= y_end,
    ">=" = x_start >= y_start
  )
}

method(`==`, list(mt_cyclical, class_any)) <- function(e1, e2) cyclical_compare("==", e1, e2)
method(`==`, list(class_any, mt_cyclical)) <- function(e1, e2) cyclical_compare("==", e1, e2)

method(`!=`, list(mt_cyclical, class_any)) <- function(e1, e2) cyclical_compare("!=", e1, e2)
method(`!=`, list(class_any, mt_cyclical)) <- function(e1, e2) cyclical_compare("!=", e1, e2)

method(`<`, list(mt_cyclical, class_any)) <- function(e1, e2) cyclical_compare("<", e1, e2)
method(`<`, list(class_any, mt_cyclical)) <- function(e1, e2) cyclical_compare("<", e1, e2)

method(`<=`, list(mt_cyclical, class_any)) <- function(e1, e2) cyclical_compare("<=", e1, e2)
method(`<=`, list(class_any, mt_cyclical)) <- function(e1, e2) cyclical_compare("<=", e1, e2)

method(`>`, list(mt_cyclical, class_any)) <- function(e1, e2) cyclical_compare(">", e1, e2)
method(`>`, list(class_any, mt_cyclical)) <- function(e1, e2) cyclical_compare(">", e1, e2)

method(`>=`, list(mt_cyclical, class_any)) <- function(e1, e2) cyclical_compare(">=", e1, e2)
method(`>=`, list(class_any, mt_cyclical)) <- function(e1, e2) cyclical_compare(">=", e1, e2)

# Ordering and de-duplication of mixtime vectors
method(xtfrm, class_mixtime) <- function(x) {
  xtfrm(vec_proxy_order(x))
}

method(unique, class_mixtime) <- function(x, incomparables = FALSE, ...) {
  vctrs::vec_unique(x)
}

method(duplicated, class_mixtime) <- function(x, incomparables = FALSE, ...) {
  vctrs::vec_duplicate_id(x) != seq_along(x)
}

method(anyDuplicated, class_mixtime) <- function(x, incomparables = FALSE, ...) {
  match(TRUE, duplicated(x), nomatch = 0L)
}

#' Tolerant comparison of mixtime time values
#'
#' @description
#' [vecvec::class_vecvec]'s `all.equal()` method (which `mixtime` inherits,
#' since a mixtime vector is a `vecvec`) handles length/`NA` mismatches and
#' groups elements by their underlying storage slot, then compares each
#' group's *raw* values with `all.equal()`. That raw comparison has no notion
#' of chronon: two elements that are `==` (e.g. `days(1)` and `hours(24)`, or
#' a `yearmonth` and an equivalent-instant `yearweek`) can end up stored with
#' different chronons or magnitudes, and so are wrongly reported as unequal.
#'
#' This method fixes that at the per-slot level: elements already `==` (which
#' is chronon-aware, see [mt_linear-compare]/[mt_duration-compare]) count as
#' equal outright; for the rest, the discrepancy is measured as the duration
#' between them in their common chronon. A time point has no "typical
#' magnitude" to scale a *relative* tolerance against the way a plain number
#' does, so - unlike [base::all.equal.numeric()] - `tolerance` is always
#' absolute, in (possibly fractional) chronon units.
#'
#' @param target,current `mt_time` vectors of the same length to compare (as
#'   passed down by [vecvec::class_vecvec]'s `all.equal()`).
#' @param tolerance Numeric tolerance, as an absolute number of `target`'s
#'   chronon units. Defaults to `sqrt(.Machine$double.eps)`, as for
#'   [base::all.equal()].
#' @param ... Ignored.
#'
#' @return `TRUE` if `target` and `current` are equal within `tolerance`,
#'   otherwise a string describing the discrepancy.
#'
#' @examples
#' all.equal(yearmonth(0), yearmonth(0))
#' all.equal(yearmonth(0), yearmonth(1))
#' all.equal(days(1), hours(24))
#'
#' @keywords internal
#' @method all.equal mixtime::mt_time
#' @export
`all.equal.mixtime::mt_time` <- function(
  target, current, tolerance = sqrt(.Machine$double.eps), ...
) {
  # `class_vecvec`'s all.equal() only filters *sparse* missingness (an NA
  # vecvec index); an NA stored inline within a slot's own data (as here)
  # still reaches us and needs handling directly - mirroring the NA check in
  # `all.equal.numeric()`.
  na_t <- is.na(target)
  na_c <- tryCatch(is.na(current), error = function(e) NULL)
  if (is.null(na_c)) {
    return(sprintf(
      "target is %s, current is %s",
      class(target)[[1L]], class(current)[[1L]]
    ))
  }
  if (!identical(na_t, na_c)) {
    return(sprintf(
      "'is.NA' value mismatch: %d in current, %d in target",
      sum(na_c), sum(na_t)
    ))
  }

  keep <- !na_t
  eq <- tryCatch(target[keep] == current[keep], error = function(e) NULL)
  if (is.null(eq)) {
    return(sprintf(
      "target is %s, current is %s",
      class(target)[[1L]], class(current)[[1L]]
    ))
  }
  if (all(eq)) return(TRUE)

  d <- target[keep][!eq] - current[keep][!eq]
  xy <- mean(abs(as.double(d)))
  if (xy > tolerance) {
    sprintf("Mean absolute difference: %s", format(xy))
  } else {
    TRUE
  }
}

index_valid <- S7::new_external_generic("tsibble", "index_valid", "x")
method(index_valid, class_mixtime) <- function(x) TRUE

interval_pull <- S7::new_external_generic("tsibble", "interval_pull", "x")
method(interval_pull, class_mixtime) <- function(x) {
  intvls <- lapply(x@x, function(x) tsibble::interval_pull(x))
  x <- new_rcrd(
    vec_rbind(!!!lapply(intvls, new_data_frame)),
    .regular = any(vapply(intvls, attr, logical(1L), which = ".regular")),
    class = c("mixtime_interval", "interval")
  )
}

method(interval_pull, mt_time) <- function(x) {
  chronon <- attr(x, "chronon")

  interval <- list(chronon@n)
  names(interval) <- time_unit_plural(chronon, 1L)

  rlang::inject(tsibble::new_interval(!!!interval))
}

#' @exportS3Method base::format mixtime_interval
format.mixtime_interval <- function(x, ...) {
  fmt <- vapply(
    vctrs::new_rcrd(unclass(x)),
    fmt_interval,
    # function(y) format(`class<-`(y, "interval")),
    character(1L)
  )
  # tsibble requires length 1 interval formatting
  paste(unique(fmt), collapse = ", ")
}

fmt_interval <- function (x, ...) {
  if (!attr(x, ".regular"))
    return("!")
  if (all(vec_proxy(x) == 0))
    return("?")
  n <- n_fields(x)
  micro <- ifelse(cli::is_utf8_output(), "\u00b5s", "us")
  defaults <- vec_c("Y", "Q", "M", "W", "D", "h", "m", "s",
                    "ms", micro, "ns", "")
  n_defaults <- vec_size(defaults)
  misc <- if (n > n_defaults)
    vec_slice(fields(x), (n_defaults + 1):n)
  else NULL
  fmt_names <- vec_c(defaults, misc)
  val <- unlist(unclass(x), use.names = FALSE)
  paste0(val[val != 0], fmt_names[val != 0], collapse = " ")
}

#' @exportS3Method base::print mixtime_interval
print.mixtime_interval <- function(x, ...) cat(format(x, ...))

#' @export
vec_proxy.mixtime_interval <- function(x, ...) {
  # Bypass tsibble:::unknown_interval length 1 checks
  apply(new_data_frame(x), 2, sum, simplify = FALSE)
}

#' @export
vec_cast.mixtime.yearweek <- function(x, to, ...) {
  yearweek(as.Date(x))
}

#' @export
vec_cast.mixtime.yearmonth <- function(x, to, ...) {
  yearmonth(as.Date(x))
}

#' @export
vec_cast.mixtime.yearquarter <- function(x, to, ...) {
  yearquarter(as.Date(x))
}

#' Extract timezone from an object
#'
#' Generic function to extract the timezone from objects that have timezone information.
#'
#' @param x An object with timezone information.
#' @param ... Additional arguments passed to methods.
#'
#' @return A character vector representing the timezone of each time point
#'   (e.g., "America/New_York", "UTC").
#'
#' @examples
#' tz_name(Sys.time())
#' tz_name(as.POSIXct("2024-06-15 12:00:00", tz = "America/New_York"))
#'
#' @export
tz_name <- S7::new_generic("tz_name", "x")
S7::method(tz_name, class_mixtime) <- function(x) {
  as.character(vecvec::vecvec_apply(x, tz_name))
}
S7::method(tz_name, mt_time) <- function(x) {
  rep_len(tz_name(attr(x, "chronon")), length(x))
}
S7::method(tz_name, mt_tz_unit) <- function(x) x@tz
S7::method(tz_name, S7::class_POSIXt) <- function(x) {
  rep_len(attr(x, "tzone") %||% NA_character_, length(x))
}
S7::method(tz_name, S7::class_any) <- function(x) {
  rep_len(NA_character_, length(x))
}

#' Get timezone offset
#'
#' Returns the UTC offset for a given datetime in its specified timezone.
#'
#' @param x A time class coercible to POSIXt with an associated time zone.
#' @param ... Additional arguments passed to methods.
#'
#' @return A `mixtime` duration vector of offsets from UTC in the same chronon
#'   (e.g. seconds for POSIXt, days for dates, etc.)
#' @export
#'
#' @examples
#' tz_offset(as.POSIXct(Sys.time(), tz = Sys.timezone()))
#' tz_offset(as.POSIXct("2024-06-15 12:00:00", tz = "America/New_York"))
tz_offset <- S7::new_generic("tz_offset", "x")
S7::method(tz_offset, S7::class_POSIXt) <- function(
  x,
  tz = tz_name(time_chronon(x)),
  ...
) {
  duration(
    get_tz_offset(x, tz),
    chronon = cal_gregorian$second(1L, tz = tz[1L])
  )
}
S7::method(tz_offset, S7::class_Date) <- function(x, ...) {
  duration(rep.int(0, length(x)), chronon = cal_gregorian$day(1L))
}
method(tz_offset, class_mixtime) <- vecvec::vecvec_apply_fn(
  tz_offset,
  SIMPLIFY = TRUE
)
method(tz_offset, mt_time) <- function(
  x,
  tz = tz_name(attr(x, "chronon")),
  ...
) {
  chronon <- attr(x, "chronon")
  duration(tz_offset_impl(as.numeric(x), chronon, tz), chronon = chronon)
}

tz_offset_impl <- function(x, chronon, tz = tz_name(chronon)) {
  offset_s <- rep(0L, length(x))

  # Naive and UTC time has no time zone offsets
  if (is.na(tz) || tz == "UTC") {
    return(offset_s)
  }

  tu_s <- cal_time_civil$second(1L, tz = "UTC")
  time_s <- chronon_convert_impl(x, chronon, tu_s, FALSE, "UTC")
  offset_s <- get_tz_offset(as.double(time_s), tz)
  nz_offset <- !is.na(offset_s) & offset_s != 0
  if (!any(nz_offset)) {
    return(rep(0, length(x)))
  }
  offset_s[nz_offset] <- offset_s[nz_offset] *
    chronon_cardinality(
      chronon,
      tu_s,
      time_s[nz_offset]
    )
  offset_s
}

# Shift a `chronon`-native numeric value from UTC into the local wall-clock
# domain of `tz`, expressed as if it were UTC. `chronon_divmod()`,
# `chronon_decompose()`/`chronon_recompose()`, and `chronon_convert_impl()`
# (when called with `tz = "UTC"`) are all tz-blind and work in this shifted
# domain - see `time_round_impl()` and `seq.mixtime::mt_time()`.
tz_to_local <- function(x, chronon, tz) {
  x + tz_offset_impl(x, chronon, tz = tz)
}

# Undo `tz_to_local()`, recovering the true (UTC) `chronon` value. A second
# pass re-evaluates the offset at the candidate raw instant rather than
# reusing the forward offset, since a value that falls in a DST gap/overlap
# can have a different offset there than at its own local-shifted value.
#
# TODO(perf): this approximates gap/overlap resolution with two
# `tz_offset_impl()` round trips (each its own `chronon_convert_impl()` +
# `get_tz_offset()` C call). The vendored `date`/`tzdb` library already has a
# purpose-built, single-lookup primitive for exactly this -
# `date::time_zone::to_sys(local_time, choose::earliest|latest)` (see
# tzdb's `include/date/tz.h`) - which is what `as.POSIXct()` itself relies on.
# Wiring up a new cpp11-registered resolver around that would both halve the
# lookup cost and replace this two-sample approximation with the library's
# real resolution logic, at the cost of touching `src/` (new registered
# function, regenerated `src/cpp11.cpp`/`R/cpp11.R` bindings, recompile).
# Left as pure R for now since the 2x cost is negligible at realistic
# `seq()`/`time_round()` sizes.
tz_to_utc <- function(x, chronon, tz, discrete = FALSE, boundary = TRUE) {
  tzo <- tz_offset_impl(
    x - tz_offset_impl(x, chronon, tz = tz),
    chronon,
    tz = tz
  )

  if (discrete) {
    if (boundary) {
      # A boundary conversion (`chronon_convert_impl()`'s multi-hop path) has
      # already resolved which `chronon` bucket the local-shifted value falls
      # into via a real divmod - the offset is spent, so only a leftover
      # *whole*-unit remainder (never seen with real-world offsets, but kept
      # for safety) still needs correcting; floor(x) - trunc(tzo) keeps that
      # correction separate from the local-bucket alignment so a fractional
      # offset can't borrow across a unit boundary and land a whole unit off.
      nudge <- 8 * .Machine$double.eps * pmax(abs(x), 1)
      as.integer(floor(x + nudge) - trunc(tzo))
    } else {
      # No divmod happened between the matching tz_to_local() shift and this
      # call (e.g. a same-chronon or pure-ratio conversion), so the offset is
      # still fully outstanding in `x` and must be subtracted whole, before
      # flooring - not floored away first and only partly corrected after.
      shifted <- x - tzo
      nudge <- 8 * .Machine$double.eps * pmax(abs(shifted), 1)
      as.integer(floor(shifted + nudge))
    }
  } else {
    x - tzo
  }
}

#' Get timezone abbreviation
#'
#' Returns the timezone abbreviation (e.g., "EST", "PDT") for a given datetime
#' in its specified timezone.
#'
#' @param x A POSIXct datetime object or something coercible to POSIXct.
#'   The timezone is extracted from this object.
#' @param tz A character vector of timezones to abbreviate at time point `x`.
#'
#' @return A character vector of timezone abbreviations.
#'
#' @examples
#' tz_abbreviation(Sys.time())
#' tz_abbreviation(as.POSIXct("2024-01-15 12:00:00", tz = "America/New_York"))
#'
#' @export
tz_abbreviation <- function(x, tz = tz_name(x)) {
  tz_abbr <- character(length(x))

  # If tz is empty, then the object has a naive local timezone
  tz_given <- !is.na(tz) & !is.na(x)

  # TODO: Handle timezone changes within chronon using [before]/[after]
  if (any(tz_given)) {
    tz_abbr[tz_given] <- get_tz_abbreviation(
      as.double(chronon_convert(x[tz_given], cal_time_civil$second(1L))),
      tz[tz_given]
    )
  }

  tz_abbr
}

#' Get timezone transitions
#'
#' Returns all timezone transitions (e.g., daylight saving time changes) that
#' occur between two datetimes. The timezone is taken from the start datetime.
#'
#' @param start A POSIXct datetime object or something coercible to POSIXct,
#'   representing the start of the time range. The timezone is extracted from this object.
#' @param end A POSIXct datetime object or something coercible to POSIXct,
#'   representing the end of the time range.
#'
#' @return A data frame with columns:
#'   * `time`: A `mixtime` linear time point (continuous, UTC seconds) giving
#'     the instant of the transition.
#'   * `offset_before`, `offset_after`: `mixtime` durations (UTC seconds)
#'     giving the UTC offset immediately before and after the transition.
#'
#' @examples
#' # Get all DST transitions in 2024 for New York
#' tz_transitions(
#'   as.POSIXct("2024-01-01", tz = "America/New_York"),
#'   as.POSIXct("2024-12-31", tz = "America/New_York")
#' )
#'
#' @export
tz_transitions <- function(start, end) {
  start <- as.POSIXct(start)
  tz <- attr(start, "tzone") %||% Sys.timezone()
  end <- as.double(as.POSIXct(end))
  start <- as.double(start)

  transitions <- get_tz_transitions(start, end, tz)

  transitions$time <- linear_time(
    transitions$time,
    chronon = cal_gregorian$second(1L, tz = "UTC"),
    discrete = FALSE
  )
  transitions$offset_before <- duration(
    transitions$offset_before,
    chronon = cal_gregorian$second(1L, tz = "UTC")
  )
  transitions$offset_after <- duration(
    transitions$offset_after,
    chronon = cal_gregorian$second(1L, tz = "UTC")
  )

  transitions
}

test_that("get_tz_offset returns correct offset values", {
  # Test with a known timezone and time
  winter <- as.POSIXct("2024-01-15 12:00:00", tz = "America/New_York")
  summer <- as.POSIXct("2024-07-15 12:00:00", tz = "America/New_York")

  winter_offset <- get_tz_offset(winter, "America/New_York")
  summer_offset <- get_tz_offset(summer, "America/New_York")

  # EST is UTC-5 (-18000 seconds), EDT is UTC-4 (-14400 seconds)
  expect_equal(winter_offset, -18000)
  expect_equal(summer_offset, -14400)

  # Test with UTC
  utc_time <- as.POSIXct("2024-01-15 12:00:00", tz = "UTC")
  expect_equal(get_tz_offset(utc_time, "UTC"), 0)

  # Test with positive offset timezone
  tokyo_time <- as.POSIXct("2024-01-15 12:00:00", tz = "Asia/Tokyo")
  expect_equal(get_tz_offset(tokyo_time, "Asia/Tokyo"), 32400) # UTC+9
})

test_that("get_tz_offset works with vector input", {
  times <- as.POSIXct(
    c("2024-01-15 12:00:00", "2024-07-15 12:00:00"),
    tz = "America/New_York"
  )
  offsets <- get_tz_offset(times, "America/New_York")

  expect_length(offsets, 2)
  expect_equal(offsets[1], -18000) # Winter
  expect_equal(offsets[2], -14400) # Summer
})

test_that("get_tz_abbreviation returns correct abbreviations", {
  winter <- as.POSIXct("2024-01-15 12:00:00", tz = "America/New_York")
  summer <- as.POSIXct("2024-07-15 12:00:00", tz = "America/New_York")

  expect_equal(get_tz_abbreviation(winter, "America/New_York"), "EST")
  expect_equal(get_tz_abbreviation(summer, "America/New_York"), "EDT")

  # Test UTC
  utc_time <- as.POSIXct("2024-01-15 12:00:00", tz = "UTC")
  expect_equal(get_tz_abbreviation(utc_time, "UTC"), "UTC")
})

test_that("get_tz_abbreviation works with vector input", {
  times <- as.POSIXct(
    c("2024-01-15 12:00:00", "2024-07-15 12:00:00"),
    tz = "America/New_York"
  )
  abbrevs <- get_tz_abbreviation(times, "America/New_York")

  expect_length(abbrevs, 2)
  expect_equal(abbrevs[1], "EST")
  expect_equal(abbrevs[2], "EDT")
})

test_that("get_tz_transitions returns DST transitions", {
  start <- as.POSIXct("2024-01-01", tz = "America/New_York")
  end <- as.POSIXct("2024-12-31", tz = "America/New_York")

  transitions <- get_tz_transitions(start, end, "America/New_York")

  # Should have 2 transitions in 2024 (spring forward, fall back)
  expect_s3_class(transitions, "data.frame")
  expect_equal(nrow(transitions), 2)
  expect_named(transitions, c("time", "offset_before", "offset_after"))

  # Check that offsets change
  expect_equal(transitions$offset_before[1], -18000) # EST
  expect_equal(transitions$offset_after[1], -14400) # EDT
  expect_equal(transitions$offset_before[2], -14400) # EDT
  expect_equal(transitions$offset_after[2], -18000) # EST
})

test_that("get_tz_transitions works with Southern Hemisphere DST", {
  start <- as.POSIXct("2024-01-01", tz = "Australia/Melbourne")
  end <- as.POSIXct("2024-12-31", tz = "Australia/Melbourne")

  transitions <- get_tz_transitions(start, end, "Australia/Melbourne")

  expect_s3_class(transitions, "data.frame")
  expect_equal(nrow(transitions), 2)

  # Southern hemisphere: transitions are reversed (fall in April, spring in October)
  expect_equal(transitions$offset_before[1], 39600) # AEDT (UTC+11)
  expect_equal(transitions$offset_after[1], 36000) # AEST (UTC+10)
})

test_that("get_tz_transitions returns empty for no transitions", {
  start <- as.POSIXct("2024-01-01", tz = "UTC")
  end <- as.POSIXct("2024-12-31", tz = "UTC")

  transitions <- get_tz_transitions(start, end, "UTC")

  expect_s3_class(transitions, "data.frame")
  expect_equal(nrow(transitions), 0)
})

test_that("timezone functions handle invalid timezone names gracefully", {
  time <- as.POSIXct("2024-01-15 12:00:00", tz = "UTC")

  # These should error with invalid timezone names
  expect_error(get_tz_offset(time, "Invalid/Timezone"))
  expect_error(get_tz_abbreviation(time, "Invalid/Timezone"))
  expect_error(get_tz_transitions(time, time, "Invalid/Timezone"))
})

test_that("converting zoned time to naive time strips the timezone", {
  zoned <- datetime("2024-04-07 02:00:00", tz = "Australia/Melbourne")
  naive <- datetime(zoned, tz = NA)

  expect_true(is.na(tz_name(naive)))
  # Wall-clock time is preserved
  expect_equal(format(naive), "2024-04-07 02:00:00")
})

test_that("converting zoned time to naive preserves wall-clock time across DST gap", {
  # Melbourne falls back at 03:00 AEDT → 02:00 AEST on 2024-04-07.
  # The ambiguous 02:xx hour appears twice; both wall-clock values are kept as-is.
  from <- datetime("2024-04-07 00:00:00", tz = "Australia/Melbourne")
  mel_dst <- seq(from, length.out = 10, by = hours(1L))
  naive <- datetime(mel_dst, tz = NA)

  expect_true(all(is.na(tz_name(naive))))
  expect_length(naive, 10)
  # The two 02:00 entries (one AEDT, one AEST) both map to 02:00 wall-clock
  expect_equal(sum(format(naive) == "2024-04-07 02:00:00"), 2L)
})

test_that("converting naive time to zoned time errors", {
  zoned <- datetime("2024-01-15 12:00:00", tz = "America/New_York")
  naive <- datetime(zoned, tz = NA)

  expect_error(
    datetime(naive, tz = "America/New_York"),
    regexp = "timezone-naive"
  )
})

test_that("converting naive time to zoned errors for any target timezone", {
  naive <- datetime(as.POSIXct("2024-07-15 09:00:00", tz = "UTC"), tz = NA)

  expect_error(datetime(naive, tz = "UTC"))
  expect_error(datetime(naive, tz = "Asia/Tokyo"))
  expect_error(datetime(naive, tz = "Europe/Berlin"))
})

test_that("get_tz_transitions handles time ranges with no transitions", {
  # Short time range with no transitions
  start <- as.POSIXct("2024-06-01", tz = "America/New_York")
  end <- as.POSIXct("2024-06-15", tz = "America/New_York")

  transitions <- get_tz_transitions(start, end, "America/New_York")

  expect_s3_class(transitions, "data.frame")
  expect_equal(nrow(transitions), 0)
})

test_that("the common chronon keeps a timezone the inputs agree on", {
  # The common chronon is found as the greatest lower bound of the inputs, which
  # is built from the granule's class alone. Properties describing the time
  # rather than its granularity (such as the timezone) must be carried across,
  # otherwise combining two Melbourne times gives a timezone-naive result which
  # can no longer be converted back to either input.
  tz <- "Australia/Melbourne"
  secs <- datetime(as.POSIXct("2015-01-01 00:00:00", tz = tz))
  hrs <- linear_time(
    as.POSIXct("2015-01-01 00:00:00", tz = tz),
    cal_gregorian$hour(1L)
  )

  expect_equal(tz_name(chronon_common(c(secs, secs))), tz)
  expect_equal(tz_name(chronon_common(c(secs, hrs))), tz)

  # Combining no longer drops the timezone, whichever way round it is done.
  expect_equal(unique(tz_name(c(secs, hrs))), tz)
  expect_equal(unique(tz_name(c(hrs, secs))), tz)
})

test_that("the common chronon of disagreeing known timezones is UTC", {
  # A common chronon that claimed one of the input timezones would silently
  # move the other input's time points. But both zones are real, resolvable
  # zones, so the disagreement can be reconciled at UTC instead of discarding
  # the timezone entirely (which left the common chronon unformattable).
  melb <- datetime(as.POSIXct(
    "2015-01-01 00:00:00",
    tz = "Australia/Melbourne"
  ))
  utc <- datetime(as.POSIXct("2015-01-01 00:00:00", tz = "UTC"))
  naive <- datetime(melb, tz = NA)

  expect_equal(tz_name(chronon_common(c(melb, utc))), "UTC")
  # Naive time cannot be represented in a timezone, so a naive input keeps the
  # common chronon naive rather than being adopted into the other's zone.
  expect_true(is.na(tz_name(chronon_common(c(melb, naive)))))
})

test_that("combining a zoned time with a naive one is consistently wall-clock", {
  # Regression test: `c()` on a zoned + a naive element with the same
  # wall-clock reading used to silently convert the zoned element to its true
  # (UTC) absolute instant while leaving the naive element untouched, landing
  # both in the same combined chronon (naive, since they disagree) but on two
  # different numeric bases.
  zoned <- linear_time(
    datetime("2024-04-07 02:00:00", tz = "Australia/Melbourne"),
    hour(1L, tz = "Australia/Melbourne")
  )
  naive <- linear_time(datetime("2024-04-07 02:00:00"), hour(1L))

  combined <- c(zoned, naive)

  expect_true(is.na(tz_name(chronon_common(combined))))
  expect_equal(as.numeric(vecvec::unvecvec(combined)), c(475682, 475682))
})

test_that("chronon_common() of mixed known timezones formats without error", {
  # Regression test for tz-merge.md: mixing two real (but different) known
  # zones used to collapse the common chronon to naive, and formatting
  # through that naive chronon directly (as ggtime's axis breaks do) crashed.
  london <- datetime(seq(
    as.POSIXct("2026-07-30 06:00:00", tz = "Europe/London"),
    by = "1 hour",
    length.out = 3
  ))
  melbourne <- datetime(seq(
    as.POSIXct("2026-07-30 06:00:00", tz = "Australia/Melbourne"),
    by = "1 hour",
    length.out = 3
  ))
  combined <- c(london, melbourne)

  expect_equal(tz_name(chronon_common(combined)), "UTC")

  x <- vecvec::unvecvec(mixtime(
    as.numeric(combined),
    chronon = chronon_common(combined),
    discrete = FALSE
  ))
  expect_no_error(format(x))
})

test_that("tz_offset() of a POSIXct states one offset per element", {
  # A POSIXct carries a single timezone for the whole vector, unlike the
  # per-element timezone that `tz_name(time_chronon(x))` reports.
  x <- as.POSIXct(c("2024-01-01", "2024-06-01"), tz = "Australia/Melbourne")
  offset <- tz_offset(x)

  expect_true(all(time_is_duration(offset)))
  expect_equal(as.numeric(offset), c(39600, 36000))
  expect_equal(tz_name(chronon_common(offset)), "Australia/Melbourne")
})

test_that("tz_transitions() of a range with no transitions is an empty table", {
  transitions <- tz_transitions(
    as.POSIXct("2024-06-01", tz = "America/New_York"),
    as.POSIXct("2024-06-15", tz = "America/New_York")
  )

  expect_equal(nrow(transitions), 0)
  expect_named(transitions, c("time", "offset_before", "offset_after"))
})

test_that("tz_to_utc() discrete rounding keeps fractional-offset midnights on the right day", {
  # Regression test for the discrete branch of tz_to_utc(): naively flooring
  # `x - tzo` (rather than `floor(x) - trunc(tzo)`) lets the fractional parts
  # of the local-shifted value and the re-derived offset interact and borrow
  # across a day boundary, silently landing on the wrong calendar day. This
  # only shows up for zones with a fractional-hour UTC offset, right near
  # local midnight, so whole-hour zones (as used elsewhere in this file)
  # don't exercise it.

  # Asia/Kolkata: UTC+5:30, just after local midnight
  kolkata <- as.POSIXct("2024-06-15 00:00:30", tz = "Asia/Kolkata")
  expect_equal(format(date(kolkata, tz = "Asia/Kolkata")), "2024-06-15 IST")

  # Pacific/Chatham: UTC+12:45, just after local midnight
  chatham <- as.POSIXct("2024-06-15 00:10:00", tz = "Pacific/Chatham")
  expect_equal(format(date(chatham, tz = "Pacific/Chatham")), "2024-06-15 +1245")

  # Pacific/Marquesas: UTC-9:30, just before local midnight
  marquesas <- as.POSIXct("2024-06-15 23:50:00", tz = "Pacific/Marquesas")
  expect_equal(format(date(marquesas, tz = "Pacific/Marquesas")), "2024-06-15 -0930")

  # America/New_York: UTC-4:00 (EDT), just before local midnight
  new_york <- as.POSIXct("2024-06-15 23:59:30", tz = "America/New_York")
  expect_equal(format(date(new_york, tz = "America/New_York")), "2024-06-15 EDT")
})

test_that("tz_to_utc() discrete rounding resolves DST gap/overlap to the correct local day", {
  # The offset used by tz_to_utc() is re-derived at the candidate raw instant
  # (a second `tz_offset_impl()` pass), rather than reused from the forward
  # shift, precisely so a value near a DST transition doesn't pick up the
  # wrong side of it.

  # America/New_York spring-forward gap, 2024-03-10: 02:00 -> 03:00, and the
  # fall-back overlap, 2024-11-03: 02:00 occurs twice
  new_york <- as.POSIXct(
    c(
      "2024-03-09 23:59:58",
      "2024-03-10 00:00:02",
      "2024-03-10 23:59:58",
      "2024-03-11 00:00:02",
      "2024-11-02 23:59:58",
      "2024-11-03 23:59:58",
      "2024-11-04 00:00:02"
    ),
    tz = "America/New_York"
  )
  expect_equal(
    format(date(new_york, tz = "America/New_York")),
    c(
      "2024-03-09 EST",
      "2024-03-10 EST",
      "2024-03-10 EST",
      "2024-03-11 EDT",
      "2024-11-02 EDT",
      "2024-11-03 EDT",
      "2024-11-04 EST"
    )
  )

  # Australia/Melbourne fall-back day itself (2015-04-05)
  melbourne <- as.POSIXct(
    c("2015-04-04 23:59:58", "2015-04-05 23:59:58", "2015-04-06 00:00:02"),
    tz = "Australia/Melbourne"
  )
  expect_equal(
    format(date(melbourne, tz = "Australia/Melbourne")),
    c("2015-04-04 AEDT", "2015-04-05 AEST", "2015-04-06 AEST")
  )
})

test_that("tz_to_utc() discrete rounding survives the shift/unshift ULP chain", {
  # The shift/unshift pair in chronon_convert_impl() (tz_to_local() before
  # the path conversion, tz_to_utc() after it) chains enough floating-point
  # operations that a value exactly on a `to`-boundary can land a few ULPs
  # below it. A bare floor() in the discrete branch then drops a whole unit -
  # this hit Europe/London specifically (seq() starting one second before
  # `from`), fixed with a small ULP-relative nudge before flooring.
  w <- cal_isoweek$week(1L)
  zones <- c(
    "UTC",
    "Etc/GMT-10",
    "Australia/Melbourne",
    "America/New_York",
    "Asia/Kolkata",
    "Europe/London"
  )

  for (tz in zones) {
    from <- time_floor(
      datetime(as.POSIXct("2015-06-01", tz = tz)),
      cal_isoweek$week(1L, tz = tz)
    )
    to <- time_ceiling(
      datetime(as.POSIXct("2015-06-22", tz = tz)),
      cal_isoweek$week(1L, tz = tz)
    )
    s <- seq(from, to, by = cal_isoweek$week(1L, tz = tz))

    expect_equal(as.numeric(s[1]), as.numeric(from), info = tz)
    expect_length(s, 5)
  }
})

test_that("comparing a zoned time against a naive one uses wall-clock time", {
  # Regression test: comparison used to convert the zoned operand to its true
  # (UTC) absolute instant while leaving the naive operand untouched, mixing
  # two different bases and producing numerically wrong results. Both sides
  # should instead be compared at face value, as `datetime(x, tz = NA)`
  # already does for a single value (wall-clock time is preserved, not the
  # absolute instant).
  zoned <- linear_time(
    datetime("2015-02-01 10:00:00", tz = "Australia/Melbourne"),
    hour(1L, tz = "Australia/Melbourne")
  )
  naive <- datetime("2015-02-01 00:00:00")

  expect_true(zoned > naive)
  expect_false(zoned < naive)
  expect_false(zoned == naive)

  # Same-zone comparisons must keep comparing true absolute instants (not be
  # affected by the wall-clock stripping used for naive/zoned comparisons).
  melb <- datetime(as.POSIXct("2015-01-01 00:00:00", tz = "Australia/Melbourne"))
  utc <- datetime(as.POSIXct("2015-01-01 00:00:00", tz = "UTC"))
  expect_true(melb < utc)
})

#' Parse text into a time point
#'
#' `time_parse()` is the inverse of [format()]: given text and the same
#' `{lin(...)}`/`{cyc(...)}` template `format` uses, it reconstructs the time
#' points that would have produced that text, via [time_compose()].
#'
#' A `format` with a `{lin(...)}` token parses to linear time; one of only
#' `{cyc(...)}` tokens parses to cyclical time, e.g. `time_parse("Feb",
#' format = "{cyc(month, year, label = TRUE)}")` recovers the same kind of
#' value as [month_of_year()].
#'
#' Granule-specific extraction and decoding labels for each token is done by
#' [linear_labels_parse()]/[cyclical_labels_parse()].
#'
#' @param x A character vector to parse.
#' @param chronon Target time granule for the result, and (with `cycle`) the
#'   source of `format` candidates when `format` is `NULL`. Its attributes
#'   (e.g. `tz`) fill in whatever `format` leaves unset, and the result is
#'   converted onto it if `format` reaches a different chronon.
#' @param cycle Target cycle granule, pairing with `chronon` for a cyclical
#'   result. Requires `chronon`.
#' @param format A glue-style format string of [lin()]/[cyc()] tokens, e.g.
#'   `"{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"` (see
#'   `vignette("time-format-strings")`), or several to try: whichever parses
#'   the most values of `x` is used for the whole vector (ties keep the
#'   earliest-listed format), and its unparsed values become `NA` (with a
#'   warning). Aborts if no format matches the shape of even one value.
#'   `time_parse(format(x, fmt), format = fmt)` round-trips back to `x`.
#'   `NULL` (the default) derives candidates from `chronon`/`cycle` via
#'   [chronon_parse_linear()]/[chronon_parse_cyclical()]; requires `chronon`.
#' @param regex If `FALSE` (the default), literal text surrounding tokens is
#'   matched exactly. If `TRUE`, it's instead used verbatim as a regular
#'   expression, e.g. `"[/-]"` to accept either `/` or `-` as a separator;
#'   `(...)` groups you write are treated as non-capturing, since capturing
#'   groups are reserved for the tokens. Ignored when `format` is derived
#'   from `chronon`, which carries its own regex mode.
#' @param na Strings to treat as missing (`NA`), checked before matching
#'   `format`. Not counted in the parsing-failure warning, unlike a value
#'   that fails to match `format`.
#' @param calendar Calendar used to resolve granule names in `format`, and
#'   to disambiguate `chronon`'s [chronon_parse_linear()] candidates when
#'   `format` is `NULL`. `NULL` (the default) uses `time_calendar(cycle)` or
#'   `time_calendar(chronon)`, whichever is supplied, else [cal_gregorian].
#' @param locale Default locale for named (`label = TRUE`) tokens that don't
#'   specify their own. `NULL` defers to each token's own scheme.
#' @param discrete Whether the result is discrete (integer chronon counts)
#'   or continuous (fractional). See [linear_time()].
#'
#' @return A `mixtime` time vector, the same length as `x`. Linear if
#'   `format` includes a `{lin(...)}` token (or `cycle` is `NULL`),
#'   cyclical otherwise.
#'
#' @seealso [format()] for the inverse direction, [time_compose()] for
#'   composing a time point from already-decoded components,
#'   [label_scheme()] for declaring how a granule's labels parse,
#'   [chronon_parse_linear()]/[chronon_parse_cyclical()] for the candidate
#'   formats derived from `chronon`/`cycle`, and `vignette("time-format-strings")`
#'   for the format string syntax.
#'
#' @examples
#' time_parse("2024-02-15", format = "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}")
#' time_parse(
#'   "15 Feb 2024",
#'   format = "{cyc(day, month)} {cyc(month, year, label = TRUE)} {lin(year)}"
#' )
#'
#' # One bad value becomes NA (with a warning) instead of aborting the batch
#' time_parse(
#'   c("2024-02-15", "not a date"),
#'   format = "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"
#' )
#'
#' # No {lin(...)} token: parses to cyclical time
#' time_parse("Feb", format = "{cyc(month, year, label = TRUE)}")
#'
#' # Several formats: whichever parses the most values is used for the whole
#' # vector; here none of the "Y-M-D" format's values match, so the "D/M/Y"
#' # format (which matches both) is used instead
#' time_parse(
#'   c("15/02/2024", "20/03/2024"),
#'   format = c(
#'     "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}",
#'     "{cyc(day, month)}/{cyc(month, year)}/{lin(year)}"
#'   )
#' )
#'
#' # regex = TRUE: match "/" or "-" as the separator, 
#' # and tolerate a trailing comment after the date.
#' time_parse(
#'   c("2024-02-15", "2024/02/15 (approx)"),
#'   format = "{lin(year)}[/-]{cyc(month, year)}[/-]{cyc(day, month)}( .*)?",
#'   regex = TRUE
#' )
#'
#' # Default format strings from the target chronon, and results with `tz`.
#' time_parse("2024-02-15 09:00:00", chronon = cal_gregorian$second(1L, tz = "America/Los_Angeles"))
#'
#' @export
time_parse <- function(
  x,
  chronon = NULL,
  cycle = NULL,
  format = NULL,
  regex = FALSE,
  na = c("", "NA"),
  calendar = NULL,
  locale = NULL,
  discrete = TRUE
) {
  new_mixtime(time_parse_impl(
    x,
    chronon = chronon,
    cycle = cycle,
    format = format,
    regex = regex,
    na = na,
    calendar = calendar,
    locale = locale,
    discrete = discrete,
    env = rlang::caller_env()
  ))
}

# The implementation of time_parse(), returning a simple time vector
# (mt_linear/mt_cyclical) rather than a mixtime() - e.g. for vec_cast()
# methods (see mt_cast_from_character() in vctrs.R) that need the parsed
# result unwrapped rather than paying for a mixtime() round-trip.
time_parse_impl <- function(
  x,
  chronon = NULL,
  cycle = NULL,
  format = NULL,
  regex = FALSE,
  na = c("", "NA"),
  calendar = NULL,
  locale = NULL,
  discrete = TRUE,
  env = rlang::caller_env()
) {
  x <- as.character(x)
  n <- length(x)

  if (!is.null(cycle) && is.null(chronon)) {
    cli::cli_abort("{.arg cycle} requires {.arg chronon} to also be supplied.", call = NULL)
  }

  if (is.null(calendar)) {
    calendar <- if (!is.null(cycle)) {
      time_calendar(cycle)
    } else if (!is.null(chronon)) {
      time_calendar(chronon)
    } else {
      cal_gregorian
    }
  }

  if (is.null(format)) {
    if (is.null(chronon)) {
      cli::cli_abort("Either {.arg chronon} or {.arg format} must be supplied.", call = NULL)
    }
    format <- if (is.null(cycle)) {
      chronon_parse_linear(chronon, calendar)
    } else {
      chronon_parse_cyclical(chronon, cycle)
    }
    regex <- attr(format, "regex") %||% FALSE
  }
  format <- as.character(format)
  if (length(format) == 0L) {
    cli::cli_abort("{.arg format} must have length 1 or more.", call = NULL)
  }

  # Validate each format's shape and resolve its chain order once, up front.
  compiled_list <- lapply(format, time_parse_compile, calendar = calendar, locale = locale, env = env, regex = regex)

  is_missing <- is.na(x) | x %in% na
  attempt <- !is_missing
  x_try <- x[attempt]
  m <- length(x_try)

  if (n == 0L) {
    result <- time_parse_na(compiled_list[[1L]]$chain, compiled_list[[1L]]$cycle, discrete, 1L)
    if (!is.null(chronon)) result <- time_parse_recast(result, chronon, cycle, discrete)
    return(result[0L])
  }
  if (m == 0L) {
    # Every value is missing: no data to choose a format by.
    result <- time_parse_na(compiled_list[[1L]]$chain, compiled_list[[1L]]$cycle, discrete, 1L)
    if (!is.null(chronon)) result <- time_parse_recast(result, chronon, cycle, discrete)
    return(result[rep(1L, n)])
  }

  # Keep whichever format parses the most values; ties keep the earliest.
  # Try likely winners first, ranked by a cheap match count on a small
  # sample, so a clear winner can end the search early.
  sample <- x_try[seq_len(min(16L, m))]
  sample_counts <- vapply(
    compiled_list,
    function(compiled) sum(time_parse_matches(compiled, sample)$matched),
    integer(1L)
  )
  search_order <- order(-sample_counts)

  best <- NULL
  for (f in search_order) {
    compiled <- compiled_list[[f]]
    matches <- time_parse_matches(compiled, x_try)
    match_count <- sum(matches$matched)
    # Match count upper-bounds parse count, so this can't beat the best yet.
    if (match_count == 0L ||
        (!is.null(best) &&
         (match_count < best$ok_count ||
          (match_count == best$ok_count && f > best$format)))) {
      next
    }

    attempted <- time_parse_attempt(compiled, x_try, matches, discrete)
    ok_count <- sum(attempted$ok)
    if (is.null(best) || ok_count > best$ok_count || (ok_count == best$ok_count && f < best$format)) {
      best <- list(format = f, attempted = attempted, ok_count = ok_count)
    }
    # A near-total match ends the search rather than checking every format.
    if (best$ok_count >= m * 0.975) {
      break
    }
  }

  # No candidate format matched the shape of even a single value.
  if (is.null(best)) {
    cli::cli_abort(
      c(
        "None of the {length(compiled_list)} candidate format{?s} in {.arg format} matched any value in {.arg x}.",
        i = "{.val {utils::head(x_try, 5L)}}{if (m > 5L) paste0(', and ', m - 5L, ' more') else ''}",
        i = "Check the format string against the shape of {.arg x} (separators, token order, numeric vs. labelled tokens)."
      ),
      class = "mixtime_parse_no_match",
      call = NULL
    )
  }

  attempted <- best$attempted
  na_one <- time_parse_na(compiled_list[[best$format]]$chain, compiled_list[[best$format]]$cycle, discrete, 1L)

  failed <- !attempted$ok
  if (any(failed)) {
    time_parse_warn_failures(x_try[failed])
  }

  # Combine parsed values and repeated NA placeholders, then reorder.
  good <- attempted$value %||% na_one[0L]
  n_bad <- sum(is_missing) + sum(failed)
  bad <- na_one[rep(1L, n_bad)]
  combined <- vctrs::vec_c(good, bad)

  perm <- integer(n)
  idx_good <- which(attempt)[attempted$ok]
  idx_bad <- c(which(is_missing), which(attempt)[failed])
  perm[idx_good] <- seq_along(idx_good)
  perm[idx_bad] <- length(idx_good) + seq_along(idx_bad)

  result <- combined[perm]
  if (is.null(chronon)) result else time_parse_recast(result, chronon, cycle, discrete)
}

# Regex-match `x_try` against a compiled format, without decoding/composing.
# A cheap pre-check for whether a format can possibly parse every value.
time_parse_matches <- function(compiled, x_try) {
  groups <- regmatches(x_try, regexec(compiled$pattern, x_try, perl = TRUE))
  list(matched = lengths(groups) > 0L, groups = groups)
}

# Decode/compose `x_try` under one compiled format. Returns list(value, ok).
time_parse_attempt <- function(compiled, x_try, matches, discrete) {
  m <- length(x_try)
  row_matched <- matches$matched
  groups <- matches$groups
  n_specs <- length(compiled$specs)

  groups[!row_matched] <- list(rep(NA_character_, n_specs + 1L))
  mat <- matrix(unlist(groups, use.names = FALSE), nrow = n_specs + 1L)
  text_cols <- lapply(seq_len(n_specs), function(k) mat[k + 1L, ])

  res <- tryCatch(
    list(value = time_parse_compose(compiled$specs, text_cols, compiled$chain_order, compiled$cycle, discrete)),
    error = function(e) list(error = e)
  )

  if (is.null(res$error)) {
    ok <- row_matched & !is.na(res$value)
    value <- res$value[ok]
  } else {
    # One bad row aborts the vectorized attempt; retry row by row to find it.
    ok <- row_matched
    row_values <- vector("list", m)
    n_ok <- 0L
    for (i in which(row_matched)) {
      row_res <- tryCatch(
        list(value = time_parse_compose(compiled$specs, lapply(text_cols, `[`, i), compiled$chain_order, compiled$cycle, discrete)),
        error = function(e) list(error = e)
      )
      if (is.null(row_res$error) && !is.na(row_res$value)) {
        n_ok <- n_ok + 1L
        row_values[[n_ok]] <- row_res$value
      } else {
        ok[[i]] <- FALSE
      }
    }
    value <- if (n_ok > 0L) vctrs::vec_c(!!!row_values[seq_len(n_ok)]) else NULL
  }

  list(value = value, ok = ok)
}

# Tokenize `format` into an anchored extraction regex plus its lin()/cyc() specs.
time_parse_compile <- function(format, calendar, locale, env, regex = FALSE) {
  mask_env <- rlang::new_environment(data = component_mask(calendar), parent = env)
  fmt <- mt_glue_fmt(format, env = mask_env)
  is_spec <- vapply(fmt, is.list, logical(1L))

  specs <- lapply(fmt[is_spec], function(tok) {
    spec <- format_token_spec(tok)
    attrs <- spec$attrs
    if (is.null(attrs$locale)) attrs$locale <- locale

    parse_spec <- if (is.null(spec$cycle)) {
      rlang::exec(linear_labels_parse, spec$chronon, !!!attrs)
    } else {
      rlang::exec(cyclical_labels_parse, spec$chronon, spec$cycle, !!!attrs)
    }
    list(chronon = spec$chronon, cycle = spec$cycle, pattern = parse_spec$pattern, decode = parse_spec$decode)
  })

  fragments <- character(length(fmt))
  spec_i <- 0L
  for (i in seq_along(fmt)) {
    if (is_spec[[i]]) {
      spec_i <- spec_i + 1L
      fragments[[i]] <- paste0("(", specs[[spec_i]]$pattern, ")")
    } else if (regex) {
      # Verbatim regex: no escaping, no whitespace tolerance. Capturing
      # groups are rewritten to non-capturing so they can't shift the
      # positional group indices used below (see time_parse_no_capture()).
      fragments[[i]] <- time_parse_no_capture(fmt[[i]])
    } else {
      fragments[[i]] <- time_parse_escape_literal(fmt[[i]])
    }
  }

  # Validate the format's shape and resolve its chain order once, up front.
  # `chain$cycle` is non-NULL for a cyc()-only (no lin()) format, giving a
  # cyclical result; see compose_chain()/compose_recompose().
  chain <- compose_chain(lapply(specs, function(s) list(chronon = s$chronon, cycle = s$cycle)))
  chain_order <- vapply(
    chain$chain,
    function(link) {
      which(vapply(
        specs,
        function(s) identical(s$chronon, link$chronon) && identical(s$cycle, link$cycle),
        logical(1L)
      ))
    },
    integer(1L)
  )

  list(
    pattern = paste0("^", paste0(fragments, collapse = ""), "$"),
    specs = specs,
    chain = chain$chain,
    cycle = chain$cycle,
    chain_order = chain_order
  )
}

# Regex-escape literal format text, collapsing runs of spaces to `\s+` for
# tolerance to whitespace variation.
time_parse_escape_literal <- function(text) {
  esc <- gsub("([.|()\\^{}+$*?\\[\\]\\\\])", "\\\\\\1", text, perl = TRUE)
  gsub(" +", "\\\\s+", esc)
}

# Rewrite a user's verbatim regex (regex = TRUE) so every capturing group it
# opens becomes non-capturing (?:...). The positional groups in the
# compiled pattern are reserved for {lin(...)}/{cyc(...)} tokens
# (time_parse_attempt() picks out each spec's text by group index), so a
# stray user capturing group would silently shift them out of alignment;
# rewriting means there's nothing for the caller to get right.
#
# `(` is left alone (already special/non-capturing) when it's escaped
# (`\(`), inside a `[...]` character class (where it's a literal char, not a
# group), or already followed by `?` (covers `(?:`, `(?=`, `(?!`, `(?<=`,
# `(?<!`, `(?<name>`).
time_parse_no_capture <- function(text) {
  if (!nzchar(text)) return(text)
  chars <- strsplit(text, "", fixed = TRUE)[[1L]]
  n <- length(chars)
  in_class <- FALSE
  escaped <- FALSE
  for (i in seq_len(n)) {
    ch <- chars[[i]]
    if (escaped) {
      escaped <- FALSE
    } else if (ch == "\\") {
      escaped <- TRUE
    } else if (in_class) {
      if (ch == "]") in_class <- FALSE
    } else if (ch == "[") {
      in_class <- TRUE
    } else if (ch == "(" && !(i < n && chars[[i + 1L]] == "?")) {
      chars[[i]] <- "(?:"
    }
  }
  paste0(chars, collapse = "")
}

# Decode each spec's text column and hand the result to compose_recompose().
# A cyc() link's decode is deferred until its coarser instance is resolved,
# except at the chain's root (position 1), which decodes outright: `at`
# defaults to NULL, so this works whether the root is a lin() or (for
# cyclical time) a cyc() token.
time_parse_compose <- function(specs, text_cols, chain_order, cycle, discrete) {
  comps <- lapply(seq_along(chain_order), function(pos) {
    k <- chain_order[[pos]]
    spec <- specs[[k]]
    text <- text_cols[[k]]
    if (pos == 1L) {
      list(chronon = spec$chronon, cycle = NULL, value = spec$decode(text))
    } else {
      list(chronon = spec$chronon, cycle = spec$cycle, value = function(at) spec$decode(text, at))
    }
  })
  compose_recompose(comps, discrete = discrete, cycle = cycle, strict = FALSE)
}

# An all-`NA` result at the chain's finest chronon, built via compose_recompose().
time_parse_na <- function(shape_chain, cycle, discrete, k) {
  chain <- shape_chain
  chain[[1L]]$value <- rep(NA_real_, k)
  for (j in seq_along(chain)[-1L]) {
    chain[[j]]$value <- function(at) rep(NA_real_, length(at))
  }
  compose_recompose(chain, discrete = discrete, cycle = cycle)
}

# Cast a time_parse_impl() result (a simple mt_linear/mt_cyclical vector) onto
# the caller-requested chronon/cycle, filling in whichever of its properties
# (tz, lat/lon, ...) were left unset.
time_parse_recast <- function(x, chronon, cycle, discrete) {
  from <- attr(x, "chronon")
  chronon <- granule_inherit_props(chronon, from)
  if (!is.null(cycle)) cycle <- granule_inherit_props(cycle, chronon)

  data <- vctrs::vec_data(x)

  # A tz-naive parse has no absolute instant to convert from - reinterpret it
  # as local wall-clock time in the target tz before any granule conversion.
  if (
    S7::S7_inherits(from, mt_tz_unit) && S7::S7_inherits(chronon, mt_tz_unit) &&
      is.na(tz_name(from)) && !is.na(tz_name(chronon))
  ) {
    from <- granule_inherit_props(from, mt_tz_unit(tz = tz_name(chronon)))
    if (tz_name(from) != "UTC") data <- tz_to_utc(data, from, tz_name(from), discrete = discrete)
  }

  if (!identical(chronon, from)) {
    data <- chronon_convert_impl(data, from, chronon, discrete)
  }

  if (is.null(cycle)) {
    mt_linear(data, chronon = chronon)
  } else {
    mt_cyclical(data, chronon = chronon, cycle = cycle)
  }
}

time_parse_warn_failures <- function(bad, show = 5L) {
  n <- length(bad)
  cli::cli_warn(
    c(
      "{n} value{?s} failed to parse and {cli::qty(n)} {?was/were} set to {.code NA}.",
      "x" = "{.val {utils::head(bad, show)}}{if (n > show) paste0(', and ', n - show, ' more') else ''}"
    ),
    class = "mixtime_parse_failure",
    call = NULL
  )
}

# Parse time sequence strings into time units
# TODO - Generalise to be calendar-aware
#        e.g. "year" being cal_gregorian$year() or cal_gregorian$isoyear()
parse_time_unit <- function(x) {
  by2 <- strsplit(x, " ", fixed = TRUE)[[1L]]
  if (length(by2) > 2L || length(by2) < 1L) 
      stop("invalid 'by' string")
  n <- if(length(by2) == 2L) as.integer(by2[[1L]]) else 1L
  x <- by2[[length(by2)]]
  
  switch(sub("s$", "", x),
    "second" = ,
    "sec" = cal_gregorian$second(n),
    "minute" = ,
    "min" = cal_gregorian$minute(n),
    "hour" = cal_gregorian$hour(n),
    "day" = cal_gregorian$day(n),
    "dstday" = cal_gregorian$day(n), # DSTdays map to regular days
    "week" = cal_isoweek$week(n),
    "month" = cal_gregorian$month(n),
    "year" = cal_gregorian$year(n),
    "quarter" = cal_gregorian$quarter(n),
    stop("Unknown time unit: '", x, "'. Valid units are: secs, mins, hours, days, weeks, months, years, DSTdays, quarters")
  )
}

# Tests for time_parse(): the inverse of format(), turning text back into a
# mixtime using the same {lin(...)}/{cyc(...)} format string. Invalid results
# behave like readr::parse_datetime() -- NA plus a summary warning -- rather
# than time_compose()'s own always-abort-on-any-bad-row behaviour.

test_that("parses a date from year/month/day tokens", {
  r <- time_parse("1980-03-15", format = "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}")

  expect_equal(format(r), "1980-03-15")
  expect_equal(
    as.numeric(vecvec::unvecvec(r)),
    as.numeric(vecvec::unvecvec(time_compose(lin(year) ~ 1980, cyc(month, year) ~ 3, cyc(day, month) ~ 15)))
  )
})

test_that("round-trips through format()", {
  x <- as_mixtime(as.Date("2024-03-15") + 0:5)
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"

  txt <- format(x, fmt)
  r <- time_parse(txt, format = fmt)

  expect_equal(as.numeric(vecvec::unvecvec(r)), as.numeric(vecvec::unvecvec(x)))
})

test_that("decodes in chain order, not text order", {
  # day comes before month/year in the text, but a day's cardinality (is the
  # 29th valid?) depends on the month/year that appear later -- decoding
  # must follow the lin()->cyc() chain, not the token's position in the text.
  fmt <- "{cyc(day, month)} {cyc(month, year, label = TRUE)} {lin(year)}"

  expect_equal(format(time_parse("29 Feb 2024", format = fmt)), "2024-02-29")
  expect_warning(time_parse("29 Feb 2023", format = fmt), "failed to parse")
})

test_that("parses abbreviated and wide named labels", {
  fmt_abbr <- "{lin(year)} {cyc(month, year, label = TRUE)} {cyc(day, month)}"
  fmt_wide <- "{lin(year)} {cyc(month, year, label = TRUE, abbreviate = FALSE)} {cyc(day, month)}"

  expect_equal(format(time_parse("2024 Feb 15", format = fmt_abbr)), "2024-02-15")
  expect_equal(format(time_parse("2024 February 15", format = fmt_wide)), "2024-02-15")
})

test_that("literal template text tolerates whitespace runs", {
  fmt <- "{lin(year)} {cyc(month, year)} {cyc(day, month)}"

  expect_equal(format(time_parse("2024   02   15", format = fmt)), "2024-02-15")
})

test_that("BC/AD year labels round-trip (a granule overriding the generics directly)", {
  expect_equal(format(time_parse("44BC", format = "{lin(year)}")), "44BC")
  expect_equal(format(time_parse("1980", format = "{lin(year)}")), "1980")
})

test_that("na values become NA without a parse-failure warning", {
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"
  r <- time_parse(c("1980-03-15", "", NA, "NA"), format = fmt)

  expect_equal(is.na(r), c(FALSE, TRUE, TRUE, TRUE))
})

test_that("a custom na vector is respected", {
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"
  r <- time_parse(c("1980-03-15", "missing"), format = fmt, na = "missing")

  expect_equal(is.na(r), c(FALSE, TRUE))
})

test_that("one bad value becomes NA with a warning, the rest still parse", {
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"

  expect_warning(
    r <- time_parse(c("1980-03-15", "not a date"), format = fmt),
    "1 value failed to parse"
  )
  expect_equal(format(r), c("1980-03-15", "NA"))
})

test_that("an invalid cardinality (day 30 of February) becomes NA with a warning", {
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"

  expect_warning(r <- time_parse("2024-02-30", format = fmt), "failed to parse")
  expect_true(is.na(r))
})

test_that("multiple bad values are all reported and set to NA", {
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"

  expect_warning(
    r <- time_parse(c("2024-02-30", "1980-03-15", "xx", "2024-13-01"), format = fmt),
    "3 values failed to parse"
  )
  expect_equal(is.na(r), c(TRUE, FALSE, TRUE, TRUE))
})

test_that("a single invalid row in a large batch doesn't fall back to retrying every row", {
  # compose_recompose(strict = FALSE) resolves its own cardinality-invalid rows
  # to NA in place (R/compose.R) rather than aborting the whole vectorized
  # compose, so this stays a single vectorized pass -- not O(m) individual
  # mixtime() constructions -- even with one bad row among many.
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"
  n <- 5000L
  x <- rep("2024-02-15", n)
  x[2500L] <- "2024-02-30"

  expect_warning(r <- time_parse(x, format = fmt), "1 value failed to parse")
  expect_equal(is.na(r), seq_len(n) == 2500L)
  expect_equal(format(r[-2500L]), rep("2024-02-15", n - 1L))
})

test_that("bad rows at different links in the chain each resolve to NA independently", {
  # "2024-13-01" is invalid at the month link; "2024-02-30" is invalid at the
  # day link (which itself depends on the already-decoded month); both must
  # become NA without disturbing the valid rows around them.
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"
  x <- c("2024-01-15", "2024-13-01", "2024-03-20", "2024-02-30", "2024-12-25")

  expect_warning(r <- time_parse(x, format = fmt), "2 values failed to parse")
  expect_equal(is.na(r), c(FALSE, TRUE, FALSE, TRUE, FALSE))
  expect_equal(format(r[c(1L, 3L, 5L)]), c("2024-01-15", "2024-03-20", "2024-12-25"))
})

test_that("a calendar-author decode() that throws still falls back to per-row retry", {
  # Simulates label_scheme()'s documented "Irregular cycles" case (R/01_labels.R),
  # where decode() aborts for text that matched the token's regex but is
  # semantically invalid given `at` -- a failure compose_recompose()'s own
  # cardinality check can't catch, since it happens inside the decode closure
  # before that check ever runs. This must still isolate to the one bad row
  # via time_parse_attempt()'s retry loop, not abort the whole call.
  old <- S7::method(cyclical_labels_parse, list(cal_gregorian$month, cal_gregorian$year))
  withr_cleanup <- function() {
    S7::method(cyclical_labels_parse, list(cal_gregorian$month, cal_gregorian$year)) <- old
  }
  on.exit(withr_cleanup())

  S7::method(cyclical_labels_parse, list(cal_gregorian$month, cal_gregorian$year)) <- function(granule, cycle, ...) {
    list(
      pattern = "\\d+",
      decode = function(text, at = NULL) {
        n <- as.integer(text)
        if (any(n == 6L, na.rm = TRUE)) {
          cli::cli_abort("Month 6 is reserved in this test scheme.")
        }
        n - 1L
      }
    )
  }

  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"
  x <- c("2024-01-15", "2024-06-01", "2024-03-20")

  expect_warning(r <- time_parse(x, format = fmt), "1 value failed to parse")
  expect_equal(is.na(r), c(FALSE, TRUE, FALSE))
  expect_equal(format(r[c(1L, 3L)]), c("2024-01-15", "2024-03-20"))
})

test_that("a format with no lin() anchor parses to cyclical time", {
  r <- time_parse("Feb", format = "{cyc(month, year, label = TRUE)}")

  expect_true(time_is_cyclical(r))
  expect_equal(format(r), "Feb")
  expect_equal(
    as.numeric(vecvec::unvecvec(r)),
    as.numeric(vecvec::unvecvec(month_of_year(as.Date("1970-02-01"))))
  )
})

test_that("a chained cyc()-only format parses to cyclical time", {
  # Collapses to (day, year): day-of-year 74, same as day_of_year() for 15 March.
  r <- time_parse("15 Mar", format = "{cyc(day, month)} {cyc(month, year, label = TRUE)}")

  expect_true(time_is_cyclical(r))
  expect_equal(format(r), "D74")
  expect_equal(
    as.numeric(vecvec::unvecvec(r)),
    as.numeric(vecvec::unvecvec(day_of_year(as.Date("1970-03-15"))))
  )
})

test_that("errors immediately for a cyc()-only format with no coarsest link", {
  expect_error(
    time_parse("1 15", format = "{cyc(month, day)} {cyc(day, month)}"),
    "coarsest link"
  )
})

test_that("errors immediately for a disconnected cyc() chain", {
  expect_error(time_parse("15", format = "{lin(year)} {cyc(day, month)}"), "does not connect")
})

test_that("an empty input returns a length-0 mixtime", {
  r <- time_parse(character(0), format = "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}")

  expect_length(r, 0L)
  expect_true(is_mixtime(r))
})

test_that("multiple formats: the first one under which every value parses is used", {
  fmts <- c(
    "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}",
    "{cyc(day, month)}/{cyc(month, year)}/{lin(year)}"
  )

  # Both values already fit the first format, so it's used outright.
  r <- time_parse(c("2024-02-15", "2024-03-20"), format = fmts)
  expect_equal(format(r), c("2024-02-15", "2024-03-20"))
})

test_that("multiple formats: a format that doesn't parse everything is skipped entirely", {
  fmts <- c(
    "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}",
    "{cyc(day, month)}/{cyc(month, year)}/{lin(year)}"
  )

  # Neither value fits the first ("Y-M-D") format, so the second ("D/M/Y")
  # is used for both instead.
  r <- time_parse(c("15/02/2024", "20/03/2024"), format = fmts)
  expect_equal(format(r), c("2024-02-15", "2024-03-20"))
})

test_that("multiple formats: formats are never mixed value-by-value", {
  fmts <- c(
    "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}",
    "{cyc(day, month)}/{cyc(month, year)}/{lin(year)}"
  )

  # "2024-02-15" only fits the first format and "15/02/2024" only fits the
  # second -- neither format parses *both*, so each parses exactly one value
  # and the tie goes to the earliest-listed format, not opportunistically
  # mixing formats value-by-value.
  expect_warning(
    r <- time_parse(c("2024-02-15", "15/02/2024"), format = fmts),
    "1 value failed to parse"
  )
  expect_equal(format(r), c("2024-02-15", "NA"))
})

test_that("multiple formats: the format that parses the most values wins, not the last one", {
  fmts <- c(
    "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}",
    "{cyc(day, month)}/{cyc(month, year)}/{lin(year)}"
  )

  # The first ("Y-M-D") format parses 2 of 3 values; the second ("D/M/Y")
  # parses only 1, so the first format wins on coverage.
  expect_warning(
    r <- time_parse(c("2024-02-15", "2024-03-20", "15/02/2024"), format = fmts),
    "1 value failed to parse"
  )
  expect_equal(format(r), c("2024-02-15", "2024-03-20", "NA"))
})

test_that("multiple formats: a partial win from a non-last format isn't discarded", {
  fmts <- c(
    "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}",
    "{cyc(day, month)}/{cyc(month, year)}/{lin(year)}"
  )

  # The first format parses 2 of 3 values; the third value's shape doesn't
  # match either format, so the second format matches nothing at all.
  expect_warning(
    r <- time_parse(c("2024-02-15", "2024-03-20", "not-a-date"), format = fmts),
    "1 value failed to parse"
  )
  expect_equal(format(r), c("2024-02-15", "2024-03-20", "NA"))
})

test_that("multiple formats: a tie is won by the earliest-listed format, even when tried later", {
  fmts <- c(
    "{lin(year)}-Q{cyc(quarter, year)}",
    "{lin(year)}-{cyc(day, year)}"
  )

  # Both formats end up parsing 2 of 5 values (a tie), but the second format
  # has a higher *sample* match count (its regex matches 3 of the 5 values,
  # vs. 2 for the first), so it's tried first. The earliest-listed format
  # must still win the tie once both are actually attempted, not whichever
  # was tried first.
  x <- c("2024-Q1", "2024-Q2", "2024-99", "2024-150", "2024-999")
  expect_warning(
    r <- time_parse(x, format = fmts),
    "3 values failed to parse"
  )
  expect_equal(format(r), c("2024 Q1", "2024 Q2", "NA", "NA", "NA"))
})

test_that("a format that matches but is invalid falls through to the next format", {
  # Both formats share the same literal shape ("{lin(year)}-{cyc(X, year)}"),
  # so "2024-60" matches the regex of both -- but 60 is not a valid quarter
  # (max 4), so it should fall through to being read as a day-of-year (max
  # 366) instead of failing outright.
  fmts <- c(
    "{lin(year)}-{cyc(quarter, year)}",
    "{lin(year)}-{cyc(day, year)}"
  )

  r <- time_parse("2024-60", format = fmts)
  expect_equal(format(r), "2024-02-29")
})

test_that("errors when no candidate format matches the shape of any value", {
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"

  expect_error(time_parse("not a date at all", format = fmt), class = "mixtime_parse_no_match")
})

test_that("multiple formats: still errors when none of them shape-match anything", {
  fmts <- c(
    "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}",
    "{cyc(day, month)}/{cyc(month, year)}/{lin(year)}"
  )

  expect_error(time_parse("nonsense", format = fmts), class = "mixtime_parse_no_match")
})

test_that("a value that shape-matches but fails to parse still becomes NA, not an error", {
  # Shape-matches fine; only the cardinality (day 30 of February) is invalid.
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"

  expect_warning(r <- time_parse("2024-02-30", format = fmt), "failed to parse")
  expect_true(is.na(r))
})

test_that("continuous (discrete = FALSE) parsing keeps a fractional chronon", {
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"
  r <- time_parse("1980-03-15", format = fmt, discrete = FALSE)

  expect_true(is.double(vecvec::unvecvec(r)))
  expect_equal(format(r), "1980-03-15 0.0%")
})

test_that("no chronon override leaves the parsed chronon as the chain reached it", {
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)} {cyc(hour, day)}:{cyc(minute, hour)}:{cyc(second, minute)}"
  r <- time_parse("2024-02-15 09:00:00", format = fmt)

  expect_true(is.na(tz_name(time_chronon(r)@x[[1L]])))
  expect_equal(format(r), "2024-02-15 09:00:00")
})

test_that("chronon with a tz treats the decoded text as clock time in that zone", {
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)} {cyc(hour, day)}:{cyc(minute, hour)}:{cyc(second, minute)}"
  r <- time_parse("2024-02-15 09:00:00", format = fmt, chronon = cal_gregorian$second(1L, tz = "America/Los_Angeles"))

  expect_equal(format(r), "2024-02-15 09:00:00 PST")
  # Storage is UTC-referenced: the same clock time in UTC is 8 hours earlier.
  r_utc <- time_parse("2024-02-15 09:00:00", format = fmt)
  expect_equal(as.numeric(vecvec::unvecvec(r)) - as.numeric(vecvec::unvecvec(r_utc)), 8 * 3600)
})

test_that("chronon with tz = \"UTC\" is a no-op on the stored value", {
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)} {cyc(hour, day)}:{cyc(minute, hour)}:{cyc(second, minute)}"
  no_tz <- time_parse("2024-02-15 09:00:00", format = fmt)
  utc <- time_parse("2024-02-15 09:00:00", format = fmt, chronon = cal_gregorian$second(1L, tz = "UTC"))

  expect_equal(vecvec::unvecvec(no_tz), vecvec::unvecvec(utc))
})

test_that("chronon applies to empty and all-missing input without erroring", {
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"
  chronon <- cal_gregorian$day(1L, tz = "Australia/Melbourne")

  r0 <- time_parse(character(0), format = fmt, chronon = chronon)
  expect_equal(length(r0), 0L)

  rna <- time_parse(NA_character_, format = fmt, chronon = chronon)
  expect_equal(length(rna), 1L)
  expect_true(is.na(rna))
})

test_that("mixtime() parses character data straight to a tz-aware chronon via time_parse()", {
  result <- mixtime("2020-06-01 05:00:00", chronon = cal_gregorian$second(1L, tz = "Australia/Melbourne"))
  expect_equal(format(result), "2020-06-01 05:00:00 AEST")
})

test_that("format = NULL derives candidate formats from chronon/cycle", {
  expect_equal(format(time_parse("2024-02-15", chronon = cal_gregorian$day(1L))), "2024-02-15")
  expect_equal(format(time_parse("3", chronon = cal_gregorian$month(1L), cycle = cal_gregorian$year(1L))), "Mar")
})

test_that("chronon converts the result when it reaches a coarser or finer chronon than format", {
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"

  r <- time_parse("2024-02-15", format = fmt, chronon = cal_gregorian$month(1L))
  expect_equal(format(r), "2024 Feb")
})

test_that("format = NULL without chronon errors", {
  expect_error(time_parse("2024-02-15"), "chronon.*format")
})

test_that("cycle without chronon errors", {
  expect_error(
    time_parse("Feb", cycle = cal_gregorian$year(1L)),
    "cycle.*chronon"
  )
})

test_that("regex = FALSE (default) treats literal metacharacters literally", {
  # "." between tokens is a literal dot, not a "match any char" wildcard.
  fmt <- "{lin(year)}.{cyc(month, year)}.{cyc(day, month)}"

  expect_equal(format(time_parse("2024.02.15", format = fmt)), "2024-02-15")
  expect_error(time_parse("2024x02x15", format = fmt), class = "mixtime_parse_no_match")
})

test_that("regex = TRUE uses literal template text as a regular expression", {
  # Same "." separator, now matched as a wildcard rather than escaped.
  fmt <- "{lin(year)}.{cyc(month, year)}.{cyc(day, month)}"

  expect_equal(format(time_parse("2024x02x15", format = fmt, regex = TRUE)), "2024-02-15")
})

test_that("regex = TRUE supports character classes for alternative separators", {
  fmt <- "{lin(year)}[/-]{cyc(month, year)}[/-]{cyc(day, month)}"

  expect_equal(format(time_parse("2024-02-15", format = fmt, regex = TRUE)), "2024-02-15")
  expect_equal(format(time_parse("2024/02/15", format = fmt, regex = TRUE)), "2024-02-15")
})

test_that("regex = TRUE supports trailing free text via a wildcard", {
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}.*"

  expect_equal(
    format(time_parse("2024-02-15 (approx)", format = fmt, regex = TRUE)),
    "2024-02-15"
  )
})

test_that("regex = TRUE does not collapse whitespace runs (unlike the default escaping)", {
  fmt <- "{lin(year)} {cyc(month, year)} {cyc(day, month)}"

  # Default: any run of spaces is tolerated.
  expect_equal(format(time_parse("2024   02   15", format = fmt)), "2024-02-15")
  # regex = TRUE: the literal " " is matched exactly, so extra spaces don't fit.
  expect_error(
    time_parse("2024   02   15", format = fmt, regex = TRUE),
    class = "mixtime_parse_no_match"
  )
  expect_equal(format(time_parse("2024 02 15", format = fmt, regex = TRUE)), "2024-02-15")
})

test_that("regex = TRUE treats a user's capturing group as non-capturing", {
  # A plain "(...)" group here would otherwise shift the positional group
  # indices used to extract the {lin(...)}/{cyc(...)} tokens' text out of
  # alignment; it must be silently treated as "(?:...)" instead.
  fmt <- "{lin(year)}(-){cyc(month, year)}-{cyc(day, month)}"

  expect_equal(format(time_parse("2024-02-15", format = fmt, regex = TRUE)), "2024-02-15")
})

test_that("regex = TRUE: an auto-converted capturing group composes with alternation and optionality", {
  fmt <- "{lin(year)}[/-]{cyc(month, year)}[/-]{cyc(day, month)}( .*)?"

  expect_equal(format(time_parse("2024-02-15", format = fmt, regex = TRUE)), "2024-02-15")
  expect_equal(
    format(time_parse("2024/02/15 (approx)", format = fmt, regex = TRUE)),
    "2024-02-15"
  )
})

test_that("regex = TRUE leaves already-special groups (non-capturing, lookaround) untouched", {
  fmt_noncapture <- "{lin(year)}(?:-){cyc(month, year)}-{cyc(day, month)}"
  fmt_lookahead <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}(?=$)"

  expect_equal(format(time_parse("2024-02-15", format = fmt_noncapture, regex = TRUE)), "2024-02-15")
  expect_equal(format(time_parse("2024-02-15", format = fmt_lookahead, regex = TRUE)), "2024-02-15")
})

test_that("regex = TRUE still needs doubled braces to write a literal '{n}' quantifier", {
  # mt_glue_fmt() treats a single "{...}" as an expression to evaluate, so a
  # literal regex quantifier like "\\d{2}" must be written with doubled
  # braces to survive as text. Trailing "\\d{2}" here isn't captured by a
  # cyc() token, so the parsed chronon stops at month.
  fmt <- "{lin(year)}-{cyc(month, year)}-\\d{{2}}"

  expect_equal(format(time_parse("2024-02-15", format = fmt, regex = TRUE)), "2024 Feb")
  expect_error(time_parse("2024-02-1", format = fmt, regex = TRUE), class = "mixtime_parse_no_match")
})

test_that("parse_format() records its regex mode as an attribute", {
  expect_equal(attr(parse_format("{lin(year)}"), "regex"), FALSE)
  expect_equal(attr(parse_format("{lin(year)}", regex = TRUE), "regex"), TRUE)
  expect_equal(
    as.character(parse_format("{lin(year)}", "{cyc(month, year)}")),
    c("{lin(year)}", "{cyc(month, year)}")
  )
})

test_that("a format's regex attribute (from parse_format()) doesn't override time_parse()'s regex argument", {
  fmt <- parse_format("{lin(year)}.{cyc(month, year)}.{cyc(day, month)}", regex = TRUE)

  # regex = TRUE is baked into `fmt` as an attribute, but time_parse() only
  # looks at its own `regex` argument, so "." is still matched literally by
  # default and "2024x02x15" doesn't match.
  expect_error(
    time_parse("2024x02x15", format = fmt),
    class = "mixtime_parse_no_match"
  )
  # Passing regex = TRUE explicitly is what's needed for "." to match as a
  # wildcard.
  expect_equal(format(time_parse("2024x02x15", format = fmt, regex = TRUE)), "2024-02-15")
})

test_that("chronon_parse_linear()/chronon_parse_cyclical() candidates are built with parse_format()", {
  # cal_gregorian$day's candidates use regex = TRUE for tolerant separators
  expect_true(attr(chronon_parse_linear(cal_gregorian$day(1L)), "regex"))
  expect_false(attr(chronon_parse_cyclical(cal_gregorian$month(1L), cal_gregorian$year(1L)), "regex"))
})

test_that("chronon_parse_linear()/chronon_parse_cyclical() tolerate irregular separators and casing", {
  # These candidates are all built with parse_format(..., regex = TRUE), so
  # regex = TRUE must be passed to time_parse() explicitly - the attribute
  # on the format vector no longer does this automatically.

  # Gregorian day: mixed separators/whitespace and D/M/Y vs. M/D/Y vs. Y/M/D
  fmt_day <- chronon_parse_linear(cal_gregorian$day(1L))
  expect_equal(format(time_parse("2024-02-15", format = fmt_day, regex = TRUE)), "2024-02-15")
  expect_equal(format(time_parse("2024/02/15", format = fmt_day, regex = TRUE)), "2024-02-15")
  expect_equal(format(time_parse("15 Feb 2024", format = fmt_day, regex = TRUE)), "2024-02-15")
  expect_equal(format(time_parse("Feb 15, 2024", format = fmt_day, regex = TRUE)), "2024-02-15")

  # Gregorian month: label/numeric and either order, various separators
  fmt_month <- chronon_parse_linear(cal_gregorian$month(1L))
  expect_equal(format(time_parse("2024 Feb", format = fmt_month, regex = TRUE)), "2024 Feb")
  expect_equal(format(time_parse("Feb/2024", format = fmt_month, regex = TRUE)), "2024 Feb")

  # ISO week: "W" designator is optional and case-insensitive
  fmt_week <- chronon_parse_cyclical(cal_isoweek$week(1L), cal_isoweek$year(1L))
  expect_equal(
    time_parse("w03", format = fmt_week, calendar = cal_isoweek, regex = TRUE),
    time_parse("W03", format = fmt_week, calendar = cal_isoweek, regex = TRUE)
  )
  expect_equal(
    time_parse("3", format = fmt_week, calendar = cal_isoweek, regex = TRUE),
    time_parse("W03", format = fmt_week, calendar = cal_isoweek, regex = TRUE)
  )

  # ISO week linear: separator between year/week and week/day is flexible
  fmt_isoday <- chronon_parse_linear(cal_isoweek$day(1L))
  expect_equal(
    time_parse("2024-W07-Mon", format = fmt_isoday, calendar = cal_isoweek, regex = TRUE),
    time_parse("2024 w07 Monday", format = fmt_isoday, calendar = cal_isoweek, regex = TRUE)
  )

  # Symmetry454 day: Y-M-W-D and D-W-M-Y agree
  fmt_sym454day <- chronon_parse_linear(cal_sym454$day(1L))
  expect_equal(
    time_parse("2024-Jan-W2-3", format = fmt_sym454day, calendar = cal_sym454, regex = TRUE),
    time_parse("3rd-W2-Jan-2024", format = fmt_sym454day, calendar = cal_sym454, regex = TRUE)
  )
})

test_that("chronon_parse_linear() for civil time-of-day chronons builds on the day chronon's own candidates", {
  # cal_gregorian$second's candidates are cal_gregorian$day's date candidates
  # (Y-M-D, D-M-Y, M-D-Y) each with a fixed "H:M:S" suffix pasted on -- the
  # parsing counterpart of chronon_format_linear()'s
  # paste(chronon_format_linear(cal$day(1L), cal), "{cyc(hour, day)}:...").
  fmt <- chronon_parse_linear(cal_gregorian$second(1L))
  expect_true(attr(fmt, "regex"))

  expect_equal(format(time_parse("2024-02-15 09:30:45", format = fmt, regex = TRUE)), "2024-02-15 09:30:45")
  # A date order other than Y-M-D only works because the suffix was combined
  # with every one of the day chronon's own parse candidates, not just its
  # single chronon_format_linear() default (which is always Y-M-D).
  expect_equal(format(time_parse("15 Feb 2024 09:30:45", format = fmt, regex = TRUE)), "2024-02-15 09:30:45")
  expect_equal(format(time_parse("Feb 15, 2024 09:30:45", format = fmt, regex = TRUE)), "2024-02-15 09:30:45")
})

test_that("chronon_parse_linear() for a plain (non-Gregorian) civil calendar falls back to a single date candidate, non-regex", {
  # cal_time_civil has no year/month, so its day chronon defers to the plain
  # mt_unit fallback -- chronon_parse_linear() should still work, combining
  # that single date candidate with each of the 24-hour/12-hour AM-PM
  # time-of-day variants, each joined by a space and by a bare "T" (4
  # candidates total: 1 date candidate x 2 clocks x 2 separators).
  fmt <- chronon_parse_linear(cal_time_civil$minute(1L))
  expect_false(isTRUE(attr(fmt, "regex")))
  expect_length(fmt, 4L)
})

test_that("chronon_parse_cyclical() for civil time-of-day-in-day chronons parses a bare clock time", {
  fmt <- chronon_parse_cyclical(cal_gregorian$second(1L), cal_gregorian$day(1L))
  r <- time_parse("09:30:45", format = fmt, calendar = cal_gregorian)

  expect_true(time_is_cyclical(r))
  expect_equal(format(r), "09:30:45")
})

test_that("mixtime() parses civil clock-time text using the new default chronon_parse_linear() candidates", {
  expect_equal(
    format(mixtime("15 Feb 2024 09:30:45", chronon = cal_gregorian$second(1L))),
    "2024-02-15 09:30:45"
  )
})

test_that("chronon_parse_linear() for civil time-of-day chronons also accepts an ISO 8601 'T' date-time separator", {
  fmt <- chronon_parse_linear(cal_gregorian$second(1L))
  expect_equal(
    format(time_parse("2024-02-15T09:30:45", format = fmt, regex = TRUE)),
    "2024-02-15 09:30:45"
  )
  # The 12-hour AM/PM suffix is also offered with a "T" separator.
  expect_equal(
    format(time_parse("2024-02-15T9:30:45 PM", format = fmt, regex = TRUE)),
    "2024-02-15 21:30:45"
  )
  expect_equal(
    format(mixtime("2024-02-15T09:30:45", chronon = cal_gregorian$second(1L))),
    "2024-02-15 09:30:45"
  )
})

test_that("chronon_parse_cyclical() for civil hour-in-day chronons parses a bare hour", {
  fmt <- chronon_parse_cyclical(cal_gregorian$hour(1L), cal_gregorian$day(1L))
  r24 <- time_parse("14h", format = fmt, calendar = cal_gregorian)
  rpm <- time_parse("2 PM", format = fmt, calendar = cal_gregorian)
  ram <- time_parse("2 AM", format = fmt, calendar = cal_gregorian)

  expect_true(time_is_cyclical(r24))
  expect_equal(format(r24, "{cyc(hour,day)}"), "14")
  expect_equal(format(rpm, "{cyc(hour,day)}"), "14")
  expect_equal(format(ram, "{cyc(hour,day)}"), "02")
})

test_that("chronon_parse_linear() for solar time-of-day chronons builds on the solar day chronon", {
  chronon <- cal_time_solar$second(1L, lat = -37.8136, lon = 144.9631)
  fmt <- chronon_parse_linear(chronon)

  r <- time_parse("5 09:30:45", format = fmt, calendar = cal_time_solar)
  r <- mixtime(r, chronon = chronon)
  expect_equal(format(r), "1970-01-06 09:30:45 [37.81S 144.96E]")
})

test_that("chronon_parse_linear() for solar arc chronons builds on the solar day chronon", {
  chronon <- cal_time_solar$arcsecond(1L, lat = -37.8136, lon = 144.9631)
  fmt <- chronon_parse_linear(chronon)

  r <- time_parse("5 090°15'30\"", format = fmt, calendar = cal_time_solar)
  r <- mixtime(r, chronon = chronon)
  expect_equal(format(r), "1970-01-06 090°15'30\" [37.81S 144.96E]")
})

test_that("chronon_parse_linear()/chronon_parse_cyclical() for civil time-of-day chronons also accept a 12-hour AM/PM clock", {
  fmt <- chronon_parse_linear(cal_gregorian$second(1L))
  expect_equal(format(time_parse("2024-02-15 9:30:45 PM", format = fmt, regex = TRUE)), "2024-02-15 21:30:45")
  expect_equal(format(time_parse("2024-02-15 9:30:45 AM", format = fmt, regex = TRUE)), "2024-02-15 09:30:45")
  # Midnight/noon are the 12-hour clock's edge cases: "12" means 0 (AM) or 12 (PM).
  expect_equal(format(time_parse("2024-02-15 12:00:00 AM", format = fmt, regex = TRUE)), "2024-02-15 00:00:00")
  expect_equal(format(time_parse("2024-02-15 12:00:00 PM", format = fmt, regex = TRUE)), "2024-02-15 12:00:00")

  fmt_cyc <- chronon_parse_cyclical(cal_gregorian$second(1L), cal_gregorian$day(1L))
  expect_equal(format(time_parse("9:30:45 PM", format = fmt_cyc, calendar = cal_gregorian)), "21:30:45")
})

test_that("chronon_parse_linear() for solar time-of-day chronons also accepts a 12-hour AM/PM clock", {
  chronon <- cal_time_solar$second(1L, lat = -37.8136, lon = 144.9631)
  fmt <- chronon_parse_linear(chronon)

  r_pm <- time_parse("5 9:30:45 PM", format = fmt, calendar = cal_time_solar)
  r_am <- time_parse("5 9:30:45 AM", format = fmt, calendar = cal_time_solar)
  expect_equal(
    as.numeric(vecvec::unvecvec(r_pm)) - as.numeric(vecvec::unvecvec(r_am)),
    12 * 3600
  )
})

# time_parse_impl() is the implementation behind time_parse(), returning a
# simple time vector (mt_linear/mt_cyclical) rather than a mixtime(); it
# exists so callers like vec_cast() (see mt_cast_from_character() in
# vctrs.R) can parse without paying for a mixtime() round-trip.
test_that("time_parse_impl() returns a simple time vector, and time_parse() is a mixtime() wrapper around it", {
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"

  r <- time_parse_impl(c("2024-02-15", "2020-01-01"), format = fmt)
  expect_true(S7::S7_inherits(r, mt_linear))
  expect_false(is_mixtime(r))

  wrapped <- time_parse(c("2024-02-15", "2020-01-01"), format = fmt)
  expect_true(is_mixtime(wrapped))
  expect_equal(wrapped, new_mixtime(r))
  expect_equal(format(wrapped), format(r))
})

test_that("time_parse_impl() returns a simple cyclical vector for a cyc()-only format", {
  r <- time_parse_impl("Feb", format = "{cyc(month, year, label = TRUE)}")
  expect_true(S7::S7_inherits(r, mt_cyclical))
  expect_false(is_mixtime(r))
})

test_that("time_parse_impl() recasts onto a simple vector when chronon is supplied", {
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"
  r <- time_parse_impl("2024-02-15", format = fmt, chronon = cal_gregorian$month(1L))
  expect_true(S7::S7_inherits(r, mt_linear))
  expect_false(is_mixtime(r))
  expect_equal(format(r), "2024 Feb")
})

test_that("time_parse_impl() edge cases (empty/all-missing input) also stay unwrapped", {
  fmt <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"

  r0 <- time_parse_impl(character(0), format = fmt)
  expect_false(is_mixtime(r0))
  expect_equal(length(r0), 0L)

  rna <- time_parse_impl(NA_character_, format = fmt)
  expect_false(is_mixtime(rna))
  expect_equal(length(rna), 1L)
  expect_true(is.na(rna))
})

register_s3_method <- function(pkg, generic, class, fun = NULL) {
  stopifnot(is.character(pkg), length(pkg) == 1)
  stopifnot(is.character(generic), length(generic) == 1)
  stopifnot(is.character(class), length(class) == 1)

  if (is.null(fun)) {
    fun <- get(paste0(generic, ".", class), envir = parent.frame())
  } else {
    stopifnot(is.function(fun))
  }

  if (pkg %in% loadedNamespaces()) {
    registerS3method(generic, class, fun, envir = asNamespace(pkg))
  }

  # Always register hook in case package is later unloaded & reloaded
  setHook(
    packageEvent(pkg, "onLoad"),
    function(...) {
      registerS3method(generic, class, fun, envir = asNamespace(pkg))
    }
  )
}

# Generated by cpp11: do not edit by hand

mt_glue_fmt <- function(input, env) {
  .Call(`_mixtime_mt_glue_fmt`, input, env)
}

approx_lunations_from_utc <- function(unix_times) {
  .Call(`_mixtime_approx_lunations_from_utc`, unix_times)
}

approx_utc_from_lunations <- function(lunation_counts) {
  .Call(`_mixtime_approx_utc_from_lunations`, lunation_counts)
}

approx_lunar_phase_from_utc <- function(unix_times) {
  .Call(`_mixtime_approx_lunar_phase_from_utc`, unix_times)
}

approx_utc_from_lunar_phase <- function(lunation_counts) {
  .Call(`_mixtime_approx_utc_from_lunar_phase`, lunation_counts)
}

approx_solar_seconds_from_utc <- function(unix_times, lat_deg, lon_deg, alt_deg) {
  .Call(`_mixtime_approx_solar_seconds_from_utc`, unix_times, lat_deg, lon_deg, alt_deg)
}

approx_utc_from_solar_seconds <- function(solar_second_counts, lat_deg, lon_deg, alt_deg) {
  .Call(`_mixtime_approx_utc_from_solar_seconds`, solar_second_counts, lat_deg, lon_deg, alt_deg)
}

approx_solar_phase_utc <- function(phase_counts, lat_deg, lon_deg, alt_deg) {
  .Call(`_mixtime_approx_solar_phase_utc`, phase_counts, lat_deg, lon_deg, alt_deg)
}

approx_solar_phase_from_utc <- function(unix_times, lat_deg, lon_deg, alt_deg) {
  .Call(`_mixtime_approx_solar_phase_from_utc`, unix_times, lat_deg, lon_deg, alt_deg)
}

approx_solar_phase_from_solar_seconds <- function(solar_second_counts, lat_deg, lon_deg, alt_deg) {
  .Call(`_mixtime_approx_solar_phase_from_solar_seconds`, solar_second_counts, lat_deg, lon_deg, alt_deg)
}

approx_solar_seconds_from_solar_phase <- function(phase_counts, lat_deg, lon_deg, alt_deg) {
  .Call(`_mixtime_approx_solar_seconds_from_solar_phase`, phase_counts, lat_deg, lon_deg, alt_deg)
}

get_tz_offset <- function(posixct, zone) {
  .Call(`_mixtime_get_tz_offset`, posixct, zone)
}

get_tz_abbreviation <- function(posixct, zone) {
  .Call(`_mixtime_get_tz_abbreviation`, posixct, zone)
}

get_tz_transitions <- function(start_posixct, end_posixct, zone_name) {
  .Call(`_mixtime_get_tz_transitions`, start_posixct, end_posixct, zone_name)
}

## TODO: These functions are crude internal implementations of graph algorithms
## that should ideally be refactored to share common code, or provided by a
## package implementing graph algorithms.

# Order a list of S7 classes (or mt_unit objects) from finest to coarsest using
# directed reachability in the chronon_cardinality method graph, where edges
# point from fine -> coarse. Returns an integer vector of indices.
# S7_order_granules <- function(classes) {
#   class_ids <- vapply(classes, S7_class_id, character(1L))

#   sigs        <- method_signatures(chronon_cardinality)
#   all_sig_cls <- vec_unique(list_unchop(sigs))
#   chr_all     <- vapply(all_sig_cls, S7_class_id, character(1L))

#   int_edges     <- vec_match(
#     unlist(lapply(sigs, function(s) vapply(s, S7_class_id, character(1L)))),
#     chr_all
#   )
#   int_edge_from <- int_edges[seq(1L, length(int_edges), by = 2L)]
#   int_edge_to   <- int_edges[seq(2L, length(int_edges), by = 2L)]

#   # Kahn's topological sort: fine nodes (in-degree 0) come first
#   n         <- length(chr_all)
#   in_degree <- tabulate(int_edge_to, nbins = n)
#   queue     <- which(in_degree == 0L)
#   topo      <- integer(0L)

#   while (length(queue) > 0L) {
#     v         <- queue[[1L]]; queue <- queue[-1L]
#     topo      <- c(topo, v)
#     nbrs      <- int_edge_to[int_edge_from == v]
#     in_degree[nbrs] <- in_degree[nbrs] - 1L
#     queue     <- c(queue, nbrs[in_degree[nbrs] == 0L])
#   }

#   int_targets <- vec_match(class_ids, chr_all)
#   order(match(int_targets, topo))
# }

method_signatures <- function(generic) {
  traverse_methods(generic@methods)
}

S7_signature_id <- function(sig) {
  # If the argument is an S7 class, return the class identifiers (package and name)
  if (!is.list(sig)) {
    return(S7_class_id(sig))
  }

  # Iterate within the signature to access S7 classes
  lapply(sig, S7_signature_id)
}

# Compile a signature list into the flattened graph structures shared by
# S7_graph_dispatch(), S7_graph_dispatch_multi() and S7_graph_glb(): the
# de-duplicated node classes, their character ids, and integer edge endpoints.
compile_signature_graph <- function(signatures) {
  classes <- vec_unique(list_unchop(signatures))
  chr_classes <- vapply(classes, S7_class_id, character(1L))

  int_edges <- vec_match(unlist(S7_signature_id(signatures)), chr_classes)

  list(
    classes = classes,
    chr_classes = chr_classes,
    edge_from = int_edges[seq(1, length(int_edges), by = 2)],
    edge_to = int_edges[seq(2, length(int_edges), by = 2)]
  )
}

# Cache of compiled signature graphs used for chronon graph dispatch.
#
# The registered S7 methods for chronon_divmod()/chronon_cardinality() are
# fixed after package load for the built-in calendars, and extensions only
# ever add methods (S7 provides no way to remove one), so the graphs
# compiled from them are cached and only rebuilt when a raw (pre-dedup)
# signature count changes. This avoids re-deriving the dispatch graph
# (unique()-ing ~35 signature objects) on every chronon_convert() call.
.chronon_signature_cache <- new.env(parent = emptyenv())

# The divmod graph only uses edges from direct chronon_divmod() methods and
# fixed (context-independent) chronon_cardinality_fixed() methods.
#
# Alongside connectivity, the graph also records which end of each edge is
# finer, so S7_graph_dispatch() can prefer a monotone path. chronon_divmod()
# signatures are direction-symmetric (methods are registered both fine -> coarse
# and coarse -> fine, since a divmod can convert either way), so they cannot
# supply this ordering - instead the ordering is provided by
# chronon_cardinality() and chronon_cardinality_fixed() signatures
chronon_divmod_graph <- function() {
  sig_divmod <- method_signatures(chronon_divmod)
  sig_fixed <- method_signatures(chronon_cardinality_fixed)
  sig_cardinality <- method_signatures(chronon_cardinality)
  n <- length(sig_divmod) + length(sig_fixed) + length(sig_cardinality)

  cache <- .chronon_signature_cache
  if (!identical(cache$divmod_n, n)) {
    cache$divmod_n <- n
    graph <- compile_signature_graph(unique(c(sig_divmod, sig_fixed)))

    fine_graph <- compile_signature_graph(unique(c(sig_fixed, sig_cardinality)))
    int_fine_nodes <- vec_match(fine_graph$chr_classes, graph$chr_classes)
    edge_fine <- int_fine_nodes[fine_graph$edge_from]
    edge_coarse <- int_fine_nodes[fine_graph$edge_to]
    # Only edges whose classes both also appear in the divmod graph are
    # usable to order its nodes; a chronon_cardinality() pair without a
    # matching chronon_divmod() connection contributes nothing there.
    keep <- !is.na(edge_fine) & !is.na(edge_coarse)
    graph$edge_fine <- edge_fine[keep]
    graph$edge_coarse <- edge_coarse[keep]

    cache$divmod_graph <- graph
  }
  cache$divmod_graph
}

chronon_cardinality_graph <- function() {
  sig_card <- method_signatures(chronon_cardinality)
  sig_fixed <- method_signatures(chronon_cardinality_fixed)
  n <- length(sig_card) + length(sig_fixed)

  cache <- .chronon_signature_cache
  if (!identical(cache$cardinality_n, n)) {
    cache$cardinality_n <- n
    cache$cardinality_graph <- compile_signature_graph(unique(c(
      sig_card,
      sig_fixed
    )))
  }
  cache$cardinality_graph
}

# The fixed-cardinality graph only uses edges from chronon_cardinality_fixed()
# methods, since these are the only relationships guaranteed to be constant
# (context-independent) and therefore safely multiplied together along a path.
chronon_cardinality_fixed_graph <- function() {
  sig_fixed <- method_signatures(chronon_cardinality_fixed)
  n <- length(sig_fixed)

  cache <- .chronon_signature_cache
  if (!identical(cache$cardinality_fixed_n, n)) {
    cache$cardinality_fixed_n <- n
    cache$cardinality_fixed_graph <- compile_signature_graph(unique(sig_fixed))
  }
  cache$cardinality_fixed_graph
}

S7_graph_dispatch <- function(graph, start, end) {
  int_node_start <- vec_match(S7_class_id(start), graph$chr_classes)
  int_node_end <- vec_match(S7_class_id(end), graph$chr_classes)
  if (is.na(int_node_start) || is.na(int_node_end)) {
    missing_units <- c(
      if (is.na(int_node_start)) S7_class_id(start),
      if (is.na(int_node_end)) S7_class_id(end)
    )
    cli::cli_abort(
      c(
        "There were no registered calendar arithmetic methods for the following classes: {paste(missing_units, collapse = \", \")}.",
        "i" = "Have you registered calendar arithmetic S7 methods for these classes?",
        ">" = "See the {.vignette mixtime::extending-mixtime} vignette for more details."
      ),
      call. = FALSE
    )
  }

  # Prefer a monotone path, where every step moves the same way through the
  # granularities: only coarser when converting to a coarser chronon, only finer
  # when converting to a finer one.
  #
  # This is safer than shorter paths in both directions (e.g. 4 quarters in a
  # year are not equally sized, so quarters -> years -> days is less safe than
  # quarters -> months -> days).
  int_path <- integer()
  if (!is.null(graph$edge_fine)) {
    int_path <- bfs_shortest_path(
      from = graph$edge_fine,
      to = graph$edge_coarse,
      start = int_node_start,
      end = int_node_end,
      directed = TRUE
    )
    if (length(int_path) == 0L) {
      int_path <- bfs_shortest_path(
        from = graph$edge_coarse,
        to = graph$edge_fine,
        start = int_node_start,
        end = int_node_end,
        directed = TRUE
      )
    }
  }

  # Conversions between units that no monotone path connects (a week is neither
  # finer nor coarser than a month) still need the unrestricted search.
  if (length(int_path) == 0L) {
    int_path <- bfs_shortest_path(
      from = graph$edge_from,
      to = graph$edge_to,
      start = int_node_start,
      end = int_node_end
    )
  }

  if (length(int_path) == 0L) {
    cli::cli_abort(
      c(
        "There is no path of registered calendar arithmetic methods between the classes {S7_class_id(start)} and {S7_class_id(end)}.",
        "i" = "Have you registered calendar arithmetic S7 methods connecting these classes?",
        ">" = "See the {.vignette mixtime::extending-mixtime} vignette for more details."
      ),
      call. = FALSE
    )
  }

  # Instantiate path of classed S7 objects for dispatch
  graph$classes[int_path]
}

# Directed reachability: is `end` reachable from `start` by following edges in
# the from -> to direction only? Used to test the fine -> coarse ordering of the
# chronon_cardinality graph (undirected BFS above would ignore direction).
directed_reachable <- function(from, to, start, end) {
  frontier <- start
  visited <- start
  while (length(frontier) > 0L) {
    reached <- to[from %in% frontier]
    if (end %in% reached) {
      return(TRUE)
    }
    frontier <- setdiff(reached, visited)
    visited <- c(visited, frontier)
  }
  FALSE
}

# TRUE if `chronon` nests within `granule` (no `chronon` block ever splits
# across a `granule` boundary). Defers to the exact ratio check when a fixed
# path connects the classes; over an irregular-only path, only `n = 1` is
# provably safe (per-`at` divisibility isn't exhaustively checkable), so
# `n > 1` errors instead of guessing.
chronon_nests_in <- function(chronon, granule) {
  if (S7_class_id(chronon) == S7_class_id(granule)) {
    return(chronon_nests_in_same_unit(chronon, granule))
  }
  if (chronon_fixed_reachable(chronon, granule)) {
    return(chronon_nests_in_fixed_ratio(chronon, granule))
  }

  if (!chronon_graph_reachable(chronon_cardinality_graph(), chronon, granule)) {
    return(FALSE)
  }
  if (chronon@n != 1L) {
    cli::cli_abort(
      c(
        "Cannot determine whether {time_granule_label(chronon)} nest within {time_granule_label(granule)}.",
        "x" = "{S7_class_id(chronon)} and {S7_class_id(granule)} are only connected by an irregular (context-dependent) cardinality, and nesting for a multi-unit chronon of one of these classes can depend on where in the calendar it falls.",
        "i" = "This is only resolvable for a single-unit ({.code n = 1}) chronon."
      ),
      call = NULL
    )
  }
  TRUE
}

# Same-class nesting: TRUE iff `granule@n %% chronon@n == 0`.
chronon_nests_in_same_unit <- function(chronon, granule) {
  granule@n %% chronon@n == 0L
}

# TRUE if `granule`'s class is reachable from `chronon`'s class along
# `graph`'s fine -> coarse edges.
chronon_graph_reachable <- function(graph, chronon, granule) {
  start <- match(S7_class_id(chronon), graph$chr_classes)
  end <- match(S7_class_id(granule), graph$chr_classes)
  !is.na(start) &&
    !is.na(end) &&
    directed_reachable(graph$edge_from, graph$edge_to, start, end)
}

# TRUE if `chronon`'s and `granule`'s classes are connected only by
# chronon_cardinality_fixed() edges.
chronon_fixed_reachable <- function(chronon, granule) {
  chronon_graph_reachable(chronon_cardinality_fixed_graph(), chronon, granule)
}

# Ratio-only nesting check, assuming chronon_fixed_reachable(chronon,
# granule) already holds: nests iff `ratio * granule@n %% chronon@n == 0`,
# where `ratio` is fixed units of `chronon` per unit of `granule` (e.g. 12
# months/year). Split out so chronon_nests_in() can reuse it without
# repeating the reachability check.
chronon_nests_in_fixed_ratio <- function(chronon, granule) {
  ratio <- chronon_cardinality_fixed(
    S7::S7_class(chronon)(1L),
    S7::S7_class(granule)(1L)
  )
  (ratio * granule@n) %% chronon@n == 0L
}

# TRUE if `chronon` nests within `granule` using only fixed (`at`-independent)
# cardinality relationships.
chronon_nests_in_fixed <- function(chronon, granule) {
  if (S7_class_id(chronon) == S7_class_id(granule)) {
    return(chronon_nests_in_same_unit(chronon, granule))
  }

  if (!chronon_fixed_reachable(chronon, granule)) {
    return(FALSE)
  }
  chronon_nests_in_fixed_ratio(chronon, granule)
}

S7_class_id <- function(x) {
  if (inherits(x, "S7_object") && !inherits(x, "S7_class")) {
    # An S7 object's class attribute already begins with its class id, so an
    # instance needs neither a class lookup nor any property access.
    return(class(x)[[1L]])
  }
  # A class not attached to a package is identified by its name alone.
  package <- x@package
  if (is.null(package)) x@name else paste0(package, "::", x@name)
}

traverse_methods <- function(x) {
  if (!is.environment(x)) {
    return(list(x@signature))
  }

  methods <- lapply(names(x), function(class) traverse_methods(x[[class]]))
  unlist(methods, recursive = FALSE)
}


bfs_shortest_path <- function(
  from = integer(),
  to = integer(),
  start = integer(),
  end = integer(),
  directed = FALSE
) {
  # Input validation
  if (length(from) != length(to)) {
    return(integer(0))
  }

  if (length(start) != 1 || length(end) != 1) {
    return(integer(0))
  }

  if (length(from) == 0) {
    return(integer(0))
  }

  # Convert to integers
  from <- as.integer(from)
  to <- as.integer(to)
  start <- as.integer(start)
  end <- as.integer(end)

  # If start equals end, return just the start vertex
  if (start == end) {
    return(start)
  }

  # Edges are walked in both directions unless `directed`. Interleaving the two
  # directions of an edge (rather than listing every forward direction and then
  # every reverse one) leaves each vertex's neighbours in edge order.
  n <- max(from, to, start, end)
  if (directed) {
    tail <- from
    head <- to
  } else {
    tail <- c(rbind(from, to))
    head <- c(rbind(to, from))
  }

  # Compressed adjacency, built with whole-vector operations rather than an
  # edge-by-edge loop: `nbr` holds every vertex's neighbours end to end, `deg`
  # how many each has, and `off` where each vertex's run starts. The explicit
  # second sort key keeps neighbours in edge order without relying on the sort
  # being stable.
  ord <- order(tail, seq_along(tail))
  nbr <- head[ord]
  deg <- tabulate(tail, nbins = n)
  off <- c(0L, cumsum(deg))

  visited <- logical(n)
  parent <- integer(n)
  visited[[start]] <- TRUE
  frontier <- start

  # Expand a whole BFS level per iteration: gathering the frontier's neighbour
  # runs in one indexing operation makes the number of interpreted iterations
  # proportional to the length of the path rather than to the size of the graph.
  while (length(frontier) > 0L && !visited[[end]]) {
    d <- deg[frontier]
    adjacent <- nbr[rep.int(off[frontier], d) + sequence(d)]
    origin <- rep.int(frontier, d)

    # Keep each vertex's first unvisited occurrence, which is the one a queue
    # popping the frontier in order would have reached first.
    new <- !visited[adjacent] & !duplicated(adjacent)
    frontier <- adjacent[new]
    parent[frontier] <- origin[new]
    visited[frontier] <- TRUE
  }

  # No path found
  if (!visited[[end]]) {
    return(integer(0))
  }

  # Reconstruct the path by walking parents back from `end`, filling a buffer
  # from the right so it needs no growing.
  path <- integer(n)
  i <- n
  node <- end
  while (node != start) {
    path[[i]] <- node
    node <- parent[[node]]
    i <- i - 1L
  }
  path[[i]] <- start
  path[i:n]
}

# Longest simple path from `start` to `end` following directed edges `from -> to`.
#
# Intended to be used with the DAG formed by chronon_cardinality methods. The
# longest path is useful for resolving invalid time points with clamping.
dag_longest_path <- function(
  from = integer(),
  to = integer(),
  start = integer(),
  end = integer()
) {
  if (length(from) != length(to)) {
    return(integer())
  }
  if (length(start) != 1L || length(end) != 1L) {
    return(integer())
  }
  if (length(from) == 0L) {
    return(integer())
  }

  from <- as.integer(from)
  to <- as.integer(to)
  start <- as.integer(start)
  end <- as.integer(end)
  if (start == end) {
    return(start)
  }

  n <- max(from, to, start, end)

  reachable <- function(seed, tail, head) {
    visited <- logical(n)
    visited[[seed]] <- TRUE
    frontier <- seed
    while (length(frontier) > 0L) {
      nbrs <- head[tail %in% frontier]
      nbrs <- unique(nbrs[!visited[nbrs]])
      visited[nbrs] <- TRUE
      frontier <- nbrs
    }
    visited
  }

  # Restrict to the sub-DAG of nodes lying on some start -> end path: irrelevant branches
  # can't affect the longest path and would otherwise need visiting during the DP below.
  fwd <- reachable(start, from, to)
  if (!fwd[[end]]) {
    return(integer())
  }
  bwd <- reachable(end, to, from)
  on_path <- fwd & bwd
  keep <- on_path[from] & on_path[to]
  sub_from <- from[keep]
  sub_to <- to[keep]

  # Kahn's topological sort of the sub-DAG (start always has in-degree 0 within it: an edge
  # into start would imply a path start -> ... -> start, i.e. a cycle).
  nodes <- which(on_path)
  in_degree <- tabulate(sub_to, nbins = n)
  queue <- nodes[in_degree[nodes] == 0L]
  topo <- integer(0L)
  while (length(queue) > 0L) {
    v <- queue[[1L]]
    queue <- queue[-1L]
    topo <- c(topo, v)
    nbrs <- sub_to[sub_from == v]
    for (nb in nbrs) {
      in_degree[[nb]] <- in_degree[[nb]] - 1L
      if (in_degree[[nb]] == 0L) queue <- c(queue, nb)
    }
  }

  # Longest-path DP: relax every edge once, in topological order.
  dist <- rep(-Inf, n)
  dist[[start]] <- 0L
  parent <- integer(n)
  for (v in topo) {
    if (!is.finite(dist[[v]])) {
      next
    }
    nbrs <- sub_to[sub_from == v]
    for (nb in nbrs) {
      if (dist[[v]] + 1L > dist[[nb]]) {
        dist[[nb]] <- dist[[v]] + 1L
        parent[[nb]] <- v
      }
    }
  }
  if (!is.finite(dist[[end]])) {
    return(integer())
  }

  path <- end
  node <- end
  while (node != start) {
    node <- parent[[node]]
    path <- c(node, path)
  }
  path
}

S7_graph_dispatch_multi <- function(
  graph,
  start,
  terminals = list(),
  groups = list()
) {
  chr_classes <- graph$chr_classes

  int_start <- vec_match(S7_class_id(start), chr_classes)

  # terminals: individual nodes that just need to appear somewhere in the tree
  int_terminals <- vec_match(
    vapply(terminals, S7_class_id, character(1L)),
    chr_classes
  )

  # groups: lists of nodes that must all co-occur on the same root-to-leaf path
  int_groups <- lapply(groups, function(group) {
    if (!is.list(group)) {
      group <- list(group)
    }
    vec_match(vapply(group, S7_class_id, character(1L)), chr_classes)
  })

  int_tree <- steiner_tree_paths(
    from = graph$edge_from,
    to = graph$edge_to,
    start = int_start,
    terminals = int_terminals,
    groups = int_groups
  )

  # Nodes are conversion waypoints rather than the granules asked for, so each
  # resolves to the finest granule of its time unit: `start` for the unit the
  # time points are measured in (which cannot be subdivided further), and a
  # single unit elsewhere. Resolving to a requested granule instead would make a
  # node as coarse as the coarsest granule requested of it, discarding the
  # remainder that finer granules of the same unit need. Callers rescale each
  # requested granule from its node, which relies on no node being coarser than
  # a granule requested of it.
  start_id <- S7_class_id(start)

  # A granule the caller already supplied is reused when it is itself no coarser
  # than a single unit: it needs no rebuilding and carries their tz / location.
  supplied <- c(terminals, unlist(groups, recursive = FALSE))
  supplied_id <- vapply(supplied, S7_class_id, character(1L))
  supplied_n <- vapply(supplied, function(granule) granule@n, numeric(1L))

  # Properties only need copying onto a constructed node when `start` holds one
  # that is not naive (e.g. a time zone) for it to inherit.
  start_informs <- any(
    !is.naive(props(start)[setdiff(names(props(start)), "n")])
  )

  resolve_tree <- function(node) {
    node_id <- chr_classes[[node$node]]
    finest <- which(supplied_id == node_id)
    if (length(finest)) {
      finest <- finest[[which.min(supplied_n[finest])]]
    }

    resolved <- if (identical(node_id, start_id)) {
      start
    } else if (length(finest) && supplied_n[[finest]] <= 1) {
      supplied[[finest]]
    } else {
      granule <- graph$classes[[node$node]](1L)
      if (start_informs) granule_inherit_props(granule, start) else granule
    }

    list(
      node = resolved,
      children = lapply(node$children, resolve_tree)
    )
  }
  resolve_tree(int_tree)
}

# Returns a nested list: list(node = int, children = list(...))
# Groups are inserted first (chained BFS ensuring co-occurrence on one path).
# Terminals are only inserted if not already present in the tree.
steiner_tree_paths <- function(
  from = integer(),
  to = integer(),
  start = integer(),
  terminals = integer(),
  groups = list()
) {
  if (length(from) != length(to)) {
    return(list())
  }
  if (length(start) != 1L) {
    return(list())
  }
  if (length(from) == 0L) {
    return(list())
  }

  # Collect all nodes currently present in the tree
  tree_nodes <- function(tree) {
    c(tree$node, unlist(lapply(tree$children, tree_nodes)))
  }

  # For a group of nodes, order by BFS depth from start (shallowest first),
  # then chain BFS segments: start -> g1 -> g2 -> ...
  group_path <- function(group) {
    if (length(group) == 1L) {
      return(bfs_shortest_path(
        from = from,
        to = to,
        start = start,
        end = group
      ))
    }
    depths <- vapply(
      group,
      function(node) {
        p <- bfs_shortest_path(from = from, to = to, start = start, end = node)
        if (length(p) == 0L) Inf else length(p) - 1L
      },
      numeric(1L)
    )
    ordered_group <- group[order(depths)]

    waypoints <- c(start, ordered_group)
    path <- integer(0)
    for (i in seq_len(length(waypoints) - 1L)) {
      segment <- bfs_shortest_path(
        from = from,
        to = to,
        start = waypoints[[i]],
        end = waypoints[[i + 1L]]
      )
      if (length(segment) == 0L) {
        return(integer(0))
      }
      path <- c(path, if (i == 1L) segment else segment[-1L])
    }
    path
  }

  insert_path <- function(tree, path) {
    if (length(path) == 0L) {
      return(tree)
    }

    node <- path[[1L]]
    rest <- path[-1L]

    child_idx <- which(vapply(
      tree$children,
      function(c) c$node == node,
      logical(1L)
    ))

    if (length(child_idx) == 0L) {
      new_child <- insert_path(list(node = node, children = list()), rest)
      tree$children <- c(tree$children, list(new_child))
    } else {
      tree$children[[child_idx]] <- insert_path(
        tree$children[[child_idx]],
        rest
      )
    }

    tree
  }

  tree <- list(node = start, children = list())

  # 1. Insert group paths first (enforce co-occurrence)
  for (group in groups) {
    path <- group_path(group)
    if (length(path) > 1L) {
      tree <- insert_path(tree, path[-1L])
    }
  }

  # 2. Insert terminals only if not already in the tree
  for (terminal in terminals) {
    if (!terminal %in% tree_nodes(tree)) {
      path <- bfs_shortest_path(
        from = from,
        to = to,
        start = start,
        end = terminal
      )
      if (length(path) > 1L) {
        tree <- insert_path(tree, path[-1L])
      }
    }
  }

  tree
}

S7_graph_glb <- function(graph, chronons) {
  int_chronons <- vec_match(
    vapply(chronons, S7_class_id, character(1L)),
    graph$chr_classes
  )

  int_glb <- greatest_lower_bound(
    from = graph$edge_from,
    to = graph$edge_to,
    nodes = int_chronons
  )

  if (rlang::is_empty(int_glb)) {
    stop(
      "One or more of the provided chronons do not share a common chronon.",
      call. = FALSE
    )
  }

  # Return glb
  graph$classes[[int_glb]]
}

# Finds the greatest lower bound that contains all `nodes` in a graph defined by
# directed edges from `from` to `to`.
greatest_lower_bound <- function(
  from = integer(),
  to = integer(),
  nodes = integer()
) {
  if (length(nodes) == 1) {
    return(nodes)
  }
  all_nodes <- unique(c(from, to))

  # Build parent map: each node -> set of its parents
  parent_map <- split(from, to)

  # Helper: Function to get all ancestors (including self) for a node
  get_ancestors <- function(node) {
    stack <- node
    visited <- logical()
    ancestors <- node
    while (length(stack) > 0) {
      current <- stack[[1]]
      stack <- stack[-1]
      parents <- parent_map[[as.character(current)]]
      # visit parents not yet visited
      if (!is.null(parents)) {
        new_parents <- setdiff(parents, ancestors)
        if (length(new_parents) > 0) {
          ancestors <- c(ancestors, new_parents)
          stack <- c(stack, new_parents)
        }
      }
    }
    ancestors
  }

  # Find all ancestors for each node in 'nodes'
  ancestors_list <- lapply(nodes, get_ancestors)

  # The GLB are the nodes that are common to all ancestor sets
  common_ancestors <- Reduce(intersect, ancestors_list)

  if (length(common_ancestors) == 0) {
    return(integer(0))
  } # No common ancestor

  # Among common ancestors, pick the one that is farthest from root (i.e., closest to nodes)
  # For this, compute for each such node the minimal distance to any of the target nodes

  node_depth <- function(target, candidate) {
    # Walk down from candidate to target
    # We'll perform BFS from candidate to target
    queue <- list(candidate)
    depth_map <- `names<-`(0, candidate)
    visited <- candidate
    while (length(queue) > 0) {
      current <- queue[[1]]
      queue <- queue[-1]
      if (current == target) {
        return(depth_map[[as.character(current)]])
      }
      # Find "children" (those where current is parent)
      children <- to[which(from == current)]
      new_children <- setdiff(children, visited)
      if (length(new_children) > 0) {
        queue <- c(queue, new_children)
        new_depths <- depth_map[[as.character(current)]] + 1
        depth_map <- c(
          depth_map,
          `names<-`(rep(new_depths, length(new_children)), new_children)
        )
        visited <- c(visited, new_children)
      }
    }
    Inf # No path found
  }

  distance_to_nodes <- function(candidate) {
    # For each node in the set, find the length of the path from candidate to node
    min(
      vapply(nodes, node_depth, numeric(1L), candidate = candidate),
      na.rm = TRUE
    )
  }

  # For each common ancestor, compute max depth to any node in 'nodes'.
  # Take the one with *minimal* (i.e., greatest lower bound, or furthest from root)
  depths <- vapply(common_ancestors, distance_to_nodes, numeric(1L))
  # The greatest lower bound is the one with minimal distance to its furthest descendant in 'nodes'
  glb <- common_ancestors[which.min(depths)]

  as.integer(glb)
}

test_that("chronon_cardinality_fixed() returns the unit-granule constant", {
  expect_equal(chronon_cardinality_fixed(cal_time_civil$hour(1L), cal_time_civil$day(1L)), 24L)
  expect_equal(chronon_cardinality_fixed(cal_isoweek$day(1L), cal_isoweek$week(1L)), 7L)

  # Variable relationships are not registered as chronon_cardinality_fixed()
  expect_error(chronon_cardinality_fixed(cal_gregorian$day(1L), cal_gregorian$month(1L)))
})

test_that("chronon_cardinality_fixed() traverses edges in either direction", {
  # A coarser `x` than `y` walks the graph against the fine -> coarse direction
  # its edges are registered in, giving the reciprocal.
  expect_equal(chronon_cardinality_fixed(cal_gregorian$year(1L), cal_gregorian$quarter(1L)), 1 / 4)
  expect_equal(chronon_cardinality_fixed(cal_time_civil$day(1L), cal_time_civil$hour(1L)), 1 / 24)

  # Multi-step paths mix the two directions
  expect_equal(chronon_cardinality_fixed(cal_time_civil$second(1L), cal_time_civil$day(1L)), 86400)
  expect_equal(chronon_cardinality_fixed(cal_time_civil$day(1L), cal_time_civil$second(1L)), 1 / 86400)
  expect_equal(chronon_cardinality_fixed(cal_gregorian$quarter(1L), cal_gregorian$month(1L)), 1 / 3)

  # A pair with no path of fixed relationships is still an error, not a guess
  expect_error(chronon_cardinality_fixed(cal_gregorian$month(1L), cal_gregorian$day(1L)))
})

test_that("chronon_cardinality() falls back to chronon_cardinality_fixed(), scaled by granule size", {
  expect_equal(chronon_cardinality(cal_time_civil$hour(1L), cal_time_civil$day(1L)), 24)
  expect_equal(chronon_cardinality(cal_time_civil$hour(2L), cal_time_civil$day(3L)), 24 * 3 / 2)

  # The inverse (coarser, finer) order is automatically derived
  expect_equal(chronon_cardinality(cal_gregorian$year(1L), cal_gregorian$month(1L)), 1 / 12)
})

test_that("only fixed cardinality (and direct divmod) relationships form divmod graph edges", {
  tu_fixed_test_a <- S7::new_class("tu_fixed_test_a", parent = mt_unit)
  tu_fixed_test_b <- S7::new_class("tu_fixed_test_b", parent = mt_unit)
  tu_variable_test_a <- S7::new_class("tu_variable_test_a", parent = mt_unit)
  tu_variable_test_b <- S7::new_class("tu_variable_test_b", parent = mt_unit)

  method(chronon_cardinality_fixed, list(tu_fixed_test_a, tu_fixed_test_b)) <- function(x, y) 5L
  method(chronon_cardinality, list(tu_variable_test_a, tu_variable_test_b)) <- function(x, y, at = NULL) {
    if (is.null(at)) cli::cli_abort("requires `at`")
    5L
  }

  edge_exists <- function(graph, a, b) {
    ia <- vec_match(S7_class_id(a), graph$chr_classes)
    ib <- vec_match(S7_class_id(b), graph$chr_classes)
    if (is.na(ia) || is.na(ib)) return(FALSE)
    any(graph$edge_from == ia & graph$edge_to == ib) ||
      any(graph$edge_from == ib & graph$edge_to == ia)
  }

  # Both relationships are visible to the cardinality graph, used for
  # ordering/topology by chronon_common() and time_parts().
  expect_true(edge_exists(chronon_cardinality_graph(), tu_fixed_test_a(1L), tu_fixed_test_b(1L)))
  expect_true(edge_exists(chronon_cardinality_graph(), tu_variable_test_a(1L), tu_variable_test_b(1L)))

  # Only the fixed relationship is safe to use for chronon_divmod() graph
  # traversal, since chronon_divmod_regular() calls chronon_cardinality()
  # without an `at`.
  expect_true(edge_exists(chronon_divmod_graph(), tu_fixed_test_a(1L), tu_fixed_test_b(1L)))
  expect_false(edge_exists(chronon_divmod_graph(), tu_variable_test_a(1L), tu_variable_test_b(1L)))
})

test_that("chronon_nests_in_fixed() checks divisibility, not just reachability", {
  # Regression test: previously ignored `@n` entirely for any cross-class
  # pair, returning TRUE whenever the classes were merely connected in the
  # fixed-cardinality graph (e.g. month -> year), regardless of whether the
  # chronon's own size tiled evenly into the granule. 12 months/year is not a
  # whole multiple of 5, so a 5-month chronon does not nest in a 1-year
  # granule - 2, 3, 4, 6, 12 do (all divisors of 12).
  expect_false(chronon_nests_in_fixed(cal_gregorian$month(5L), cal_gregorian$year(1L)))
  expect_false(chronon_nests_in_fixed(cal_gregorian$month(7L), cal_gregorian$year(1L)))
  expect_false(chronon_nests_in_fixed(cal_gregorian$month(13L), cal_gregorian$year(1L)))
  for (n in c(1L, 2L, 3L, 4L, 6L, 12L)) {
    expect_true(
      chronon_nests_in_fixed(cal_gregorian$month(n), cal_gregorian$year(1L)),
      info = paste("n =", n)
    )
  }

  # Scaling the granule's own `@n` changes what divides evenly too (e.g. 5
  # months nests in 5 years: 60 months / 5 == 12).
  expect_true(chronon_nests_in_fixed(cal_gregorian$month(5L), cal_gregorian$year(5L)))
  expect_true(chronon_nests_in_fixed(cal_gregorian$month(4L), cal_gregorian$year(2L)))

  # Multi-hop fixed chain (month -> quarter -> year), not just a directly
  # registered edge.
  expect_true(chronon_nests_in_fixed(cal_gregorian$month(1L), cal_gregorian$quarter(1L)))
  expect_true(chronon_nests_in_fixed(cal_gregorian$quarter(1L), cal_gregorian$year(1L)))

  # Same-class case is untouched (n divisibility only).
  expect_true(chronon_nests_in_fixed(cal_gregorian$month(2L), cal_gregorian$month(6L)))
  expect_false(chronon_nests_in_fixed(cal_gregorian$month(5L), cal_gregorian$month(6L)))

  # No registered fixed relationship at all (day <-> month is irregular).
  expect_false(chronon_nests_in_fixed(cal_gregorian$day(1L), cal_gregorian$month(1L)))
})

test_that("chronon_nests_in() matches chronon_nests_in_fixed() for any fixed-reachable pair", {
  # A fixed path's ratio is constant regardless of `at`, so if it doesn't
  # divide evenly, the pair never nests - no other (necessarily irregular)
  # path could make it nest either. chronon_nests_in() defers to
  # chronon_nests_in_fixed()'s exact answer for these pairs rather than
  # stopping at plain reachability.
  expect_false(chronon_nests_in(cal_gregorian$month(5L), cal_gregorian$year(1L)))
  expect_true(chronon_nests_in(cal_gregorian$month(4L), cal_gregorian$year(1L)))
})

test_that("chronon_nests_in() checks @n for pairs only reachable via an irregular edge", {
  # day <-> month has no fixed cardinality (see chronon_nests_in_fixed() test
  # above), so every day <-> month/year path crosses an irregular edge. A
  # single-unit chronon always nests (a day is never split across a month
  # boundary), regardless of the granule's own `@n`.
  expect_true(chronon_nests_in(cal_gregorian$day(1L), cal_gregorian$month(5L)))
  expect_true(chronon_nests_in(cal_gregorian$day(1L), cal_gregorian$year(1L)))

  # A multi-unit chronon's blocks can straddle a month boundary depending on
  # where in the calendar they fall (e.g. day(2) blocks land differently
  # depending on the parity of the month's start day-of-epoch) - this can't be
  # resolved without a specific `at`, so it errors rather than guessing.
  expect_error(chronon_nests_in(cal_gregorian$day(2L), cal_gregorian$month(1L)))
  expect_error(chronon_nests_in(cal_gregorian$day(7L), cal_gregorian$year(1L)))

  # No registered relationship at all is still a plain FALSE, not an error.
  expect_false(chronon_nests_in(cal_gregorian$month(1L), cal_time_lunar$phase(1L)))
})

#' Default parsing format strings for chronons
#'
#' Provides candidate format strings for parsing text into a given chronon
#' (finest time granule), for use as the `format` argument of
#' [time_parse()]. Dispatches the same way as
#' [chronon_format_linear()]/[chronon_format_cyclical()], but returns every
#' common format instead of a single default, so [time_parse()] can try each
#' in turn and keep whichever parses the most values. Methods should build
#' their return value with [parse_format()].
#'
#' @inheritParams chronon_format_linear
#'
#' @return A character vector of format templates, ordered from most to
#'   least common, built with [parse_format()]. The first element typically
#'   matches [chronon_format_linear()] (for `chronon_parse_linear()`) or
#'   [chronon_format_cyclical()] (for `chronon_parse_cyclical()`) for the
#'   same chronon.
#'
#' @seealso [parse_format()] for building candidate format strings (what
#'   these methods are built from), [time_parse()] for using the result as
#'   candidate formats, [chronon_format_linear()]/[chronon_format_cyclical()]
#'   for the single default format these are based on.
#'
#' @export
#' @examples
#' chronon_parse_linear(cal_gregorian$year(1L))
#' chronon_parse_linear(cal_gregorian$month(1L))
#' chronon_parse_linear(cal_gregorian$day(1L))
#' chronon_parse_linear(cal_isoweek$day(1L))
#'
#' @rdname chronon_parse
chronon_parse_linear <- new_generic("chronon_parse_linear", c("x", "cal"), function(x, cal = time_calendar(x), ...) {
  S7::S7_dispatch()
})
method(chronon_parse_linear, list(mt_unit, class_any)) <- function(x, cal) {
  chronon_format_linear(x, cal)
}

#' @examples
#' chronon_parse_cyclical(cal_gregorian$month(1L), cal_gregorian$year(1L))
#' chronon_parse_cyclical(cal_isoweek$day(1L), cal_isoweek$week(1L))
#'
#' @rdname chronon_parse
#' @export
chronon_parse_cyclical <- new_generic("chronon_parse_cyclical", c("x", "y"))
method(chronon_parse_cyclical, list(mt_unit, mt_unit)) <- function(x, y) {
  chronon_format_cyclical(x, y)
}

#' Candidate format strings for parsing time
#'
#' Combines one or more format strings into a character vector for use as
#' the `format` argument of [time_parse()]. When multiple format strings are
#' given, [time_parse()] tries each in turn and keeps whichever parses the
#' most values, so `parse_format()` is the usual way to build up a set of
#' candidates to try (it's also how [chronon_parse_linear()]/
#' [chronon_parse_cyclical()] methods build theirs). It can optionally mark
#' its candidates as using regex syntax rather than literal text; see the
#' `regex` argument below.
#'
#' @param ... Format strings, as for the `format` argument of [time_parse()].
#' @param regex Whether the literal (non-token) text in `...` should be
#'   matched as regular expression syntax rather than escaped literally
#'   (e.g. `"[/-]"` to accept either `/` or `-` as a separator). This is a
#'   niche option for irregular text - most formats leave it at the
#'   default, `FALSE`. See the `regex` argument of [time_parse()] for
#'   details.
#'
#' @return A character vector of format strings suitable for [time_parse()],
#'   with a `"regex"` attribute for the `regex` argument the parser.
#' 
#' @seealso [time_parse()] for using the result as `format`,
#'   [chronon_parse_linear()]/[chronon_parse_cyclical()] for calendar-specific
#'   candidates built this way.
#' @export
#' @examples
#' parse_format("{lin(year)}-{cyc(month, year)}-{cyc(day, month)}")
#'
#' # Multiple candidates are tried in turn, keeping whichever parses most values
#' parse_format(
#'   "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}",
#'   "{lin(year)}/{cyc(month, year)}/{cyc(day, month)}"
#' )
#'
#' # regex = TRUE treats the surrounding text as regular expression syntax
#' parse_format(
#'   "{lin(year)}[/-]{cyc(month, year)}[/-]{cyc(day, month)}",
#'   regex = TRUE
#' )
parse_format <- function(..., regex = FALSE) {
  structure(as.character(c(...)), regex = regex)
}

# Helper function for inheriting the 'date' parsing string and extending it
# with time parsing strings for cal_time_* time systems
parse_format_time_of_day <- function(cal, suffix, suffix_regex = suffix) {
  date_fmts <- chronon_parse_linear(cal$day(1L), cal)
  regex <- isTRUE(attr(date_fmts, "regex"))
  suffixes <- if (regex) suffix_regex else suffix
  parse_format(as.vector(outer(date_fmts, suffixes, paste0)), regex = regex)
}

#' Find the common chronon of a time object
#'
#' @description
#'
#' This utility function takes a set of chronons and identifies a common chronon
#' of the finest granularity that can represent all input chronons without loss
#' of information. This is useful for operations that require a shared time
#' granule, such as combining or comparing different time measured at different
#' precisions.
#'
#' The result is obtained by finding the greatest lower bound (GLB) of the input
#' chronons using the ordered relationships defined by `chronon_cardinality()`
#' methods. The GLB represents the finest chronon that can represent all input
#' chronons without loss of information.
#'
#' @param x A time object (typically a [`mixtime`]).
#' @param .ptype If NULL, the default, the output returns the common chronon
#' across all chronons of `x`. Alternatively, a prototype chronon can be
#' supplied to `.ptype` to demand a specific chronon is used. If the supplied
#' `.ptype` cannot represent all input chronons without loss of information,
#' an error is raised.
#'
#' @param ... Additional arguments for methods.
#'
#' @return A time granule object representing the common chronon.
#'
#' @examples
#' # The common chronon between a year-month and a day is a day
#' chronon_common(c(yearmonth(Sys.Date()), date(Sys.Date())))
#'
#' # The common chronon between a Gregorian month and an ISO week is a day
#' chronon_common(c(yearmonth(Sys.Date()), yearweek(Sys.Date())))
#'
#' # The common chronon between a ISO week and an hour is an hour
#' chronon_common(c(yearweek(Sys.Date()), linear_time(Sys.time(), hour(1L))))
#'
#' @name chronon_common
#' @export
chronon_common <- S7::new_generic("chronon_common", "x")


#' @rdname chronon_common
chronon_common.mixtime <- function(x, .ptype = NULL, ...) {
  cli::cli_abort(
    c(
      "This method is for documentation purposes only. The actual method is S7::method(chronon_common, class_mixtime).",
      i = "This documentation workaround is pending https://github.com/r-lib/roxygen2/issues/1872"
    ),
    call = NULL
  )
}

S7::method(chronon_common, class_mixtime) <- function(x, .ptype = NULL, ...) {
  chronon_common_impl(
    lapply(x@x, attr, "chronon"),
    .ptype = .ptype
  )
}

S7::method(chronon_common, class_any) <- function(x, .ptype = NULL, ...) {
  chronon_common_impl(
    list(attr(time_chronon(x[1L])@x[[1L]], "chronon")),
    .ptype = .ptype
  )
}

chronon_common_impl <- function(chronons, .ptype = NULL) {
  # TODO: Validate that the supplied .ptype can represent all input chronons
  if (!is.null(.ptype)) {
    return(.ptype)
  }

  chronons <- unique(chronons)
  if (vec_size(chronons) == 1L) {
    return(chronons[[1L]])
  }

  # Search strategy:
  # The cardinality methods are directional
  # (such that `x` is a finer chronon than `y`)
  # Construct a graph of the methods to find the common root chronon.
  glb <- S7_graph_glb(chronon_cardinality_graph(), chronons)(1L)

  # Restore properties of the common chronon from the input chronons.
  # If the properties (e.g. timezone) vary across the input chronons, they are
  # dropped from the common chronon (resulting in a timezone-naive chronon).
  common <- granule_inherit_shared_props(glb, chronons)

  # If the common chronon has a naive time zone, then prevent it from inheriting
  # timezone attributes in downstream operations.
  granule_harden_naive(common)
}

#' Round, floor and ceiling transformations for time objects
#'
#' A family of helpers to round date/time objects to a specified time granule
#' such as second, minute, hour, or day. These functions preserve the input 
#' time class, as rounded by the attributes of the `granule`.
#'
#' @name time_round
#' @aliases time_round time_floor time_ceiling trunc_time
#'
#' @param x A date/time object to be rounded. Accepted types include Date,
#'   POSIXct, POSIXlt and other objects that inherit from POSIXt. The returned
#'   object will be of the same class as the input.
#' @param granule A time granule (or object coercible to a time granule, e.g. "day").
#' @param ... Additional arguments passed to specific implementations.
#'
#' @return An object of the same class as x with its time components adjusted
#'   to the requested granule.
#'
#' @examples
#' # Round POSIXct to the nearest minute (preserving tz)
#' t <- as.POSIXct("2020-01-01 12:34:56", tz = "UTC")
#' time_round(t, granule = cal_gregorian$minute(1L))
#'
#' # Floor to the nearest hour
#' time_floor(t, granule = cal_gregorian$hour(1L))
#'
#' # Ceiling a Date (treated as midnight-of-day rounding)
#' d <- as.Date("2020-01-01")
#' time_ceiling(d, granule = cal_gregorian$month(1L))
#'
#' @seealso [base::round], [lubridate::round_date]
#' @export
time_round <- new_generic("time_round", "x", function(x, granule, ...) {
  S7::S7_dispatch()
})
method(time_round, class_mixtime) <- vecvec::vecvec_apply_fn(time_round)
method(time_round, S7::class_any) <- function(x, granule, ...) {
  time_round_impl(x, granule, round)
}

#' @rdname time_round
#' @export
time_ceiling <- new_generic("time_ceiling", "x", function(x, granule, ...) {
  S7::S7_dispatch()
})
method(time_ceiling, class_mixtime) <- vecvec::vecvec_apply_fn(time_ceiling)
method(time_ceiling, S7::class_any) <- function(x, granule, ...) {
  # Ceiling time (floor + 0.5 to round .0 up)
  time_round_impl(x, granule, function(res) ceiling(floor(res) + 0.5))
}

#' @rdname time_round
#' @export
time_floor <- new_generic("time_floor", "x", function(x, granule, ...) {
  S7::S7_dispatch()
})
method(time_floor, class_mixtime) <- vecvec::vecvec_apply_fn(time_floor)
method(time_floor, S7::class_any) <- function(x, granule, ...) {
  time_round_impl(x, granule, floor)
}

# Shared boundary-rounding logic for time_round()/time_ceiling()/time_floor().
#
# The timezone offset is applied while `x` is still expressed in its own
# native chronon (which always converts to/from itself exactly - a 1:1
# ratio), and only then converted into the granule's units for rounding.
# Applying the offset *after* converting into the granule would require
# scaling it by `chronon_cardinality()`, which for granules like `quarter()`
# is a variable, context-dependent conversion (the number of seconds in a
# quarter isn't fixed) - and can only be approximated (e.g. by an average
# quarter length), leaking drift into the rounded boundary.
time_round_impl <- function(x, granule, round_fn) {
  if (is.character(granule)) granule <- parse_time_unit(granule)
  if (S7_inherits(granule, mt_unit)) granule <- duration(1L, granule)
  if (length(granule) != 1L) {
    cli::cli_abort("{.var granule} must be a single time duration", call. = FALSE)
  }
  by <- attr(granule@x[[1L]], "chronon")
  by@n <- by@n * as.numeric(granule)

  # Inherit non-naive attributes from chronon
  by <- granule_inherit_props(by, chronon <- chronon_common(x))

  # Native numeric value of `x`, in its own chronon (an exact conversion)
  res <- chronon_convert(x, chronon)

  # Apply timezone offsets (UTC -> tz) while still in the exact native unit
  res <- tz_to_local(res, chronon, tz_name(by))

  # Convert the (now local wall-clock) instant into the granule's units and round it
  res <- chronon_convert_impl(res, from = chronon, to = by, discrete = FALSE, tz = "UTC")
  res <- round_fn(res)

  # Convert the rounded boundary back to the native chronon (still local wall-clock)
  res <- chronon_convert_impl(res, from = by, to = chronon, discrete = FALSE, tz = "UTC")

  # Undo timezone offsets (tz -> UTC)
  res <- tz_to_utc(res, chronon, tz_name(by))

  if (is.integer(x)) res <- as.integer(res)
  attributes(res) <- attributes(x)
  res
}

#' Create a new calendar
#'
#' Define a new calendar as a collection of time units. Calendars are the
#' foundation for representing dates and times in terms of human-readable
#' components like years, months, days, hours, minutes, and seconds. Each
#' calendar is defined by specifying the time units it contains, which
#' determine how time values can be interpreted and manipulated.
#'
#' Time units are typically S7 class definitions that inherit from [`mt_unit`]
#' for standard units, [`mt_tz_unit`] for timezone-aware units (civil time), or 
#' [`mt_loc_unit`] for location-aware units (astronomical time). The calendar
#' object provides a namespace for accessing these unit constructors and
#' defines the relationships between them for calendar arithmetic.
#'
#' @param ... Named time unit class definitions. Each argument should be a
#'   time unit class (typically created with `S7::new_class()`) that inherits
#'   from `mt_unit` or `mt_tz_unit`. The names define the calendar's fields
#'   and are used to access unit constructors (e.g., `calendar$year()`).
#' @param inherit Optional calendar to inherit time units from. Units
#'   defined in `...` will override inherited units with the same name.
#' @param class Character vector of additional classes for the calendar object.
#'
#' @return A calendar object of class `c(class, "mt_calendar")`, consisting of a
#'   named list containing the specified time unit classes.
#'
#' @seealso [linear_time()], [cyclical_time()]
#'
#' @examples
#' # Create a simple calendar with year and month units
#' # (inheriting from civil time units for day, hour, minute, second, ...)
#' cal_simple <- new_calendar(
#'   year = new_time_unit("tu_year", parent = mt_tz_unit),
#'   month = new_time_unit("tu_month", parent = mt_tz_unit),
#'   inherit = cal_time_civil,
#'   class = "cal_simple"
#' )
#'
#' # Create time granules from the calendar
#' cal_simple$year(1L)
#' cal_simple$month(1L)
#' 
#' @importFrom rlang list2
#' @export
new_calendar <- function(..., inherit = NULL, class = character()) {
  vctrs::vec_assert(class, character())
  time_units <- list2(...)
  if (!is.null(inherit)) {
    # Add units from inherit that aren't already in time_units
    new_names <- setdiff(names(inherit), names(time_units))
    time_units <- c(time_units, inherit[new_names])
  }
  
  cal <- structure(
    time_units,
    class = c(class, "mt_calendar")
  )
  
  # Create a reference environment that wraps the calendar
  cal_env <- new.env(parent = emptyenv())
  
  # Add the environment reference to each time unit
  for (i in seq_along(cal)) {
    attr(cal[[i]], "cal") <- cal_env
  }
  
  # Add reference to self in environment
  cal_env$calendar <- cal
  
  cal
}

#' @exportS3Method base::print mt_calendar
print.mt_calendar <- function(x, ...) {
  # Get calendar class (excluding mt_calendar base class)
  cal_classes <- setdiff(class(x), "mt_calendar")
  
  # Header
  if (length(cal_classes) > 0) {
    cat("<", paste(cal_classes, collapse = ", "), ">\n", sep = "")
  } else {
    cat("<mt_calendar>\n")
  }
  
  # List time units
  if (length(x) > 0) {
    cat("Time units:\n")
    for (name in names(x)) {
      cat("  - ", name, "\n", sep = "")
    }
  } else {
    cat("No time units defined\n")
  }
  
  invisible(x)
}

#' Obtain the calendar of a time object
#' 
#' This S7 generic function extracts the calendar system from a time object.
#' The calendar defines the collection of time units (years, months, days, etc.)
#' used to interpret the time representation.
#' 
#' @param x A time object (e.g., [base::Date], [base::POSIXct], [linear_time()], etc.)
#' @param ... Additional arguments for methods.
#' 
#' @return A calendar object (e.g., `cal_gregorian`, `cal_isoweek`)
#' 
#' @examples
#' 
#' # The calendar of a Date object is the Gregorian calendar
#' time_calendar(Sys.Date())
#' 
#' # The calendar of a POSIXct object is also Gregorian
#' time_calendar(Sys.time())
#' 
#' # The calendar of a yearweek object is the ISO week calendar
#' time_calendar(yearweek(Sys.Date()))
#' 
#' # A mixed time object returns a list of calendars
#' time_calendar(c(yearmonth(Sys.Date()), Sys.Date()))
#' 
#' @export
time_calendar <- S7::new_generic("calendar", c("x"))

method(time_calendar, S7::class_Date) <- function(x) cal_gregorian
method(time_calendar, S7::class_POSIXt) <- function(x) cal_gregorian
method(time_calendar, mt_unit) <- function(x) {
  attr(S7::S7_class(x), "cal")$calendar
}
method(time_calendar, class_mixtime) <- function(x) {
  time_calendar(chronon_common(x))
}
method(time_calendar, mt_time) <- function(x) time_calendar(time_chronon(x))
method(time_calendar, S7::class_any) <- function(x) cal_gregorian

#' @importFrom S7 props props<-
granule_inherit_props <- function(x, parent) {
  naive_props <- is.naive(props <- props(x))
  parent_props <- props(parent)
  inherit_props <- intersect(names(props)[naive_props], names(parent_props))
  props[inherit_props] <- parent_props[inherit_props]
  props(x) <- props
  x
}

# Inherit naive properties on which every parent shares the same attribute.
granule_inherit_shared_props <- function(x, parents) {
  # Build a stand-in parent by filling x's own naive properties with the
  # value every parent agrees on, then let granule_inherit_props() perform
  # the actual copy (a no-op for properties left untouched below).
  shared <- x
  shared_props <- props(shared)

  for (prop in names(shared_props)[is.naive(shared_props)]) {
    values <- lapply(parents, function(parent) props(parent)[[prop]])

    is_unset <- vapply(values, function(value) {
      # `tz` (and other naive-fillable properties) can carry an explicit NA
      # rather than the classed `mt_naive` sentinel (e.g. `datetime(x, tz =
      # NA)`) -- both mean "unknown", so both must block reconciliation.
      is.null(value) || inherits(value, "mt_naive") || (length(value) == 1L && is.na(value))
    }, logical(1L))

    if (any(is_unset)) {
      next
    }

    if (prop == "tz") {
      # Differing known zones can always be reconciled at a common zone (UTC,
      # since every named zone converts to it without loss); only a genuinely
      # unknown zone (handled above) leaves the result naive.
      shared_props[[prop]] <- if (length(unique(values)) == 1L) values[[1L]] else "UTC"
      next
    }

    if (!all(vapply(values, identical, logical(1L), values[[1L]]))) {
      next
    }

    shared_props[[prop]] <- values[[1L]]
  }

  props(shared) <- shared_props
  granule_inherit_props(x, shared)
}

# Convert a chronon's *default*-naive properties (the classed `mt_naive`
# sentinel, e.g. an unset `tz`) into an explicit NA of the same underlying
# type (e.g. `datetime(x, tz = NA)`'s naive result)
granule_harden_naive <- function(x) {
  props <- props(x)
  naive_props <- is.naive(props)
  props[naive_props] <- lapply(props[naive_props], unclass)
  props(x) <- props
  x
}

# Compute numeric time parts from a time object
#
# `chronon_parts()` decomposes a time object into its numeric linear and cyclical
# components by repeatedly applying `chronon_divmod()` along the call path of a
# minimal Steiner tree. The linear parts are the `$div` result, which is re-used
# along the divmod path where `$mod` results form the required cyclical parts.
#
# @param x A `mt_time` object (an uni-typed mixtime vector)
# @param linear A list of time granules representing linear (non-repeating)
#   components, e.g. `list(cal_gregorian$year(1L))`.
# @param cyclical A list of cyclical component specifications, where each
#   element is a list of two time granules `list(from, to)` (chronon and cycle),
#   e.g. `list(list(cal_gregorian$month(1L), cal_gregorian$year(1L)))` for
#   month-of-year.
#
# @return A list with three elements:
#   - `$linear`: a list of integer vectors, one per element of `linear`.
#   - `$cyclical`: a list of integer vectors, one per element of `cyclical`.
#   - `$cyclical_at`: a list of integer vectors, one per element of `cyclical`,
#     giving the linear position of the cycle granule.
chronon_parts <- function(x, linear = list(), cyclical = list()) {
  start_tu <- attr(x, "chronon")

  # Shift the time points into local time, truncating the offset for discrete
  # time models so they remain whole chronons. A continuous time model keeps
  # its fractional position instead of flooring it away here.
  x <- vec_data(x)
  discrete <- is.integer(x)
  offset <- tz_offset_impl(x, start_tu)
  x <- x + if (discrete) trunc(offset) else offset
  if (discrete) x <- floor(x)

  # The two granules of a cyclical part: the chronon is the finer unit (the
  # desired output unit) and the cycle the coarser unit it repeats within.
  cyclical_chronon <- lapply(cyclical, `[[`, 1L)
  cyclical_cycle   <- lapply(cyclical, `[[`, 2L)

  # Parts are recognised by matching class ids against the node reached at each
  # step of the traversal, so each set of targets is keyed by position in a
  # vector of the class id it is matched on (NA where the match cannot apply).
  linear_ids  <- vapply(linear, S7_class_id, character(1L))
  chronon_ids <- vapply(cyclical_chronon, S7_class_id, character(1L))
  cycle_ids   <- vapply(cyclical_cycle, S7_class_id, character(1L))

  # Multi-unit cycles (e.g. a month within 15 months) repeat within a coarser
  # granule of the *same* time unit, so their chronon and cycle resolve to one
  # node: a self-loop in the class-keyed cardinality graph rather than a path
  # through it. They ask the graph only to reach that node (a terminal, with no
  # co-occurrence to enforce) and are matched on their chronon, since matching
  # them on their cycle would pair them with the degenerate divmod of a granule
  # with itself.
  self_cycle <- chronon_ids == cycle_ids
  self_cycle_ids <- chronon_ids
  self_cycle_ids[!self_cycle] <- NA_character_
  cycle_ids[self_cycle] <- NA_character_

  # Prepare results to be filled via recursive divmod execution
  linear_results   <- vector("list", length(linear))
  cyclical_results <- vector("list", length(cyclical))
  cyclical_at      <- vector("list", length(cyclical))

  # Resolve parts with finer granules than the chronon for continuous time data
  cycle_is_root <- !self_cycle &
    cycle_ids == S7_class_id(start_tu) &
    vapply(cyclical_cycle, function(cy) cy@n == start_tu@n, logical(1L))
  if (any(cycle_is_root)) {
    whole <- if (discrete) x else floor(x)
    for (i in which(cycle_is_root)) {
      chronon_i <- cyclical_chronon[[i]]
      cyclical_results[[i]] <- if (discrete) {
        rep(NA_integer_, length(x))
      } else {
        chronon_divmod(start_tu, chronon_i, x)$div -
          chronon_divmod(start_tu, chronon_i, whole)$div
      }
      cyclical_at[[i]] <- whole
    }
    cycle_ids[cycle_is_root] <- NA_character_
  }

  # Find a suitable tree of chronon_divmod() steps that computes all cyclical
  # and linear parts. Cross-unit cycles must reach both of their granules on one
  # root-to-leaf path, while the other targets need only be reached.
  tree <- S7_graph_dispatch_multi(
    graph      = chronon_cardinality_graph(),
    start      = start_tu,
    terminals  = c(linear, cyclical_chronon[self_cycle]),
    groups     = cyclical[!self_cycle & !cycle_is_root]
  )

  # Traverse the divmod path to compute parts. `parent_id` is the caller's
  # `child_id`, so it is handed down rather than looked up again.
  traverse <- function(node, parent_tu, parent_id, x) {
    child_tu <- node$node
    child_id <- S7_class_id(child_tu)

    dm <- chronon_divmod(
      from = parent_tu,
      to   = child_tu,
      x    = x
    )

    # Collect linear result: div when child matches a linear target
    for (i in which(child_id == linear_ids)) {
      linear_results[[i]] <<- count_in_granule(dm$div, child_tu, linear[[i]]) +
        chronon_epoch(child_tu)
    }

    # Complete multi-unit cycles rooted at this granule. `dm$div` is the time
    # point in `child_tu` units (at the root, an identity divmod of `x`), so the
    # position within the cycle is a single divmod away.
    for (i in which(child_id == self_cycle_ids)) {
      dm_self <- chronon_divmod(
        from = cyclical_chronon[[i]],
        to   = cyclical_cycle[[i]],
        x    = count_in_granule(dm$div, child_tu, cyclical_chronon[[i]])
      )
      cyclical_results[[i]] <<- dm_self$mod
      cyclical_at[[i]]      <<- dm_self$div
    }

    # Recurse each child with $div as the new time point (now in child_tu units)
    incomplete <- unlist(lapply(node$children, traverse, child_tu, child_id, dm$div))

    # Unwind recursion with backward conversion for cyclical parts: an
    # incomplete result counts `child_tu` units so far, so re-expressing it in
    # the finer `parent_tu` multiplies by the cardinality at the current `dm$div`
    # position and adds this step's remainder. The cardinality is a property of
    # this step alone, so it is found once for all of them.
    if (length(incomplete) > 0L) {
      cardinality <- chronon_cardinality(parent_tu, child_tu, dm$div)
      for (i in incomplete) {
        cyclical_results[[i]] <<- cyclical_results[[i]] * cardinality + dm$mod
      }
    }

    # Cycles whose coarser granule is this node start counting here
    starting <- which(child_id == cycle_ids)
    if (length(starting) > 0L) {
      cyclical_results[starting] <<- list(dm$mod)
      cyclical_at[starting]      <<- list(dm$div)
      incomplete <- c(incomplete, starting)
    }

    # Cyclical parts accumulate in `parent_tu`, the finest granule of their
    # chronon's unit, so those completing here are rescaled to the size asked for.
    complete <- chronon_ids[incomplete] == parent_id
    for (i in incomplete[complete]) {
      cyclical_results[[i]] <<- count_in_granule(
        cyclical_results[[i]], parent_tu, cyclical_chronon[[i]]
      )
    }

    # Return vector of which cyclical targets are still incomplete after this step
    incomplete[!complete]
  }
  traverse(tree, start_tu, S7_class_id(start_tu), x)

  # Check if all targets were found
  abort_missing_parts(linear_results, linear, time_granule_label, "linear")
  abort_missing_parts(cyclical_results, cyclical, cycle_label, "cyclical")

  # Return list of the same order as input
  list(linear = linear_results, cyclical = cyclical_results, cyclical_at = cyclical_at)
}

# Re-express `value`, a count of `from` granules, as a count of `to` granules of
# the same time unit (e.g. months at the month node into quarters for
# `month(3L)`). Nodes hold the finest granule of their time unit, so `to` is
# never finer than `from` and the conversion is a division away.
count_in_granule <- function(value, from, to) {
  if (from@n == to@n) value else chronon_divmod(from, to, value)$div
}

# Abort when the traversal did not reach every requested part, describing those
# it missed with `label()` applied to the corresponding element of `targets`.
abort_missing_parts <- function(results, targets, label, type) {
  found <- !vapply(results, is.null, logical(1L))
  if (all(found)) return(invisible(NULL))

  missed <- vapply(targets[!found], label, character(1L))
  cli::cli_abort(
    c(
      "The following {type} time parts could not be computed from the input time object:",
      i = "{missed}",
      i = "All requested {type} parts need to be included as granules for {.code linear_time()}"
    ),
    call = NULL
  )
}

# Describe a cyclical part for error messages, e.g. "1 month -> 15 months".
cycle_label <- function(x) {
  paste(time_granule_label(x[[1L]]), "->", time_granule_label(x[[2L]]))
}

# Fixed reference: 2024-01-15 09:20:30 UTC
t_ref <- datetime(as.POSIXct("2024-01-15 09:20:30", tz = "UTC"))

# ---- Linear parts -----------------------------------------------------------

test_that("chronon_parts() returns correct linear parts", {
  # chronon_parts is internally applied to mt_time vectors, not mixed-type vectors
  x <- vecvec::unvecvec(t_ref)
  
  # Singular linear parts
  result <- chronon_parts(x, linear = list(cal_gregorian$second(1L)))
  expect_equal(result$linear[[1L]], 1705310430L)

  result <- chronon_parts(x, linear = list(cal_gregorian$year(1L)))
  expect_equal(result$linear[[1L]], 2024L)

  result <- chronon_parts(x, linear = list(cal_gregorian$day(1L)))
  expect_equal(result$linear[[1L]], 19737L)

  # Multiple linear parts together
  result <- chronon_parts(
    x,
    linear = list(
      cal_gregorian$second(1L),
      cal_gregorian$minute(1L),
      cal_gregorian$hour(1L),
      cal_gregorian$day(1L),
      cal_gregorian$month(1L),
      cal_gregorian$year(1L)
    )
  )
  expect_equal(result$linear[[1L]], 1705310430L) # total seconds
  expect_equal(result$linear[[2L]], 28421840L)   # total minutes
  expect_equal(result$linear[[3L]], 473697L)     # total hours
  expect_equal(result$linear[[4L]], 19737L)      # total days
  expect_equal(result$linear[[5L]], 648L)        # total months
  expect_equal(result$linear[[6L]], 2024L)       # year

  # No linear parts should return empty list
  result <- chronon_parts(x)
  expect_identical(result$linear, list())
})

# ---- Cyclical parts ---------------------------------------------------------

test_that("chronon_parts() returns correct cyclical parts", {
  # chronon_parts is internally applied to mt_time vectors, not mixed-type vectors
  x <- vecvec::unvecvec(t_ref)
  
  # Singular cyclical parts
  result <- chronon_parts(
    x,
    cyclical = list(list(cal_gregorian$second(1L), cal_gregorian$minute(1L)))
  )
  expect_equal(result$cyclical[[1L]], 30L)

  result <- chronon_parts(
    x,
    cyclical = list(list(cal_gregorian$minute(1L), cal_gregorian$hour(1L)))
  )
  expect_equal(result$cyclical[[1L]], 20L)

  result <- chronon_parts(
    x,
    cyclical = list(list(cal_gregorian$hour(1L), cal_gregorian$day(1L)))
  )
  expect_equal(result$cyclical[[1L]], 9L)

  result <- chronon_parts(
    x,
    cyclical = list(list(cal_gregorian$month(1L), cal_gregorian$year(1L)))
  )
  expect_equal(result$cyclical[[1L]], 0L)

  result <- chronon_parts(
    x,
    cyclical = list(list(cal_gregorian$day(1L), cal_gregorian$month(1L)))
  )
  expect_equal(result$cyclical[[1L]], 14L)

  # Multiple cyclical parts together
  result <- chronon_parts(
    x,
    cyclical = list(
      list(cal_gregorian$second(1L), cal_gregorian$minute(1L)),
      list(cal_gregorian$minute(1L), cal_gregorian$hour(1L)),
      list(cal_gregorian$hour(1L), cal_gregorian$day(1L))
    )
  )
  expect_equal(result$cyclical[[1L]], 30L) # second-of-minute
  expect_equal(result$cyclical[[2L]], 20L) # minute-of-hour
  expect_equal(result$cyclical[[3L]], 9L)  # hour-of-day
})

# ---- Multi-step cyclical (e.g. second-of-hour, second-of-day) ---------------

test_that("chronon_parts() correctly computes multi-step cyclical parts", {
  # chronon_parts is internally applied to mt_time vectors, not mixed-type vectors
  x <- vecvec::unvecvec(t_ref)
  
  # 09:20:30 => second-of-hour = 20*60 + 30 = 1230
  result <- chronon_parts(
    x,
    cyclical = list(list(cal_gregorian$second(1L), cal_gregorian$hour(1L)))
  )
  expect_equal(result$cyclical[[1L]], 1230L)

  # 09:20:30 => 9*3600 + 20*60 + 30 = 33630
  result <- chronon_parts(
    x,
    cyclical = list(list(cal_gregorian$second(1L), cal_gregorian$day(1L)))
  )
  expect_equal(result$cyclical[[1L]], 33630L)

  # Both together
  result <- chronon_parts(
    x,
    cyclical = list(
      list(cal_gregorian$second(1L), cal_gregorian$minute(1L)),
      list(cal_gregorian$second(1L), cal_gregorian$hour(1L)),
      list(cal_gregorian$second(1L), cal_gregorian$day(1L))
    )
  )
  expect_equal(result$cyclical[[1L]], 30L)    # second-of-minute
  expect_equal(result$cyclical[[2L]], 1230L)  # second-of-hour
  expect_equal(result$cyclical[[3L]], 33630L) # second-of-day
})

test_that("chronon_parts() returns parts at the requested granule size", {
  # Tree nodes hold the finest granule of their time unit, so parts requested at
  # a coarser size are rescaled from their node rather than returned unscaled.
  x <- vecvec::unvecvec(linear_time(0:16, chronon = cal_gregorian$month(1L)))

  # Linear: quarters, not months, since the epoch
  result <- chronon_parts(x, linear = list(cal_gregorian$month(3L)))
  expect_equal(result$linear[[1L]], (0:16) %/% 3L)

  d <- as.Date("1970-01-01") + c(0L, 40L, 100L, 400L, 500L)
  x <- vecvec::unvecvec(as_mixtime(d))

  # Cyclical: quarter-of-year, reached by traversing from a day chronon
  result <- chronon_parts(
    x,
    cyclical = list(list(cal_gregorian$month(3L), cal_gregorian$year(1L)))
  )
  expect_equal(result$cyclical[[1L]], c(0L, 0L, 1L, 0L, 1L))

  # A coarser granule of the same unit must not anchor the node and cost the
  # finer granules sharing it their remainder.
  result <- chronon_parts(
    x,
    linear   = list(cal_gregorian$month(3L)),
    cyclical = list(list(cal_gregorian$month(1L), cal_gregorian$month(15L)))
  )
  expect_equal(result$linear[[1L]], c(0L, 0L, 1L, 4L, 5L))
  expect_equal(result$cyclical[[1L]], c(0L, 1L, 3L, 13L, 1L))
})

test_that("chronon_parts() correctly computes multi-unit cyclical parts", {
  # A cycle of several of the chronon's own time unit (e.g. a month within 15
  # months) is a self-loop in the class-keyed cardinality graph, not a path.

  # Chronon of the cycle is the time vector's own chronon
  x <- vecvec::unvecvec(linear_time(0:20, chronon = cal_gregorian$month(1L)))
  result <- chronon_parts(
    x,
    cyclical = list(list(cal_gregorian$month(1L), cal_gregorian$month(15L)))
  )
  expect_equal(result$cyclical[[1L]], (0:20) %% 15L)

  # Chronon of the cycle is reached by traversing from a finer chronon
  d <- as.Date("1970-01-01") + c(0L, 40L, 100L, 400L, 500L)
  x <- vecvec::unvecvec(as_mixtime(d))
  result <- chronon_parts(
    x,
    cyclical = list(list(cal_gregorian$month(1L), cal_gregorian$month(15L)))
  )
  expect_equal(result$cyclical[[1L]], c(0L, 1L, 3L, 13L, 1L))

  # Alongside the linear and (cross-unit) cyclical parts of the same traversal
  result <- chronon_parts(
    x,
    linear   = list(cal_gregorian$year(1L)),
    cyclical = list(
      list(cal_gregorian$month(1L), cal_gregorian$month(15L)),
      list(cal_gregorian$month(1L), cal_gregorian$year(1L))
    )
  )
  expect_equal(result$linear[[1L]], c(1970L, 1970L, 1970L, 1971L, 1971L))
  expect_equal(result$cyclical[[1L]], c(0L, 1L, 3L, 13L, 1L))
  expect_equal(result$cyclical[[2L]], c(0L, 1L, 3L, 1L, 4L))
})

# ---- Mixed linear and cyclical ----------------------------------------------
test_that("chronon_parts() returns correct linear and cyclical parts together", {
  # chronon_parts is internally applied to mt_time vectors, not mixed-type vectors
  x <- vecvec::unvecvec(t_ref)
  
  result <- chronon_parts(
    x,
    linear = list(cal_gregorian$year(1L), cal_gregorian$day(1L)),
    cyclical = list(
      list(cal_gregorian$hour(1L), cal_gregorian$day(1L)),
      list(cal_gregorian$minute(1L), cal_gregorian$hour(1L))
    )
  )
  expect_equal(result$linear[[1L]], 2024L)  # year
  expect_equal(result$linear[[2L]], 19737L) # total days
  expect_equal(result$cyclical[[1L]], 9L)   # hour-of-day
  expect_equal(result$cyclical[[2L]], 20L)  # minute-of-hour
})

# ---- Timezones --------------------------------------------------------------

test_that("chronon_parts() applies timezone offsets to granules", {
  # 2024-01-15 09:20:30 UTC = 2024-01-15 20:20:30 AEDT
  x <- vecvec::unvecvec(datetime(t_ref, tz = "Australia/Melbourne"))
  result <- chronon_parts(
    x,
    cyclical = list(list(cal_gregorian$hour(1L), cal_gregorian$day(1L)))
  )
  expect_equal(result$cyclical[[1L]], 20L)

  # 2024-01-14 20:20:00 UTC = 2024-01-14 15:20:00 EST (UTC-5) = 2024-01-15 07:20:00 AEDT (UTC+11)
  t <- as.POSIXct("2024-01-14 20:20:00", tz = "UTC")
  x_utc <- vecvec::unvecvec(datetime(t, tz = "UTC"))
  x_est <- vecvec::unvecvec(datetime(t, tz = "America/New_York"))
  x_melb <- vecvec::unvecvec(datetime(t, tz = "Australia/Melbourne"))

  r_est <- chronon_parts(x_est, cyclical = list(
    list(cal_gregorian$minute(1L), cal_gregorian$hour(1L)),
    list(cal_gregorian$hour(1L), cal_gregorian$day(1L)),
    list(cal_gregorian$day(1L), cal_gregorian$month(1L))
  ))
  r_utc <- chronon_parts(x_utc, cyclical = list(
    list(cal_gregorian$minute(1L), cal_gregorian$hour(1L)),
    list(cal_gregorian$hour(1L), cal_gregorian$day(1L)),
    list(cal_gregorian$day(1L), cal_gregorian$month(1L))
  ))
  r_melb <- chronon_parts(x_melb, cyclical = list(
    list(cal_gregorian$minute(1L), cal_gregorian$hour(1L)),
    list(cal_gregorian$hour(1L), cal_gregorian$day(1L)),
    list(cal_gregorian$day(1L), cal_gregorian$month(1L))
  ))

  # Minute-of-hour should be the same across timezones since it's a sub-hour unit
  expect_equal(r_utc$cyclical[[1L]], 20L)
  expect_equal(r_est$cyclical[[1L]], 20L)
  expect_equal(r_melb$cyclical[[1L]], 20L)

  expect_equal(r_utc$cyclical[[2L]], 20L) # UTC: 20h on same day
  expect_equal(r_est$cyclical[[2L]], 15L) # EST: 15h on same day
  expect_equal(r_melb$cyclical[[2L]], 7L) # AEDT: 07h next day

  expect_equal(r_utc$cyclical[[3L]], 13L) # UTC: 14th day of month
  expect_equal(r_est$cyclical[[3L]], 13L) # EST: 14th day of month
  expect_equal(r_melb$cyclical[[3L]], 14L) # AEDT: 15th day of month
})

# ---- Vectorised inputs ------------------------------------------------------

test_that("chronon_parts() handles vectorised inputs", {
  t_mult <- as.POSIXct(
    c("2024-01-15 09:20:30", "2024-06-21 15:45:00"),
    tz = "UTC"
  )
  x_mult <- vecvec::unvecvec(datetime(t_mult, tz = "UTC"))

  result <- chronon_parts(
    x_mult,
    cyclical = list(
      list(cal_gregorian$hour(1L), cal_gregorian$day(1L)),
      list(cal_gregorian$second(1L), cal_gregorian$hour(1L))
    )
  )
  
  # Should return both hours: 9 (09:20:30) and 15 (15:45:00)
  expect_equal(result$cyclical[[1L]], c(9L, 15L))

  # 09:20:30 => 20*60+30 = 1230; 15:45:00 => 45*60+0 = 2700
  expect_equal(result$cyclical[[2L]], c(1230L, 2700L))

  t_mult <- as.POSIXct(
    c("2024-01-15 09:20:30", "2024-06-21 15:45:00", "2024-12-31 23:59:59"),
    tz = "UTC"
  )
  x_mult <- vecvec::unvecvec(datetime(t_mult, tz = "UTC"))
  result <- chronon_parts(
    x_mult,
    cyclical = list(
      list(cal_gregorian$hour(1L), cal_gregorian$day(1L)),
      list(cal_gregorian$minute(1L), cal_gregorian$hour(1L)),
      list(cal_gregorian$second(1L), cal_gregorian$minute(1L))
    )
  )

  expect_equal(result$cyclical[[1L]], c(9L, 15L, 23L))    # hour-of-day
  expect_equal(result$cyclical[[2L]], c(20L, 45L, 59L))   # minute-of-hour
  expect_equal(result$cyclical[[3L]], c(30L, 0L, 59L))    # second-of-minute

  t_mult <- as.POSIXct(
    c("2020-03-01 00:00:00", "2023-07-04 12:00:00", "2024-01-15 09:20:30"),
    tz = "UTC"
  )
  x_mult <- vecvec::unvecvec(datetime(t_mult, tz = "UTC"))
  result <- chronon_parts(
    x_mult,
    linear = list(cal_gregorian$year(1L))
  )
  expect_equal(result$linear[[1L]], c(2020L, 2023L, 2024L))
})

#' Default formatting strings for chronons
#'
#' Provides default linear time formatting strings for a given chronon (finest 
#' time granule). The format strings use placeholders like `{lin(year(1L))}`,
#' `{cyc(month(1L), year(1L)}` and `{cyc(day(1L), month(1L)}`, which are
#' evaluated in the context of the data's [`time_calendar()`].
#' 
#' @param x A time granule for the chronon.
#' @param y A time granule for the cycle
#' @param cal The calendar of the chronon, used to disambiguate suitable format 
#'   strings for time units that are shared across calendars (e.g. 
#'   `cal_gregorian$day` and `cal_isoweek$day`).
#' @param ... Additional arguments for methods.
#'
#' @return A character string containing the default format template for the chronon.
#'
#' @export
#' @examples
#' chronon_format_linear(cal_gregorian$year(1L))
#' chronon_format_linear(cal_gregorian$month(1L))
#' chronon_format_linear(cal_gregorian$day(1L))
#' chronon_format_linear(cal_isoweek$day(1L))
#' 
#' @rdname chronon_format
chronon_format_linear <- new_generic("chronon_format_linear", c("x", "cal"), function(x, cal = time_calendar(x), ...) {
  S7::S7_dispatch()
})
method(chronon_format_linear, list(mt_unit, class_any)) <- function(x, cal) {
  paste0(time_unit_abbr(x), "{lin(chronon_common(.time))}")
}

#' @examples
#' chronon_format_cyclical(cal_gregorian$month(1L), cal_gregorian$year(1L))
#' chronon_format_cyclical(cal_gregorian$day(1L), cal_gregorian$month(1L))
#' chronon_format_cyclical(cal_isoweek$day(1L), cal_isoweek$week(1L))
#' chronon_format_cyclical(cal_isoweek$week(1L), cal_isoweek$year(1L))
#' 
#' @rdname chronon_format
#' @export
chronon_format_cyclical <- new_generic("chronon_format_cyclical", c("x", "y"))
method(chronon_format_cyclical, list(mt_unit, mt_unit)) <- function(x, y) {
  paste0(time_unit_abbr(x), "{cyc(chronon_common(.time), chronon_common(time_cycle(.time)))}")
}

#' @examples
#' chronon_format_duration(cal_gregorian$year(1L))
#' chronon_format_duration(cal_gregorian$month(1L))
#' chronon_format_duration(cal_gregorian$day(1L))
#' 
#' @rdname chronon_format
#' @export
chronon_format_duration <- new_generic("chronon_format_duration", "x")
method(chronon_format_duration, mt_unit) <- function(x) {
  "{duration_label(.time)}"
}

# A duration as text, e.g. "3 months" or "1.5 days". This is time_granule_label()
# over the magnitudes of a duration vector, except that continuous (double)
# durations show a fractional magnitude and always read as plural.
duration_label <- function(x) {
  n <- vec_data(x)
  chronon <- chronon_common(x)
  if (is.double(n)) {
    paste(format(n, nsmall = 1L), time_unit_plural(chronon, 2L))
  } else {
    time_granule_label(chronon, n)
  }
}

#' Default formatting strings for chronon attributes
#' 
#' Provides suffixes for default formatting strings for a given chronon (time granule).
#' This provides useful information such as timezones or locations in the string.
#' 
#' @param x A chronon (time granule) object.
#' @param ... Additional arguments for methods.
#'
#' @return A character string containing the default format suffix for the chronon.
#' 
#' @export
chronon_format_attr <- new_generic("chronon_format_attr", "x")
method(chronon_format_attr, mt_unit) <- function(x) {
  ""
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, message = FALSE---------------------------------------------------
library(mixtime)
library(S7)

## -----------------------------------------------------------------------------
.onLoad <- function(...) {
  methods_register()
}

## ----time-units---------------------------------------------------------------
# Timezone-aware Symmetry454 year and month units
new_time_unit("tu_symmetry454_year",  parent = mt_tz_unit)
new_time_unit("tu_symmetry454_month", parent = mt_tz_unit)

## ----new-calendar-------------------------------------------------------------
cal_symmetry454 <- new_calendar(
  year  = new_time_unit("tu_symmetry454_year",  parent = mt_tz_unit),
  month = new_time_unit("tu_symmetry454_month", parent = mt_tz_unit),
  week  = cal_isoweek$week,
  # Inherit civil-time units (day, hour, minute, second, ...)
  inherit = cal_time_civil,
  class = "cal_symmetry454"
)

cal_symmetry454

## ----calendar-access----------------------------------------------------------
cal_symmetry454$year(1L)
cal_symmetry454$month(1L, tz = "UTC")

## -----------------------------------------------------------------------------
chronon_cardinality(cal_isoweek$day(1L), cal_isoweek$week(1L))

## ----cardinality-fixed--------------------------------------------------------
# Each Symmetry454 year has 12 months
method(chronon_cardinality_fixed, list(cal_symmetry454$month, cal_symmetry454$year)) <-
  function(x, y) {
    12L
  }
chronon_cardinality(cal_symmetry454$month(1L), cal_symmetry454$year(1L))

## -----------------------------------------------------------------------------
chronon_cardinality(cal_gregorian$day(1L), cal_gregorian$month(1L), at = 0L) # Jan 1970
chronon_cardinality(cal_gregorian$day(1L), cal_gregorian$month(1L), at = 1L) # Feb 1970
chronon_cardinality(cal_gregorian$day(1L), cal_gregorian$month(1L), at = 25L) # Feb 1972 (leap year)

## ----cardinality-variable-----------------------------------------------------
method(chronon_cardinality, list(cal_symmetry454$week, cal_symmetry454$month)) <-
  function(x, y, at = NULL) {
    # The number of weeks in each n-month period
    month_size <- y@n
    nweeks_cyc <- circsum(c(4L, 5L, 4L), month_size)

    # Find which n-month period we're in based on the "at" position (months since epoch)
    period <- at %% length(nweeks_cyc) + 1L

    nweeks <- nweeks_cyc[period]

    # Add the extra week to December for Symmetry454 leap years.
    # A year is a leap year when (52*year + 146) %% 293 < 52.
    m1 <- at * month_size
    contains_dec <- which((m1 %% 12L) >= (12L - month_size))
    year <- 1970L + m1[contains_dec] %/% 12L
    is_leap_year <- ((52 * year + 146L) %% 293L) < 52L

    nweeks[contains_dec[is_leap_year]] <- nweeks[contains_dec[is_leap_year]] + 1L

    # Scale by the number of weeks in the n-week time granule
    nweeks / x@n
  }

# The 4-5-4 cycle across a full Symmetry454 year
chronon_cardinality(cal_symmetry454$week(1L), cal_symmetry454$month(1L), at = 0:11)

# 1970 is a leap year (53 weeks), so December has 5 weeks instead of 4
chronon_cardinality(cal_symmetry454$week(1L), cal_symmetry454$month(1L), at = 11L)

# Non-leap year: December has the usual 4 weeks
chronon_cardinality(cal_symmetry454$week(1L), cal_symmetry454$month(1L), at = 12:23) # Dec 1971

# The number of weeks in a multi-month period is the sum of the weeks in each month
chronon_cardinality(cal_symmetry454$week(1L), cal_symmetry454$month(2L), at = 0:5)

## -----------------------------------------------------------------------------
# The number of weeks in a Symmetry454 year (derived from week → month → year)
# Note that the leap year in 1970 (at = 0) produces 53 weeks via week → month.
chronon_cardinality(cal_symmetry454$week(1L), cal_symmetry454$year(1L), at = 0:4)

# The number of days in a Symmetry454 year (derived from day → week → month → year)
chronon_cardinality(cal_symmetry454$day(1L), cal_symmetry454$year(1L), at = 0:4)

## -----------------------------------------------------------------------------
# The number of Symmetry454 years in a month is the inverse of months → years (1/12)
chronon_cardinality(cal_symmetry454$year(1L), cal_symmetry454$month(1L))

# The number of Symmetry454 months in 2 weeks (requires `at` since weeks → months is irregular)
chronon_cardinality(cal_symmetry454$month(1L), cal_symmetry454$week(2L), at = 0:4)

## -----------------------------------------------------------------------------
chronon_divmod(cal_gregorian$day(1L), cal_gregorian$month(1L), 45L)

## ----divmod-------------------------------------------------------------------
method(chronon_divmod, list(cal_symmetry454$week, cal_symmetry454$month)) <-
  function(from, to, x) {
    # Most of this code works on 1-week granules
    week_size <- from@n
    x <- x * week_size  # convert n-weeks to 1-weeks

    # 1. Account for leap weeks by regularising x to have a fixed 52 weeks per year

    # The symmetrical sub-cycles of the 293-year leap week cycle are:
    #   17+11+17 + 17+17+11+17+17 + 17+11+17 + 17+17+11+17+17 + 17+11+17
    #   = 45 + 79 + 45 + 79 + 45 = 293
    # Primary (length 17) sub-cycles have 3 leap years: 00100000100000100
    # Secondary (length 11) sub-cycles have 2 leap years: 00100000100
    leaps_cycle_17 <- function(w) (w >= 157L) + (w >= 470L) + (w >= 783L)
    leaps_cycle_11 <- function(w) (w >= 157L) + (w >= 470L)
    leaps_cycle_45 <- function(w) {
      ifelse(w < 887L,  leaps_cycle_17(w),
      ifelse(w < 1461L, 3L + leaps_cycle_11(w - 887L),
                        5L + leaps_cycle_17(w - 1461L)))
    }
    leaps_cycle_79 <- function(w) {
      ifelse(w < 887L,  leaps_cycle_17(w),
      ifelse(w < 1774L, 3L  + leaps_cycle_17(w - 887L),
      ifelse(w < 2348L, 6L  + leaps_cycle_11(w - 1774L),
      ifelse(w < 3235L, 8L  + leaps_cycle_17(w - 2348L),
                        11L + leaps_cycle_17(w - 3235L)))))
    }
    leaps_symmetry454 <- function(x) {
      # There are 15288 weeks in a full 293-year cycle (293*52 + 52 leap weeks)
      w <- x %% 15288L
      x %/% 15288L * 52L +
      ifelse(w < 2348L,  leaps_cycle_45(w),
      ifelse(w < 6470L,  8L  + leaps_cycle_79(w - 2348L),
      ifelse(w < 8818L,  22L + leaps_cycle_45(w - 6470L),
      ifelse(w < 12940L, 30L + leaps_cycle_79(w - 8818L),
                         44L + leaps_cycle_45(w - 12940L)))))
    }

    # Offset x to align with the nearest 293-year cycle boundary before the epoch.
    # There are 349 leap years between year 1-W1 and the 1970-W1 epoch, and the
    # nearest cycle start is (1969*52 + 349) %% (293*52 + 52) = 11009 weeks before epoch.
    x_cyc <- x + 11009L + week_size  # right align multi-week granules
    n_leaps <- leaps_symmetry454(x_cyc)

    # Regularise x to have exactly 52 weeks per year by subtracting leap weeks.
    # (37 leap years occur in the cycle before the epoch, so we add 37 back.)
    x_reg <- x - n_leaps + 37L

    # 2. Use the 4-5-4 pattern to find the month (div) and week remainder (mod)
    ## The number of weeks in each n-month period
    month_size <- to@n
    weeks_len <- circsum(c(4L, 5L, 4L), month_size)

    ## The total weeks in a full n-month cycle
    weeks_tot <- sum(weeks_len)

    ## Find which n-month cycle we're in based on the regularised week count
    period_full <- x_reg %/% weeks_tot

    ## Find which part within the n-month cycle we're in
    weeks_seq  <- cumsum(weeks_len[-length(weeks_len)])
    period_part <- rowSums(outer(x_reg %% weeks_tot, weeks_seq, ">="))

    # div: total complete n-month cycles + complete n-months within the current cycle
    div <- period_full * length(weeks_len) + period_part
    # mod: remaining (regularised) weeks within the current n-month period
    mod <- x_reg %% weeks_tot - c(0L, weeks_seq)[period_part + 1L]


    # 3. Adjust the remainder to account for leap weeks that were removed during regularisation.

    # Identify leap weeks re-using cumulative in-cycle leap week counts: leaps_symmetry454()
    # If the week added a leap week from the previous, then it itself must be a leap week.
    # Applied only to regularised 52/53rd weeks of the year (only these weeks can be leap weeks)
    last_weeks  <- which(x_reg %% 52L >= 51L)
    leap_weeks  <- last_weeks[n_leaps[last_weeks] - leaps_symmetry454(x_cyc[last_weeks] - week_size) > 0L]
    mod[leap_weeks] <- mod[leap_weeks] + 1L  # restore leap week to remainder

    # Scale mod back to the original n-unit week size
    mod <- mod %/% week_size

    # Return the divmod result
    list(div = div, mod = mod)
  }

## -----------------------------------------------------------------------------
# Week 19 (0-indexed) of 1970 is the 2nd week (div=1) of May 1970 (mod=4)
with(cal_symmetry454, chronon_divmod(week(1L), month(1L), 18L))

# Week 52 (0-indexed) is the leap week of 1970; it is the 5th week (div=4) of Dec 1970 (mod=11)
with(cal_symmetry454, chronon_divmod(week(1L), month(1L), 52L))

## -----------------------------------------------------------------------------
# The 5th fortnight of 1970 is the 3rd fortnight (div=2) of Feb 1970 (mod=1)
with(cal_symmetry454, chronon_divmod(week(2L), month(1L), 4L))

## -----------------------------------------------------------------------------
# Divmod for converting days → years is derived from days → weeks → months → years
# Gregorian day 839 since unix epoch (1972-04-19) is symmetry454 day 101 (mod=100) of year 1972 (div=2)
with(cal_symmetry454, chronon_divmod(day(1L), year(1L), 839L))

## -----------------------------------------------------------------------------
method(chronon_epoch, cal_symmetry454$year) <- function(x) 1970L

## -----------------------------------------------------------------------------
method(time_unit_full, cal_symmetry454$year)  <- function(x) "Symmetry454 year{?/s}"
method(time_unit_abbr, cal_symmetry454$year)  <- function(x) "Y"
method(time_unit_full, cal_symmetry454$month) <- function(x) "Symmetry454 month{?/s}"
method(time_unit_abbr, cal_symmetry454$month) <- function(x) "M"

## -----------------------------------------------------------------------------
linear_time(as.Date("1955-11-12"), chronon = cal_symmetry454$year(1L))

## -----------------------------------------------------------------------------
year(as.Date("1955-11-12"), calendar = cal_symmetry454)

## -----------------------------------------------------------------------------
# Week of the month
cyclical_time(as.Date("1955-11-12"), chronon = week(1L), cycle = month(1L), calendar = cal_symmetry454)

# Month of the year
month_of_year(as.Date("1955-11-12"), calendar = cal_symmetry454)

## -----------------------------------------------------------------------------
method(linear_labels_format, cal_symmetry454$year) <- function(granule, i, ...) {
  ifelse(i <= 0L, paste0(-i + 1L, "BC"), i)
}
method(linear_labels_parse, cal_symmetry454$year) <- function(granule, ...) {
  list(
    pattern = "\\d+(?:BC)?",
    decode = function(text, at = NULL) {
      bc <- grepl("BC$", text)
      n <- as.integer(sub("BC$", "", text))
      ifelse(bc, 1L - n, n) - chronon_epoch(granule)
    }
  )
}

## -----------------------------------------------------------------------------
year(-1:2, calendar = cal_symmetry454)

## -----------------------------------------------------------------------------
# Labels for months of the year, essentially the same as Gregorian months in years.
method(cyclical_labels, list(cal_symmetry454$month, cal_symmetry454$year)) <-
  label_scheme(
    start = 1L,
    width = 2L,
    vocab = vocab_table(
      `en-GB` = list(wide = month.name, abbreviated = month.abb)
    )
  )

# Labels for weeks of the month are simply 1-indexed (e.g. "1", "2", ...)
method(cyclical_labels, list(cal_symmetry454$week, cal_symmetry454$month)) <-
  label_scheme(start = 1L)

## -----------------------------------------------------------------------------
# Month of year
month_of_year(as.Date("1955-11-12"), calendar = cal_symmetry454)
# Week of month
cyclical_time(as.Date("1955-11-12"), chronon = cal_symmetry454$week(1L), cycle = cal_symmetry454$month(1L))

# Day of week (inherited from cal_isoweek)
day_of_week(as.Date("1955-11-12"), calendar = cal_symmetry454)
# Day of year (inherited from cal_time_civil)
day_of_year(as.Date("1955-11-12"), calendar = cal_symmetry454)

## -----------------------------------------------------------------------------
date(as.Date("1955-11-12"), calendar = cal_symmetry454)
yearweek(as.Date("1955-11-12"), calendar = cal_symmetry454)
yearmonth(as.Date("1955-11-12"), calendar = cal_symmetry454)

## -----------------------------------------------------------------------------
# Simply display years as numbers (e.g. "1970", "1971", ...)
method(
  chronon_format_linear,
  list(cal_symmetry454$year, S7::class_any)
) <- function(x, cal) "{lin(year(1L))}"

# Display labelled months as month within year (e.g. "1970 Jan", "1970 Feb", ...)
method(
  chronon_format_linear,
  list(cal_symmetry454$month, S7::class_any)
) <- function(x, cal) "{lin(year(1L))}-{cyc(month(1L), year(1L), label=TRUE, abbreviate=TRUE)}"

# Display weeks as week in month in year (e.g. "1970-01-W1", "1970-01-W2", ...)
method(
  chronon_format_linear,
  list(cal_symmetry454$week, S7::new_S3_class("cal_symmetry454"))
) <- function(x, cal) "{lin(year(1L))}-{cyc(month(1L), year(1L), label=TRUE, abbreviate=TRUE)}-W{cyc(week(1L), month(1L))}"

# Format days as day in week in month in year (e.g. "1970-Jan-W1-Mon", "1970-Jan-W1-Tue", ...)
method(
  chronon_format_linear,
  list(cal_symmetry454$day, S7::new_S3_class("cal_symmetry454"))
) <- function(x, cal) "{lin(year(1L))}-{cyc(month(1L), year(1L), label=TRUE, abbreviate=TRUE)}-W{cyc(week(1L), month(1L))}-{cyc(day(1L), week(1L), label=TRUE, abbreviate=TRUE)}"

## -----------------------------------------------------------------------------
# Years are formatted as YYYY
year(as.Date("1955-11-12"), calendar = cal_symmetry454)

# Months are formatted as YYYY Mon
yearmonth(as.Date("1955-11-12"), calendar = cal_symmetry454)

# Weeks are formatted as YYYY-MM-WW
yearweek(as.Date("1955-11-12"), calendar = cal_symmetry454)

# Days are formatted as YYYY-MM-WW-DD
linear_time(as.Date(c("1985-10-26", "1955-11-05", "1955-11-12")), chronon = cal_symmetry454$day(1L))

## -----------------------------------------------------------------------------
# Format months in years as abbreviated month labels (e.g. "Jan", "Feb", ...)
method(
  chronon_format_cyclical,
  list(cal_symmetry454$month, cal_symmetry454$year)
) <- function(x, y) "{cyc(month,year,label=TRUE,abbreviate=TRUE)}"

## -----------------------------------------------------------------------------
month_of_year(as.Date("1955-11-12"), calendar = cal_symmetry454)

## -----------------------------------------------------------------------------
duration(1L, cal_symmetry454$year(1L))
duration(3L, cal_symmetry454$month(1L))

## -----------------------------------------------------------------------------
method(chronon_parse_linear, list(cal_symmetry454$month, S7::class_any)) <- function(x, cal) {
  parse_format(
    "{lin(year(1L))}-{cyc(month(1L), year(1L), label=TRUE, abbreviate=TRUE)}",
    "{lin(year(1L))}-{cyc(month(1L), year(1L))}"
  )
}

## -----------------------------------------------------------------------------
time_parse(c("1970-Jan", "1970-03"), format = chronon_parse_linear(cal_symmetry454$month(1L)))

#' Extract linear and cyclical time components
#'
#' `time_components()` decomposes a time vector into its constituent parts using
#' `dplyr::mutate()`-like semantics. Each named expression is built from the
#' [lin()] and [cyc()] helpers (the same vocabulary used in [format()] strings)
#' and produces a component time vector:
#'
#' - `lin(<granule>)` extracts a **linear** component (a non-repeating count,
#'   e.g. the year), returning a linear time vector.
#' - `cyc(<granule>, <cycle>)` extracts a **cyclical** component (a repeating
#'   position within a larger cycle, e.g. the month within the year), returning a
#'   cyclical time vector.
#'
#' All requested components are computed together in a single decomposition of
#' the underlying time vector (via `chronon_parts()`), reusing the shared
#' recursive `chronon_divmod()` results rather than converting each component
#' independently.
#'
#' @param x A `mixtime` (or an object coercible to one via [as_mixtime()], such
#'   as a `Date` or `POSIXct`).
#' @param ... Named expressions using [lin()] and [cyc()] describing the
#'   components to extract. The granule names (e.g. `year`, `month`, `day`) are
#'   resolved in the calendar of `x`.
#' @param calendar Calendar system used to resolve granule names, overlaid on the
#'   calendar of `x`. Defaults to `time_calendar(x)`. Supply e.g. [cal_isoweek]
#'   to make ISO `week`-based components available.
#'
#' @return A data frame with one column per requested component. `lin()` columns
#'   are linear (`mt_linear`) time vectors and `cyc()` columns are cyclical
#'   (`mt_cyclical`) time vectors.
#'
#' @seealso [lin()] and [cyc()] for the component helpers, [linear_time()] and
#'   [cyclical_time()] for constructing individual component vectors, and
#'   [format()] for the string counterpart of this interface.
#'
#' @examples
#' t <- yearmonth(as.Date("2026-02-14") + c(0, 40, 400))
#'
#' # Extract the year (linear) and month-of-year (cyclical)
#' time_components(t, yr = lin(year), mth = cyc(month, year))
#'
#' # Components can be named automatically from the expression
#' time_components(as.Date("2025-12-15") + 0:3, cyc(day, cal_isoweek$week))
#'
#' @export
time_components <- function(x, ..., calendar = time_calendar(x)) {
  specs <- enquos(..., .named = TRUE)
  if (length(specs) == 0L) {
    cli::cli_abort(
      c(
        "Must supply at least one component to extract.",
        i = "Use {.fn lin} for linear components, e.g. {.code lin(year)}.",
        i = "Use {.fn cyc} for cyclical components, e.g. {.code cyc(month, year)}."
      ),
      call = NULL
    )
  }

  x <- as_mixtime(x)

  # Decompose each underlying (uni-typed) mt_time part once, computing every
  # requested component together via a single chronon_parts() traversal.
  decomp <- lapply(x@x, component_parts, specs = specs, calendar = calendar)

  # Rebuild each component as a mixtime column, reusing x's ordering index by
  # swapping the vecvec parts (each decomposed part matches the original length).
  cols <- lapply(seq_along(specs), function(j) {
    col <- x
    col@x <- lapply(decomp, `[[`, j)
    col
  })

  vctrs::new_data_frame(
    rlang::set_names(cols, names(specs)),
    class = c("tbl_df", "tbl")
  )
}

#' Linear and cyclical component helpers
#'
#' `lin()` and `cyc()` name the time components addressed by mixtime's
#' *component-aware contexts* — a single vocabulary shared across [format()] (and
#' parsing) format strings and [time_components()] expressions. They are only
#' meaningful inside one of these contexts; calling them directly is an error.
#'
#' A linear and a cyclical component of the same granule store the *same* value:
#' the count of that chronon since the Unix epoch. The reduction to a within-cycle
#' position happens only when a cyclical vector is formatted. `cyc()` therefore
#' behaves like `lin()` on its finest granule but additionally records the cycle.
#'
#' @param granule The time granule to address, given as a granule generator (e.g.
#'   `year`) or a sized time unit (e.g. `year(1L)`). Resolved in the calendar of
#'   the time vector being formatted or decomposed.
#' @param cycle The coarser granule defining the cycle a `cyc()` component repeats
#'   within (e.g. `year` in `cyc(month, year)`).
#'
#' @return A component specification, consumed internally by the component-aware
#'   context (e.g. [format()] or [time_components()]).
#'
#' @seealso [format()] and [time_components()]
#'
#' @name component_helpers
#' @export
lin <- function(granule) {
  cli::cli_abort(
    c(
      "{.fn lin} must be used inside a component-aware context.",
      i = "Use it in a {.fn format} string, when parsing, or in a {.fn time_components} expression, e.g. {.code lin(year)}."
    ),
    call = NULL
  )
}

#' @rdname component_helpers
#' @export
cyc <- function(granule, cycle) {
  cli::cli_abort(
    c(
      "{.fn cyc} must be used inside a component-aware context.",
      i = "Use it in a {.fn format} string, when parsing, or in a {.fn time_components} expression, e.g. {.code cyc(month, year)}."
    ),
    call = NULL
  )
}

# Decompose a single mt_time into the requested lin()/cyc() component vectors.
# Returns a list of mt_linear / mt_cyclical vectors, one per element of `specs`,
# in the same order.
#
# Every component is obtained from the `$linear` decomposition of its finest
# granule (see the note on [lin()]/[cyc()]); `cyc()` merely also records the
# cycle. All granules are extracted together in a single `chronon_parts()` call.
component_parts <- function(x_time, specs, calendar) {
  chronon <- attr(x_time, "chronon")
  mask <- component_mask(calendar, chronon)

  evaled <- lapply(specs, eval_component, mask = mask)

  # Drop special values (NA / Inf) before decomposition, as format() does.
  xd <- vec_data(x_time)
  keep <- !(is.na(x_time) | is.infinite(xd))
  n <- length(x_time)

  # Match the discreteness of the input: discrete (integer) time yields integer
  # components, continuous (double) time keeps fractional positions.
  discrete <- is.integer(xd)
  coerce <- if (discrete) as.integer else identity

  # A component finer than the input chronon is temporally indeterminate in
  # discrete time (the month of a year() is undefined), so its value is NA
  # everywhere and it takes no part in the decomposition; continuous time
  # resolves finer granules exactly. This is time_is_determinate_at()'s per-part
  # rule (specials are handled separately below via `keep`).
  determinable <- vapply(
    evaled,
    function(spec) !discrete || chronon_nests_in(chronon, spec[[1L]]),
    logical(1L)
  )

  # Extract the finest granule (chronon) of every determinable component as a
  # linear part in a single shared traversal, then re-align to all specs (the
  # indeterminate ones stay NULL and become NA below).
  linear_vals <- vector("list", length(evaled))
  linear_vals[determinable] <- chronon_parts(
    x_time[keep],
    linear = lapply(evaled[determinable], `[[`, 1L)
  )$linear

  .mapply(
    function(spec, val, ok) {
      if (ok) {
        # chronon_parts() adds chronon_epoch() to linear parts for display; undo
        # it so the stored value is a count of chronons since the Unix epoch.
        val <- coerce(val - chronon_epoch(spec[[1L]]))
      }
      if (length(spec) == 2L) {
        # Cyclical: NA for an indeterminate component, and for a special
        # (NA / Inf) input which has no cyclical position.
        vals <- if (ok) {
          component_scatter(val, n, keep)
        } else {
          xd[rep(NA_integer_, n)]
        }
        mt_cyclical(vals, chronon = spec[[1L]], cycle = spec[[2L]])
      } else {
        # Linear: NA for an indeterminate component; otherwise an infinite time
        # has an infinite count, so carry the input's +/-Inf (and NA) through.
        vals <- if (ok) {
          component_scatter(val, n, keep, special = xd)
        } else {
          xd[rep(NA_integer_, n)]
        }
        mt_linear(vals, chronon = spec[[1L]])
      }
    },
    dots = list(evaled, linear_vals, determinable),
    MoreArgs = NULL
  )
}

# Build the lin()/cyc() data-mask shared by component-aware contexts.
#
# `chronon` (if specified) provides the primary source of time units that form
# granules for the components, and `calendar` adds additional units not present
# in the calendar of `chronon` (useful for cyclical granularities between
# calendars, e.g. day-of-week).
component_mask <- function(calendar, chronon = NULL) {
  cal <- calendar
  if (!is.null(chronon)) {
    cal <- time_calendar(chronon)
    extra <- setdiff(names(calendar), names(cal))
    cal[extra] <- calendar[extra]
  }

  as_tu <- function(g) {
    if (!S7_inherits(g, mt_unit)) {
      g <- g(1L)
    }
    if (!is.null(chronon)) {
      g <- granule_inherit_props(g, chronon)
    }
    g
  }

  # `...` carries options (e.g. label arguments) attached as attribute
  c(
    cal,
    list(
      lin = function(granule, ...) structure(list(as_tu(granule)), ...),
      cyc = function(granule, cycle, ...) {
        structure(list(as_tu(granule), as_tu(cycle)), ...)
      }
    )
  )
}

# Evaluate a single component expression in the lin()/cyc() mask, adding the
# familiar hint when a unit (commonly `week`) is missing from the calendar.
eval_component <- function(quo, mask) {
  tryCatch(
    eval_tidy(quo, data = mask),
    error = function(e) {
      msg <- conditionMessage(e)
      if (
        grepl("week", msg, fixed = TRUE) &&
          grepl("not found", msg, fixed = TRUE)
      ) {
        cli::cli_abort(
          c(
            msg,
            i = "The {.code week} unit is not in this calendar.",
            ">" = "Supply the ISO week calendar, e.g. {.code calendar = cal_isoweek}."
          ),
          call = NULL
        )
      }
      cli::cli_abort(msg, call = NULL)
    }
  )
}

# Scatter computed component values back into their original positions. Dropped
# (special) positions are filled from `special`, a length-n template that
# defaults to NA of `val`'s type; linear components pass the input data so that
# +/-Inf carries through unchanged (and NA stays NA).
component_scatter <- function(
  val,
  n,
  keep,
  special = val[rep(NA_integer_, n)]
) {
  if (all(keep)) {
    return(val)
  }
  out <- special
  out[keep] <- val
  out
}

# Tests for time_components(): mutate-like extraction of linear (lin()) and
# cyclical (cyc()) components from time vectors.
#
# A linear and a cyclical component of the same granule store the same value
# (the chronon count since the Unix epoch); the difference is only in how a
# cyclical vector is formatted. So a lin()/cyc() component must be identical to
# the corresponding linear_time()/cyclical_time() helper.

test_that("lin() extracts linear components matching the helper functions", {
  t <- yearmonth(as.Date("2026-02-14") + c(0, 40, 400))

  res <- time_components(t, yr = lin(year), ym = lin(month))

  expect_s3_class(res, "data.frame")
  expect_named(res, c("yr", "ym"))
  expect_true(all(time_is_linear(res$yr)))
  expect_identical(res$yr, year(t))
  expect_identical(res$ym, yearmonth(t))
})

test_that("cyc() extracts cyclical components matching the helper functions", {
  t <- yearmonth(as.Date("2026-02-14") + c(0, 40, 400))

  res <- time_components(t, m = cyc(month, year), q = cyc(quarter, year))

  expect_true(all(time_is_cyclical(res$m)))
  expect_identical(res$m, month_of_year(t))
  expect_identical(res$q, cyclical_time(t, chronon = quarter(1L), cycle = year(1L)))
})

test_that("cyc() with an alternate calendar (ISO week) works", {
  d <- as.Date("2025-12-15") + 0:6  # Mon..Sun

  res <- time_components(d, dow = cyc(day, week), calendar = cal_isoweek)

  expect_identical(res$dow, day_of_week(d))
  expect_equal(format(res$dow), format(day_of_week(d)))
})

test_that("time-of-day components extract from a datetime (second chronon)", {
  dt <- datetime(as.POSIXct(c("2026-02-14 06:30:00", "2026-02-14 18:45:00"), tz = "UTC"))

  res <- time_components(dt, tod = cyc(second, day), h = cyc(hour, day))

  expect_identical(res$tod, time_of_day(dt))
  expect_equal(format(res$h), c("h06", "h18"))
})

test_that("columns follow the ... order and support automatic naming", {
  t <- yearmonth(as.Date("2026-02-14"))

  res <- time_components(t, cyc(month, year), yr = lin(year))

  expect_length(res, 2L)
  expect_identical(names(res)[[2L]], "yr")
  # First column auto-named from its expression.
  expect_match(names(res)[[1L]], "cyc\\(month, year\\)")
})

test_that("special values (NA) propagate to all components", {
  t <- yearmonth(date(c("2026-02-14", NA, "2027-03-01")))

  res <- time_components(t, yr = lin(year), m = cyc(month, year))

  expect_true(is.na(res$yr[2L]))
  expect_true(is.na(res$m[2L]))
  expect_equal(format(res$yr), c("2026", "NA", "2027"))
})

test_that("Inf inputs give infinite linear and NA cyclical components", {
  t <- yearmonth(c(0, NA, Inf, -Inf))

  res <- time_components(t, yr = lin(year), m = cyc(month, year))

  # Linear: an infinite time is an infinite count; NA stays NA.
  expect_identical(is.na(res$yr), c(FALSE, TRUE, FALSE, FALSE))
  expect_equal(trimws(format(res$yr)), c("1970 0.0%", "NA", "Inf", "-Inf"))

  # Cyclical: no cyclical position for a special value, so NA (incl. Inf).
  expect_true(all(is.na(res$m[2:4])))
})

test_that("mixed-granularity mixtimes decompose each part correctly", {
  mx <- c(yearmonth(date("2026-02-14")), yearquarter(date("2026-08-01")))

  res <- time_components(mx, yr = lin(year), q = cyc(quarter, year))

  expect_equal(format(res$yr), format(year(mx)))
  expect_equal(format(res$q), c("Q1", "Q3"))
})

test_that("discrete components finer than the input chronon are NA", {
  # Discrete (integer) year(2020L) has no determinate month.
  mx <- c(year(2020L), yearmonth(as.Date("2020-02-01")))

  res <- time_components(mx, yr = lin(year), mth = cyc(month, year), m = lin(month))

  # Month components are NA for the year(); the yearmonth() resolves to February.
  expect_identical(is.na(res$mth), c(TRUE, FALSE))
  expect_identical(is.na(res$m), c(TRUE, FALSE))
  expect_equal(format(res$mth), c("NA", "Feb"))

  # The coarser (year) component is determinable for both.
  expect_false(any(is.na(res$yr)))
  expect_equal(format(res$yr), c("2020", "2020"))
})

test_that("continuous time resolves finer components exactly (no NA)", {
  # Continuous (double) year(2020) is 0% of the way through 2020, i.e. the start
  # of January -> the month is determinate, not NA.
  res <- time_components(year(2020), mth = cyc(month, year), m = lin(month))

  expect_false(is.na(res$mth))
  expect_false(is.na(res$m))
  expect_equal(format(res$mth), "Jan 0.0%")
})

test_that("components are extracted at multi-unit granule sizes", {
  d <- as.Date(c("2023-06-21", "2024-06-21", "2025-06-21", "2026-06-21"))

  # Two-year granules: the calendar's own day -> year divmod converts to single
  # years only, so the size is resolved by rescaling from the year granule.
  res <- time_components(d, y = lin(cal_gregorian$year(2L)))
  expect_equal(format(res$y), c("2022", "2024", "2024", "2026"))

  # Quarters expressed as three months are counted in quarters, not months
  res <- time_components(d, q = lin(cal_gregorian$month(3L)))
  expect_equal(
    vctrs::vec_data(vecvec::unvecvec(res$q)),
    ((2023:2026 - 1970) * 12L + 5L) %/% 3L
  )
})

test_that("time_components() errors informatively on misuse", {
  t <- yearmonth(date("2026-02-14"))

  expect_error(time_components(t), "at least one component")
  expect_error(
    time_components(as.Date("2025-12-15"), cyc(day, week)),
    "week"
  )
})

# Tests for time_compose(): the inverse of time_components(), building a
# single mixtime time point back up from lin()/cyc() components supplied as
# `spec ~ value` formulas or as already-tagged linear/cyclical time vectors.
# A cyc() formula's value is the 1-indexed position within the cycle (e.g.
# month 3 is March, day 15 is the 15th), matching everyday counting rather
# than the raw 0-indexed position chronon_parts()/time_components() use
# internally.

test_that("composes a date from year/month/day formulas", {
  r <- time_compose(lin(year) ~ 1980, cyc(month, year) ~ 3, cyc(day, month) ~ 15)

  expect_equal(format(r), "1980-03-15")
  expect_equal(as.numeric(vecvec::unvecvec(r)), as.numeric(as.Date("1980-03-15")))
})

test_that("a lin() anchor alone is a valid, coarser time point", {
  r <- time_compose(lin(year) ~ 1980)

  expect_equal(format(r), "1980")
  expect_true(time_is_linear(r))
})

test_that("round-trips through time_components() output columns", {
  parts <- time_components(as.Date("2024-03-15"), yr = lin(year), mth = cyc(month, year))

  r <- time_compose(parts$yr, parts$mth)

  expect_equal(format(r), "2024 Mar")
})

test_that("mixes formulas and already-tagged vectors", {
  r <- time_compose(lin(year) ~ 1980, month_of_year(as.Date("1980-03-01")))

  expect_equal(format(r), "1980 Mar")
})

test_that("cyc() formula values are 1-indexed, not the raw internal position", {
  # cyc(month, year) ~ 3 is the 3rd month (March), matching everyday counting
  # -- not chronon_parts()'s raw 0-indexed position (which would be April).
  r <- time_compose(lin(year) ~ 1980, cyc(month, year) ~ 3)

  expect_equal(format(r), "1980 Mar")
})

test_that("multi-unit (self-referencing) cycles compose correctly", {
  # Block 3 (0-indexed) of 3-month blocks is months 9-11 since epoch; the 3rd
  # (1-indexed) month within it is month 11 -> December 1970.
  r <- time_compose(
    lin(cal_gregorian$month(3L)) ~ 3,
    cyc(cal_gregorian$month(1L), cal_gregorian$month(3L)) ~ 3
  )

  expect_equal(as.numeric(vecvec::unvecvec(r)), 11)
  expect_equal(format(r), "1970 Dec")
})

test_that("a multi-unit cyc() only connects to a matching-size frontier", {
  expect_error(
    time_compose(
      lin(year) ~ 1980,
      cyc(month, year) ~ 3,
      cyc(cal_gregorian$month(1L), cal_gregorian$month(3L)) ~ 3
    ),
    "does not connect"
  )
})

test_that("vectorised components recycle to a common length", {
  r <- time_compose(lin(year) ~ 1980, cyc(month, year) ~ 1:3, cyc(day, month) ~ 1)

  expect_equal(format(r), c("1980-01-01", "1980-02-01", "1980-03-01"))
})

test_that("NA in any component propagates to NA in the result", {
  r1 <- time_compose(lin(year) ~ c(1980, NA), cyc(month, year) ~ 3, cyc(day, month) ~ 15)
  expect_identical(is.na(r1), c(FALSE, TRUE))

  r2 <- time_compose(lin(year) ~ 1980, cyc(month, year) ~ c(3, NA), cyc(day, month) ~ 15)
  expect_identical(is.na(r2), c(FALSE, TRUE))
})

test_that("Inf in the anchor carries through as an infinite count", {
  r <- time_compose(lin(year) ~ c(1980, Inf, -Inf), cyc(month, year) ~ 3, cyc(day, month) ~ 15)

  expect_identical(is.na(r), c(FALSE, FALSE, FALSE))
  expect_equal(as.numeric(vecvec::unvecvec(r)), c(3726, Inf, -Inf))
})

test_that("a single cyc() with no lin() anchor composes a cyclical time point", {
  r <- time_compose(cyc(month, year) ~ 3)

  expect_true(time_is_cyclical(r))
  expect_equal(format(r), "Mar")
  expect_equal(
    as.numeric(vecvec::unvecvec(r)),
    as.numeric(vecvec::unvecvec(month_of_year(as.Date("1970-03-01"))))
  )
})

test_that("a chained cyc() sequence with no lin() anchor composes a cyclical time point", {
  # Rooted at cyc(month, year); day 15 of month 3 collapses to (day, year).
  r <- time_compose(cyc(day, month) ~ 15, cyc(month, year) ~ 3)

  expect_true(time_is_cyclical(r))
  expect_equal(format(r), "D74")
  expect_equal(
    as.numeric(vecvec::unvecvec(r)),
    as.numeric(vecvec::unvecvec(day_of_year(as.Date("1970-03-15"))))
  )
})

test_that("cyclical time_compose() output round-trips through time_compose() again", {
  # 1981, a non-leap year, keeps day-of-year 74 on 15 March.
  anchored <- time_compose(lin(year) ~ 1981, cyc(month, year) ~ 3, cyc(day, month) ~ 15)
  cyclical <- time_compose(cyc(day, month) ~ 15, cyc(month, year) ~ 3)

  r <- time_compose(lin(year) ~ 1981, cyclical)
  expect_equal(format(r), format(anchored))
})

test_that("errors when every cyc() component's cycle is itself supplied (no coarsest link)", {
  # Neither component is coarser than the other, so neither can root the chain.
  expect_error(
    time_compose(cyc(day, month) ~ 15, cyc(month, day) ~ 1),
    "coarsest link"
  )
})

test_that("errors on disconnected cyc()-only chains with no lin() anchor", {
  expect_error(
    time_compose(cyc(month, year) ~ 3, cyc(quarter, year) ~ 1),
    "disconnected"
  )
})

test_that("errors when multiple lin() anchors are supplied", {
  expect_error(time_compose(lin(year) ~ 1980, lin(month) ~ 5), "exactly one")
})

test_that("errors when a cyc() component doesn't connect to the chain", {
  expect_error(
    time_compose(lin(year) ~ 1980, cyc(day, month) ~ 15),
    "does not connect"
  )
})

test_that("errors when two cyc() components branch from the same frontier", {
  expect_error(
    time_compose(lin(year) ~ 1980, cyc(month, year) ~ 3, cyc(quarter, year) ~ 1),
    "Multiple.*cyc"
  )
})

test_that("errors on a duplicate cyc() component", {
  expect_error(
    time_compose(lin(year) ~ 1980, cyc(month, year) ~ 3, cyc(month, year) ~ 4),
    "Multiple.*cyc"
  )
})

test_that("errors when a cyc() value is out of range", {
  expect_error(
    time_compose(lin(year) ~ 1980, cyc(month, year) ~ 3, cyc(day, month) ~ 32),
    "not a valid"
  )
})

test_that("errors when a cyc() value is out of range for a boundaried cycle (12-hour AM/PM clock)", {
  # Valid 1-indexed hour-within-ampm positions are 1-12; 13 overflows into
  # the next ampm and is rejected rather than silently wrapping.
  expect_error(
    time_compose(lin(cal_time_civil$ampm) ~ 0, cyc(cal_time_civil$hour, cal_time_civil$ampm) ~ 13),
    "not a valid"
  )
})

test_that("errors for an input that is neither a formula nor a time vector", {
  expect_error(time_compose(42), "formula")
})

test_that("errors on an empty call", {
  expect_error(time_compose(), "at least one component")
})

#' Constructor for mixtime vectors
#' 
#' Creates a `mixtime` vector, which can contain time points of different 
#' granularities (e.g. monthly and quarterly) in a single vector via `vecvec`.
#' 
#' @param x A mixtime time vector (created with [new_time()]) to wrap in a mixtime class.
#' 
#' @return A `mixtime` object, which allows mixed-type time vectors to coexist in a single vector.
#' 
#' @export
new_mixtime <- function(x = mt_linear()) {
  stopifnot(S7_inherits(x, mt_time))
  class_mixtime(list(x))
}

#' Create a mixtime vector
#'
#' A mixtime is a vector which describes a point in time. It uses a calendar
#' definition to translate a vector of numbers into a point in time.
#'
#' @param data A vector of time values. This can be a character vector (e.g. "2024-01-01"),
#'  a numeric vector (e.g. seconds since epoch), or a time class (e.g. Date, POSIXct, yearmonth, etc.).
#' @param chronon A time granule object representing the smallest indivisible time granule (chronon) for the mixtime. This is used to interpret the numeric values in `data` and to define the time resolution of the mixtime. If not provided, it will be inferred from `data`.
#' @param cycle An optional time granule object representing the cycle for cyclical time. This is used to define the repeating cycle for cyclical time representations (e.g. day-of-week, month-of-year). If not provided, the mixtime will be treated as linear time.
#' @param discrete A logical indicating whether the time values should be treated as discrete (integer) or continuous (fractional). This affects how numeric values are interpreted and how time arithmetic is performed. The default is `TRUE` (discrete).
#' 
#' @return A `mixtime` object representing the time values in `data` according to the specified `chronon` and `cycle`.
#' 
#' @examples
#' # Create a mixtime for today
#' mixtime(Sys.Date())
#' 
#' # Create a mixtime for the current date and time
#' mixtime(Sys.time())
#' 
#' # Convert time from tsibble classes to mixtime
#' mixtime(tsibble::yearmonth("2024 Jan"))
#' 
#' # Create a mixtime for the time of day (cyclical time)
#' mixtime(Sys.time(), cycle = cal_gregorian$day(1L))
#' 
#' # Specify a timezone for the chronon
#' mixtime(Sys.time(), chronon = cal_gregorian$second(1L, tz = Sys.timezone()))
#' mixtime(Sys.time(), chronon = cal_gregorian$second(1L, tz = "Pacific/Honolulu"))
#' mixtime(Sys.time(), chronon = cal_gregorian$second(1L, tz = "Australia/Melbourne"))
#' 
#' # Dates (and all granularities) can have timezones
#' mixtime(Sys.time(), chronon = cal_gregorian$day(1L, tz = Sys.timezone()))
#' mixtime(Sys.time(), chronon = cal_gregorian$day(1L, tz = "Pacific/Honolulu"))
#' mixtime(Sys.time(), chronon = cal_gregorian$day(1L, tz = "Australia/Melbourne"))
#' 
#' # Continuous time tracks progress within the chronon
#' mixtime(Sys.time(), chronon = cal_gregorian$day(1L, tz = Sys.timezone()), discrete = FALSE)
#' 
#' # Mixtime can combine different granularities and timezones in a vector
#' now <- Sys.time()
#' c(
#'   # Datetime (second chronon) in UTC
#'   mixtime(now),
#'   # Date (minute chronon) in local timezone
#'   mixtime(now, chronon = cal_gregorian$minute(1L, tz = Sys.timezone())),
#'   # Month (month chronon) in UTC
#'   mixtime(now, chronon = cal_gregorian$month(1L))
#' )
#' @export
mixtime <- function(data, chronon = time_chronon(data), cycle = time_cycle(data), discrete = TRUE) {
  # Handle duration-based chronon and cycle inputs
  if (is_mixtime(chronon)) {
    if (length(chronon@x) > 1L) {
      cli::cli_abort(
        c(
          "mixtime() currently only supports singular granule chronons",
          i = "To mix chronons, combine several mixtime vectors with {.fun c}."
        ),
        call = NULL
      )
    }
    chronon_duration <- chronon@x[[1L]]
    chronon <- attr(chronon_duration, "chronon")
    if (length(chronon_duration) == 1L) {
      # Scale the granule by the chronon duration (e.g. 3 months -> 3 * month(1L))
      chronon@n <- chronon@n * as.numeric(chronon_duration)
    }
  }
  if (is_mixtime(cycle)) {
    if (length(cycle@x) > 1L) {
      cli::cli_abort(
        c(
          "mixtime() currently only supports singular granule cycles",
          i = "To mix cycles, combine several mixtime vectors with {.fun c}."
        ),
        call = NULL
      )
    }
    cycle_duration <- cycle@x[[1L]]
    if (anyNA(cycle_duration)) {
      cycle <- NULL
    } else {
      cycle <- attr(cycle_duration, "chronon")
      if (length(cycle_duration) == 1L) {
        # Scale the granule by the cycle duration (e.g. 3 months -> 3 * month(1L))
        cycle@n <- cycle@n * as.numeric(cycle_duration)
      }
    }
  }
  
  # Add default granule properties if not given in chronon or cycle
  chronon <- granule_inherit_props(chronon, chronon_common(data))
  if (!is.null(cycle)) {
    cycle <- granule_inherit_props(cycle, chronon)
  }

  # Validate time granules
  if (!inherits(chronon, "mixtime::mt_unit")) {
    cli::cli_abort("{.var chronon} must be a time granule object.", call. = FALSE)
  }
  if (!is.null(cycle) && !inherits(cycle, "mixtime::mt_unit")) {
    cli::cli_abort("{.var cycle} must be a time granule object.", call. = FALSE)
  }

  # Parse text data
  if (is.character(data)) {
    return(time_parse(data, chronon = chronon, cycle = cycle, discrete = discrete))
  } else {
    # Apply origin offset for numeric data
    if (!is_mixtime(data) && is.numeric(data) && (epoch <- chronon_epoch(chronon)) != 0L) {
      data <- data - epoch
    }

    # Cast from Date, POSIXct, etc.
    if (!is.numeric(data) || !is.null(attributes(data))) {
      data <- chronon_convert(
        data,
        chronon,
        discrete = discrete
      )
    }
  }

  new_mixtime(
    if (is.null(cycle)) {
      mt_linear(data, chronon = chronon)
    } else {
      mt_cyclical(data, chronon = chronon, cycle = cycle)
    }
  )
}

#' Check if an object is a mixtime
#'
#' Tests whether `x` inherits from the `mixtime` class.
#'
#' @param x An object to test.
#'
#' @return A scalar logical: `TRUE` if `x` is a `mixtime` vector, `FALSE`
#'   otherwise.
#'
#' @seealso [as_mixtime()] to coerce objects to `mixtime`,
#'   [mixtime()] to construct a `mixtime`.
#'
#' @examples
#' is_mixtime(Sys.Date())
#' is_mixtime(mixtime(Sys.Date()))
#'
#' @export
is_mixtime <- function(x) {
  S7_inherits(x, class_mixtime)
}

#' @method Summary mixtime::mixtime
#' @export
`Summary.mixtime::mixtime` <- function(..., na.rm = FALSE) {
  new_mixtime(NextMethod())
}

method(c, class_mixtime) <- function(...) {
  dots <- unname(rlang::list2(...))
  is_mt <- vapply(dots, is_mixtime, logical(1L))
  dots[!is_mt] <- lapply(dots[!is_mt], vec_cast_to_mixtime, to = new_mixtime())
  do.call(utils::getS3method("c", "vecvec::vecvec"), dots)
}

test_that("Linear time handles text parsing with set timezones", {
  # Test that date preserves the date in the specified timezone
  result <- date("2020-01-01", tz = "America/Los_Angeles")
  
  # Should display as 2020 Jan 1 in PST timezone
  expect_equal(format(result), "2020-01-01 PST")
  
  # Test with another timezone
  result_utc <- date("2020-01-01", tz = "UTC")
  expect_equal(format(result_utc), "2020-01-01")
  
  # Test with Europe/Berlin
  result_berlin <- date("2020-01-01", tz = "Europe/Berlin")
  expect_equal(format(result_berlin), "2020-01-01 CET")
})

test_that("linear_time() creates hour and minute times with timezone", {
  # Create a POSIXct in America/Los_Angeles
  base_time <- as.POSIXct("2020-01-15 14:30:00", tz = "America/Los_Angeles")
  result <- linear_time(base_time, minute(1L, tz = "America/Los_Angeles"))
  
  # Should include date and time in Pacific timezone
  expect_true(is_mixtime(result))
  expect_match(format(result), "2020-01-15 14:30 PST")
  
  # Test with UTC
  base_utc <- as.POSIXct("2020-01-15 09:15:00", tz = "UTC")
  result_utc <- linear_time(base_utc, minute(1L, tz = "UTC"))
  expect_match(format(result_utc), "2020-01-15 09:15")
  
  # Test with Australia/Melbourne
  base_melb <- as.POSIXct("2020-01-15 23:45:00", tz = "Australia/Melbourne")
  result_melb <- linear_time(base_melb, minute(1L, tz = "Australia/Melbourne"))
  expect_match(format(result_melb), "2020-01-15 23:45 AEDT")
  
  # Test with hour precision
  base_berlin <- as.POSIXct("2020-01-15 16:00:00", tz = "Europe/Berlin")
  result_hour <- linear_time(base_berlin, hour(1L, tz = "Europe/Berlin"))
  expect_match(format(result_hour), "2020-01-15 16h CET")
})

test_that("linear_time() converts between timezones showing date differences", {
  # Create a POSIXct representing 2020-01-01 09:00 in Australia/Melbourne
  melbourne_time <- as.POSIXct("2020-01-01 09:00:00", tz = "Australia/Melbourne")
  
  # Convert to America/Los_Angeles - should be previous day (2019-12-31)
  pacific_ymd <- date(melbourne_time, tz = "America/Los_Angeles")
  expect_equal(format(pacific_ymd), "2019-12-31 PST")
  
  # Convert to Europe/Berlin - should also be previous day (2019-12-31)
  berlin_ymd <- date(melbourne_time, tz = "Europe/Berlin")
  expect_equal(format(berlin_ymd), "2019-12-31 CET")
  
  # The Melbourne date should still be 2020-01-01
  melbourne_ymd <- date(melbourne_time)
  expect_equal(format(melbourne_ymd), "2020-01-01 AEDT")
})

test_that("linear_time() preserves instant across timezone conversions", {
  # Start with a specific moment in America/Los_Angeles
  pacific_time <- as.POSIXct("2020-06-15 09:00:00", tz = "America/Los_Angeles")
  
  # Convert to Australia/Melbourne
  melbourne_converted <- linear_time(pacific_time, chronon = second(1L, tz = "Australia/Melbourne"))
  
  # Convert back to America/Los_Angeles
  pacific_back <- linear_time(melbourne_converted, chronon = second(1L, tz = "America/Los_Angeles"))
  
  # Should represent the same instant in time
  expect_equal(as.numeric(pacific_time), as.numeric(pacific_back))
  
  # But the calendar dates differ
  expect_equal(
    format(date(melbourne_converted)),
    "2020-06-16 AEST"  # Next day in Melbourne
  )
  expect_equal(
    format(date(pacific_time)),
    "2020-06-15 PDT"  # Original day in Pacific
  )
})

test_that("linear_time() with mixtime inputs", {
  expect_equal(
    year(yearquarter(1L)),
    year(1970L)
  )
})

# Tests for Math group generics (round, floor, ceiling) on time objects.
#
# These operations apply to the inner numeric value of the time object,
# preserving the time class and its attributes.

# ---------------------------------------------------------------------------
# round / floor / ceiling on durations
# ---------------------------------------------------------------------------

test_that("round() on a duration rounds the inner value", {
  expect_equal(as.numeric(round(days(1.3))),  1)
  expect_equal(as.numeric(round(days(1.5))),  2)
  expect_equal(as.numeric(round(days(1.7))),  2)
  expect_equal(as.numeric(round(days(-1.3))), -1)
})

test_that("round() on a duration preserves the time class", {
  result <- round(days(1.3))
  expect_true(is_mixtime(result))
})

test_that("floor() on a duration floors the inner value", {
  expect_equal(as.numeric(floor(days(1.3))),  1)
  expect_equal(as.numeric(floor(days(1.9))),  1)
  expect_equal(as.numeric(floor(days(-1.3))), -2)
})

test_that("floor() on a duration preserves the time class", {
  result <- floor(days(1.3))
  expect_true(is_mixtime(result))
})

test_that("ceiling() on a duration ceils the inner value", {
  expect_equal(as.numeric(ceiling(days(1.1))),  2)
  expect_equal(as.numeric(ceiling(days(1.0))),  1)
  expect_equal(as.numeric(ceiling(days(-1.3))), -1)
})

test_that("ceiling() on a duration preserves the time class", {
  result <- ceiling(days(1.3))
  expect_true(is_mixtime(result))
})

# ---------------------------------------------------------------------------
# round / floor / ceiling on linear time
# ---------------------------------------------------------------------------

test_that("round() on a linear time rounds the inner value", {
  expect_equal(as.numeric(round(yearmonth(1.3))),  1)
  expect_equal(as.numeric(round(yearmonth(1.5))),  2)
  expect_equal(as.numeric(round(yearmonth(1.7))),  2)
})

test_that("round() on a linear time preserves the time class", {
  result <- round(yearmonth(1.3))
  expect_true(is_mixtime(result))
})

test_that("floor() on a linear time floors the inner value", {
  expect_equal(as.numeric(floor(yearmonth(1.3))), 1)
  expect_equal(as.numeric(floor(yearmonth(1.9))), 1)
})

test_that("ceiling() on a linear time ceils the inner value", {
  expect_equal(as.numeric(ceiling(yearmonth(1.1))), 2)
  expect_equal(as.numeric(ceiling(yearmonth(1.0))), 1)
})

# ---------------------------------------------------------------------------
# round() with digits argument
# ---------------------------------------------------------------------------

test_that("round() respects the digits argument for durations", {
  expect_equal(as.numeric(round(days(1.567), digits = 1)), 1.6)
  expect_equal(as.numeric(round(days(1.567), digits = 2)), 1.57)
})

#' Cardinality between time granules
#'
#' This S7 generic function defines the calendrical relationships between two 
#' chronons, and is one of the building block for defining calendars in mixtime.
#' It calculates how many `x` chronons fit into the `y` chronon. Some chronon 
#' sizes are context-dependent (such as the number of days in a month),
#' and so an optional time point defined in terms of `y` chronons can be
#' provided with `at`.
#' 
#' @param x The finer time granule (e.g. `cal_gregorian$month(1L)`)
#' @param y The coarser time granule (e.g. `cal_gregorian$year(1L)`)
#' @param ... Additional arguments for methods.
# #' @param at Optional time point for context-dependent cardinality, defined in
# #' terms of `y` (e.g., if `y` is `month()`, then `at` could be a 
# #' `yearmonth()`)
#'
#' @return Numeric describing how many `x` time granules fit into `y` at time `at`.
#' 
#' @details
#' 
#' The methods are dispatched based on the shortest path along defined methods.
#' This allows for defining only the direct relationships between adjacent
#' time units, and relying on graph traversal to find how to convert between
#' more distant units. For example the number of seconds in an hour can be
#' calculated from the number of seconds in a minute and then number of minutes
#' in an hour.
#' 
#' If a method is defined for converting between time units of different 
#' calendar systems (e.g., Gregorian calendar days to Chinese calendar days),
#' then that method can be used to convert times at any granularity between the
#' two systems.
#' 
#' @examples
#' # There are 12 months in a year
#' with(cal_gregorian, chronon_cardinality(month(1L), year(1L)))
#' 
#' # There are 7 days in a week
#' with(cal_isoweek, chronon_cardinality(day(1L), week(1L)))
#' 
#' # There are 3600 seconds in an hour
#' with(cal_gregorian, chronon_cardinality(second(1L), hour(1L)))
#' 
#' # There are 18 "2 months" in 3 years
#' with(cal_gregorian, chronon_cardinality(month(2L), year(3L)))
#' 
#' # There are 365 days in 2025 (a common year)
#' chronon_cardinality(
#'   cal_gregorian$day(1L), cal_gregorian$year(1L),
#'   at = year(as.Date("2025-01-01"))
#' )
#' 
#' # There are 366 days in 2024 (a leap year)
#' chronon_cardinality(
#'   cal_gregorian$day(1L), cal_gregorian$year(1L), 
#'   at = mixtime::year(as.Date("2024-01-01"))
#' )
#' 
#' # There are 29 days in February 2024 (a leap year)
#' chronon_cardinality(
#'   cal_gregorian$day(1L), cal_gregorian$month(1L), 
#'   at = yearmonth(as.Date("2024-02-01"))
#' )
#'
#' @export
chronon_cardinality <- S7::new_generic("chronon_cardinality", c("x", "y"))

#' Fixed cardinality between time granules
#'
#' A restricted variant of [chronon_cardinality()] for time granule pairs
#' whose relationship is a constant, context-independent number (e.g., 60
#' seconds in a minute, 24 hours in a day). Unlike `chronon_cardinality()`,
#' methods for this generic do not receive (and must not need) an `at` time
#' point, and should return the number of unit (`n = 1L`) `x` granules that
#' fit within one unit `y` granule.
#'
#' Defining a method for `chronon_cardinality_fixed()` automatically provides
#' a `chronon_cardinality()` method for the pair (scaled by the requested
#' granule sizes via the `list(mt_unit, mt_unit)` fallback method), and marks
#' the relationship as safe to use for [chronon_divmod()]'s graph traversal,
#' where variable (context-dependent) cardinalities cannot be used since no
#' `at` is available mid-traversal.
#'
#' @inheritParams chronon_cardinality
#'
#' @return A single number describing how many unit `x` time granules fit
#'   into a unit `y` time granule.
#'
#' @keywords internal
#' @export
chronon_cardinality_fixed <- S7::new_generic("chronon_cardinality_fixed", c("x", "y"))

# #' @rdname chronon_cardinality
method(chronon_cardinality, list(mt_unit, mt_unit)) <- function(x, y, at = NULL) {
  # Check if x and y are the same class
  if (S7_class_id(x) == S7_class_id(y)) {
    return(y@n/x@n)
  }

  # Use a registered fixed-cardinality method if available, scaling the
  # constant (unit granule) ratio by the requested granule sizes.
  if (!is.null(chronon_cardinality_fixed@methods[[S7_class_id(x)]][[S7_class_id(y)]])) {
    return(chronon_cardinality_fixed(x, y) * y@n / x@n)
  }
  if (!is.null(chronon_cardinality_fixed@methods[[S7_class_id(y)]][[S7_class_id(x)]])) {
    return(y@n / (chronon_cardinality_fixed(y, x) * x@n))
  }

  # Try to find a method with arguments swapped
  # (This feels unsafe for finding exact matching of S7 dispatch.)

  if (!is.null(y_env <- chronon_cardinality@methods[[S7_class_id(y)]])) {
    if (S7_class_id(x) %in% names(y_env)) {
      # Matching inverse method found, use it with inversion.
      # TODO: Convert at to x units?
      return(1/chronon_cardinality(y, x, at = at))
    }
  }

  # No specific method defined between these classes
  # Attempt graph traversal to find a sequence of methods
  path <- S7_graph_dispatch(chronon_cardinality_graph(), y, x)

  path[[1]] <- y
  path[[length(path)]] <- x
  # Initialise intermediate classes with 1L and adjacent properties
  path[c(-1, -length(path))] <- lapply(path[c(-1, -length(path))], function(tu){
    # Ideally inherit from `x`, but if incomplete inherit from `y`
    granule_inherit_props(granule_inherit_props(tu(1L), x), y)
  })
  
  result <- path[[1]]
  for (i in seq(2, length.out = length(path)-1)) {
    ## QUESTION: Why does this not work with `generic` instead of `chronon_cardinality`? S7 bug?

    result <- chronon_cardinality(path[[i]], result, at)
    if (!is.null(at)) at <- at * result
    # Class the result with the next class in the path
    result <- S7::S7_class(path[[i]])(result)
  }

  result@n
  # vec_data(result)
}

# #' @rdname chronon_cardinality_fixed
method(chronon_cardinality_fixed, list(mt_unit, mt_unit)) <- function(x, y) {
  # Check if x and y are the same class
  if (S7_class_id(x) == S7_class_id(y)) {
    return(1)
  }

  # No specific method defined between these classes.
  # Attempt graph traversal using only registered fixed-cardinality edges,
  # multiplying the unit-granule cardinalities along the path.
  path <- S7_graph_dispatch(chronon_cardinality_fixed_graph(), y, x)

  result <- 1
  for (i in seq(2, length.out = length(path) - 1)) {
    fine <- path[[i]](1L)
    coarse <- path[[i - 1]](1L)
    # Allow the fixed cardinality method to be defined in either direction.
    result <- result * if (is.null(chronon_cardinality_fixed@methods[[S7_class_id(fine)]][[S7_class_id(coarse)]])) {
      1 / chronon_cardinality_fixed(coarse, fine)
    } else {
      chronon_cardinality_fixed(fine, coarse)
    }
  }
  result
}

# --- shared helpers ---------------------------------------------------------------

# Shift a time point by a duration
#
# Shifts involving time and duration chronons with irregular cardinalities are
# computed with calendar-field arithmetic, clamping any components that are
# invalid (e.g. 2026-01-31 + months(1) -> 2026-02-28, not 2026-03-03).
duration_shift <- function(time, dur) {
  time_data <- S7_data(time)
  time_chronon <- time@chronon
  dur_chronon <- dur@chronon

  if (chronon_needs_clamping(time_chronon, dur_chronon)) {
    path <- chronon_longest_path(time_chronon, dur_chronon)
    decomposed <- chronon_decompose(as.numeric(time_data), path)
    div <- decomposed$div[[length(path)]] + S7_data(dur)
    new_time_data <- as.numeric(
      chronon_recompose(div, decomposed$mod, path)$value
    )
  } else {
    dur_chronon <- granule_inherit_props(dur_chronon, time_chronon)
    at <- chronon_convert_impl(
      as.numeric(time_data),
      time_chronon,
      dur_chronon,
      discrete = FALSE
    )
    new_time_data <- chronon_convert_impl(
      at + S7_data(dur),
      dur_chronon,
      time_chronon,
      discrete = FALSE
    )
  }

  shift <- as.numeric(new_time_data) - time_data
  # Preserve integer type for discrete time points
  if (is.integer(time_data)) {
    shift <- as.integer(round(shift))
  }
  shift
}

# Combine two durations (or divide them) via their common chronon.
duration_combine <- function(e1, e2, op) {
  x_chronon <- e1@chronon
  y_chronon <- e2@chronon
  tu <- chronon_common_impl(list(x_chronon, y_chronon))
  # Scale magnitudes to common chronon units before performing arithmetic
  xd <- S7_data(e1) * chronon_cardinality(tu, x_chronon)
  yd <- S7_data(e2) * chronon_cardinality(tu, y_chronon)
  res <- op(xd, yd)
  # duration / duration = numeric ratio; otherwise a duration with the common chronon
  if (identical(op, `/`)) {
    return(res)
  }
  mt_duration(res, chronon = tu)
}

# --- math ---------------------------------------------------------------------
#' @importFrom vctrs vec_math
#' @export
S7::method(vec_math, mt_time) <- function(.fn, .x, ...) {
  if (.fn == "mean") {
    res <- vctrs::vec_math_base(.fn, .x, ...)
    if (is.integer(.x)) {
      res <- as.integer(res)
    }
    return(vec_restore(res, .x))
  }
  if (.fn %in% c("is.nan", "is.finite", "is.infinite")) {
    return(vctrs::vec_math_base(.fn, .x, ...))
  }
  if (.fn %in% c("round", "floor", "ceiling")) {
    res <- vctrs::vec_math_base(.fn, .x, ...)
    return(vec_restore(res, .x))
  }
  stop(
    sprintf("Math function '%s' not supported for continuous time", .fn),
    call. = FALSE
  )
}

# --- differences in time produce durations in the common chronon --------------

method(`-`, list(mt_time, mt_time)) <- function(e1, e2) {
  x_chronon <- e1@chronon
  y_chronon <- e2@chronon
  tu <- chronon_common_impl(list(x_chronon, y_chronon))
  cx <- chronon_convert(e1, tu, discrete = FALSE)
  cy <- chronon_convert(e2, tu, discrete = FALSE)
  # When a TZ-aware operand uses a coarser granule than the common chronon, the conversion
  # produces a wall-clock value rather than a UTC-aligned value (the TZ offset cancels in
  # same-TZ coarser->finer conversions). Subtract the TZ offset at the wall-clock position
  # to realign to UTC.
  cx <- tz_wall_clock_to_utc(cx, x_chronon, tu)
  cy <- tz_wall_clock_to_utc(cy, y_chronon, tu)
  mt_duration(cx - cy, chronon = tu)
}

# --- relative time (linear time and durations) --------------------------------

method(`+`, list(mt_time, mt_duration)) <- function(e1, e2) {
  S7_data(e1) <- S7_data(e1) + duration_shift(e1, e2)
  e1
}

method(`-`, list(mt_time, mt_duration)) <- function(e1, e2) {
  # Negate the duration itself to keep duration_shift() simpler
  S7_data(e2) <- -S7_data(e2)

  S7_data(e1) <- S7_data(e1) + duration_shift(e1, e2)
  e1
}

method(`+`, list(mt_duration, mt_time)) <- function(e1, e2) {
  S7_data(e2) <- duration_shift(e2, e1) + S7_data(e2)
  e2
}

method(`-`, list(mt_duration, mt_time)) <- function(e1, e2) {
  S7_data(e2) <- duration_shift(e2, e1) - S7_data(e2)
  e2
}

# --- combining durations (resulting in common chronon durations) --------------

method(`+`, list(mt_duration, mt_duration)) <- function(e1, e2) {
  duration_combine(e1, e2, `+`)
}
method(`-`, list(mt_duration, mt_duration)) <- function(e1, e2) {
  duration_combine(e1, e2, `-`)
}
method(`/`, list(mt_duration, mt_duration)) <- function(e1, e2) {
  duration_combine(e1, e2, `/`)
}

# --- implicit chronons (time and numeric) -------------------------------------

method(`+`, list(mt_time, S7::class_numeric)) <- function(e1, e2) {
  S7_data(e1) <- S7_data(e1) + e2
  e1
}
method(`-`, list(mt_time, S7::class_numeric)) <- function(e1, e2) {
  S7_data(e1) <- S7_data(e1) - e2
  e1
}
method(`+`, list(S7::class_numeric, mt_time)) <- function(e1, e2) {
  S7_data(e2) <- e1 + S7_data(e2)
  e2
}
method(`-`, list(S7::class_numeric, mt_time)) <- function(e1, e2) {
  S7_data(e2) <- e1 - S7_data(e2)
  e2
}

# --- numeric scaling of durations (e.g. days(3) * 2) --------------------------

method(`*`, list(mt_duration, S7::class_numeric)) <- function(e1, e2) {
  S7_data(e1) <- S7_data(e1) * e2
  e1
}
method(`/`, list(mt_duration, S7::class_numeric)) <- function(e1, e2) {
  S7_data(e1) <- S7_data(e1) / e2
  e1
}
method(`*`, list(S7::class_numeric, mt_duration)) <- function(e1, e2) {
  S7_data(e2) <- e1 * S7_data(e2)
  e2
}

# `mt_cyclical` has no dedicated numeric method: its stored value is a raw chronon count
# (like `mt_linear`), and the cyclical "position within the cycle" is purely a display-time
# computation from that raw count (see format.R). So `mt_cyclical + <number>` falls through
# to the general `(mt_time, class_numeric)` method above, which just advances the raw
# chronon and lets formatting recompute the position - this also makes variable-length
# cycles (e.g. day_of_month()) advance correctly, since there is no fixed period to wrap
# against.

# --- friendly errors for unsupported operator/operand combinations ----------------
# Registered on `class_any` so any operand pair that isn't handled above lands here. The
# concrete methods above are more specific and always take precedence.

method(`+`, list(mt_time, S7::class_any)) <- function(e1, e2) {
  cli::cli_abort(
    "Addition is only supported between a time point and a duration, or between durations.",
    call. = FALSE
  )
}
method(`+`, list(S7::class_any, mt_time)) <- function(e1, e2) {
  cli::cli_abort(
    "Addition is only supported between a time point and a duration, or between durations.",
    call. = FALSE
  )
}
method(`*`, list(mt_time, S7::class_any)) <- function(e1, e2) {
  cli::cli_abort(
    "Multiplication is only supported between a duration and a number.",
    call. = FALSE
  )
}
method(`*`, list(S7::class_any, mt_time)) <- function(e1, e2) {
  cli::cli_abort(
    "Multiplication is only supported between a duration and a number.",
    call. = FALSE
  )
}
method(`/`, list(mt_time, S7::class_any)) <- function(e1, e2) {
  cli::cli_abort(
    "Division is only supported between two durations, or a duration divided by a number.",
    call. = FALSE
  )
}
method(`/`, list(S7::class_any, mt_time)) <- function(e1, e2) {
  cli::cli_abort(
    "Division is only supported between two durations, or a duration divided by a number.",
    call. = FALSE
  )
}

# --- helpers ----------------------------------------------------------------------

# Correct wall-clock -> UTC alignment for coarser-granule TZ-aware operands.
# When from_chronon has a non-UTC TZ and is a different (coarser) class than to_chronon,
# the chronon_convert result is a wall-clock value (the offset was added then canceled).
# Subtracting the TZ offset at that position restores the true UTC-equivalent value.
tz_wall_clock_to_utc <- function(value, from_chronon, to_chronon) {
  if (!S7_inherits(from_chronon, mt_tz_unit)) {
    return(value)
  }
  from_tz <- tz_name(from_chronon)
  if (is.na(from_tz) || from_tz == "UTC") {
    return(value)
  }
  # Same granule class: offset already cancelled correctly (no correction needed)
  if (
    identical(S7::S7_class(from_chronon)@name, S7::S7_class(to_chronon)@name)
  ) {
    return(value)
  }
  # Coarser granule: subtract the TZ offset at the wall-clock position
  tzo <- get_tz_offset(as.double(value), from_tz)
  value - tzo
}

#' Check the time type of values
#'
#' Test whether elements of a `mixtime` vector are linear, cyclical, or durations.
#'
#' These helpers return a logical vector the same length as `x` identifying the
#' type of time represented by each element.
#'
#' @param x A time object (typically a `mixtime` vector).
#' @param ... Additional arguments for methods.
#'
#' @return A logical vector the same length as `x`.
#'
#' @examples
#' t <- c(yearmonth(0), month_of_year(0), months(0L))
#' time_is_linear(t)
#' time_is_cyclical(t)
#' time_is_duration(t)
#'
#' @name is_time
#' @export
time_is_linear <- S7::new_generic("time_is_linear", "x")
method(time_is_linear, class_mixtime) <- vecvec::vecvec_apply_fn(time_is_linear, logical())
method(time_is_linear, class_any) <- function(x) rep.int(FALSE, length(x))
method(time_is_linear, mt_linear) <- function(x) rep.int(TRUE, length(x))

#' @rdname is_time
#' @export
time_is_cyclical <- S7::new_generic("time_is_cyclical", "x")
method(time_is_cyclical, class_mixtime) <- vecvec::vecvec_apply_fn(time_is_cyclical, logical())
method(time_is_cyclical, class_any) <- function(x) rep.int(FALSE, length(x))
method(time_is_cyclical, mt_cyclical) <- function(x) rep.int(TRUE, length(x))

#' @rdname is_time
#' @export
time_is_duration <- S7::new_generic("time_is_duration", "x")
method(time_is_duration, class_mixtime) <- vecvec::vecvec_apply_fn(time_is_duration, logical())
method(time_is_duration, class_any) <- function(x) rep.int(FALSE, length(x))
method(time_is_duration, mt_duration) <- function(x) rep.int(TRUE, length(x))
method(time_is_duration, new_S4_class("Period", package = "lubridate")) <- function(x) rep.int(TRUE, length(x))

#' @rdname is_time
#' @export
is_time_linear <- function(x, ...) {
  lifecycle::deprecate_stop("0.3.0", "is_time_linear()", "time_is_linear()")
}

#' @rdname is_time
#' @export
is_time_cyclical <- function(x, ...) {
  lifecycle::deprecate_stop("0.3.0", "is_time_cyclical()", "time_is_cyclical()")
}

#' @rdname is_time
#' @export
is_time_duration <- function(x, ...) {
  lifecycle::deprecate_stop("0.3.0", "is_time_duration()", "time_is_duration()")
}

#' Test whether time is determinate at a granule's precision
#'
#' `time_is_determinate_at()` tests, for each element of a `mixtime` vector, whether
#' the time point is well-defined at the precision of `granule`.
#'
#' Discrete (integer) time cannot resolve a granule finer than its own chronon (a
#' `year()` has no determinate month), so those elements are `FALSE`. Continuous
#' (fractional) time tracks progress within its chronon and so resolves finer
#' granules exactly (0% through 2020 is 0% through January), giving `TRUE`.
#' Coarser-or-equal granules are always determinate. Missing (`NA`) and infinite
#' times give `NA`.
#'
#' @param x A time object (typically a `mixtime` vector).
#' @param granule The time granule whose precision to test, given as a granule
#'   generator (e.g. `cal_gregorian$month`) or a sized time unit (e.g.
#'   `cal_gregorian$month(1L)`).
#' @param ... Additional arguments for methods.
#'
#' @return A logical vector the same length as `x`.
#'
#' @examples
#' # Discrete: a year has no determinate month
#' time_is_determinate_at(year(2020L), cal_gregorian$month(1L))
#'
#' # Continuous: 0% through 2020 is 0% through January
#' time_is_determinate_at(year(2020), cal_gregorian$month(1L))
#'
#' # A coarser granule is always determinate
#' time_is_determinate_at(yearmonth(as.Date("2020-02-01")), cal_gregorian$year(1L))
#'
#' @export
time_is_determinate_at <- S7::new_generic(
  "time_is_determinate_at", "x",
  function(x, granule, ...) S7::S7_dispatch()
)
method(time_is_determinate_at, class_mixtime) <-
  vecvec::vecvec_apply_fn(time_is_determinate_at, logical())
method(time_is_determinate_at, class_any) <- function(x, granule, ...) {
  time_is_determinate_at(as_mixtime(x), granule, ...)
}
method(time_is_determinate_at, mt_time) <- function(x, granule, ...) {
  # Coerce a bare granule generator (e.g. cal_gregorian$month) into a unit.
  if (!S7_inherits(granule, mt_unit)) granule <- granule(1L)

  # Continuous (fractional) time tracks progress within its chronon and so
  # resolves any finer granule exactly; discrete (integer) time cannot subdivide
  # below its chronon, so a finer granule is temporally indeterminate.
  ok <- !is.integer(vec_data(x)) || chronon_nests_in(attr(x, "chronon"), granule)
  out <- rep(ok, length(x))
  # A missing or infinite time is not a determinate instant of any granule.
  out[is.na(x) | is.infinite(vec_data(x))] <- NA
  out
}

#' Test whether a granule is completed at a time point
#'
#' `time_is_complete_at()` tests, for each element of a `mixtime` vector, whether
#' the coarser `granule` that element falls into is fully observed *by the vector
#' as a whole* -- that is, whether every finer chronon making up that granule is
#' present somewhere in `x`.
#'
#' Unlike [time_is_determinate_at()], completeness is a collective property: an
#' element is `TRUE` only when the other elements needed to fill its granule are
#' also present. For example, in `year(1L)` the months of `2020 Jan : 2020 Oct`
#' are all `FALSE` (November and December are missing, so 2020 is incomplete),
#' whereas in `2020 Jan : 2021 Mar` the twelve months of 2020 are `TRUE` (they
#' complete 2020) while the three months of 2021 remain `FALSE`.
#'
#' A granule equal to `x`'s own chronon is completed by each point on its own
#' (`TRUE`). A granule finer than `x` cannot be completed by coarser points
#' (`FALSE`). Missing (`NA`) and infinite times give `NA`.
#'
#' Completeness is only defined within a single time granularity. Mixed-type
#' `mixtime` vectors (e.g. months alongside days) are not yet supported and
#' raise an error.
#'
#' @inheritParams time_is_determinate_at
#'
#' @return A logical vector the same length as `x`.
#'
#' @seealso [time_is_determinate_at()]
#'
#' @examples
#' # 2020 Jan : 2020 Oct does not complete the year -> all FALSE
#' time_is_complete_at(yearmonth(as.Date("2020-01-01")) + 0:9, cal_gregorian$year(1L))
#'
#' # 2020 Jan : 2021 Mar completes 2020 (TRUE) but not 2021 (FALSE)
#' time_is_complete_at(yearmonth(as.Date("2020-01-01")) + 0:14, cal_gregorian$year(1L))
#'
#' @export
time_is_complete_at <- S7::new_generic(
  "time_is_complete_at", "x",
  function(x, granule, ...) S7::S7_dispatch()
)
method(time_is_complete_at, class_mixtime) <- function(x, granule, ...) {
  # Completeness is only defined within a single granularity: mixed types (e.g.
  # months alongside days) would each fill different granules, which is not yet
  # implemented. `x@x` holds one part per distinct time type.
  if (length(x@x) > 1L) {
    cli::cli_abort(
      c(
        "{.fn time_is_complete_at} does not yet support mixed-granularity {.cls mixtime} vectors.",
        i = "To compute completeness within each granularity, include grouping over the `time_chronon()`."
      ),
      call = NULL
    )
  }
  if (length(x@x) == 0L) return(logical())
  time_is_complete_at(x@x[[1L]], granule, ...)
}
method(time_is_complete_at, class_any) <- function(x, granule, ...) {
  time_is_complete_at(as_mixtime(x), granule, ...)
}
method(time_is_complete_at, mt_time) <- function(x, granule, ...) {
  # Coerce a bare granule generator (e.g. cal_gregorian$year) into a unit.
  if (!S7_inherits(granule, mt_unit)) granule <- granule(1L)
  chronon <- attr(x, "chronon")
  granule <- granule_inherit_props(granule, chronon)

  xd <- vec_data(x)
  out <- logical(length(x))
  # A missing or infinite time completes no granule.
  special <- is.na(x) | is.infinite(xd)
  out[special] <- NA
  keep <- !special
  if (!any(keep)) return(out)

  # Completeness only applies when x is finer than (or equal to) the granule: a
  # granule finer than x's chronon cannot be filled by coarser points.
  if (!chronon_nests_in(chronon, granule)) {
    out[keep] <- FALSE
    return(out)
  }

  # Chronon position of each observation and the granule instance it falls into.
  pos <- floor(xd[keep])
  g <- chronon_convert_impl(pos, chronon, granule, discrete = TRUE)
  grp <- unique(g)

  # A granule instance is complete when the distinct x chronons observed within
  # it reach the number of x chronons that fill one granule (its cardinality).
  expected <- chronon_cardinality(chronon, granule, at = grp)
  observed <- vapply(
    split(pos, match(g, grp)),
    function(p) length(unique(p)),
    integer(1L)
  )
  complete <- observed >= expected

  out[keep] <- complete[match(g, grp)]
  out
}

# Calendar-field clamping
#
# Shared machinery for resolving invalid time points produced by temporal
# operations between granularities with irregular cardinality (e.g. Jan 31 + 1
# month). Only the whole-unit part of the offending remainder is clamped;
# fractional (e.g. time-of-day) and regular-cardinality remainders are left
# alone.

# --- longest path -----------------------------------------------------------

# Longest fine -> coarse chain of registered edges from `from` to `to`, so
# clamping visits every intermediate boundary (day -> month -> year) instead of
# skipping via a shorter direct edge (day -> year) with no "month" to clamp
# against. Falls back to S7_graph_dispatch()'s shortest path when no such DAG
# exists to walk.
chronon_longest_path <- function(from, to) {
  # Same time unit is always one hop - avoids collapsing to a node that can't
  # hold both `n`s.
  if (identical(S7_class_id(from), S7_class_id(to))) {
    return(list(from, to))
  }

  graph <- chronon_divmod_graph()
  int_from <- vec_match(S7_class_id(from), graph$chr_classes)
  int_to <- vec_match(S7_class_id(to), graph$chr_classes)

  int_path <- integer()
  if (!is.null(graph$edge_fine) && !is.na(int_from) && !is.na(int_to)) {
    int_path <- dag_longest_path(
      graph$edge_fine,
      graph$edge_coarse,
      int_from,
      int_to
    )
  }

  path <- if (length(int_path) == 0L) {
    # No monotone route - fall back to shortest path.
    S7_graph_dispatch(chronon_divmod_graph(), from, to)
  } else {
    graph$classes[int_path]
  }
  path[[1L]] <- from
  path[[length(path)]] <- to
  path[c(-1L, -length(path))] <- lapply(
    path[c(-1L, -length(path))],
    function(tu) {
      granule_inherit_props(granule_inherit_props(tu(1L), to), from)
    }
  )
  path
}

# --- component composition ----------------------------------------------------

# TRUE if from<->to has a fixed (context-independent) cardinality ratio - never
# needs clamping.
chronon_hop_is_regular <- function(from, to) {
  from_id <- S7_class_id(from)
  to_id <- S7_class_id(to)
  fixed_methods <- chronon_cardinality_fixed@methods

  identical(from_id, to_id) ||
    !is.null(fixed_methods[[from_id]][[to_id]]) ||
    !is.null(fixed_methods[[to_id]][[from_id]])
}

# TRUE when the component machinery below is worth using: `to` must be a coarser
# field containing `from` (chronon_nests_in()), and the conversion must not
# already be exact via chronon_cardinality_fixed() alone
# (!chronon_nests_in_fixed()). Otherwise callers should fall back to a plain
# chronon_convert_impl()/chronon_divmod() round trip.
chronon_needs_clamping <- function(from, to) {
  chronon_nests_in(from, to) && !chronon_nests_in_fixed(from, to)
}

# Forward walk: decompose `x` (path[[1]] units) into the (div, mod) pair at
# every hop of `path`. `div[[i]]` is `x` in path[[i]] units; `mod[[i]]` is the
# remainder from converting path[[i+1]] -> path[[i]].
chronon_decompose <- function(x, path) {
  n <- length(path)
  div <- vector("list", n)
  mod <- vector("list", n - 1L)
  div[[1L]] <- x
  for (i in seq(2L, length.out = n - 1L)) {
    result <- chronon_divmod_dispatch(path[[i - 1L]], path[[i]], div[[i - 1L]])
    div[[i]] <- result$div
    mod[[i - 1L]] <- result$mod
  }
  list(div = div, mod = mod)
}

# Backward walk: reconstruct a path[[1]]-unit value from `div`/`mod`
# (chronon_decompose()'s output, `div` possibly shifted). Clamps each irregular
# hop's whole-unit remainder to its paired coarser count's valid range.
# `clamped` flags where clamping changed a value (for warning on overflow).
chronon_recompose <- function(div, mod, path, clamp = TRUE) {
  n <- length(path)
  value <- div
  clamped <- rep(FALSE, length(value))
  for (i in seq(n, by = -1L, length.out = n - 1L)) {
    m <- mod[[i - 1L]]
    if (clamp && !chronon_hop_is_regular(path[[i - 1L]], path[[i]])) {
      card <- chronon_cardinality(path[[i - 1L]], path[[i]], at = value)
      m_int <- floor(m)
      m_clamped <- pmin(card - 1, m_int)
      clamped <- clamped | (m_clamped != m_int)
      m <- m_clamped + (m - m_int)
    }
    value <- chronon_divmod_dispatch(path[[i]], path[[i - 1L]], value)$div + m
  }
  list(value = value, clamped = clamped)
}

#' Solar time unit classes
#' 
#' Time unit constructors for the transit-based solar time system, where the
#' boundary of each day is at apparent solar midnight. Solar events define the
#' `ampm` (midnight and noon) and `illumination` (dawn, sunrise, sunset, dusk)
#' units. `cal_time_solar` is an alias for `cal_time_solar_transit`.
#'
#' @format A location-based calendar containing transit-based solar time units.
#' 
#' @details
#' The following time units are available in the solar calendar systems.
#' 
#' - `day()`: Day unit
#' - `ampm()`: Half-day units (AM = before solar noon, PM = after solar noon)
#' - `hour()`: Hour units within the solar day
#' - `minute()`: Minute units within the solar hour
#' - `second()`: Second units within the solar minute
#' - `degree()`: Solar angle units within the day
#' - `arcminute()`: Arcminute units within the solar degree
#' - `arcsecond()`: Arcsecond units within the solar arcminute
#' - `illumination()`: Illumination phases (night, astronomical/nautical/civil dawn, day, civil/nautical/astronomical dusk)
#' 
#' ## AM/PM half-days
#' 
#' The `ampm` unit divides each solar day into two halves between solar noon and
#' solar midnight:
#' 
#' | Half | Period                  | Description                              |
#' |------|-------------------------|------------------------------------------|
#' | AM   | Solar midnight to noon  | Morning half; before solar transit       |
#' | PM   | Solar noon to midnight  | Afternoon half; after solar transit      |
#' 
#' ## Solar illumination phases
#' 
#' Phases describe the illumination state of the sky and correspond to standard
#' twilight definitions used in astronomy and navigation. Each phase is bounded
#' by a pair of solar altitude thresholds:
#' 
#' | Phase              | Solar altitude range  | Description                                                 |
#' |--------------------|-----------------------|-------------------------------------------------------------|
#' | Night              | < -18°                | Sky fully dark; from last dusk to first dawn (spans noon)   |
#' | Astronomical dawn  | -18° to -12°          | Astronomical twilight before sunrise; faint objects obscured|
#' | Nautical dawn      | -12° to -6°           | Nautical twilight before sunrise; horizon visible at sea    |
#' | Civil dawn         | -6° to -0.833°        | Civil twilight before sunrise; sky brightening in the east  |
#' | Day                | > -0.833°             | Sun above the horizon; spans solar noon                     |
#' | Civil dusk         | -0.833° to -6°        | Civil twilight after sunset; sky fading in the west         |
#' | Nautical dusk      | -6° to -12°           | Nautical twilight after sunset; horizon visible at sea      |
#' | Astronomical dusk  | -12° to -18°          | Astronomical twilight after sunset; faint objects obscured  |
#' 
#' The -0.833° threshold for sunrise and sunset accounts for the mean angular
#' radius of the solar disc (0.267°) plus the standard atmospheric refraction
#' at the horizon (0.566°). Noon and midnight are derived from the equation of
#' time rather than a fixed altitude. Locations that experience polar day or
#' polar night (civil days where sunrise does not occur) are not currently 
#' supported, it is recommended to use an alternative reference location.
#' 
#' @return An S3 list of class `c("cal_time_solar", "mt_calendar")` containing
#'   the named time unit classes of the solar transit calendar. Each unit is
#'   accessible via `$` notation and calling it with a step size and location
#'   produces a time granule (e.g., 1 solar day granule as
#'   `cal_time_solar$day(1L, lat = 0, lon = 0)`). Because solar day boundaries
#'   depend on the observer's position, each unit constructor requires `lat`
#'   and `lon` arguments.
#'
#' @seealso [`cal_time_civil`], [`cal_time_lunar`]
#' 
#' @examples
#' # Find the current solar time in Melbourne
#' datetime(Sys.time(), calendar = cal_time_solar, lat = -37.8136, lon = 144.9631)
#' 
#' @name calendar_time_solar
#' @export
cal_time_solar_transit <- new_calendar(
  day = S7::new_class("tu_solar_day", parent = mt_loc_unit),
  ampm = S7::new_class("tu_solar_ampm", parent = mt_loc_unit),
  hour = S7::new_class("tu_solar_hour", parent = mt_loc_unit),
  minute = S7::new_class("tu_solar_minute", parent = mt_loc_unit),
  second = S7::new_class("tu_solar_second", parent = mt_loc_unit),
  degree = S7::new_class("tu_solar_degree", parent = mt_loc_unit),
  arcminute = S7::new_class("tu_solar_arcminute", parent = mt_loc_unit),
  arcsecond = S7::new_class("tu_solar_arcsecond", parent = mt_loc_unit),
  illumination = S7::new_class("tu_solar_lum", parent = mt_loc_unit)
)

#' @export
#' @rdname calendar_time_solar
cal_time_solar <- cal_time_solar_transit


# Time unit labels
method(time_unit_full, cal_time_solar$day) <- function(x) "day{?/s}"
method(time_unit_abbr, cal_time_solar$day) <- function(x) "D"
method(time_unit_full, cal_time_solar$ampm) <- function(x) "halfday{?/s}"
method(time_unit_abbr, cal_time_solar$ampm) <- function(x) "HD"
method(time_unit_full, cal_time_solar$hour) <- function(x) "hour{?/s}"
method(time_unit_abbr, cal_time_solar$hour) <- function(x) "h"
method(time_unit_full, cal_time_solar$minute) <- function(x) "minute{?/s}"
method(time_unit_abbr, cal_time_solar$minute) <- function(x) "m"
method(time_unit_full, cal_time_solar$second) <- function(x) "second{?/s}"
method(time_unit_abbr, cal_time_solar$second) <- function(x) "s"
method(time_unit_full, cal_time_solar$degree) <- function(x) "degree{?/s}"
method(time_unit_abbr, cal_time_solar$degree) <- function(x) "\u00b0"
method(time_unit_full, cal_time_solar$arcminute) <- function(x) "arcminute{?/s}"
method(time_unit_abbr, cal_time_solar$arcminute) <- function(x) "'"
method(time_unit_full, cal_time_solar$arcsecond) <- function(x) "arcsecond{?/s}"
method(time_unit_abbr, cal_time_solar$arcsecond) <- function(x) "\""
method(time_unit_full, cal_time_solar$illumination) <- function(x) "illumination phase{?/s}"
method(time_unit_abbr, cal_time_solar$illumination) <- function(x) "I"

# Default formats
method(chronon_format_linear, list(cal_time_solar$day, class_any)) <- function(x, cal) "{lin(day)}"
method(chronon_format_linear, list(cal_time_solar$ampm, class_any)) <- function(x, cal) "{lin(day)} {cyc(ampm,day)}"
method(chronon_format_linear, list(cal_time_solar$hour, class_any)) <- function(x, cal) paste(chronon_format_linear(cal$day(1L), cal), "{cyc(hour,day)}h")
method(chronon_format_linear, list(cal_time_solar$minute, class_any)) <- function(x, cal) paste(chronon_format_linear(cal$day(1L), cal), "{cyc(hour,day)}:{cyc(minute,hour)}")
method(chronon_format_linear, list(cal_time_solar$second, class_any)) <- function(x, cal) paste(chronon_format_linear(cal$day(1L), cal), "{cyc(hour,day)}:{cyc(minute,hour)}:{cyc(second,minute)}")
method(chronon_format_linear, list(cal_time_solar$degree, class_any)) <- function(x, cal) paste(chronon_format_linear(cal$day(1L), cal), "{cyc(degree,day)}\u00b0")
method(chronon_format_linear, list(cal_time_solar$arcminute, class_any)) <- function(x, cal) paste(chronon_format_linear(cal$day(1L), cal), "{cyc(degree,day)}\u00b0{cyc(arcminute,degree)}'")
method(chronon_format_linear, list(cal_time_solar$arcsecond, class_any)) <- function(x, cal) paste(chronon_format_linear(cal$day(1L), cal), "{cyc(degree,day)}\u00b0{cyc(arcminute,degree)}'{cyc(arcsecond,arcminute)}\"")
method(chronon_format_linear, list(cal_time_solar$illumination, class_any)) <- function(x, cal) "{lin(day)} {lin(illumination)}"

# Candidate parsing formats
method(chronon_parse_linear, list(cal_time_solar$hour, class_any)) <- function(x, cal) {
  parse_format_time_of_day(
    cal,
    c(
      " {cyc(hour,day)}h",
      " {cyc(hour,ampm)} {cyc(ampm,day,label=TRUE)}"
    )
  )
}
method(chronon_parse_linear, list(cal_time_solar$minute, class_any)) <- function(x, cal) {
  parse_format_time_of_day(
    cal,
    c(
      " {cyc(hour,day)}:{cyc(minute,hour)}",
      " {cyc(hour,ampm)}:{cyc(minute,hour)} {cyc(ampm,day,label=TRUE)}"
    )
  )
}
method(chronon_parse_linear, list(cal_time_solar$second, class_any)) <- function(x, cal) {
  parse_format_time_of_day(
    cal,
    c(
      " {cyc(hour,day)}:{cyc(minute,hour)}:{cyc(second,minute)}",
      " {cyc(hour,ampm)}:{cyc(minute,hour)}:{cyc(second,minute)} {cyc(ampm,day,label=TRUE)}"
    )
  )
}
method(chronon_parse_linear, list(cal_time_solar$degree, class_any)) <- function(x, cal) {
  parse_format_time_of_day(cal, " {cyc(degree,day)}\u00b0")
}
method(chronon_parse_linear, list(cal_time_solar$arcminute, class_any)) <- function(x, cal) {
  parse_format_time_of_day(cal, " {cyc(degree,day)}\u00b0{cyc(arcminute,degree)}'")
}
method(chronon_parse_linear, list(cal_time_solar$arcsecond, class_any)) <- function(x, cal) {
  parse_format_time_of_day(cal, " {cyc(degree,day)}\u00b0{cyc(arcminute,degree)}'{cyc(arcsecond,arcminute)}\"")
}

# Solar time
method(chronon_cardinality_fixed, list(cal_time_solar$ampm, cal_time_solar$day)) <- function(x, y) {
  2L
}
method(chronon_cardinality_fixed, list(cal_time_solar$hour, cal_time_solar$day)) <- function(x, y) {
  24L
}
method(chronon_cardinality_fixed, list(cal_time_solar$hour, cal_time_solar$ampm)) <- function(x, y) {
  12L
}
method(chronon_cardinality_fixed, list(cal_time_solar$minute, cal_time_solar$hour)) <- function(x, y) {
  60L
}
method(chronon_cardinality_fixed, list(cal_time_solar$second, cal_time_solar$minute)) <- function(x, y) {
  60L
}

# Solar arc time
method(chronon_cardinality_fixed, list(cal_time_solar$degree, cal_time_solar$day)) <- function(x, y) {
  360L
}
method(chronon_cardinality_fixed, list(cal_time_solar$arcminute, cal_time_solar$degree)) <- function(x, y) {
  60L
}
method(chronon_cardinality_fixed, list(cal_time_solar$arcsecond, cal_time_solar$arcminute)) <- function(x, y) {
  60L
}

# Solar illumination phases
method(chronon_cardinality, list(cal_time_solar$second, cal_time_solar$illumination)) <- function(x, y, at = NULL) {
  (approx_solar_seconds_from_solar_phase((at + 1L) * y@n, y@lat, y@lon, y@alt) -
     approx_solar_seconds_from_solar_phase(at * y@n, y@lat, y@lon, y@alt)) / x@n
}

# The number of UTC seconds in a solar second
method(chronon_cardinality, list(cal_time_civil$second, cal_time_solar$second)) <- function(x, y, at = NULL) {
  at_ss <- approx_solar_seconds_from_utc(at, y@lat, y@lon, y@alt)
  (approx_utc_from_solar_seconds((at_ss + 1L) * y@n, y@lat, y@lon, y@alt) -
     approx_utc_from_solar_seconds(at_ss * y@n, y@lat, y@lon, y@alt)) / x@n
}

# Convert UTC seconds → continuous solar second count
method(chronon_divmod, list(cal_time_civil$second, cal_time_solar$second)) <- function(from, to, x) {
  list(
    div = approx_solar_seconds_from_utc(as.double(x) * from@n, to@lat, to@lon, to@alt) / to@n,
    mod = 0
  )
}

# Convert solar second count → UTC seconds
method(chronon_divmod, list(cal_time_solar$second, cal_time_civil$second)) <- function(from, to, x) {
  list(
    div = approx_utc_from_solar_seconds(as.double(x) * from@n, from@lat, from@lon, from@alt) / to@n,
    mod = 0
  )
}

# Convert solar seconds → continuous solar phase count.
# A single C++ call converts solar seconds directly to phases without a UTC round-trip.
method(chronon_divmod, list(cal_time_solar$second, cal_time_solar$illumination)) <- function(from, to, x) {
  list(
    div = approx_solar_phase_from_solar_seconds(as.double(x) * from@n, to@lat, to@lon, to@alt) / to@n,
    mod = 0
  )
}
# Convert solar phase count → solar seconds.
# Integer phase counts return the phase start boundary; fractional counts interpolate.
method(chronon_divmod, list(cal_time_solar$illumination, cal_time_solar$second)) <- function(from, to, x) {
  list(
    div = approx_solar_seconds_from_solar_phase(as.double(x) * from@n, from@lat, from@lon, from@alt) / to@n,
    mod = 0
  )
}

## Time labels
### Linear labels for solar days (re-use Gregorian dates)
S7::method(linear_labels_format, cal_time_solar$day) <- function(granule, i, ...) {
  return(format(.Date(i * granule@n)))
}

### Linear labels for solar time granules
# (not cyclical over solar days due to midnight day boundary splitting nights)
solar_illumination_phases <- c(
  "astronomical dawn", "nautical dawn", "civil dawn", "day",
  "civil dusk", "nautical dusk", "astronomical dusk", "night"
)
method(linear_labels_format, cal_time_solar$illumination) <- function(granule, i, ...) {
  solar_illumination_phases[floor((i * granule@n)%%8) + 1L]
}

# method(cyclical_labels_format, list(cal_time_solar$illumination, cal_time_solar$day)) <- function(granule, cycle, i) {
#   solar_illumination_phases[i + 1L]
# }

method(cyclical_labels, list(cal_time_solar$ampm, cal_time_solar$day)) <- label_scheme(
  start = 0L,
  vocab = vocab_table(`en-GB` = list(wide = c("AM", "PM"), abbreviated = c("AM", "PM"))),
  transform = list(
    encode = function(i, at = NULL) i + 1L,
    decode = function(d, at = NULL) d - 1L
  )
)

method(cyclical_labels_format, list(cal_time_solar$hour, S7::class_any)) <- function(granule, cycle, i, at = NULL, ...) {
  if (S7_inherits(cycle, cal_time_solar$ampm)) {
    # 12 hours count with 12,1,2,...,11
    sprintf("%02d", (i-1L)%%12L + 1L)
  } else {
    # 24 hours count with 0-indexing
    sprintf("%02d", i)
  }
}
method(cyclical_labels_parse, list(cal_time_solar$hour, S7::class_any)) <- function(granule, cycle, ...) {
  list(
    pattern = "\\d+",
    decode = function(text, at = NULL) {
      d <- as.integer(text)
      if (S7_inherits(cycle, cal_time_solar$ampm)) d %% 12L else d
    }
  )
}

# Minutes/seconds count with 0-indexing -- plain numeric (tier 1).
method(cyclical_labels, list(cal_time_solar$minute, S7::class_any)) <- label_scheme(start = 0L, width = 2L)
method(cyclical_labels, list(cal_time_solar$second, S7::class_any)) <- label_scheme(start = 0L, width = 2L)

method(cyclical_labels_format, list(cal_time_solar$degree, S7::class_any)) <- function(granule, cycle, i, at = NULL, ...) {
  sprintf("%03d", (i - 180L*chronon_cardinality(cycle, cal_time_solar$day(1L))) %% 360L)
}
method(cyclical_labels_parse, list(cal_time_solar$degree, S7::class_any)) <- function(granule, cycle, ...) {
  list(
    pattern = "\\d+",
    decode = function(text, at = NULL) {
      shift <- 180L * chronon_cardinality(cycle, cal_time_solar$day(1L))
      (as.integer(text) + shift) %% 360L
    }
  )
}

# Arcminutes/arcseconds count with 0-indexing -- plain numeric (tier 1).
method(cyclical_labels, list(cal_time_solar$arcminute, S7::class_any)) <- label_scheme(start = 0L, width = 2L)
method(cyclical_labels, list(cal_time_solar$arcsecond, S7::class_any)) <- label_scheme(start = 0L, width = 2L)

#' Linear time function factory
#' 
#' `new_linear_time_fn()` creates a linear time function for a specified
#' chronon. A chronon is the smallest indivisible time granule (e.g., days, 
#' hours).
#' 
#' @param chronon A bare call for a time granule object representing the chronon
#'   (e.g., `day(1)`)
#' @param default_calendar A default calendar used to find the time units for 
#'   conversion if they don't exist in the calendar of the input data (e.g., 
#'   `cal_isoweek` for week chronons to work with gregorian calendar inputs).
#' 
#' @return A function used to create linear time points with a specific chronon.
#' 
#' @examples
#' 
#' # Linear time with 1 month granules as the chronon
#' ym <- new_linear_time_fn(month(1L))
#' ym(Sys.Date())
#' 
#' # Linear time with 1 day granules as the chronon
#' yd <- new_linear_time_fn(day(1L))
#' yd(Sys.Date())
#' 
#' # Linear time with 1 week granules as the chronon, using the ISO week calendar
#' yw <- new_linear_time_fn(week(1L), default_calendar = cal_isoweek)
#' yw(Sys.Date())
#' 
#' # Linear time with 1 hour granules as the chronon
#' ymd_h <- new_linear_time_fn(hour(1L))
#' ymd_h(Sys.time())
#' 
#' @export
new_linear_time_fn <- function(chronon, default_calendar = cal_gregorian) {
  chronon <- rlang::new_quosure(
    enexpr(chronon), 
    env = rlang::as_data_mask(default_calendar)
  )

  function(
    data, discrete = TRUE, calendar = time_calendar(data), ...
  ) {
    # Add tz / loc to chronon
    chronon <- quo_add_dots(chronon, ...)

    linear_time(
      data, chronon = !!chronon, discrete = discrete, calendar = calendar
    )
  }
}

#' Linear time points
#' 
#' `linear_time()` creates a vector of linear time points with a specified
#' chronon (smallest time granule). This function is useful for creating custom
#' time representations that aren't covered by the convenience functions like
#' [yearmonth()] or [yearweek()].
#' 
#' @param data Input data to convert to linear time. Can be:
#'   - Numeric values (interpreted as chronons since Unix epoch)
#'   - Character strings (parsed as dates/times)
#'   - Date or POSIXct objects
#'   - Other time objects
#' @param chronon A time granule expression representing the chronon (smallest 
#'   indivisible time granule), evaluated in the context of `calendar`. Use 
#'   unquoted expressions like `month(1L)` or `hour(1L)`. Chronons from a
#'   specific calendar can also be used (e.g. `cal_isoweek$week(1L)`).
#'   Defaults to the time chronon of the input `data` (`time_chronon(data)`).
#' @param discrete Logical. If `TRUE` (default), returns integer chronons since
#'   Unix epoch (discrete time model). If `FALSE`, returns fractional chronons 
#'   allowing representation of fractional time granules (continuous time model).
#' @param calendar Calendar system used to evaluate `chronon` and `granules`.
#'   Defaults to `time_calendar(data)` for existing time objects. Common options
#'   include [cal_gregorian] and [cal_isoweek].
#' 
#' @return A `mixtime` time vector containing an `mt_linear` vector.
#' 
#' @seealso 
#' - [new_linear_time_fn()] for creating reusable linear time functions
#' - [yearmonth()], [yearquarter()], [year()] for Gregorian time representations
#' - [yearweek()] for ISO 8601 week-based time
#' - [cal_gregorian], [cal_isoweek] for calendar systems
#' 
#' @examples
#' # Hourly time
#' linear_time(
#'   Sys.time(),
#'   chronon = hour(1L)
#' )
#' 
#' # Monthly time
#' linear_time(
#'   Sys.Date(),
#'   chronon = month(1L)
#' )
#' 
#' # Discrete vs continuous time
#' linear_time(Sys.time(), chronon = day(1L), discrete = TRUE)
#' linear_time(Sys.time(), chronon = day(1L), discrete = FALSE)
#' 
#' # ISO week calendar with week-day structure
#' linear_time(
#'   Sys.Date(),
#'   chronon = day(1L),
#'   calendar = cal_isoweek
#' )
#' 
#' @export
linear_time <- function(
  data, chronon = time_chronon(data), discrete = TRUE, 
  calendar = time_calendar(data)
) {
  # Evaluate chronon and granules with a calendar mask
  quo_chronon <- enquo(chronon)
  tryCatch({
    chronon <- eval_tidy(quo_chronon, data = calendar, env = emptyenv())
  }, error = function(e) {
    # Special hint for common error of 'week' unit not found in Gregorian calendar
    if (e$message == "could not find function \"week\"") {
      e$message <- c(
        e$message,
        "i" = "This error often occurs when trying to use a 'week' chronon without the ISO week calendar.\n",
        ">" = "Try specifying the calendar explicitly, e.g. `calendar = cal_isoweek`."
      )
    }
    cli::cli_abort(e$message, call = NULL)
  })

  mixtime(data, chronon = chronon, cycle = NULL, discrete = discrete)
}

#' Linear time helper functions
#' 
#' Convenience functions for creating common linear time representations. These
#' functions work with different calendar systems and adapt based on the input
#' data's calendar.
#' 
#' @param data A vector of time points (e.g. [base::Date], [base::POSIXt])
#' @param discrete If `TRUE`, the number of chronons since Unix epoch that
#' `data` falls into is returned as an integer. If `FALSE`, a fractional number
#'  of chronons is returned (analagous to time using a continuous time model).
#' @param calendar A calendar used to evaluate the time units. Defaults to the
#'   calendar of the input data. Common options include [cal_gregorian] and
#'   [cal_isoweek].
#' @param ... Additional arguments for [linear_time()], such as `tz` for timezones.
#' 
#' @details
#' These functions create linear time representations with different chronons
#' and granules:
#' 
#' - `year()`: Represents time in whole years. The chronon is one year.
#' - `yearquarter()`: Represents time in quarters, grouped by year. The chronon
#'   is one quarter, with years as the granule.
#' - `yearmonth()`: Represents time in months, grouped by year. The chronon is
#'   one month, with years as the granule.
#' - `yearweek()`: Represents time in weeks, grouped by year. The chronon is
#'   one week, with years as the granule. Defaults to ISO 8601 week calendar.
#' 
#' @section Calendar flexibility:
#' These functions adapt to the calendar system of the input data. For example:
#'
#' - `year(date("2025-12-29"))` returns a Gregorian year
#' - `year(yearweek(date("2025-12-29")))` returns an ISO week-based year
#'
#' You can also explicitly specify a calendar using the `calendar` argument:
#'
#' \preformatted{
#' year(date("2025-12-29"), calendar = cal_isoweek)
#' }
#' 
#' @section Custom linear time representations:
#' For more complex time structures, use [linear_time()] or [new_linear_time_fn()]
#' to create custom representations with any combination of chronons and granules.
#' 
#' @return A `mixtime` time vector containing an `mt_linear` vector with chronons matching the function used.
#' 
#' @seealso 
#' - [linear_time()] for creating custom linear time representations
#' - [new_linear_time_fn()] for creating reusable linear time functions
#' - [cal_gregorian], [cal_isoweek] for calendar systems
#' 
#' @examples
#' 
#' # Gregorian year
#' year(Sys.Date())
#' year(Sys.Date(), discrete = FALSE)
#' 
#' # ISO week-based year
#' year(yearweek(Sys.Date()))
#' 
#' # Year-quarter
#' yearquarter(Sys.Date())
#' yearquarter(Sys.Date(), discrete = FALSE)
#' 
#' # Year-month
#' yearmonth(Sys.Date())
#' yearmonth(Sys.Date(), discrete = FALSE)
#' 
#' # Year-week (ISO 8601)
#' yearweek(Sys.Date())
#' yearweek(0:52)
#' 
#' @name linear_time_helpers
#' @export
year <- new_linear_time_fn(
  chronon = year(1L)
)

#' @rdname linear_time_helpers
#' @export
yearquarter <- new_linear_time_fn(
  chronon = quarter(1L)
)

#' @rdname linear_time_helpers
#' @export
yearmonth <- new_linear_time_fn(
  chronon = month(1L)
)

#' @rdname linear_time_helpers
#' @export
yearweek <- new_linear_time_fn(
  chronon = week(1L),
  default_calendar = cal_isoweek
)

#' @rdname linear_time_helpers
#' @export
date <- new_linear_time_fn(
  chronon = day(1L)
)

#' @rdname linear_time_helpers
#' @export
datetime <- new_linear_time_fn(
  chronon = second(1L)
)

# ----------------------------------------------------------------------------
# Vector size (length(), vctrs::vec_size())
# ----------------------------------------------------------------------------

test_that("Size of mixtime vectors", {
  expect_equal(length(yearmonth(1:10L)), 10L)
  expect_equal(length(yearmonth(1L)),   1L)
  expect_equal(length(yearmonth(integer())), 0L)

  expect_equal(length(month_of_year(1:5L)), 5L)
  expect_equal(length(month_of_year(1L)), 1L)
  expect_equal(length(month_of_year(integer())), 0L)

  expect_equal(vctrs::vec_size(yearmonth(1:7L)), 7L)
  expect_equal(vctrs::vec_size(yearweek(1:4L)), 4L)
})

# ----------------------------------------------------------------------------
# Formatting (format())
# ----------------------------------------------------------------------------

test_that("Formatting mixtime vectors", {
  x <- yearmonth(648:650L)
  expect_equal(format(x), c("2024 Jan", "2024 Feb", "2024 Mar"))

  x <- yearweek(2818:2820L)
  expect_equal(format(x), c("2024 W01", "2024 W02", "2024 W03"))
})

# ---------------------------------------------------------------------------
# Missing values (is.na())
# ---------------------------------------------------------------------------

test_that("Missing values in mixtime vectors", {
  x <- yearmonth(1:4L)
  expect_equal(is.na(x), rep(FALSE, 4L))

  x <- yearmonth(c(1L, NA, 3L))
  expect_equal(is.na(x), c(FALSE, TRUE, FALSE))
})

# ---------------------------------------------------------------------------
# Arithmetic (+, -)
# ---------------------------------------------------------------------------

test_that("Arithmetic - addition and subtraction of integers", {
  x <- yearmonth(648:650L)
  result <- x + 2L
  expect_s7_class(result, class_mixtime)
  expect_equal(format(result), c("2024 Mar", "2024 Apr", "2024 May"))

  x <- yearweek(2818:2820L)
  result <- x + 2L
  expect_s7_class(result, class_mixtime)
  expect_equal(format(result), c("2024 W03", "2024 W04", "2024 W05"))
})

test_that("Arithmetic - subtraction of mixtime vectors", {
  x <- yearmonth(648:652L)
  result <- x - x
  expect_equal(as.integer(result), rep(0L, 5L))
})

# ---------------------------------------------------------------------------
#  Comparison (<, <=, ==, >=, >)
# ---------------------------------------------------------------------------

test_that("Comparison - inequalities", {
  x <- yearmonth(648:650L)
  expect_equal(x < yearmonth(648:650L), rep(FALSE, 3L))  # equal, not less

  x <- yearmonth(648:652L)
  y <- x + 1L
  expect_equal(y > x, rep(TRUE, 5L))

  y <- x - 1L
  expect_equal(y < x, rep(TRUE, 5L))
})

test_that("Comparison - equality", {
  x <- yearmonth(648:650L)
  expect_equal(x == x, rep(TRUE, 3L))
})

test_that("Comparison - mixed granularity discrete intervals (#95)", {
  # {1980} = [1980-01-01, 1980-12-31], {1980 Dec} = [1980-12-01, 1980-12-31]
  y1980 <- year(1980L)
  dec1980 <- yearmonth(as.integer((1980 - 1970) * 12 + 11))

  expect_false(y1980 == dec1980)
  expect_false(y1980 < dec1980)
  expect_false(y1980 > dec1980)
  expect_true(y1980 <= dec1980)   # same right endpoint (end of Dec 1980)
  expect_false(y1980 >= dec1980)  # year starts earlier than December

  # A month entirely before another year is strictly less than it
  expect_true(yearmonth(as.integer((1979 - 1970) * 12)) < y1980)
  # A month entirely after another year is strictly greater than it
  expect_true(yearmonth(as.integer((1981 - 1970) * 12)) > y1980)
})

test_that("Comparison - identical granules compare directly", {
  x <- yearmonth(648:650L)
  y <- yearmonth(649:651L)
  expect_equal(x < y, rep(TRUE, 3L))
  expect_equal(x == yearmonth(648:650L), rep(TRUE, 3L))
})

test_that("Comparison - continuous linear time compares as points", {
  x <- year(as.Date("2020-06-01"), discrete = FALSE)
  y <- yearmonth(as.Date("2020-01-01"), discrete = FALSE)
  expect_true(x > y)
  expect_true(y < x)
})

test_that("Comparison - continuous point within a discrete span", {
  point <- linear_time(as.Date("1980-06-15"), chronon = month(1L), discrete = FALSE)
  y1980 <- year(1980L)
  expect_true(point <= y1980)
  expect_true(point >= y1980)
  expect_false(point < y1980)
  expect_false(point > y1980)
})


# ---------------------------------------------------------------------------
#  Indexing and subsetting ([, [[)
# ---------------------------------------------------------------------------


test_that("Subsetting - single-element", {
  x <- yearmonth(648:650L)
  expect_s7_class(x[1L], class_mixtime)
  expect_equal(length(x[1L]), 1L)
  expect_equal(format(x[1L]), "2024 Jan")
})

test_that("Subsetting - multi-element", {
  x <- yearmonth(648:652L)
  sub <- x[c(3L, 1L, 2L)]
  expect_s7_class(sub, class_mixtime)
  expect_equal(length(sub), 3L)
  expect_equal(format(sub), c("2024 Mar", "2024 Jan", "2024 Feb"))
})

test_that("Subsetting - negative indexing", {
  x <- yearmonth(648:650L)
  result <- x[-1L]
  expect_equal(length(result), 2L)
  expect_equal(format(result), c("2024 Feb", "2024 Mar"))
})

test_that("Subsetting - logical indexing", {
  x <- yearmonth(648:651L)
  result <- x[c(TRUE, FALSE, TRUE, FALSE)]
  expect_equal(length(result), 2L)
  expect_equal(format(result), c("2024 Jan", "2024 Mar"))
})

# ---------------------------------------------------------------------------
# Uniqueness (unique())
# ---------------------------------------------------------------------------

test_that("Uniqueness - unique() removes duplicates", {
  x <- yearmonth(648:652L)
  # rep() itself will fail until subsetting is fixed, so construct via + 0L
  duped <- c(x, x)
  result <- unique(duped)
  expect_s7_class(result, class_mixtime)
  expect_equal(length(result), 5L)

  x <- yearmonth(650:652L)
  expect_equal(length(unique(x)), 3L)
})

# ---------------------------------------------------------------------------
# Ordering (rev(), sort())
# ---------------------------------------------------------------------------

test_that("rev() reverses mixtime yearmonth", {
  x <- yearmonth(648:652L)
  result <- rev(x)
  expect_s7_class(result, class_mixtime)
  expect_equal(format(result), rev(format(x)))
})

test_that("sort() orders mixtime vectors", {
  x <- yearmonth(648:652L)
  x_rev <- rev(x)
  # Construct reversed by arithmetic rather than subsetting
  expect_equal(x, sort(x_rev))
  
  expect_equal(x_rev, sort(x, decreasing = TRUE))
})

# ---------------------------------------------------------------------------
# Combinations (c())
# ---------------------------------------------------------------------------

test_that("Combinations - c() combines same type vectors", {
  x <- yearmonth(1:2L)
  y <- yearmonth(3:5L)
  result <- c(x, y)
  expect_s7_class(result, class_mixtime)
  expect_equal(length(result), 5L)
})

test_that("Combinations - c() combines mixed type vectors", {
  x <- yearmonth(1:2L)
  y <- yearweek(3:5L)
  result <- c(x, y)
  expect_s7_class(result, class_mixtime)
  expect_equal(length(result), 5L)
})

test_that("Combinations - c() with NA", {
  x <- yearmonth(1:2L)
  result <- c(x, NA)
  expect_s7_class(result, class_mixtime)
  expect_equal(length(result), 3L)
  expect_equal(is.na(result), c(FALSE, FALSE, TRUE))
})

test_that("Cardinality of the Symmetric454 calendar: cal_sym454", {
  # 1970 is a leap year, so Jan has 4 weeks, Feb has 5 weeks, Mar has 4 weeks, ..., Dec has 5 weeks
  expect_equal(
    with(cal_sym454, chronon_cardinality(week(1L), month(1L), at = 0:11L)),
    c(4L, 5L, 4L, 4L, 5L, 4L, 4L, 5L, 4L, 4L, 5L, 5L)
  )
  # 1971 is not a leap year; its months are at positions 12..23 from the epoch
  expect_equal(
    with(cal_sym454, chronon_cardinality(week(1L), month(1L), at = 12:23L)),
    c(4L, 5L, 4L, 4L, 5L, 4L, 4L, 5L, 4L, 4L, 5L, 4L)
  )

  # 1970 is a leap year and should have 53 weeks
  expect_equal(
    with(cal_sym454, chronon_cardinality(week(1L), year(1L), at = 0L)),
    53L
  )
  # 1971 is not a leap year and should have 52 weeks
  expect_equal(
    with(cal_sym454, chronon_cardinality(week(1L), year(1L), at = 12L)),
    52L
  )

  # Each quarter (3 months) should have 13 weeks except the last having 14 weeks in a leap year
  expect_equal(
    with(cal_sym454, chronon_cardinality(week(1L), month(3L), at = 0:3L)),
    c(13L, 13L, 13L, 14L)
  )

  # Each quarter (3 months) should have 13 weeks in a non-leap year
  expect_equal(
    with(cal_sym454, chronon_cardinality(week(1L), month(3L), at = 4:7L)),
    c(13L, 13L, 13L, 13L)
  )
})

test_that("Divmod of the Symmetric454 calendar is invertible via Date round-trip", {
  # Converting Sym454 year integers -> Date -> Sym454 year (continuous) should
  # recover the original values exactly, since each integer maps to the first
  # day of that Sym454 year.
  years <- 1970:2000
  sym454_years <- linear_time(years, chronon = cal_sym454$year(1L))
  dates <- as.Date(sym454_years)
  recovered <- linear_time(dates, chronon = cal_sym454$year(1L), discrete = FALSE)
  expect_equal(as.double(recovered), as.double(sym454_years))
})


test_that("Divmod of the Symmetric454 calendar: cal_sym454", {
  # Regular 4-5-4 pattern in 1970 (leap year)
  expect_equal(
    with(cal_sym454, chronon_divmod(week(1L), month(1L), c(0L, 4L, 9L, 13L, 51L, 52L))),
    list(div = c(0L, 1L, 2L, 3L, 11L, 11L), mod = c(0L, 0L, 0L, 0L, 3L, 4L))
  )
  # Regular 4-5-4 pattern in 1971 (non-leap year)
  expect_equal(
    with(cal_sym454, chronon_divmod(week(1L), month(1L), c(53L, 55L, 62L, 66L, 104L))),
    list(div = c(12L, 12L, 14L, 15L, 23L), mod = c(0L, 2L, 0L, 0L, 3L))
  )

  # 2-week units (fortnights)
  expect_equal(
    with(cal_sym454, chronon_divmod(week(2L), month(1L), c(0L, 4L, 5L, 9L, 13L))),
    list(div = c(0L, 1L, 2L, 4L, 6L), mod = c(0L, 2L, 0L, 0L, 0L))
  )

  # Multi-month units (quarters)
  expect_equal(
    with(cal_sym454, chronon_divmod(week(1L), month(3L), c(0L, 13L, 26L, 39L))),
    list(div = c(0L, 1L, 2L, 3L), mod = c(0L, 0L, 0L, 0L))
  )

  # Fortnight units with multi-month units (2-week quarters)
  expect_equal(
    with(cal_sym454, chronon_divmod(week(2L), month(3L), c(0L, 3L, 13L, 26L, 39L))),
    list(div = c(0L, 0L, 2L, 3L, 5L), mod = c(0L, 3L, 0L, 6L, 6L))
  )
})

test_that("Divmod of the Symmetric454 calendar (month -> week) agrees with cardinality, including leap-year December", {
  # month -> week is the forward direction of the tests above. Its result
  # (the week-index a month starts on) must equal the cumulative sum of
  # week -> month cardinalities. This must hold across a leap-year December
  # (months 549:552 = Oct 2015..Jan 2016; 2015 is a Sym454 leap year), where
  # a previous bug double-counted December's own leap week, making its
  # starting week (and everything from Jan 2016 onwards) one week too late.
  months_idx <- 548:554
  fwd <- with(cal_sym454, chronon_divmod(month(1L), week(1L), months_idx))$div
  card <- with(cal_sym454, chronon_cardinality(week(1L), month(1L), at = 0:554))
  expected <- cumsum(c(0L, card))[months_idx + 1L]
  expect_equal(fwd, expected)

  # Same check for multi-month (quarter) units spanning the same leap-year
  # December -- quarter 183 = Oct/Nov/Dec 2015, quarter 184 = Jan/Feb/Mar 2016.
  expect_equal(
    with(cal_sym454, chronon_divmod(month(3L), week(1L), 182:185))$div,
    c(2374L, 2387L, 2401L, 2414L)
  )

  # Systematic check across 1970-2028: every month's forward divmod must
  # agree with the cumulative cardinality, not just leap-year Decembers.
  months_idx <- 0:700
  fwd <- with(cal_sym454, chronon_divmod(month(1L), week(1L), months_idx))$div
  card <- with(cal_sym454, chronon_cardinality(week(1L), month(1L), at = months_idx))
  expected <- cumsum(c(0L, card))[seq_along(months_idx)]
  expect_equal(fwd, expected)
})

date1 <- as.integer(as.Date("2026-02-26"))
date2 <- as.integer(as.Date("2026-08-19"))
expect_equal_time <- function(object, expected, tolerance = 120, ...) {
  tdiff <- abs(as.numeric(object) - as.numeric(expected))
  expect_lte(max(tdiff), tolerance, ...)
}

# ---------------------------------------------------------------------------
# Phase boundary index (new altitude-event scheme):
#
#  p=0  b=0  astro dawn    -18° morning  (astro twilight begins, night ends)
#  p=1  b=1  nautical dawn -12° morning
#  p=2  b=2  civil dawn    -6°  morning
#  p=3  b=3  sunrise       -0.833° morning (day begins)
#  p=4  b=4  sunset        -0.833° evening (day ends)
#  p=5  b=5  civil dusk    -6°  evening
#  p=6  b=6  nautical dusk -12° evening
#  p=7  b=7  astro dusk    -18° evening  (night begins)
#
#  Night (p=7) straddles solar midnight: astro dusk to next day's astro dawn.
#  Phase count for UTC day d, phase p: d * 8L + p
# ---------------------------------------------------------------------------

# ---------------------------------------------------------------------------
# All 8 phase boundaries: Melbourne, date1 (2026-02-26, AEDT = UTC+11)
#
# Critical edge case: Melbourne solar noon is ~02:30 UTC, so morning events
# (p=0..3) occur on UTC Feb 25 even though the solar day is indexed as Feb 26.
# ---------------------------------------------------------------------------
test_that("All 8 phase boundaries correct: Melbourne, 2026-02-26", {
  mel <- cal_time_solar$illumination(1L, lat = -37.8136, lon = 144.9631)

  # p=0: astronomical dawn (sun crosses -18° ascending)
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 0L, mel), tz = "Australia/Melbourne"),
    as.POSIXct("2026-02-26 05:28:00", tz = "Australia/Melbourne")
  )
  # p=1: nautical dawn (sun crosses -12° ascending)
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 1L, mel), tz = "Australia/Melbourne"),
    as.POSIXct("2026-02-26 06:01:00", tz = "Australia/Melbourne")
  )
  # p=2: civil dawn (sun crosses -6° ascending)
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 2L, mel), tz = "Australia/Melbourne"),
    as.POSIXct("2026-02-26 06:34:00", tz = "Australia/Melbourne")
  )
  # p=3: sunrise (sun crosses -0.833° ascending)
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 3L, mel), tz = "Australia/Melbourne"),
    as.POSIXct("2026-02-26 07:01:00", tz = "Australia/Melbourne")
  )
  # p=4: sunset (sun crosses -0.833° descending)
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 4L, mel), tz = "Australia/Melbourne"),
    as.POSIXct("2026-02-26 20:04:00", tz = "Australia/Melbourne")
  )
  # p=5: civil dusk (sun crosses -6° descending)
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 5L, mel), tz = "Australia/Melbourne"),
    as.POSIXct("2026-02-26 20:31:00", tz = "Australia/Melbourne")
  )
  # p=6: nautical dusk (sun crosses -12° descending)
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 6L, mel), tz = "Australia/Melbourne"),
    as.POSIXct("2026-02-26 21:03:00", tz = "Australia/Melbourne")
  )
  # p=7: astronomical dusk / night begins (sun crosses -18° descending)
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 7L, mel), tz = "Australia/Melbourne"),
    as.POSIXct("2026-02-26 21:36:00", tz = "Australia/Melbourne")
  )
})

# ---------------------------------------------------------------------------
# All 8 phase boundaries: Melbourne, date2 (2026-08-19, AEST = UTC+10)
# Southern Hemisphere winter: shorter days, different twilight lengths.
# ---------------------------------------------------------------------------
test_that("All 8 phase boundaries correct: Melbourne, 2026-08-19", {
  mel <- cal_time_solar$illumination(1L, lat = -37.8136, lon = 144.9631)

  # p=0: astronomical dawn
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 0L, mel), tz = "Australia/Melbourne"),
    as.POSIXct("2026-08-19 05:32:00", tz = "Australia/Melbourne")
  )
  # p=1: nautical dawn
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 1L, mel), tz = "Australia/Melbourne"),
    as.POSIXct("2026-08-19 06:02:00", tz = "Australia/Melbourne")
  )
  # p=2: civil dawn
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 2L, mel), tz = "Australia/Melbourne"),
    as.POSIXct("2026-08-19 06:33:00", tz = "Australia/Melbourne")
  )
  # p=3: sunrise
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 3L, mel), tz = "Australia/Melbourne"),
    as.POSIXct("2026-08-19 07:00:00", tz = "Australia/Melbourne")
  )
  # p=4: sunset
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 4L, mel), tz = "Australia/Melbourne"),
    as.POSIXct("2026-08-19 17:47:00", tz = "Australia/Melbourne")
  )
  # p=5: civil dusk
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 5L, mel), tz = "Australia/Melbourne"),
    as.POSIXct("2026-08-19 18:14:00", tz = "Australia/Melbourne")
  )
  # p=6: nautical dusk
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 6L, mel), tz = "Australia/Melbourne"),
    as.POSIXct("2026-08-19 18:45:00", tz = "Australia/Melbourne")
  )
  # p=7: astronomical dusk / night begins
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 7L, mel), tz = "Australia/Melbourne"),
    as.POSIXct("2026-08-19 19:16:00", tz = "Australia/Melbourne")
  )
})

# ---------------------------------------------------------------------------
# All 8 phase boundaries: Munich, date1 (2026-02-26, CET = UTC+1)
# Northern Hemisphere, late winter.
# ---------------------------------------------------------------------------
test_that("All 8 phase boundaries correct: Munich, 2026-02-26", {
  muc <- cal_time_solar$illumination(1L, lat = 48.133, lon = 11.583)

  # p=0: astronomical dawn
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 0L, muc)),
    as.POSIXct("2026-02-26 05:17:00", tz = "Europe/Berlin")
  )
  # p=1: nautical dawn
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 1L, muc)),
    as.POSIXct("2026-02-26 05:53:00", tz = "Europe/Berlin")
  )
  # p=2: civil dawn
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 2L, muc)),
    as.POSIXct("2026-02-26 06:29:00", tz = "Europe/Berlin")
  )
  # p=3: sunrise
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 3L, muc)),
    as.POSIXct("2026-02-26 07:00:00", tz = "Europe/Berlin")
  )
  # p=4: sunset
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 4L, muc)),
    as.POSIXct("2026-02-26 17:53:00", tz = "Europe/Berlin")
  )
  # p=5: civil dusk
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 5L, muc)),
    as.POSIXct("2026-02-26 18:24:00", tz = "Europe/Berlin")
  )
  # p=6: nautical dusk
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 6L, muc)),
    as.POSIXct("2026-02-26 19:00:00", tz = "Europe/Berlin")
  )
  # p=7: astronomical dusk / night begins
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 7L, muc)),
    as.POSIXct("2026-02-26 19:36:00", tz = "Europe/Berlin")
  )
})

# ---------------------------------------------------------------------------
# All 8 phase boundaries: Munich, date2 (2026-08-19, CEST = UTC+2)
# Northern Hemisphere, late summer: long days and notably extended twilights
# at 48°N (civil ~34 min, nautical ~46 min, astro ~66 min per band).
# ---------------------------------------------------------------------------
test_that("All 8 phase boundaries correct: Munich, 2026-08-19", {
  muc <- cal_time_solar$illumination(1L, lat = 48.133, lon = 11.583)

  # p=0: astronomical dawn
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 0L, muc)),
    as.POSIXct("2026-08-19 04:10:00", tz = "Europe/Berlin")
  )
  # p=1: nautical dawn
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 1L, muc)),
    as.POSIXct("2026-08-19 04:57:00", tz = "Europe/Berlin")
  )
  # p=2: civil dawn
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 2L, muc)),
    as.POSIXct("2026-08-19 05:39:00", tz = "Europe/Berlin")
  )
  # p=3: sunrise
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 3L, muc)),
    as.POSIXct("2026-08-19 06:13:00", tz = "Europe/Berlin")
  )
  # p=4: sunset
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 4L, muc)),
    as.POSIXct("2026-08-19 20:20:00", tz = "Europe/Berlin")
  )
  # p=5: civil dusk
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 5L, muc)),
    as.POSIXct("2026-08-19 20:53:00", tz = "Europe/Berlin")
  )
  # p=6: nautical dusk
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 6L, muc)),
    as.POSIXct("2026-08-19 21:35:00", tz = "Europe/Berlin")
  )
  # p=7: astronomical dusk / night begins
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 7L, muc)),
    as.POSIXct("2026-08-19 22:22:00", tz = "Europe/Berlin")
  )
})

# ---------------------------------------------------------------------------
# All 8 phase boundaries: Honolulu (Hawaii), date1 (2026-02-26, HST = UTC-10)
# West-longitude location; Hawaii does not observe DST.
# ---------------------------------------------------------------------------
test_that("All 8 phase boundaries correct: Honolulu, 2026-02-26", {
  hni <- cal_time_solar$illumination(1L, lat = 21.3069, lon = -157.8583)

  # p=0: astronomical dawn
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 0L, hni)),
    as.POSIXct("2026-02-26 05:39:00", tz = "Pacific/Honolulu")
  )
  # p=1: nautical dawn
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 1L, hni)),
    as.POSIXct("2026-02-26 06:05:00", tz = "Pacific/Honolulu")
  )
  # p=2: civil dawn
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 2L, hni)),
    as.POSIXct("2026-02-26 06:31:00", tz = "Pacific/Honolulu")
  )
  # p=3: sunrise
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 3L, hni)),
    as.POSIXct("2026-02-26 06:54:00", tz = "Pacific/Honolulu")
  )
  # p=4: sunset
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 4L, hni)),
    as.POSIXct("2026-02-26 18:34:00", tz = "Pacific/Honolulu")
  )
  # p=5: civil dusk
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 5L, hni)),
    as.POSIXct("2026-02-26 18:57:00", tz = "Pacific/Honolulu")
  )
  # p=6: nautical dusk
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 6L, hni)),
    as.POSIXct("2026-02-26 19:22:00", tz = "Pacific/Honolulu")
  )
  # p=7: astronomical dusk / night begins
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 8L + 7L, hni)),
    as.POSIXct("2026-02-26 19:48:00", tz = "Pacific/Honolulu")
  )
})

# ---------------------------------------------------------------------------
# All 8 phase boundaries: Honolulu (Hawaii), date2 (2026-08-19, HST = UTC-10)
# Northern Hemisphere summer: longer days than February.
# ---------------------------------------------------------------------------
test_that("All 8 phase boundaries correct: Honolulu, 2026-08-19", {
  hni <- cal_time_solar$illumination(1L, lat = 21.3069, lon = -157.8583)

  # p=0: astronomical dawn
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 0L, hni)),
    as.POSIXct("2026-08-19 04:53:00", tz = "Pacific/Honolulu")
  )
  # p=1: nautical dawn
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 1L, hni)),
    as.POSIXct("2026-08-19 05:21:00", tz = "Pacific/Honolulu")
  )
  # p=2: civil dawn
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 2L, hni)),
    as.POSIXct("2026-08-19 05:48:00", tz = "Pacific/Honolulu")
  )
  # p=3: sunrise
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 3L, hni)),
    as.POSIXct("2026-08-19 06:11:00", tz = "Pacific/Honolulu")
  )
  # p=4: sunset
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 4L, hni)),
    as.POSIXct("2026-08-19 18:58:00", tz = "Pacific/Honolulu")
  )
  # p=5: civil dusk
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 5L, hni)),
    as.POSIXct("2026-08-19 19:21:00", tz = "Pacific/Honolulu")
  )
  # p=6: nautical dusk
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 6L, hni)),
    as.POSIXct("2026-08-19 19:48:00", tz = "Pacific/Honolulu")
  )
  # p=7: astronomical dusk / night begins
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 8L + 7L, hni)),
    as.POSIXct("2026-08-19 20:15:00", tz = "Pacific/Honolulu")
  )
})

# ---------------------------------------------------------------------------
# Night phase spans midnight correctly: Honolulu
# ---------------------------------------------------------------------------
test_that("Night phase spans midnight correctly: Honolulu", {
  hni <- cal_time_solar$illumination(1L, lat = 21.3069, lon = -157.8583)

  hni_midnight_d1 <- as.POSIXct("2026-02-27 00:00:00", tz = "Pacific/Honolulu")
  hni_midnight_d2 <- as.POSIXct("2026-08-20 00:00:00", tz = "Pacific/Honolulu")

  # date1: p=7 (astro dusk) is before midnight
  expect_lt(
    as.numeric(as.POSIXct(linear_time(date1 * 8L + 7L, hni))),
    as.numeric(hni_midnight_d1)
  )
  # date1: next day p=0 (astro dawn) is after midnight
  expect_gt(
    as.numeric(as.POSIXct(linear_time((date1 + 1L) * 8L + 0L, hni))),
    as.numeric(hni_midnight_d1)
  )

  # date2: p=7 (astro dusk) is before midnight
  expect_lt(
    as.numeric(as.POSIXct(linear_time(date2 * 8L + 7L, hni))),
    as.numeric(hni_midnight_d2)
  )
  # date2: next day p=0 (astro dawn) is after midnight
  expect_gt(
    as.numeric(as.POSIXct(linear_time((date2 + 1L) * 8L + 0L, hni))),
    as.numeric(hni_midnight_d2)
  )
})

# ---------------------------------------------------------------------------
# Solar noon calculations (unchanged: uses ampm unit, not illumination phase)
# ---------------------------------------------------------------------------
test_that("Solar noon calculations", {
  # Noon = ampm half 1 (start of PM half = solar transit)
  mel_ampm <- cal_time_solar$ampm(1L, lat = -37.8136, lon = 144.9631)
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 2L + 1L, mel_ampm)),
    as.POSIXct("2026-02-26 13:33:00", tz = "Australia/Melbourne")
  )
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 2L + 1L, mel_ampm)),
    as.POSIXct("2026-08-19 12:23:00", tz = "Australia/Melbourne")
  )

  muc_ampm <- cal_time_solar$ampm(1L, lat = 48.133, lon = 11.583)
  expect_equal_time(
    as.POSIXct(linear_time(date1 * 2L + 1L, muc_ampm)),
    as.POSIXct("2026-02-26 12:26:00", tz = "Europe/Berlin")
  )
  expect_equal_time(
    as.POSIXct(linear_time(date2 * 2L + 1L, muc_ampm)),
    as.POSIXct("2026-08-19 13:17:00", tz = "Europe/Berlin")
  )
})

# ---------------------------------------------------------------------------
# Night phase spans midnight correctly
#
# p=7 (astro dusk / night start) must be before local midnight of day d.
# p=0 of day d+1 (next astro dawn) must be after local midnight of day d+1.
# This confirms night straddles solar midnight.
# ---------------------------------------------------------------------------
test_that("Night phase spans midnight correctly", {
  mel <- cal_time_solar$illumination(1L, lat = -37.8136, lon = 144.9631)
  muc <- cal_time_solar$illumination(1L, lat = 48.133,   lon = 11.583)

  mel_midnight_d1  <- as.POSIXct("2026-02-27 00:00:00", tz = "Australia/Melbourne")
  mel_midnight_d2  <- as.POSIXct("2026-08-20 00:00:00", tz = "Australia/Melbourne")
  muc_midnight_d1  <- as.POSIXct("2026-02-27 00:00:00", tz = "Europe/Berlin")
  muc_midnight_d2  <- as.POSIXct("2026-08-20 00:00:00", tz = "Europe/Berlin")

  # Melbourne date1: p=7 (astro dusk) is before midnight
  expect_lt(
    as.numeric(as.POSIXct(linear_time(date1 * 8L + 7L, mel))),
    as.numeric(mel_midnight_d1)
  )
  # Melbourne date1: next day p=0 (astro dawn) is after midnight
  expect_gt(
    as.numeric(as.POSIXct(linear_time((date1 + 1L) * 8L + 0L, mel))),
    as.numeric(mel_midnight_d1)
  )

  # Melbourne date2: p=7 (astro dusk) is before midnight
  expect_lt(
    as.numeric(as.POSIXct(linear_time(date2 * 8L + 7L, mel))),
    as.numeric(mel_midnight_d2)
  )
  # Melbourne date2: next day p=0 (astro dawn) is after midnight
  expect_gt(
    as.numeric(as.POSIXct(linear_time((date2 + 1L) * 8L + 0L, mel))),
    as.numeric(mel_midnight_d2)
  )

  # Munich date1: p=7 (astro dusk) is before midnight
  expect_lt(
    as.numeric(as.POSIXct(linear_time(date1 * 8L + 7L, muc))),
    as.numeric(muc_midnight_d1)
  )
  # Munich date1: next day p=0 (astro dawn) is after midnight
  expect_gt(
    as.numeric(as.POSIXct(linear_time((date1 + 1L) * 8L + 0L, muc))),
    as.numeric(muc_midnight_d1)
  )

  # Munich date2: p=7 (astro dusk) is before midnight
  expect_lt(
    as.numeric(as.POSIXct(linear_time(date2 * 8L + 7L, muc))),
    as.numeric(muc_midnight_d2)
  )
  # Munich date2: next day p=0 (astro dawn) is after midnight
  expect_gt(
    as.numeric(as.POSIXct(linear_time((date2 + 1L) * 8L + 0L, muc))),
    as.numeric(muc_midnight_d2)
  )
})

# ---------------------------------------------------------------------------
# Round-trip: UTC → illumination mixtime → UTC
#
# Take a known UTC POSIXct at roughly the midpoint of each phase, convert to
# a continuous illumination mixtime via linear_time(..., discrete = FALSE),
# then back to UTC via as.POSIXct(), and verify the result matches the input
# within tolerance. Tests Melbourne, Munich, and Honolulu across all 8 phases.
# ---------------------------------------------------------------------------
test_that("Round-trip UTC → phase count → UTC", {
  lat <- -37.8136; lon <- 144.9631
  mel <- cal_time_solar$illumination(1L, lat = lat, lon = lon)

  # Phase midpoints: local Melbourne time (AEDT) for solar day date1
  midpoints_utc <- as.POSIXct(c(
    "2026-02-26 05:52:00",  # p=0 midpoint: between astro dawn and nautical dawn
    "2026-02-26 06:21:00",  # p=1 midpoint: between nautical dawn and civil dawn
    "2026-02-26 06:47:00",  # p=2 midpoint: between civil dawn and sunrise
    "2026-02-26 13:33:00",  # p=3 midpoint: solar noon (midpoint of day)
    "2026-02-26 20:17:00",  # p=4 midpoint: between sunset and civil dusk
    "2026-02-26 20:44:00",  # p=5 midpoint: between civil dusk and nautical dusk
    "2026-02-26 21:13:00",  # p=6 midpoint: between nautical dusk and astro dusk
    "2026-02-27 01:33:00"   # p=7 midpoint: solar midnight (middle of night)
  ), tz = "Australia/Melbourne")

  illumination <- linear_time(midpoints_utc, mel, discrete = FALSE)
  recovered <- as.POSIXct(illumination)
  expect_equal_time(recovered, midpoints_utc)

  lat <- 48.133; lon <- 11.583
  muc <- cal_time_solar$illumination(1L, lat = lat, lon = lon)

  # Phase midpoints: local Munich time (CET) for solar day date1
  midpoints_utc <- as.POSIXct(c(
    "2026-02-26 05:38:00",  # p=0 midpoint: between astro dawn and nautical dawn
    "2026-02-26 06:13:00",  # p=1 midpoint: between nautical dawn and civil dawn
    "2026-02-26 06:44:00",  # p=2 midpoint: between civil dawn and sunrise
    "2026-02-26 12:26:00",  # p=3 midpoint: solar noon (midpoint of day)
    "2026-02-26 18:08:00",  # p=4 midpoint: between sunset and civil dusk
    "2026-02-26 18:39:00",  # p=5 midpoint: between civil dusk and nautical dusk
    "2026-02-26 19:16:00",  # p=6 midpoint: between nautical dusk and astro dusk
    "2026-02-27 00:26:00"   # p=7 midpoint: solar midnight (middle of night)
  ), tz = "Europe/Berlin")

  illumination <- linear_time(midpoints_utc, muc, discrete = FALSE)
  recovered <- as.POSIXct(illumination)
  expect_equal_time(recovered, midpoints_utc)


  lat <- 21.3069; lon <- -157.8583
  hni <- cal_time_solar$illumination(1L, lat = lat, lon = lon)

  # Phase midpoints: local Honolulu time (HST) for solar day date1
  midpoints_utc <- as.POSIXct(c(
    "2026-02-26 05:53:00",  # p=0 midpoint: between astro dawn and nautical dawn
    "2026-02-26 06:19:00",  # p=1 midpoint: between nautical dawn and civil dawn
    "2026-02-26 06:43:00",  # p=2 midpoint: between civil dawn and sunrise
    "2026-02-26 12:44:00",  # p=3 midpoint: solar noon (midpoint of day)
    "2026-02-26 18:46:00",  # p=4 midpoint: between sunset and civil dusk
    "2026-02-26 19:10:00",  # p=5 midpoint: between civil dusk and nautical dusk
    "2026-02-26 19:36:00",  # p=6 midpoint: between nautical dusk and astro dusk
    "2026-02-27 00:44:00"   # p=7 midpoint: solar midnight (middle of night)
  ), tz = "Pacific/Honolulu")

  illumination <- linear_time(midpoints_utc, hni, discrete = FALSE)
  recovered <- as.POSIXct(illumination)
  expect_equal_time(recovered, midpoints_utc)
})

# ---------------------------------------------------------------------------
# Fractional phase interpolation
#
# d * 8 + 3.5 is the midpoint of the day phase (between sunrise and sunset).
# This should be close to solar noon (within a few minutes), since solar noon
# is the midpoint of the day by definition.
# ---------------------------------------------------------------------------
test_that("Fractional phase interpolation: day midpoint approx solar noon", {
  mel_illum <- cal_time_solar$illumination(1L, lat = -37.8136, lon = 144.9631)
  mel_ampm  <- cal_time_solar$ampm(1L,         lat = -37.8136, lon = 144.9631)
  muc_illum <- cal_time_solar$illumination(1L, lat = 48.133,   lon = 11.583)
  muc_ampm  <- cal_time_solar$ampm(1L,         lat = 48.133,   lon = 11.583)

  # Melbourne date1: midpoint of day phase vs solar noon
  mel_day_mid <- as.POSIXct(linear_time(date1 * 8 + 3.5, mel_illum))
  mel_noon    <- as.POSIXct(linear_time(date1 * 2 + 1,   mel_ampm))
  expect_equal(as.numeric(mel_day_mid), as.numeric(mel_noon), tolerance = 300)

  # Melbourne date2
  mel_day_mid2 <- as.POSIXct(linear_time(date2 * 8 + 3.5, mel_illum))
  mel_noon2    <- as.POSIXct(linear_time(date2 * 2 + 1,   mel_ampm))
  expect_equal(as.numeric(mel_day_mid2), as.numeric(mel_noon2), tolerance = 300)

  # Munich date1
  muc_day_mid <- as.POSIXct(linear_time(date1 * 8 + 3.5, muc_illum))
  muc_noon    <- as.POSIXct(linear_time(date1 * 2 + 1,   muc_ampm))
  expect_equal(as.numeric(muc_day_mid), as.numeric(muc_noon), tolerance = 300)

  # Munich date2
  muc_day_mid2 <- as.POSIXct(linear_time(date2 * 8 + 3.5, muc_illum))
  muc_noon2    <- as.POSIXct(linear_time(date2 * 2 + 1,   muc_ampm))
  expect_equal(as.numeric(muc_day_mid2), as.numeric(muc_noon2), tolerance = 300)
})

test_that("time_floor/time_ceiling/time_round work for hourly linear_time with 6-hour granule", {
  # hour 13 since epoch (1970-01-01 13h) -- 1/3 of the way into a 6-hour block
  h13 <- linear_time(13L, cal_gregorian$hour(1L))
  h12 <- linear_time(12L, cal_gregorian$hour(1L))
  h18 <- linear_time(18L, cal_gregorian$hour(1L))

  expect_equal(time_floor(h13,   granule = cal_gregorian$hour(6L)), h12)
  expect_equal(time_ceiling(h13, granule = cal_gregorian$hour(6L)), h18)
  expect_equal(time_round(h13,   granule = cal_gregorian$hour(6L)), h12) # closer to 12 than 18

  # hour 16 -- past the midpoint, so round goes up
  h16 <- linear_time(16L, cal_gregorian$hour(1L))
  expect_equal(time_floor(h16,   granule = cal_gregorian$hour(6L)), h12)
  expect_equal(time_ceiling(h16, granule = cal_gregorian$hour(6L)), h18)
  expect_equal(time_round(h16,   granule = cal_gregorian$hour(6L)), h18) # closer to 18
})

test_that("time_floor/time_ceiling/time_round on exact 6-hour boundary", {
  # hour 12 is exactly on a 6-hour boundary
  h12 <- linear_time(12L, cal_gregorian$hour(1L))
  h18 <- linear_time(18L, cal_gregorian$hour(1L))

  expect_equal(time_floor(h12,   granule = cal_gregorian$hour(6L)), h12)
  # ceiling always advances an exact boundary (per implementation intent)
  expect_equal(time_ceiling(h12, granule = cal_gregorian$hour(6L)), h18)
  expect_equal(time_round(h12,   granule = cal_gregorian$hour(6L)), h12)
})

test_that("time_floor/time_ceiling/time_round work for hourly linear_time with 1-hour granule", {
  # Rounding to the same chronon granule is an identity for floor/round;
  # ceiling advances exact boundaries.
  h13 <- linear_time(13L, cal_gregorian$hour(1L))
  h14 <- linear_time(14L, cal_gregorian$hour(1L))

  expect_equal(time_floor(h13,   granule = cal_gregorian$hour(1L)), h13)
  expect_equal(time_ceiling(h13, granule = cal_gregorian$hour(1L)), h14) # exact → advances
  expect_equal(time_round(h13,   granule = cal_gregorian$hour(1L)), h13)
})

test_that("time_floor/time_ceiling/time_round work for yearmonth with quarter granule", {
  # Feb 2020 is the 2nd month of Q1 2020
  ym_feb <- yearmonth(as.Date("2020-02-01"))
  q1_2020 <- yearmonth(as.Date("2020-01-01"))  # month 600 since epoch
  q2_2020 <- yearmonth(as.Date("2020-04-01"))  # month 603

  expect_equal(time_floor(ym_feb,   granule = cal_gregorian$quarter(1L)), q1_2020)
  expect_equal(time_ceiling(ym_feb, granule = cal_gregorian$quarter(1L)), q2_2020)
  expect_equal(time_round(ym_feb,   granule = cal_gregorian$quarter(1L)), q1_2020)

  # Aug 2020 (month 607): 607/3 = 202.33 — rounds down to Q3 (Jul)
  ym_aug <- yearmonth(as.Date("2020-08-01"))
  q3_2020 <- yearmonth(as.Date("2020-07-01"))  # month 606
  q4_2020 <- yearmonth(as.Date("2020-10-01"))  # month 609

  expect_equal(time_floor(ym_aug,   granule = cal_gregorian$quarter(1L)), q3_2020)
  expect_equal(time_ceiling(ym_aug, granule = cal_gregorian$quarter(1L)), q4_2020)
  expect_equal(time_round(ym_aug,   granule = cal_gregorian$quarter(1L)), q3_2020)

  # Sep 2020 (month 608): 608/3 = 202.67 — rounds up to Q4 (Oct)
  ym_sep <- yearmonth(as.Date("2020-09-01"))
  expect_equal(time_floor(ym_sep,   granule = cal_gregorian$quarter(1L)), q3_2020)
  expect_equal(time_ceiling(ym_sep, granule = cal_gregorian$quarter(1L)), q4_2020)
  expect_equal(time_round(ym_sep,   granule = cal_gregorian$quarter(1L)), q4_2020)
})

test_that("time_floor/time_ceiling/time_round work for yearmonth with year granule", {
  # Jun 2020 (month 605): 605/12 = 50.42 — floors to year 50 (2020)
  ym_jun <- yearmonth(as.Date("2020-06-01"))
  yr_2020 <- yearmonth(as.Date("2020-01-01"))  # month 600
  yr_2021 <- yearmonth(as.Date("2021-01-01"))  # month 612

  expect_equal(time_floor(ym_jun,   granule = cal_gregorian$year(1L)), yr_2020)
  expect_equal(time_ceiling(ym_jun, granule = cal_gregorian$year(1L)), yr_2021)
  expect_equal(time_round(ym_jun,   granule = cal_gregorian$year(1L)), yr_2020) # < 6 months past Jan

  # Aug 2020 (month 607): 607/12 = 50.58 — rounds up to 2021
  ym_aug <- yearmonth(as.Date("2020-08-01"))
  expect_equal(time_floor(ym_aug,   granule = cal_gregorian$year(1L)), yr_2020)
  expect_equal(time_ceiling(ym_aug, granule = cal_gregorian$year(1L)), yr_2021)
  expect_equal(time_round(ym_aug,   granule = cal_gregorian$year(1L)), yr_2021) # > 6 months past Jan
})

test_that("time_floor/time_ceiling/time_round work for yearquarter with year granule", {
  # Q2 2020 = quarter 201 since epoch; 201/4 = 50.25, floors to year 2020
  yq_q2 <- yearquarter(as.Date("2020-05-01"))  # quarter 201
  yq_q1_2020 <- yearquarter(as.Date("2020-01-01"))  # quarter 200
  yq_q1_2021 <- yearquarter(as.Date("2021-01-01"))  # quarter 204

  expect_equal(time_floor(yq_q2,   granule = cal_gregorian$year(1L)), yq_q1_2020)
  expect_equal(time_ceiling(yq_q2, granule = cal_gregorian$year(1L)), yq_q1_2021)
  expect_equal(time_round(yq_q2,   granule = cal_gregorian$year(1L)), yq_q1_2020) # Q2 is closer to Q1
})

test_that("time_floor/time_ceiling/time_round handle NA in vectorised mt_linear input", {
  ym_vec <- yearmonth(c(as.Date("2020-02-01"), as.Date("2020-08-01"), NA))

  res_floor   <- time_floor(ym_vec,   granule = cal_gregorian$quarter(1L))
  res_ceiling <- time_ceiling(ym_vec, granule = cal_gregorian$quarter(1L))
  res_round   <- time_round(ym_vec,   granule = cal_gregorian$quarter(1L))

  expect_equal(res_floor[1],   yearmonth(as.Date("2020-01-01")))  # Feb → Q1 start
  expect_equal(res_floor[2],   yearmonth(as.Date("2020-07-01")))  # Aug → Q3 start
  expect_true(is.na(res_floor[3]))

  expect_equal(res_ceiling[1], yearmonth(as.Date("2020-04-01")))  # Feb → Q2 start
  expect_equal(res_ceiling[2], yearmonth(as.Date("2020-10-01")))  # Aug → Q4 start
  expect_true(is.na(res_ceiling[3]))

  expect_equal(res_round[1],   yearmonth(as.Date("2020-01-01")))  # Feb rounds to Q1
  expect_equal(res_round[2],   yearmonth(as.Date("2020-07-01")))  # Aug rounds to Q3
  expect_true(is.na(res_round[3]))
})

test_that("time_floor/time_ceiling/time_round handle vectorised hourly linear_time with NA", {
  h_vec <- linear_time(c(13L, 16L, NA), cal_gregorian$hour(1L))

  res_floor   <- time_floor(h_vec,   granule = cal_gregorian$hour(6L))
  res_ceiling <- time_ceiling(h_vec, granule = cal_gregorian$hour(6L))
  res_round   <- time_round(h_vec,   granule = cal_gregorian$hour(6L))

  expect_equal(res_floor[1],   linear_time(12L, cal_gregorian$hour(1L)))
  expect_equal(res_floor[2],   linear_time(12L, cal_gregorian$hour(1L)))
  expect_true(is.na(res_floor[3]))

  expect_equal(res_ceiling[1], linear_time(18L, cal_gregorian$hour(1L)))
  expect_equal(res_ceiling[2], linear_time(18L, cal_gregorian$hour(1L)))
  expect_true(is.na(res_ceiling[3]))

  expect_equal(res_round[1],   linear_time(12L, cal_gregorian$hour(1L)))  # 13 → 12
  expect_equal(res_round[2],   linear_time(18L, cal_gregorian$hour(1L)))  # 16 → 18
  expect_true(is.na(res_round[3]))
})


test_that("time_floor/time_ceiling/time_round handle vectorised hourly linear_time with timezones", {
  melb_hr <- cal_gregorian$hour(1L, tz = "Australia/Melbourne")
  
  # Time points across the 2020 DST change (+11 -> +10)
  h_vec <- linear_time(c(440557:440581), melb_hr)
  res_floor   <- time_floor(h_vec,   granule = cal_gregorian$day(1L))
  res_ceiling <- time_ceiling(h_vec, granule = cal_gregorian$day(1L))
  res_round   <- time_round(h_vec,   granule = cal_gregorian$day(1L))

  expect_equal(res_floor,   linear_time(rep(440557L, 25), melb_hr))
  expect_equal(res_ceiling,   linear_time(rep(440582L, 25), melb_hr))
  expect_equal(res_round,   linear_time(c(rep(440557L, 13), rep(440582L, 12)), melb_hr))
})

test_that("time_round/time_floor/time_ceiling work for POSIXct with tz preservation", {
  t <- as.POSIXct("2020-01-01 12:34:56", tz = "UTC")

  # expected values computed via epoch arithmetic (seconds)
  expected_round <- as.POSIXct(round(as.numeric(t) / 60) * 60, origin = "1970-01-01", tz = "UTC")
  expected_floor <- as.POSIXct(floor(as.numeric(t) / 60) * 60, origin = "1970-01-01", tz = "UTC")
  expected_ceiling <- as.POSIXct(ceiling(as.numeric(t) / 60) * 60, origin = "1970-01-01", tz = "UTC")

  # using package minute unit object (if available) and character granule
  expect_equal(
    time_round(t, granule = cal_gregorian$minute(1L)), 
    expected_round
  )
  expect_equal(
    time_floor(t, granule = cal_gregorian$minute(1L)), 
    expected_floor
  )
  expect_equal(
    time_ceiling(t, granule = cal_gregorian$minute(1L)), 
    expected_ceiling
  )

  expect_equal(
    time_round(t, granule = "1 minute"),
    expected_round
  )
  expect_equal(
    time_floor(t, granule = "1 minute"),
    expected_floor
  )
  expect_equal(
    time_ceiling(t, granule = "1 minute"),
    expected_ceiling
  )

  # tz preserved
  res <- time_round(t, granule = "1 minute")
  expect_equal(attr(res, "tzone"), attr(t, "tzone"))
  expect_s3_class(res, "POSIXct")

  # vectorized inputs and NA preservation
  t2 <- as.POSIXct("2020-01-01 12:34:20", tz = "UTC")
  vec_in <- c(t, t2, NA)
  attr(vec_in, "tzone") <- attr(t, "tzone")
  out_round <- time_round(vec_in, granule = "1 minute")
  expect_equal(
    out_round[1], 
    expected_round
  )
  expect_equal(
    out_round[2], 
    as.POSIXct(round(as.numeric(t2) / 60) * 60, origin = "1970-01-01", tz = "UTC")
  )
  expect_true(is.na(out_round[3]))
})

test_that("time_floor/time_ceiling work for Date", {
  d <- as.Date("2020-01-15")

  first_of_month <- as.Date(format(d, "%Y-%m-01"))
  first_of_next_month <- seq(first_of_month, by = "1 month", length.out = 2)[2]

  expect_equal(
    time_floor(d, granule = cal_gregorian$month(1L)), 
    first_of_month
  )
  expect_equal(
    time_ceiling(d, granule = cal_gregorian$month(1L)), 
    first_of_next_month
  )

  expect_equal(
    time_floor(d, granule = "month"), 
    first_of_month
  )
  expect_equal(
    time_ceiling(d, granule = "month"), 
    first_of_next_month
  )

  # class preserved
  res_floor <- time_floor(d, granule = "month")
  expect_s3_class(res_floor, "Date")
})

# ── datetime: civil (Gregorian + timezone) ────────────────────────────────────

test_that("time_floor/time_ceiling/time_round work for civil datetime rounded to day", {
  # 2020-06-21 10:30:00 UTC = 2020-06-21 20:30:00 AEST (UTC+10, no DST)
  t_utc <- as.POSIXct("2020-06-21 10:30:00", tz = "UTC")
  now_civ <- datetime(t_utc, tz = "Australia/Melbourne")

  fl_exp <- datetime(as.POSIXct("2020-06-21 00:00:00", tz = "Australia/Melbourne"),
                     tz = "Australia/Melbourne")
  cl_exp <- datetime(as.POSIXct("2020-06-22 00:00:00", tz = "Australia/Melbourne"),
                     tz = "Australia/Melbourne")

  expect_equal(time_floor(now_civ,   granule = cal_gregorian$day(1L)), fl_exp)
  expect_equal(time_ceiling(now_civ, granule = cal_gregorian$day(1L)), cl_exp)
  # 20:30 is past 12:00 midpoint → rounds up to next day
  expect_equal(time_round(now_civ,   granule = cal_gregorian$day(1L)), cl_exp)
})

test_that("time_floor/time_ceiling/time_round work for civil datetime rounded to hour", {
  # 2020-06-21 20:30:00 AEST → floor to 20:00, ceiling to 21:00, round to 20:00 (:30 = half)
  t_utc  <- as.POSIXct("2020-06-21 10:30:00", tz = "UTC")
  now_civ <- datetime(t_utc, tz = "Australia/Melbourne")

  fl_exp <- datetime(as.POSIXct("2020-06-21 20:00:00", tz = "Australia/Melbourne"),
                     tz = "Australia/Melbourne")
  cl_exp <- datetime(as.POSIXct("2020-06-21 21:00:00", tz = "Australia/Melbourne"),
                     tz = "Australia/Melbourne")

  expect_equal(time_floor(now_civ,   granule = cal_gregorian$hour(1L)), fl_exp)
  expect_equal(time_ceiling(now_civ, granule = cal_gregorian$hour(1L)), cl_exp)
  # :30 is at the midpoint — convention rounds to even (floor) → 20:00
  expect_equal(time_round(now_civ,   granule = cal_gregorian$hour(1L)), fl_exp)
})

test_that("time_floor/time_ceiling work for civil datetime rounded to quarter across a UTC offset", {
  # Quarters have a variable number of days, so the timezone offset can't be
  # converted exactly into quarter units (only approximated by an average
  # quarter length) - it must be applied while still in seconds. Regression
  # test for a bug where the boundary landed ~760s away from local midnight.
  melb <- datetime(as.POSIXct("2015-02-15", tz = "Australia/Melbourne"),
                    tz = "Australia/Melbourne")
  expect_equal(
    time_floor(melb, granule = cal_gregorian$quarter(1L)),
    datetime(as.POSIXct("2015-01-01 00:00:00", tz = "Australia/Melbourne"),
             tz = "Australia/Melbourne")
  )
  expect_equal(
    time_ceiling(melb, granule = cal_gregorian$quarter(1L)),
    datetime(as.POSIXct("2015-04-01 00:00:00", tz = "Australia/Melbourne"),
             tz = "Australia/Melbourne")
  )

  ny <- datetime(as.POSIXct("2015-02-15", tz = "America/New_York"),
                 tz = "America/New_York")
  expect_equal(
    time_floor(ny, granule = cal_gregorian$quarter(1L)),
    datetime(as.POSIXct("2015-01-01 00:00:00", tz = "America/New_York"),
             tz = "America/New_York")
  )
  expect_equal(
    time_ceiling(ny, granule = cal_gregorian$quarter(1L)),
    datetime(as.POSIXct("2015-04-01 00:00:00", tz = "America/New_York"),
             tz = "America/New_York")
  )
})

test_that("time_floor/time_ceiling/time_round work for civil datetime early in the day", {
  # 2020-06-22 02:15:00 AEST → rounds DOWN to 2020-06-22 00:00 (< 12h into day)
  t_utc   <- as.POSIXct("2020-06-21 16:15:00", tz = "UTC")
  now_early <- datetime(t_utc, tz = "Australia/Melbourne")

  fl_exp <- datetime(as.POSIXct("2020-06-22 00:00:00", tz = "Australia/Melbourne"),
                     tz = "Australia/Melbourne")
  cl_exp <- datetime(as.POSIXct("2020-06-23 00:00:00", tz = "Australia/Melbourne"),
                     tz = "Australia/Melbourne")

  expect_equal(time_floor(now_early,   granule = cal_gregorian$day(1L)), fl_exp)
  expect_equal(time_ceiling(now_early, granule = cal_gregorian$day(1L)), cl_exp)
  # 02:15 is less than 12h past midnight → rounds DOWN
  expect_equal(time_round(now_early,   granule = cal_gregorian$day(1L)), fl_exp)
})

# ── datetime: solar (apparent-solar-day calendar) ─────────────────────────────

test_that("time_floor/time_ceiling/time_round work for solar datetime rounded to solar day", {
  # 2020-06-21 10:30:00 UTC ≈ 2020-06-21 20:07:58 solar Melbourne
  # (solar midnight for that day ≈ 2020-06-20 14:22 UTC)
  lat <- -37.8136; lon <- 144.9631
  t_utc   <- as.POSIXct("2020-06-21 10:30:00", tz = "UTC")
  now_sol <- datetime(t_utc, calendar = cal_time_solar, lat = lat, lon = lon)

  fl_exp  <- time_floor(now_sol,   cal_time_solar$day(1L))  # 2020-06-21 solar midnight
  cl_exp  <- time_ceiling(now_sol, cal_time_solar$day(1L))  # 2020-06-22 solar midnight

  expect_equal(time_floor(now_sol,   granule = cal_time_solar$day(1L)), fl_exp)
  expect_equal(time_ceiling(now_sol, granule = cal_time_solar$day(1L)), cl_exp)
  # 20:07 solar is past the 12:00 midpoint → rounds up to next solar day
  expect_equal(time_round(now_sol,   granule = cal_time_solar$day(1L)), cl_exp)
})

test_that("time_floor/time_ceiling/time_round work for solar datetime rounded to solar hour", {
  # 2020-06-21 20:07:58 solar → floor 20:00, ceiling 21:00, round 20:00 (7 min < 30)
  lat <- -37.8136; lon <- 144.9631
  t_utc   <- as.POSIXct("2020-06-21 10:30:00", tz = "UTC")
  now_sol <- datetime(t_utc, calendar = cal_time_solar, lat = lat, lon = lon)

  fl_exp <- time_floor(now_sol,   cal_time_solar$hour(1L))  # 20:00 solar
  cl_exp <- time_ceiling(now_sol, cal_time_solar$hour(1L))  # 21:00 solar

  expect_equal(time_floor(now_sol,   granule = cal_time_solar$hour(1L)), fl_exp)
  expect_equal(time_ceiling(now_sol, granule = cal_time_solar$hour(1L)), cl_exp)
  # :07 is well before the half-hour midpoint → rounds DOWN to 20:00
  expect_equal(time_round(now_sol,   granule = cal_time_solar$hour(1L)), fl_exp)
})

test_that("time_floor/time_ceiling/time_round work for solar datetime early in solar day", {
  # ~00:38 solar time (40 min after solar midnight) → rounds DOWN to same solar day
  lat <- -37.8136; lon <- 144.9631
  t_utc      <- as.POSIXct("2020-06-20 15:00:00", tz = "UTC")
  now_sol_early <- datetime(t_utc, calendar = cal_time_solar, lat = lat, lon = lon)

  fl_exp <- time_floor(now_sol_early,   cal_time_solar$day(1L))
  cl_exp <- time_ceiling(now_sol_early, cal_time_solar$day(1L))

  expect_equal(time_floor(now_sol_early,   granule = cal_time_solar$day(1L)), fl_exp)
  expect_equal(time_ceiling(now_sol_early, granule = cal_time_solar$day(1L)), cl_exp)
  # 00:38 is well before the 12h midpoint → round goes DOWN to same solar midnight
  expect_equal(time_round(now_sol_early,   granule = cal_time_solar$day(1L)), fl_exp)
})

# ── Cross-calendar rounding ───────────────────────────────────────────────────

test_that("civil datetime can be rounded using a solar day granule", {
  # Floor a civil (AEST) datetime to the solar day boundary for Melbourne.
  # The solar midnight is slightly offset from civil midnight.
  lat <- -37.8136; lon <- 144.9631
  t_utc   <- as.POSIXct("2020-06-21 10:30:00", tz = "UTC")
  now_civ <- datetime(t_utc, tz = "Australia/Melbourne")

  fl_exp <- time_floor(now_civ,   cal_time_solar$day(1L, lat = lat, lon = lon))
  cl_exp <- time_ceiling(now_civ, cal_time_solar$day(1L, lat = lat, lon = lon))

  expect_equal(time_floor(now_civ,   granule = cal_time_solar$day(1L, lat = lat, lon = lon)), fl_exp)
  expect_equal(time_ceiling(now_civ, granule = cal_time_solar$day(1L, lat = lat, lon = lon)), cl_exp)
  # 20:30 AEST is past the midpoint of the solar day → rounds up to next solar midnight
  expect_equal(time_round(now_civ, granule = cal_time_solar$day(1L, lat = lat, lon = lon)), cl_exp)
})

test_that("solar datetime can be rounded using a civil day granule", {
  # Floor a solar datetime to the civil midnight boundary (Melbourne tz).
  # Result is expressed in solar time but snaps to civil midnight.
  lat <- -37.8136; lon <- 144.9631
  t_utc   <- as.POSIXct("2020-06-21 10:30:00", tz = "UTC")
  now_sol <- datetime(t_utc, calendar = cal_time_solar, lat = lat, lon = lon)

  fl_exp <- time_floor(now_sol,   cal_gregorian$day(1L, tz = "Australia/Melbourne"))
  cl_exp <- time_ceiling(now_sol, cal_gregorian$day(1L, tz = "Australia/Melbourne"))

  expect_equal(time_floor(now_sol,   granule = cal_gregorian$day(1L, tz = "Australia/Melbourne")), fl_exp)
  expect_equal(time_ceiling(now_sol, granule = cal_gregorian$day(1L, tz = "Australia/Melbourne")), cl_exp)
  # 20:07 solar is past the civil-midnight midpoint (expressed in solar) → rounds up
  expect_equal(time_round(now_sol, granule = cal_gregorian$day(1L, tz = "Australia/Melbourne")), cl_exp)
})

# Lunation 1276: new moon 2026-02-17 12:03 UTC
# Lunation 1281: new moon 2026-07-14 09:45 UTC
# Tolerance of 5 minutes (~300s) for approximate lunation algorithm

expect_equal_time <- function(object, expected, tolerance = 300, ...) {
  expect_equal(as.numeric(object), as.numeric(expected), tolerance = tolerance, ...)
}

test_that("New moon times (synodic month)", {
  # Lunation 1276: 2026-02-17 ~12:03 UTC (new moon)
  expect_equal_time(
    as.POSIXct(linear_time(1276L, cal_time_lunar$month(1L)), tz = "UTC"),
    as.POSIXct("2026-02-17 12:03:00", tz = "UTC")
  )
  # Local timezone representations of the same instant
  expect_equal_time(
    as.POSIXct(linear_time(1276L, cal_time_lunar$month(1L))),
    as.POSIXct("2026-02-17 23:03:19", tz = "Australia/Melbourne")
  )
  expect_equal_time(
    as.POSIXct(linear_time(1276L, cal_time_lunar$month(1L))),
    as.POSIXct("2026-02-17 13:03:19", tz = "Europe/Berlin")
  )

  # Lunation 1281: 2026-07-14 ~09:45 UTC (new moon)
  expect_equal_time(
    as.POSIXct(linear_time(1281L, cal_time_lunar$month(1L)), tz = "UTC"),
    as.POSIXct("2026-07-14 09:45:00", tz = "UTC")
  )
  expect_equal_time(
    as.POSIXct(linear_time(1281L, cal_time_lunar$month(1L))),
    as.POSIXct("2026-07-14 19:45:18", tz = "Australia/Melbourne")
  )
  expect_equal_time(
    as.POSIXct(linear_time(1281L, cal_time_lunar$month(1L))),
    as.POSIXct("2026-07-14 11:45:18", tz = "Europe/Berlin")
  )
})

test_that("Full moon times (synodic phase index 4)", {
  # Lunation 1276: full moon ~2026-03-04 06:44 UTC
  expect_equal_time(
    as.POSIXct(linear_time(1276L * 8 + 4, cal_time_lunar$phase(1L)), tz = "UTC"),
    as.POSIXct("2026-03-04 06:44:00", tz = "UTC")
  )
  expect_equal_time(
    as.POSIXct(linear_time(1276L * 8 + 4, cal_time_lunar$phase(1L))),
    as.POSIXct("2026-03-04 17:44:34", tz = "Australia/Melbourne")
  )
  expect_equal_time(
    as.POSIXct(linear_time(1276L * 8 + 4, cal_time_lunar$phase(1L))),
    as.POSIXct("2026-03-04 07:44:34", tz = "Europe/Berlin")
  )

  # Lunation 1281: full moon ~2026-07-29 01:41 UTC
  expect_equal_time(
    as.POSIXct(linear_time(1281L * 8 + 4, cal_time_lunar$phase(1L)), tz = "UTC"),
    as.POSIXct("2026-07-29 01:41:00", tz = "UTC")
  )
  expect_equal_time(
    as.POSIXct(linear_time(1281L * 8 + 4, cal_time_lunar$phase(1L))),
    as.POSIXct("2026-07-29 11:41:40", tz = "Australia/Melbourne")
  )
  expect_equal_time(
    as.POSIXct(linear_time(1281L * 8 + 4, cal_time_lunar$phase(1L))),
    as.POSIXct("2026-07-29 03:41:40", tz = "Europe/Berlin")
  )
})

test_that("Consecutive new moons are ~29.5 days apart", {
  posix_times <- as.POSIXct(linear_time(320:330, cal_time_lunar$month(1L)))
  gaps_days <- as.numeric(diff(posix_times), units = "days")
  expect_true(all(gaps_days > 29.2 & gaps_days < 29.9))
})

# Tests for time_is_determinate_at(): whether a time point is well-defined at the
# precision of a given granule (see also the indeterminacy handling exercised in
# test-components.R).

test_that("discrete time is indeterminate below its chronon", {
  # A discrete (integer) year has no determinate month.
  expect_false(time_is_determinate_at(year(2020L), cal_gregorian$month))
  # A coarser-or-equal granule is always determinate.
  expect_true(time_is_determinate_at(year(2020L), cal_gregorian$year))
  expect_true(time_is_determinate_at(yearmonth(as.Date("2020-02-01")), cal_gregorian$year))
})

test_that("granules of the same time unit are compared by size", {
  # A quarter, expressed as three months, cannot resolve the month within it.
  q <- linear_time(0:2, chronon = cal_gregorian$month(3L))
  expect_false(time_is_determinate_at(q, cal_gregorian$month(1L))[[1L]])
  expect_true(time_is_determinate_at(q, cal_gregorian$month(3L))[[1L]])
  expect_true(time_is_determinate_at(q, cal_gregorian$month(15L))[[1L]])

  # ... and so cannot complete one either.
  expect_identical(
    time_is_complete_at(q, cal_gregorian$month(1L)),
    rep(FALSE, 3L)
  )
})

test_that("continuous time resolves finer granules exactly", {
  # 0% through 2020 is 0% through January -> determinate.
  expect_true(time_is_determinate_at(year(2020), cal_gregorian$month))
})

test_that("granule accepts a generator or a sized unit", {
  expect_identical(
    time_is_determinate_at(year(2020L), cal_gregorian$month),
    time_is_determinate_at(year(2020L), cal_gregorian$month(1L))
  )
})

test_that("determinacy is computed per element of a mixtime", {
  mx <- c(year(2020L), yearmonth(as.Date("2020-02-01")), year(2019L))
  expect_identical(
    time_is_determinate_at(mx, cal_gregorian$month),
    c(FALSE, TRUE, FALSE)
  )
})

test_that("missing times give NA", {
  mx <- c(year(2020L), yearmonth(as.Date(NA)))
  expect_identical(time_is_determinate_at(mx, cal_gregorian$month), c(FALSE, NA))
})

# Tests for time_is_complete_at(): whether the observed vector fills the coarser
# granule that each element falls into.

test_that("an incomplete granule is FALSE for all its points", {
  # 2020 Jan : 2020 Oct does not complete the year (Nov, Dec missing).
  jan_oct <- yearmonth(as.Date("2020-01-01")) + 0:9
  expect_identical(
    time_is_complete_at(jan_oct, cal_gregorian$year),
    rep(FALSE, 10L)
  )
})

test_that("only completed granules are TRUE", {
  # 2020 is complete (Jan:Dec), 2021 is not (only Jan:Mar).
  jan20_mar21 <- yearmonth(as.Date("2020-01-01")) + 0:14
  expect_identical(
    time_is_complete_at(jan20_mar21, cal_gregorian$year),
    c(rep(TRUE, 12L), rep(FALSE, 3L))
  )
})

test_that("granule accepts a generator or a sized unit", {
  x <- yearmonth(as.Date("2020-01-01")) + 0:11
  expect_identical(
    time_is_complete_at(x, cal_gregorian$year),
    time_is_complete_at(x, cal_gregorian$year(1L))
  )
})

test_that("completeness handles unordered and duplicated points", {
  # Same twelve months, shuffled and with a duplicate, still complete.
  x <- yearmonth(as.Date("2020-01-01")) + c(11:0, 5L)
  expect_true(all(time_is_complete_at(x, cal_gregorian$year)))
  # Dropping one month leaves the year incomplete.
  x <- yearmonth(as.Date("2020-01-01")) + c(0:4, 6:11)
  expect_true(all(!time_is_complete_at(x, cal_gregorian$year)))
})

test_that("duplicated points are not double-counted towards completeness", {
  # Eleven distinct months (December missing) plus a duplicate of January: twelve
  # observations but only eleven distinct months, so the year is not complete.
  x <- yearmonth(as.Date("2020-01-01")) + c(0:10, 0L)
  expect_true(all(!time_is_complete_at(x, cal_gregorian$year)))
})

test_that("a granule equal to the chronon is completed by each point", {
  x <- yearmonth(as.Date("2020-01-01")) + 0:2
  expect_identical(
    time_is_complete_at(x, cal_gregorian$month),
    rep(TRUE, 3L)
  )
})

test_that("a granule finer than the chronon is never complete", {
  # A yearly point cannot complete a month.
  expect_identical(
    time_is_complete_at(year(2020L), cal_gregorian$month),
    FALSE
  )
})

test_that("variable cardinality (days in a month) is respected", {
  # All 29 days of February 2024 (a leap year) complete the month.
  feb <- as.Date("2024-02-01") + 0:28
  expect_true(all(time_is_complete_at(feb, cal_gregorian$month)))
  # 28 days do not complete a 29-day February.
  expect_true(all(!time_is_complete_at(feb[-1L], cal_gregorian$month)))
})

test_that("completeness is computed per granule across a mixtime", {
  mx <- c(
    yearmonth(as.Date("2020-01-01")) + 0:11, # complete 2020
    yearmonth(as.Date("2021-01-01")) + 0:2   # incomplete 2021
  )
  expect_identical(
    time_is_complete_at(mx, cal_gregorian$year),
    c(rep(TRUE, 12L), rep(FALSE, 3L))
  )
})

test_that("mixed-granularity mixtimes are not supported", {
  mx <- c(yearmonth(as.Date("2020-01-01")), as.Date("2020-01-01"))
  expect_error(
    time_is_complete_at(mx, cal_gregorian$year),
    "mixed-granularity"
  )
})

test_that("missing times give NA and do not count towards completeness", {
  x <- c(yearmonth(as.Date("2020-01-01")) + 0:11, yearmonth(as.Date(NA)))
  expect_identical(
    time_is_complete_at(x, cal_gregorian$year),
    c(rep(TRUE, 12L), NA)
  )
})

#' Describe a granule's label scheme
#'
#' Builds the function a [linear_labels()]/[cyclical_labels()] method is:
#' assign its result directly, e.g. `method(cyclical_labels, list(granule,
#' cycle)) <- label_scheme(...)`. Three levels, from plain to most irregular:
#'
#' 1. Plain numeric. No `vocab`: labels are `i + start`, optionally
#'    zero-padded to `width`. The default for most granules (minutes,
#'    seconds).
#' 2. Named, regular. `vocab` supplies the names (see [vocab_table()]);
#'    `start` positions raw index 0 in both the numeric and vocab-indexed
#'    rendering. Covers most named calendar units (months, weekdays).
#' 3. Named, irregular. `transform` overrides the plain `start` shift with
#'    hand-written `encode`/`decode`, for cycles where raw index and name
#'    aren't a constant offset apart (e.g. a leap month splitting one name
#'    into two). See "Irregular cycles" below.
#'
#' A scheme is only consulted for named (`label = TRUE`) rendering and its
#' inverse; numeric rendering is always the plain `i + start` shift (see
#' [linear_labels_format()]). Labels that aren't index-shaped at all (e.g.
#' "1BC") override the generics directly, documented in [linear_labels()]
#' instead.
#'
#' @param start Author-fixed. Raw index 0 displays as `start`, shared by
#'   numeric rendering and vocab indexing (e.g. `start = 1L` makes January,
#'   raw index 0, display as `"1"` and index `vocab()[[1]]`). A caller can't
#'   override this: get it wrong and dates are wrong, not just styled
#'   differently.
#' @param vocab `NULL` (numeric only), or a function `function(type = NULL,
#'   locale = NULL)` returning either the full named list of renderings for
#'   one locale (`type = NULL`), or one rendering (e.g.
#'   `vocab("abbreviated")`). [vocab_table()] builds this from a plain
#'   lookup table; see "Locale specific labels" below for how to specify labels.
#' @param transform `NULL` (plain `start` shift), or `list(encode =
#'   function(i, at = NULL) -> vocab index, decode = function(d, at = NULL)
#'   -> raw i)`, overriding the indexing used for named rendering only. Both
#'   functions are authored scalar (`if`/`return`) and vectorized
#'   internally. See "Irregular cycles" below.
#' @param width,locale Caller-overridable defaults for `cyc(granule, cycle,
#'   width = ...)`/`locale = ...` at format/parse time, not calendar facts.
#'   `width` zero-pads numeric rendering (e.g. `width = 2L` for `"02"`);
#'   `locale` selects which `vocab()` entry to use by default.
#'
#' @return A function `function(granule, cycle) list(start =, vocab =,
#'   transform =, width =, locale =)`, assignable directly as a
#'   [linear_labels()]/[cyclical_labels()] method. `granule`/`cycle` are only
#'   used for dispatch: a scheme describes a granule class, not the instance
#'   a method happens to be called with.
#'
#' @section Locale specific labels:
#' `vocab` is a plain closure, not a shared registry, so two calendar
#' packages can't collide on it. To add locale support to an existing
#' granule, redefine the `linear_labels()`/`cyclical_labels()` method and
#' call the previous one for anything you're not changing: the normal S7
#' rule that the last `method(...) <-` wins. That needs a hand-written
#' wrapper rather than assigning `label_scheme()`'s result directly, since
#' overriding one field means fetching the old scheme and changing it:
#'
#' ```r
#' scheme <- cyclical_labels(cal_gregorian$month(1L), cal_gregorian$year(1L))
#' method(cyclical_labels, list(cal_gregorian$month, cal_gregorian$year)) <-
#'   function(granule, cycle) {
#'     scheme$vocab <- function(type = NULL, locale = NULL) {
#'       if (identical(locale %||% "en-GB", "fr-FR")) {
#'         fr <- list(
#'           wide = c("janvier", "février", "mars", "avril", "mai", "juin",
#'                    "juillet", "août", "septembre", "octobre", "novembre", "décembre"),
#'           abbreviated = c("janv.", "févr.", "mars", "avr.", "mai", "juin",
#'                           "juil.", "août", "sept.", "oct.", "nov.", "déc.")
#'         )
#'         return(if (is.null(type)) fr else fr[[type]])
#'       }
#'       scheme$vocab(type, locale)
#'     }
#'     scheme
#'   }
#' ```
#'
#' Locale tags follow BCP 47 (`"en-GB"`, `"fr-FR"`), like ICU, `stringi` and
#' `clock`. Type names follow CLDR's wide/abbreviated/narrow, so CLDR data
#' (e.g. `stringi::stri_datetime_symbols()`) works with [vocab_table()]
#' unchanged. Other `type` values are allowed too, e.g. `"emoji"` for lunar
#' phases, picked at render time via [cyclical_labels_format()]'s `type`
#' argument.
#'
#' @section Irregular cycles:
#' `at` (in `transform`'s `encode`/`decode`, and in
#' [cyclical_labels_format()]/[cyclical_labels_parse()]) is the coarser
#' granule instance's raw position, not epoch-shifted, with `n = 1`. It
#' follows the same convention as [chronon_cardinality()]. It lets
#' `transform` pick the right name when that name depends on which cycle
#' instance a raw index falls in, e.g. the Hebrew calendar's leap year
#' splitting "Adar" into "Adar I"/"Adar II":
#'
#' ```r
#' month_names <- c("Tishrei", "Cheshvan", "Kislev", "Tevet", "Shevat",
#'                   "Adar", "Adar I", "Adar II",
#'                   "Nisan", "Iyar", "Sivan", "Tammuz", "Av", "Elul")
#'
#' method(cyclical_labels, list(cal_hebrew$month, cal_hebrew$year)) <- label_scheme(
#'   start = 1L,
#'   vocab = vocab_table(`en-GB` = list(wide = month_names, abbreviated = month_names)),
#'   transform = list(
#'     encode = function(i, at) {
#'       leap <- is_hebrew_leap_year(at)
#'       if (i <= 4L) return(i + 1L)
#'       if (!leap) return(if (i == 5L) 6L else i + 3L)
#'       if (i == 5L) return(7L)
#'       if (i == 6L) return(8L)
#'       i + 2L
#'     },
#'     decode = function(d, at) {
#'       leap <- is_hebrew_leap_year(at)
#'       if (d <= 5L) return(d - 1L)
#'       if (!leap) {
#'         if (d == 6L) return(5L)
#'         if (d %in% c(7L, 8L)) cli::cli_abort("'Adar I'/'Adar II' are not valid outside a leap year.")
#'         return(d - 3L)
#'       }
#'       if (d == 6L) cli::cli_abort("'Adar' is ambiguous in a leap year, use 'Adar I' or 'Adar II'.")
#'       if (d == 7L) return(5L)
#'       if (d == 8L) return(6L)
#'       d - 2L
#'     }
#'   )
#' )
#' ```
#'
#' The aborts in `decode()` matter: without them, "Adar II" typed in a
#' common year would fall through to the same offset as plain "Adar" and
#' silently parse to the wrong month. Only the label layer, which knows
#' both the text and `at`, can catch that. `transform`'s leap predicate must
#' also be the same one [chronon_cardinality()] uses for this pair; put it
#' in one shared helper.
#'
#' @seealso [linear_labels()]/[cyclical_labels()], the generics a scheme is
#'   returned from; [vocab_table()] for the common hand-listed name table
#'   case.
#' @export
label_scheme <- function(
  start = 0L,
  vocab = NULL,
  transform = NULL,
  width = NULL,
  locale = NULL
) {
  fields <- list(
    start = start,
    vocab = vocab,
    transform = transform,
    width = width,
    locale = locale
  )
  # granule/cycle are unused, only here so this can be assigned directly as
  # a linear_labels()/cyclical_labels() method. S7 forbids default values on
  # dispatch args, so this can't be called with no arguments either.
  function(granule, cycle) fields
}

#' A granule's label scheme
#'
#' `linear_labels()` and `cyclical_labels()` are S7 generics returning a
#' granule's label scheme: a plain list describing how its internal position
#' renders as text and parses back (`start`, `vocab`, `transform`, `width`,
#' `locale`; see [label_scheme()]).
#'
#' [linear_labels_format()]/[linear_labels_parse()] (and their cyclical
#' counterparts) do the actual rendering/parsing, with the default method
#' looking up the scheme and applying it. A granule with labels that are
#' incompatible with label schemes (e.g. 2BC, 1BC, 1, 2, ...) registers a method
#' directly on those generics instead, skipping the scheme system. See
#' "Overriding the generics directly" below.
#'
#' @param granule A time granule instance (e.g. `month(1L)`).
#' @param cycle For `cyclical_labels()`, the coarser granule instance this
#'   cycle repeats within (e.g. `year(1L)`).
#' @param ... Unused; only present for S7 dispatch.
#'
#' @return A plain list, as constructed by [label_scheme()].
#'
#' @section Overriding the format and parsing methods:
#' `linear_labels_format()`/`linear_labels_parse()` are ordinary S7
#' generics with a default method, not the only place formatting logic can
#' live. So escaping the scheme system for one granule is just registering
#' a method directly:
#'
#' ```r
#' method(linear_labels_format, cal_gregorian$year) <- function(granule, i, ...) {
#'   ifelse(i <= 0L, paste0(-i + 1L, "BC"), i)
#' }
#' method(linear_labels_parse, cal_gregorian$year) <- function(granule, ...) {
#'   list(
#'     pattern = "\\d+(?:BC)?",
#'     decode = function(text, at = NULL) {
#'       bc <- grepl("BC$", text)
#'       n <- as.integer(sub("BC$", "", text))
#'       ifelse(bc, 1L - n, n) - chronon_epoch(granule)
#'     }
#'   )
#' }
#' ```
#'
#' `decode()` needs `- chronon_epoch(granule)` because this method bypasses
#' the default, which normally does that step (see [linear_labels_format()]'s
#' "Epoch shift" section).
#'
#' The `...` on `_format()` is not just style: `format()` forwards whatever
#' [component_helpers]' `lin()`/`cyc()` were called with (`label`,
#' `abbreviate`, `type`, `width`, `locale`, ...) as named arguments. A
#' method without `...` errors with "unused argument" as soon as a caller
#' passes something it doesn't declare.
#'
#' Registering only `linear_labels_format()` is fine: that's the same "no
#' parsing yet" state a fresh granule starts in. But registering it without
#' `linear_labels_parse()` leaves parsing silently wrong instead of just
#' missing: text like `"1BC"` falls through to whatever the granule's
#' scheme says, usually plain numeric decoding, which doesn't understand
#' "BC" at all.
#'
#' @seealso [label_scheme()] for the fields a method returns;
#'   [linear_labels_format()]/[cyclical_labels_format()] for the generics
#'   that interpret it; [vocab_table()] for the common hand-listed name
#'   table case.
#'
#' @examples
#' # A scheme for months of the year, using R's localised month names.
#' S7::method(cyclical_labels, list(cal_gregorian$month, cal_gregorian$year)) <- label_scheme(
#'   start = 1L, width = 2L,
#'   vocab = vocab_table(`en-GB` = list(wide = month.name, abbreviated = month.abb))
#' )
#'
#' @rdname label_scheme
#' @export
linear_labels <- S7::new_generic("linear_labels", "granule")

#' @rdname label_scheme
#' @export
cyclical_labels <- S7::new_generic("cyclical_labels", c("granule", "cycle"))

method(linear_labels, mt_unit) <- label_scheme(start = 0L)

# start = 0L here because this fallback must leave raw i unchanged,
# matching the old behaviour before schemes existed. start = 1L is a good
# default when an author declares a scheme, not when nobody has customised
# anything.
method(cyclical_labels, list(mt_unit, mt_unit)) <- label_scheme(start = 0L)

#' Render and parse a granule's labels
#'
#' The generics turning a granule's internal position into display text
#' (`linear_labels_format()` for a non-repeating position, e.g. the year;
#' `cyclical_labels_format()` for a position within a larger cycle, e.g. the
#' month within the year) and back (`linear_labels_parse()`/
#' `cyclical_labels_parse()`).
#'
#' Each has one default method (on the base `mt_unit` class) that looks up
#' the granule's scheme via [linear_labels()]/[cyclical_labels()] and
#' applies it (see [label_scheme()] for the fields). Calendar authors
#' usually declare a scheme instead of writing a method here directly.
#' Writing one directly overrides the generics for labels that aren't
#' index-shaped at all, see [linear_labels()]'s "Overriding the generics
#' directly" section.
#'
#' @param granule A time granule object representing the granule (e.g.
#'   `month(1L)`).
#' @param cycle A time granule object representing the cycle (e.g.
#'   `year(1L)`).
#' @param ... Passed on to the method. The default method (see below) takes:
#'   - `i`: Integer vector: the position along the linear axis, or within
#'     the cycle for `cyclical_labels_format()`.
#'   - `at`: The linear position of the cycle granule, letting a method
#'     produce labels specific to that cycle instance (mainly for irregular
#'     cycles, e.g. a leap month). Same convention as `at` in
#'     [chronon_cardinality()].
#'   - `label`: If `TRUE`, return named labels (e.g. "February"). If
#'     `FALSE`, return the numeric position as character.
#'   - `abbreviate`: If `TRUE`, return abbreviated labels (e.g. "Feb"). If
#'     `FALSE`, return full labels (e.g. "February").
#'   - `type`: Overrides `abbreviate`'s abbreviated/wide choice of
#'     [vocab_table()] type outright, e.g. `type = "emoji"`. `NULL` (default)
#'     defers to `abbreviate`.
#'   - `width`,`locale`: Call-time overrides of the scheme's `width`/`locale`
#'     defaults (see [label_scheme()]); `NULL` defers to the scheme.
#'
#' @return Character vector of labels for the time point.
#'
#' @section Epoch shift:
#' `linear_labels_format()`'s `i` arrives already epoch-shifted for display
#' (`chronon_parts()` adds [chronon_epoch()] before calling it, e.g. a
#' Gregorian year's `i` is `2024`, not `54`). The default
#' `linear_labels_parse()` method shifts `decode()`'s output back, by
#' subtracting [chronon_epoch()] after the scheme's own decode runs. A
#' hand-written override method skips this and must do the same
#' subtraction itself if the granule has a non-zero epoch; see the example
#' in [linear_labels()].
#'
#' @seealso [linear_labels()]/[cyclical_labels()] for the authoring
#'   interface these generics' default methods interpret.
#'
#' @examples
#' # Labels for years on a linear axis
#' with(cal_gregorian, linear_labels_format(year(1L), 2020:2025))
#'
#' # Labels for months in a year
#' with(cal_gregorian, cyclical_labels_format(month(1L), year(1L), 0:11))
#'
#' @rdname label_format
#' @export
linear_labels_format <- S7::new_generic("linear_labels_format", "granule")

method(linear_labels_format, mt_unit) <- function(
  granule,
  i,
  label = FALSE,
  abbreviate = TRUE,
  type = NULL,
  width = NULL,
  locale = NULL,
  ...
) {
  scheme <- linear_labels(granule)
  label_render(
    i,
    at = NULL,
    label = label,
    abbreviate = abbreviate,
    type = type,
    width = width %||% scheme$width,
    locale = locale %||% scheme$locale,
    start = scheme$start,
    vocab = scheme$vocab,
    transform = scheme$transform,
    # No vocab: fall back to the pre-scheme "<i><abbr>" style (e.g. "5Y").
    fallback_label = function(i) {
      paste0(i, if (abbreviate) time_unit_abbr(granule) else time_unit_plural(granule, i))
    }
  )
}

#' @rdname label_format
#' @export
cyclical_labels_format <- S7::new_generic(
  "cyclical_labels_format",
  c("granule", "cycle")
)

method(cyclical_labels_format, list(mt_unit, mt_unit)) <- function(
  granule,
  cycle,
  i,
  at = NULL,
  label = FALSE,
  abbreviate = TRUE,
  type = NULL,
  width = NULL,
  locale = NULL,
  ...
) {
  scheme <- cyclical_labels(granule, cycle)
  label_render(
    i,
    at = at,
    label = label,
    abbreviate = abbreviate,
    type = type,
    width = width %||% scheme$width,
    locale = locale %||% scheme$locale,
    start = scheme$start,
    vocab = scheme$vocab,
    transform = scheme$transform,
    # Same fallback as linear, "<abbr><i>" order (e.g. "M5" not "5M").
    fallback_label = function(i) {
      paste0(if (abbreviate) time_unit_abbr(granule) else time_unit_plural(granule, i), i)
    }
  )
}

#' @rdname label_format
#' @export
linear_labels_parse <- S7::new_generic("linear_labels_parse", "granule")

method(linear_labels_parse, mt_unit) <- function(granule, ...) {
  scheme <- linear_labels(granule)
  spec <- label_parse_spec(
    start = scheme$start,
    vocab = scheme$vocab,
    transform = scheme$transform
  )
  # Undo chronon_parts()'s epoch shift, to match time_compose()'s raw convention.
  spec$decode <- local({
    inner <- spec$decode
    function(text, at = NULL) inner(text, at) - chronon_epoch(granule)
  })
  spec
}

#' @rdname label_format
#' @export
cyclical_labels_parse <- S7::new_generic(
  "cyclical_labels_parse",
  c("granule", "cycle")
)

method(cyclical_labels_parse, list(mt_unit, mt_unit)) <- function(
  granule,
  cycle,
  ...
) {
  scheme <- cyclical_labels(granule, cycle)
  label_parse_spec(
    start = scheme$start,
    vocab = scheme$vocab,
    transform = scheme$transform
  )
}

# Wraps a scalar transform (written with if/return) so it can be called
# with vectors, one at value per element. force(f) because callers
# reassign the result onto the same name; without forcing, f would stay
# lazy and recurse into itself.
vectorize_over_at <- function(f) {
  force(f)
  function(x, at = NULL) {
    n <- length(x)
    at <- if (is.null(at)) vector("list", n) else as.list(rep_len(at, n))
    unlist(.mapply(f, list(as.list(x), at), NULL))
  }
}

# Shared rendering logic behind the *_format() default methods. Numeric
# rendering never uses transform or locale. Only label = TRUE rendering
# needs vocab (and transform$encode for irregular cycles). With no vocab,
# fallback_label(i) runs instead.
label_render <- function(
  i,
  at,
  label,
  abbreviate,
  type,
  width,
  locale,
  start,
  vocab,
  transform,
  fallback_label
) {
  if (!label) {
    idx <- i + start
    if (!is.null(width)) {
      return(sprintf(paste0("%0", width, "d"), idx))
    }
    return(as.character(idx))
  }
  if (is.null(vocab)) {
    return(fallback_label(i))
  }
  encode <- if (is.null(transform)) {
    function(i, at = NULL) i + start
  } else {
    vectorize_over_at(transform$encode)
  }

  # type lets a caller pick a vocab_table() type other than wide/abbreviated
  # (e.g. "emoji").
  type <- type %||% if (abbreviate) "abbreviated" else "wide"
  names <- vocab(type, locale)
  unname(names[encode(i, at)])
}

# Shared spec-builder behind the *_parse() default methods (see
# linear_labels_format()'s docs for the pattern/decode shape returned).
# Mirrors label_render(): a numeric token always decodes as d - start and
# skips transform; only a vocab-name token goes through transform$decode.
label_parse_spec <- function(start, vocab, transform) {
  if (is.null(vocab)) {
    return(list(
      pattern = "\\d+",
      decode = function(text, at = NULL) as.integer(text) - start
    ))
  }

  # vocab() with no type returns every rendering for the default locale.
  # Types are parallel-indexed, so a match in any one gives the same vocab
  # index.
  vocab_all <- vocab()
  all_names <- unlist(vocab_all, use.names = FALSE)
  uniq_names <- unique(all_names)
  # Escape regex metacharacters (e.g. the "." in French "janv.") via a
  # bracket expression + backreference before matching them literally.
  escaped_names <- gsub(
    "([.|()\\^{}+$*?\\[\\]\\\\])",
    "\\\\\\1",
    uniq_names[order(-nchar(uniq_names))],
    perl = TRUE
  )
  pattern <- paste(c(escaped_names, "\\d+"), collapse = "|")

  decode <- if (is.null(transform)) {
    function(d, at = NULL) d - start
  } else {
    vectorize_over_at(transform$decode)
  }

  list(
    pattern = pattern,
    decode = function(text, at = NULL) {
      out <- integer(length(text))
      numeric <- grepl("^\\d+$", text)
      if (any(numeric)) {
        out[numeric] <- as.integer(text[numeric]) - start
      }
      if (any(!numeric)) {
        text_lower <- tolower(text[!numeric])
        idx_by_type <- lapply(vocab_all, function(v) match(text_lower, tolower(v)))
        vocab_idx <- Reduce(function(a, b) ifelse(is.na(a), b, a), idx_by_type)
        if (anyNA(vocab_idx)) {
          cli::cli_abort(
            "Unrecognised label {.val {text[!numeric][is.na(vocab_idx)][1L]}}."
          )
        }
        out[!numeric] <- decode(vocab_idx, at[!numeric])
      }
      out
    }
  )
}

#' Build a `vocab` function from a plain name table
#'
#' The common case for a [label_scheme()] `vocab` argument: a hand-listed
#' table of names, one entry per locale, each a named list of renderings. By
#' convention these follow CLDR's wide/abbreviated/narrow, but `type` can be
#' anything. `vocab_table()` wraps it in the `function(type = NULL, locale =
#' NULL)` shape `vocab` requires.
#'
#' For vocab backed by an external i18n source instead of a hand-listed
#' table, write that `function(type, locale)` directly instead. Same shape,
#' no list needed.
#'
#' @param ... Named entries, one per locale (e.g. `` `en-GB` = list(wide =
#'   month.name, abbreviated = month.abb)``). Locale tags follow BCP 47.
#' @param default_locale The locale returned when `locale` is unsupplied (or
#'   explicitly `NULL`, meaning "caller didn't ask", since
#'   `time_parse()`/`format()` always pass some value for `locale`).
#'
#' @return A function `function(type = NULL, locale = NULL)`, suitable as
#'   the `vocab` field of a [label_scheme()].
#'
#' @examples
#' month_vocab <- vocab_table(`en-GB` = list(wide = month.name, abbreviated = month.abb))
#' month_vocab("abbreviated")
#' month_vocab(locale = "en-GB")
#'
#' @seealso [label_scheme()]
#' @export
vocab_table <- function(..., default_locale = "en-GB") {
  table <- list(...)
  function(type = NULL, locale = NULL) {
    locale <- locale %||% default_locale
    entry <- table[[locale]] %||%
      cli::cli_abort("Unsupported locale: {.val {locale}}")
    if (is.null(type)) {
      return(entry)
    }
    entry[[type]] %||%
      cli::cli_abort(
        "Unsupported type {.val {type}} for locale {.val {locale}}"
      )
  }
}

#' Gregorian time unit classes
#'
#' Time unit constructors for the Gregorian calendar system. These units can be
#' used with [linear_time()] to create custom time representations.
#'
#' @format A civil-based calendar containing Gregorian time units.
#'
#' @details
#' The following time units are available in the Gregorian calendar (`cal_gregorian$`).
#'
#' - `year()`: Year unit
#' - `quarter()`: Quarter (3-month period) unit
#' - `month()`: Month unit
#' - `day()`: Day unit
#' - `hour()`: Hour unit
#' - `minute()`: Minute unit
#' - `second()`: Second unit
#' - `millisecond()`: Millisecond unit
#'
#' These units form a hierarchy where conversions between adjacent units follow
#' the Gregorian calendar rules. For units that don't have a fixed relationship
#' (e.g., months to days), the conversion requires a time context.
#'
#' @return An S3 list of class `c("cal_gregorian", "mt_calendar")` containing
#'   the named time unit classes of the Gregorian calendar. Each unit is
#'   accessible via `$` notation and calling it with a step size produces a
#'   time granule (e.g., 1 month granule as `cal_gregorian$month(1L)`).
#'
#' @seealso [linear_time()] for creating linear time points.
#'
#' @examples
#' # Create a custom time representation using Gregorian units
#' linear_time(
#'   Sys.time(),
#'   chronon = hour(1L)
#' )
#'
#' @name calendar_gregorian
#' @export
cal_gregorian <- new_calendar(
  year = new_class("tu_year", parent = mt_tz_unit),
  quarter = new_class("tu_quarter", parent = mt_tz_unit),
  month = new_class("tu_month", parent = mt_tz_unit),
  inherit = cal_time_civil,
  class = "cal_gregorian"
)

# Time unit labels
method(time_unit_full, cal_gregorian$year) <- function(x) "year{?/s}"
method(time_unit_abbr, cal_gregorian$year) <- function(x) "Y"
method(time_unit_full, cal_gregorian$quarter) <- function(x) "quarter{?/s}"
method(time_unit_abbr, cal_gregorian$quarter) <- function(x) "Q"
method(time_unit_full, cal_gregorian$month) <- function(x) "month{?/s}"
method(time_unit_abbr, cal_gregorian$month) <- function(x) "M"

# Epoch for years
method(chronon_epoch, cal_gregorian$year) <- function(x) 1970L

# Default formats
method(chronon_format_linear, list(cal_gregorian$year, class_any)) <- 
  function(x, cal) "{lin(year)}"
method(chronon_format_linear, list(cal_gregorian$quarter, class_any)) <- 
  function(x, cal) "{lin(year)} Q{cyc(quarter,year)}"
method(chronon_format_linear, list(cal_gregorian$month, class_any)) <- 
  function(x, cal) "{lin(year)} {cyc(month,year,label=TRUE,abbreviate=TRUE)}"
method(chronon_format_linear,
  list(cal_gregorian$day, S7::new_S3_class("cal_gregorian"))) <- 
  function(x, cal) "{lin(year)}-{cyc(month,year)}-{cyc(day,month)}"

method(chronon_format_cyclical, 
  list(cal_gregorian$month, cal_gregorian$year)) <-
  function(x, y) "{cyc(month,year,label=TRUE,abbreviate=TRUE)}"

# Candidate parsing formats
method(chronon_parse_linear, list(cal_gregorian$month, class_any)) <-
  function(x, cal) {
    parse_format(
      # e.g. "2024-03", "2024 Mar", "2024, March"
      "{lin(year)}[-/,\\s]+{cyc(month,year)}",
      # e.g. "03-2024", "Mar 2024", "March, 2024"
      "{cyc(month,year)}[-/,\\s]+{lin(year)}",
      regex = TRUE
    )
  }
method(chronon_parse_linear, list(cal_gregorian$quarter, class_any)) <-
  function(x, cal) {
    parse_format(
      # e.g. "2024 Q1", "2024-Q1", "2024/1"
      "{lin(year)}[-/,\\s]+[Qq]?{cyc(quarter,year)}",
      # e.g. "Q1 2024", "Q1, 2024", "1 2024"
      "[Qq]?{cyc(quarter,year)}[-/,\\s]+{lin(year)}",
      regex = TRUE
    )
  }
method(
  chronon_parse_linear,
  list(cal_gregorian$day, S7::new_S3_class("cal_gregorian"))
) <- function(x, cal) {
  parse_format(
    # e.g. "2024-03-15", "2024 Mar 15", "2024, March, 15th"
    "{lin(year)}[-/,\\s]+{cyc(month,year)}[-/,\\s]+{cyc(day,month)}(?:st|nd|rd|th)?",
    # e.g. "15-03-2024", "15th Mar 2024", "15 March, 2024"
    "{cyc(day,month)}(?:st|nd|rd|th)?[-/,\\s]+{cyc(month,year)}[-/,\\s]+{lin(year)}",
    # e.g. "03-15-2024", "Mar 15th, 2024", "March 15, 2024"
    "{cyc(month,year)}[-/,\\s]+{cyc(day,month)}(?:st|nd|rd|th)?[-/,\\s]+{lin(year)}",
    regex = TRUE
  )
}
method(chronon_parse_cyclical, list(cal_gregorian$month, cal_gregorian$year)) <-
  function(x, y) {
    # e.g. "3", "Mar", "March"
    parse_format("{cyc(month,year)}")
  }

### Calendar algebra methods for Gregorian time units
method(
  chronon_cardinality_fixed,
  list(cal_gregorian$quarter, cal_gregorian$year)
) <- function(x, y) {
  4L
}
method(
  chronon_cardinality_fixed,
  list(cal_gregorian$month, cal_gregorian$year)
) <- function(x, y) {
  12L
}
method(
  chronon_cardinality,
  list(cal_gregorian$day, cal_gregorian$year)
) <- function(x, y, at = NULL) {
  if (is.null(at)) {
    stop(
      "The number of days in a year requires the time context `at`.",
      call. = FALSE
    )
  }
  if (y@n != 1L) {
    cli::cli_abort(
      "The cardinality of days in multi-year chronons is not yet supported."
    )
  }
  (is_leap_year(1970L + as.integer(at)) + 365L) / x@n
}
method(
  chronon_cardinality_fixed,
  list(cal_gregorian$month, cal_gregorian$quarter)
) <- function(x, y) {
  3L
}

monthdays <- c(31L, 28L, 31L, 30L, 31L, 30L, 31L, 31L, 30L, 31L, 30L, 31L)

is_leap_year <- function(year) {
  year <- floor(year)
  (year %% 4L == 0L & year %% 100L != 0L) | (year %% 400L == 0L)
}

# Days since epoch (1970-01-01) to the 1st of the Gregorian month
# `month_index` months after epoch (0-indexed: 0 is 1970-01, -1 is 1969-12,
# ...). Shared by `chronon_cardinality()` and `chronon_divmod()` below so
# that both agree on where a multi-month window starts and ends.
month_start_days <- function(month_index) {
  year <- fdiv(month_index, 12L) + 1970L
  month <- month_index - 12L * (year - 1970L) + 1L
  ly <- as.integer(is_leap_year(year))

  # Years since 1970
  365L *
    (year - 1970L) +
    # Leap days since 1970
    fdiv(year - 1968L, 4L) -
    fdiv(year - 1900L, 100L) +
    fdiv(year - 1600L, 400L) +
    # Days this year before this month
    fdiv(367L * month - 362L, 12L) +
    (month > 2L) * (-2L + ly) -
    ly
}

method(
  chronon_cardinality,
  list(cal_gregorian$day, cal_gregorian$month)
) <- function(x, y, at = NULL) {
  if (is.null(at)) {
    stop(
      "The number of days in a month requires the time context `at`.",
      call. = FALSE
    )
  }

  n_months <- y@n
  at <- as.integer(at)

  # `at` is a block index in `y`-month units; its window covers the
  # `abs(n_months)` calendar months starting at absolute month `at *
  # n_months`, regardless of `n_months`'s sign.
  start_idx <- at * n_months
  md <- month_start_days(start_idx + abs(n_months)) - month_start_days(start_idx)

  md / x@n
}

### Chronon casting between Gregorian time granules
method(
  chronon_divmod,
  list(cal_gregorian$day, cal_gregorian$month)
) <- function(from, to, x) {
  # Scale `x` to be 1 day increments
  x_scale <- from@n
  x <- x_scale * x
  x_int <- floor(x)
  x_frac <- x - x_int

  # Shift to days since 0000-03-01 (algorithm anchor)
  z <- x_int + 719468L

  # (day) -> (year, month, day) arithmetic
  era <- (z >= 0L) * fdiv(z, 146097L) + (z < 0L) * fdiv(z - 146096L, 146097L)
  doe <- z - era * 146097L # day-of-era [0, 146096]
  yoe <- fdiv(
    doe - fdiv(doe, 1460L) + fdiv(doe, 36524L) - fdiv(doe, 146096L),
    365L
  )
  doy <- doe - (365L * yoe + fdiv(yoe, 4L) - fdiv(yoe, 100L) + fdiv(yoe, 400L))
  mp <- fdiv(5L * doy + 2L, 153L) # month prime [0, 11]
  day <- doy - fdiv(153L * mp + 2L, 5L) # day [0, 30]
  month <- mp + 3L - 12L * (mp >= 10L) # month [1, 12]
  year <- yoe + era * 400L + (month <= 2L) # year (proleptic Gregorian)

  # Absolute month index of `x` (0-indexed months since epoch)
  res <- (year - 1970L) * 12L + month - 1L

  # Block index: the integer `div` whose window [div*n, div*n + |n|)
  # contains `res`. The window always runs forward from `div * n`, so for
  # n > 0 that's floor(res / n); for n < 0 it's -floor(res / -n).
  n <- to@n
  div <- if (n > 0L) fdiv(res, n) else -fdiv(res, -n)

  # Day offset from the window's start to `x`.
  list(
    div = div,
    mod = (x_int - month_start_days(div * n) + x_frac) / x_scale
  )
}
method(
  chronon_divmod,
  list(cal_gregorian$month, cal_gregorian$day)
) <- function(from, to, x) {
  # Convert to months since epoch
  x <- from@n * x

  # Separate integer and fractional parts to correctly handle fractional months
  x_int <- floor(x)
  x_frac <- x - x_int

  year <- fdiv(x_int, 12L) + 1970L
  ly <- as.integer(is_leap_year(year))
  month <- x_int - 12L * (year - 1970L) + 1L

  # Start of the month in days since epoch
  result <-
    # Years since 1970
    365 *
    (year - 1970) +
    # Leap days since 1970
    fdiv(year - 1968, 4) -
    fdiv(year - 1900, 100) +
    fdiv(year - 1600, 400) +
    # Days this year before this month
    fdiv(367 * month - 362, 12) +
    (month > 2) * (-2 + ly) -
    ly

  # Convert fractional months to fractional days using days in current month
  days_in_month <- monthdays[month] + (month == 2L) * ly
  result <- result + x_frac * days_in_month

  # Scale by `to` day chronons
  result <- result / to@n

  list(
    div = result,
    mod = 0L
  )
}


method(chronon_divmod, list(cal_gregorian$day, cal_gregorian$year)) <- function(
  from,
  to,
  x
) {
  # Modulo arithmetic to convert from days to years
  if (to@n != 1L) {
    stop(
      "Converting to non-year chronons from days not yet supported",
      call. = FALSE
    )
  }

  # Scale `x` to be 1 day increments
  x_scale <- from@n
  x <- x_scale * x
  x_int <- floor(x)
  x_frac <- x - x_int

  # Shift to days since 0000-03-01 (algorithm anchor)
  z <- x_int + 719468L

  era <- (z >= 0L) * fdiv(z, 146097L) + (z < 0L) * fdiv(z - 146096L, 146097L)
  doe <- z - era * 146097L # day-of-era [0, 146096]
  yoe <- fdiv(
    doe - fdiv(doe, 1460L) + fdiv(doe, 36524L) - fdiv(doe, 146096L),
    365L
  )
  doy <- doe - (365L * yoe + fdiv(yoe, 4L) - fdiv(yoe, 100L) + fdiv(yoe, 400L))
  mp <- fdiv(5L * doy + 2L, 153L) # month prime [0, 11]
  month <- mp + 3L - 12L * (mp >= 10L) # month [1, 12]
  year <- yoe + era * 400L + (month <= 2L) # year (proleptic Gregorian)

  ly <- is_leap_year(year)
  yday <- (doy + 59 + ly) %% (365L + ly)

  list(
    div = year - 1970L,
    mod = (yday + x_frac) / x_scale
  )
}
method(chronon_divmod, list(cal_gregorian$year, cal_gregorian$day)) <- function(
  from,
  to,
  x
) {
  # Convert to years since epoch
  x <- from@n * x

  floor_int <- function(x) as.integer(floor(x))

  # Days since epoch
  d <- 365L *
    x +
    # Leap days since epoch
    floor_int((x + 1L) / 4) -
    # Subtract century years (not leap years)
    floor_int((x + 69L) / 100) +
    # Add back quad-century years (leap years)
    floor_int((x + 369L) / 400L) +
    # Add fractional part of leap year
    is_leap_year(x + 1970L) * (x - floor_int(x))

  # Scale by `to` day chronons
  d <- d / to@n

  list(
    div = d,
    mod = 0L
  )
}

## Time labels
bc_ad_format <- function(granule, i, ...) {
  ifelse(i <= 0L, paste0(-i + 1L, "BC"), i)
}
bc_ad_parse <- function(granule, ...) {
  list(
    pattern = "\\d+(?:BC)?",
    decode = function(text, at = NULL) {
      bc <- grepl("BC$", text)
      n <- as.integer(sub("BC$", "", text))
      # This override bypasses linear_labels_parse()'s default method, so
      # (unlike a linear_labels() scheme) it must undo the epoch shift
      # chronon_parts() applies for display itself -- see
      # linear_labels_format()'s "Epoch shift" section.
      ifelse(bc, 1L - n, n) - chronon_epoch(granule)
    }
  )
}
method(linear_labels_format, cal_gregorian$year) <- bc_ad_format
method(linear_labels_parse, cal_gregorian$year) <- bc_ad_parse

### Cyclical labels for Gregorian time granules
method(cyclical_labels, list(cal_gregorian$quarter, S7::class_any)) <- label_scheme(start = 1L)
method(cyclical_labels, list(cal_gregorian$month, cal_gregorian$year)) <- label_scheme(
  start = 1L,
  width = 2L,
  vocab = vocab_table(`en-GB` = list(wide = month.name, abbreviated = month.abb))
)

#' Obtain the cycle of a time object
#' 
#' This S7 generic function extracts the cycle (the cyclical time granule) from
#' a time object, such as cyclical time representations.
#' 
#' @param x A time object (e.g., [base::Date], [base::POSIXct], [linear_time()], etc.)
#' @param ... Additional arguments for methods.
#' 
#' @return A time [duration()] object representing the cycle of each value
#'   (e.g. `weeks(1L)`), or `NA` if the object has no cyclical component.
#' 
#' @examples
#' 
#' # Non-cyclical objects return NA
#' time_cycle(Sys.Date())
#' 
#' # The cycle of a cyclical time object
#' time_cycle(month_of_year(Sys.Date()))
#' 
#' @export
time_cycle <- S7::new_generic("time_cycle", "x")

S7::method(time_cycle, S7::class_any) <- function(x) {
  duration(rep(NA_integer_, length(x)), mt_unit(1L))
}

S7::method(time_cycle, class_mixtime) <- function(x) {
  x@x <- lapply(x@x, function(x) time_cycle(x)@x[[1L]])
  x
}

S7::method(time_cycle, mt_time) <- function(x) {
  if (is.null(attr(x, "cycle"))) {
    return(duration(NA_real_, mt_unit(1L)))
  }
  
  duration(rep(1L, length(x)), attr(x, "cycle"))
}

# {hms} time class
S7::method(time_cycle, S7::new_S3_class("hms")) <- function(x) {
  duration(rep(1L, length(x)), chronon = cal_gregorian$day(1L))
}

# Coercion with `as.*()` family.
#' @export
S7::method(as.character, mt_time) <- function(x, ...) vctrs::vec_cast(x, character())
#' @export
S7::method(as.integer, mt_time) <- function(x, ...) vctrs::vec_cast(x, integer())
#' @export
S7::method(as.double, mt_time) <- function(x, ...) vctrs::vec_cast(x, double())
#' @export
S7::method(as.Date, mt_linear) <- function(x, ...) vctrs::vec_cast(x, vctrs::new_date())
#' @export
S7::method(as.POSIXct, mt_linear) <- function(x, tz = "", ...) {
  vctrs::vec_cast(x, vctrs::new_datetime(tzone = tz))
}

#' Convert a time class into a mixtime
#'
#' Coerces a time object (e.g. `Date`, `POSIXct`, `yearmonth`) to a `mixtime`
#' vector using [vctrs::vec_cast()]. The chronon and cycle are inferred from
#' `x` via [time_chronon()] and [time_cycle()].
#'
#' @param x A time value to convert to a `mixtime`. Any time class with a defined `time_chronon()` method can be converted (e.g. `Date`, `POSIXct`, `yearmonth`, etc.).
#' @param ... Additional arguments passed to the underlying [vec_cast()] method.
#'
#' @return A `mixtime` object corresponding to `x`.
#'
#' @seealso [mixtime()] for constructing a `mixtime` directly from data,
#'   [is_mixtime()] for testing if an object is a `mixtime`.
#'
#' @examples
#' as_mixtime(Sys.Date())
#' as_mixtime(Sys.time())
#'
#' @export
as_mixtime <- function(x, ...) {
  vec_cast(x, new_mixtime())
}

#' Extract locations from an object
#'
#' Generic function to extract the location from objects that have location information.
#' 
#' @param x An object with location information.
#' @param ... Additional arguments passed to methods.
#'
#' @return A numeric value representing the location (e.g., longitude, latitude, etc.).
#'
#' @examples
#' t <- linear_time(
#'   1:3, 
#'   cal_time_solar$day(1L, lat = -37.8136, lon = 144.9631)
#' )
#' 
#' loc_longitude(t)
#' loc_latitude(t)
#' loc_altitude(t)
#' 
#' @name mixtime_location
#' @export
loc_latitude <- S7::new_generic("loc_latitude", "x")
S7::method(loc_latitude, mt_loc_unit) <- function(x) x@lat
S7::method(loc_latitude, mt_time) <- function(x) {
  rep_len(loc_latitude(attr(x, "chronon")), length(x))
}
S7::method(loc_latitude, class_mixtime) <- function(x) {
  as.numeric(vecvec::vecvec_apply(x, loc_latitude))
}
S7::method(loc_latitude, S7::class_any) <- function(x) NA_real_

#' @name mixtime_location
#' @export
loc_longitude <- S7::new_generic("loc_longitude", "x")
S7::method(loc_longitude, mt_loc_unit) <- function(x) x@lon
S7::method(loc_longitude, mt_time) <- function(x) {
  rep_len(loc_longitude(attr(x, "chronon")), length(x))
}
S7::method(loc_longitude, class_mixtime) <- function(x) {
  as.numeric(vecvec::vecvec_apply(x, loc_longitude))
}
S7::method(loc_longitude, S7::class_any) <- function(x) NA_real_

#' @name mixtime_location
#' @export
loc_altitude <- S7::new_generic("loc_altitude", "x")
S7::method(loc_altitude, mt_loc_unit) <- function(x) x@alt
S7::method(loc_altitude, mt_time) <- function(x) {
  rep_len(loc_altitude(attr(x, "chronon")), length(x))
}
S7::method(loc_altitude, class_mixtime) <- function(x) {
  as.numeric(vecvec::vecvec_apply(x, loc_altitude))
}
S7::method(loc_altitude, S7::class_any) <- function(x) NA_real_

test_that("index_valid returns TRUE for mixtime objects", {
  expect_true(tsibble::index_valid(yearmonth(1:3)))
  expect_true(tsibble::index_valid(c(yearmonth(1:3), year(2000:2002))))
})

test_that("interval_pull works for a single-chronon mixtime", {
  intvl <- tsibble::interval_pull(yearmonth(360L))
  expect_s3_class(intvl, "interval")
  expect_equal(format(intvl), "1M")
})

test_that("interval_pull works for a mixed-chronon mixtime", {
  # This was the original failing case: @n accessed on a duration wrapper
  intvl <- tsibble::interval_pull(c(yearmonth(360L), year(2000L)))
  expect_s3_class(intvl, "interval")
  expect_equal(format(intvl), "1M, 1Y")
})

test_that("yearweek creation from integers works", {
  # Test basic creation
  yw <- yearweek(0:52)
  expect_length(yw, 53)
  
  # Test single value
  yw_single <- yearweek(1)
  expect_length(yw_single, 1)
})

test_that("yearweek creation from dates works", {
  # Test with multiple dates
  dates <- as.Date(c("2023-01-01", "2023-06-15", "2023-12-31"))
  yw <- yearweek(dates)
  expect_length(yw, 3)
})

test_that("yearweek handles edge cases around year boundaries", {
  # December 31, 2024 is a Tuesday, belongs to 2025-W1
  expect_equal(format(yearweek(as.Date("2024-12-31"))), "2025 W01")
  
  # January 1, 2024 is a Monday, starts 2024-W1
  expect_equal(format(yearweek(as.Date("2024-01-01"))), "2024 W01")
  
  # January 1, 2023 is a Sunday, belongs to 2022-W52
  expect_equal(format(yearweek(as.Date("2023-01-01"))), "2022 W52")
  
  # January 2, 2023 is a Monday, starts 2023-W1
  expect_equal(format(yearweek(as.Date("2023-01-02"))), "2023 W01")
})

test_that("yearweek handles years with 53 weeks", {
  # 2020 had 53 weeks (leap year starting on Wednesday)
  expect_equal(format(yearweek(as.Date("2020-12-31"))), "2020 W53")
  
  # 2015 had 53 weeks (started on Thursday)
  expect_equal(format(yearweek(as.Date("2015-12-31"))), "2015 W53")
})

test_that("yearweek Thursday rule is correct", {
  # ISO 8601 rule: Week 1 is the first week with a Thursday
  
  # 2019-01-01 is Tuesday, so it's in 2019-W01 (Thu is Jan 3)
  expect_equal(format(yearweek(as.Date("2019-01-01"))), "2019 W01")
  
  # 2021-01-01 is Friday, so it's in 2020-W53 (no Thu in that week of 2021)
  expect_equal(format(yearweek(as.Date("2021-01-01"))), "2020 W53")
  
  # 2022-01-01 is Saturday, so it's in 2021-W52
  expect_equal(format(yearweek(as.Date("2022-01-01"))), "2021 W52")
})

test_that("yearweek handles various date ranges without errors", {
  # Test that converting dates gives consistent results
  dates <- seq(as.Date("2020-01-01"), as.Date("2024-12-31"), by = "7 days")
  yw <- yearweek(dates)
  expect_false(any(is.na(yw)))
  expect_length(yw, length(dates))
  
  # Leap year dates
  leap_dates <- as.Date(c("2020-02-28", "2020-02-29", "2020-03-01"))
  yw_leap <- yearweek(leap_dates)
  expect_length(yw_leap, 3)
  expect_false(any(is.na(yw_leap)))
  
  # Non-leap year dates
  non_leap_dates <- as.Date(c("2021-02-28", "2021-03-01"))
  yw_non_leap <- yearweek(non_leap_dates)
  expect_length(yw_non_leap, 2)
  expect_false(any(is.na(yw_non_leap)))
})

test_that("yearweek handles Unix epoch correctly", {
  # Unix epoch: 1970-01-01 is a Thursday, should be 1970-W01
  expect_equal(format(yearweek(as.Date("1970-01-01"))), "1970 W01")
})



test_that("yearweek handles all days of week consistently", {
  # Test a full week to ensure all weekdays are handled
  week_start <- as.Date("2023-06-05") # A Monday
  full_week <- seq(week_start, by = "1 day", length.out = 7)
  yw_week <- yearweek(full_week)
  
  # All days in the same week should have the same yearweek
  expect_equal(length(unique(yw_week)), 1)
})

test_that("yearweek handles year boundary transitions", {
  # Comprehensive test of dates around year boundaries for multiple years
  test_years <- 2020:2025
  
  for (year in test_years) {
    # Test last few days of December
    dec_dates <- as.Date(paste0(year, "-12-", 29:31))
    yw_dec <- yearweek(dec_dates)
    expect_false(any(is.na(yw_dec)))
    
    # Test first few days of January
    jan_dates <- as.Date(paste0(year, "-01-", 1:3))
    yw_jan <- yearweek(jan_dates)
    expect_false(any(is.na(yw_jan)))
  }
})

test_that("cal_isoweek$week chronon_cardinality works correctly", {
  week_unit <- cal_isoweek$week(1L)
  day_unit <- cal_isoweek$day(1L)
  
  # 1 week = 7 days
  expect_equal(chronon_cardinality(day_unit, week_unit), 7)
  
  # 2 weeks = 14 days
  expect_equal(chronon_cardinality(day_unit, cal_isoweek$week(2L)), 14)
})

test_that("chronon_divmod from days to weeks works", {
  day_unit <- cal_isoweek$day(1L)
  week_unit <- cal_isoweek$week(1L)
  
  # Test with multiples of 7 (should have remainder 3 for Thursday epoch)
  result <- chronon_divmod(day_unit, week_unit, 7L)
  expect_equal(result$mod, 3L)
  
  # Test with non-multiples
  result2 <- chronon_divmod(day_unit, week_unit, 10L)
  expect_equal(result2$div, 1L)
  expect_equal(result2$mod, 6L)
})

test_that("chronon_divmod from weeks to days works", {
  week_unit <- cal_isoweek$week(1L)
  day_unit <- cal_isoweek$day(1L)
  
  # 1 week should convert to 7 days (plus epoch offset)
  result <- chronon_divmod(week_unit, day_unit, 1L)
  expect_equal(result$div %% 7, 4L)  # Check epoch alignment
  expect_equal(result$mod, 0L)
})

test_that("all.equal() on time values succeeds instead of erroring", {
  expect_true(isTRUE(all.equal(yearmonth(0), yearmonth(0))))
  expect_false(isTRUE(all.equal(yearmonth(0), yearmonth(1))))
  expect_match(all.equal(yearmonth(0), yearmonth(1)), "Mean absolute difference")
})

test_that("all.equal() agrees with == across differing chronons", {
  # These are `==` (24 hours is exactly 1 day), so all.equal() must treat
  # them as equal too, despite differing raw storage/chronon.
  expect_true(days(1) == hours(24))
  expect_true(isTRUE(all.equal(days(1), hours(24))))

  expect_true(isTRUE(all.equal(days(1), days(1))))
  expect_false(isTRUE(all.equal(days(1), days(2))))
})

test_that("all.equal() tolerance is absolute, in chronon units", {
  a <- yearmonth(Sys.Date(), discrete = FALSE)
  b <- a + 1e-10
  expect_true(isTRUE(all.equal(a, b)))
  expect_false(isTRUE(all.equal(a, b, tolerance = 1e-12)))
})

test_that("all.equal() reports length and NA mismatches", {
  expect_match(
    all.equal(yearmonth(0:2), yearmonth(0:1)),
    "Lengths \\(3, 2\\) differ"
  )
  expect_match(
    all.equal(yearmonth(c(0, NA)), yearmonth(c(0, 1))),
    "is.NA"
  )
  expect_true(isTRUE(all.equal(yearmonth(c(0, NA)), yearmonth(c(0, NA)))))
})

test_that("all.equal() works via testthat::expect_equal()", {
  expect_equal(days(1), hours(24))
  expect_failure(expect_equal(yearmonth(0), yearmonth(1)))
})

---
title: "Time format strings"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Time format strings}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, message = FALSE}
library(mixtime)
```

mixtime format strings describe how a point in time is formatted as text. They are analagous to the time format strings used in `strptime()` and `strftime()`.

## Time component codes {#time-codes}

Where `strftime()` uses `%`-prefixed codes tied to the Gregorian calendar (e.g. `%Y` for year, `%m` for month), mixtime uses [glue](https://glue.tidyverse.org/)-style `{...}` placeholders that refer explicitly to time components defined by linear and cyclical granules. 

| Code | Description |
|---|---|
| `lin(granule)` | A linear granularity is typically used for the coarsest granule (e.g. years). |
| `cyc(granule, cycle)` | A cyclical granularity displays finer granules within a coarser cycle (e.g. months within years). |

The `granule` and `cycle` arguments accept time granules from a calendar, for example the format codes for `yearmonth()` are `lin(year(1L))` and `cyc(month(1L), year(1L))`. These format codes are combined in the time format string with glue substitutions: `"{lin(year(1L))}-{cyc(month(1L), year(1L))}"`.

```{r}
format(date(Sys.Date()), format = "{lin(year(1L))}-{cyc(month(1L), year(1L))}")
```

The linear labels are produced by `linear_labels_format()`, and the cyclical labels are produced by `cyclical_labels_format()` -- see `linear_labels()`/`cyclical_labels()` for how a calendar declares these (and their text-parsing inverse) in one step. When formatting mixtime vectors, additional arguments passed to `lin()` and `cyc()` are forwarded to the label helper functions. Common arguments include `label = TRUE` to get named labels (e.g. "January"), and `abbreviate = TRUE` for short names (e.g. "Jan" instead of "January").

```{r}
format(date(Sys.Date()), format = "{lin(year)} {cyc(month, year, label = TRUE, abbreviate = TRUE)}")
```

Default format strings for each chronon (and cycle) are provided by methods of the `chronon_format_linear()` and `chronon_format_cyclical()` generics (see `vignette("extending-mixtime")`). For example, the default format string for a time with a 1-month chronon is `"{lin(year)}-{cyc(month, year)}"`:

```{r}
format(linear_time(Sys.Date(), chronon = cal_gregorian$month(1L)))
```

The granules used in the format string are calendar-aware, so they inherit attributes from the data being formatted. This includes the calendar (e.g. `cal_gregorian`) and granule attributes (e.g. time zone for civil time and location for astronomical time). Explicitly specifying the calendar or granule attributes in the format string allows different calendars and granule attributes to be combined in the same format string. 

<!-- 
For example, here we display the time in Berlin alongside Melbourne:

```{r}
# Default format for a second-resolution chronon is "YYYY-MM-DD HH:MM:SS"
fmt_time <- chronon_format_linear(cal_gregorian$second(1L))
mel <- "Australia/Melbourne"
fmt_time_mel <- "({cyc(hour(1L, tz=mel), day(1L, tz = mel))}:{cyc(minute(1L, tz=mel), hour(1L, tz = mel))}:{cyc(second(1L, tz=mel), minute(1L, tz = mel))} {tz(.time, mel)})"
format(
  datetime(Sys.time(), tz = "UTC"), 
  format = paste(fmt_time, "{tz(.time)}", fmt_time_mel)
)
```
-->

For example, the lunar phase can be added to a date format string by including the `phase` granule from the `cal_time_lunar` calendar. We can also set `type = "emoji"` to get a lunar phase emoji instead of the named phase:

```{r}
fmt_date  <- "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"
fmt_lunar <- "{with(cal_time_lunar, cyc(phase, month, label = TRUE, type = \"emoji\"))}"
format(date(Sys.Date()), format = paste(fmt_date, fmt_lunar))
```

This makes the relationship between the format string and the calendar structure transparent, and means the same format-string idiom works unchanged across calendars.

## Advanced time formatting

In addition to the `lin()` and `cyc()` codes, mixtime format strings support glue-style expressions that can be used to insert arbitrary text and perform calculations on the time components. The `.time` variable is available within the format string and refers to the time being formatted. This variable is commonly used to display time attributes, including:

| Code | Description |
|---|---|
| `tz(.time)` | Time zone (e.g. `AEST`) |
| `loc(.time)` | Location (e.g. `37.81S 144.96E`) |
| `frac(.time)` | Fractional chronons (e.g. `39.2%`)|

The `.time` argument can be used in any calculation within the format string.

## Comparison with POSIX time format strings

POSIX time format strings are concise but require memorising a large number of `%`-prefixed codes which are platform-dependent and and mostly specific to the Gregorian calendar. In contrast, mixtime format strings are more verbose but more intuitive and generalisable to different calendars and time formats.

An ISO-8601 date (`YYYY-MM-DD`) is represented with the mixtime format string `"{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"`. While more verbose than the equivalent `strftime` format string (`"%Y-%m-%d"`), the mixtime string is more descriptive and generalisable to other combinations of granules.

```{r}
# YYYY-MM-DD format for a Gregorian date
format(date(Sys.Date()), format = "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}")
format(Sys.Date(), format = "%Y-%m-%d")
```

This simplifies the construction of less common time formats, for example the year-day format (`YYYY-DDD`), which is `"{lin(year)}-D{cyc(day, year)}"` in mixtime but `"%Y-D%j"` in `strftime` (where `%j` is a non-obvious code for "day of year").

```{r}
# YYY-DDD format for a Gregorian date
format(date(Sys.Date()), format = "{lin(year)}-D{cyc(day, year)}")
format(Sys.Date(), format = "%Y-D%j")
```

Format strings for non-Gregorian calendars are also more intuitive, since the granules are specified in reference to units of calendar systems. For example, the format string for ISO week dates (`YYYY-WW-DDD`) is `"{lin(year)}-W{cyc(week, year)}-{cyc(day, week, label = TRUE)}"` in mixtime. The equivalent `strftime` format string (`"%G-W%V-%a"`) requires knowing that `%G`/`%V` are the ISO week variants of `%Y`/`%W` - a nuanced distinction that is easy to get wrong.

```{r}
# YYYY-WW-DD format for a ISO week date
format(
  date(Sys.Date(), calendar = cal_isoweek),
  format = "{lin(year)}-W{cyc(week, year)}-{cyc(day, week, label = TRUE)}"
)
format(Sys.Date(), format = "%G-W%V-%a")
```

The table below maps every `strftime`/`strptime` conversion specification to its mixtime equivalent. There are many more possible mixtime format strings that don't have a direct `strftime` equivalent.

<details>
<summary>Show/hide POSIX and mixtime equivalent code table</summary>

| POSIX | mixtime | Description |
|-------|----------------------------------|-------------|
| `%Y` | `{lin(year)}` | Year with century (e.g. 2026) |
| `%G` | `{lin(cal_isoweek$year)}` | ISO 8601 week-based year (1-53) |
| `%m` | `{cyc(month, year)}` | Month as decimal number (01–12) |
| `%B` | `{cyc(month, year, label = TRUE)}` | Full month name (e.g. January) |
| `%b`, `%h` | `{cyc(month, year, label = TRUE, abbreviate = TRUE)}` | Abbreviated month name (e.g. Jan) |
| `%d` | `{cyc(day, month)}` | Day of month (01–31) |
| `%j` | `{cyc(day, year)}` | Day of year (001–366) |
| `%H` | `{cyc(hour, day)}` | Hour, 24-hour clock (00–23) |
| `%I` | `{cyc(hour, ampm)}` | Hour, 12-hour clock (01–12) |
| `%p` | `{cyc(ampm, day, label = TRUE)}` | AM/PM indicator |
| `%M` | `{cyc(minute, hour)}` | Minute (00–59) |
| `%S` | `{cyc(second, minute)}` | Second as integer (00–61) |
| `%A` | `{cyc(day, cal_isoweek$week, label = TRUE)}` | Full weekday name (e.g. Monday) |
| `%a` | `{cyc(day, cal_isoweek$week, label = TRUE, abbreviate = TRUE)}` | Abbreviated weekday name (e.g. Mon) |
| `%u` | `{cyc(day, cal_isoweek$week)}` | Weekday as numeric (1 = Monday, 7 = Sunday) |
| `%V` | `{with(cal_isoweek, cyc(week, year))}` | ISO 8601 week number (01–53) |
| `%Z` | `{tz(.time)}` | Time zone abbreviation (e.g. `AEST`) |
| `%z` | `+{tz_offset(.time)}` | UTC offset (e.g. `+1000`) |
| `%s` | `{as.numeric(as.POSIXt(.time))}` | Seconds since the Unix epoch |

<!-- 
Composite types
| `%F` | `{lin(year)}-{cyc(month, year)}-{cyc(day, month)}` | ISO 8601 date (`%Y-%m-%d`) |
| `%T` | `{cyc(hour, day)}:{cyc(minute, hour)}:{cyc(second, minute)}` | Time (`%H:%M:%S`) |
| `%R` | `{cyc(hour, day)}:{cyc(minute, hour)}` | Hours and minutes (`%H:%M`) |
| `%D` | `{cyc(month, year)}/{cyc(day, month)}/{lin(year) %% 100}` | US date (`%m/%d/%y`) |
 -->

<!-- 
Unsupported types (TODO)
| `%OS` | `{cyc(second, minute)}{frac(.time)}` | Fractional seconds |
| `%e` | `{cyc(day, month)}` | Day of month with leading space for single digits |
| `%W` | `{cyc(week, year)}` | Week of year, Monday as first day (00–53) |
| `%y` | — | Year without century (00–99); no direct equivalent |
| `%C` | — | Century (year ÷ 100); no direct equivalent |
| `%g` | — | Last two digits of ISO week-based year; no direct equivalent |
| `%w` | — | Weekday as decimal (0 = Sunday); no direct equivalent |
| `%U` | — | Week of year, Sunday as first day (00–53); no direct equivalent | 
-->

</details>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cal0_time_solar.R
\name{calendar_time_solar}
\alias{calendar_time_solar}
\alias{cal_time_solar_transit}
\alias{cal_time_solar}
\title{Solar time unit classes}
\format{
A location-based calendar containing transit-based solar time units.
}
\usage{
cal_time_solar_transit

cal_time_solar
}
\value{
An S3 list of class \code{c("cal_time_solar", "mt_calendar")} containing
the named time unit classes of the solar transit calendar. Each unit is
accessible via \code{$} notation and calling it with a step size and location
produces a time granule (e.g., 1 solar day granule as
\code{cal_time_solar$day(1L, lat = 0, lon = 0)}). Because solar day boundaries
depend on the observer's position, each unit constructor requires \code{lat}
and \code{lon} arguments.
}
\description{
Time unit constructors for the transit-based solar time system, where the
boundary of each day is at apparent solar midnight. Solar events define the
\code{ampm} (midnight and noon) and \code{illumination} (dawn, sunrise, sunset, dusk)
units. \code{cal_time_solar} is an alias for \code{cal_time_solar_transit}.
}
\details{
The following time units are available in the solar calendar systems.
\itemize{
\item \code{day()}: Day unit
\item \code{ampm()}: Half-day units (AM = before solar noon, PM = after solar noon)
\item \code{hour()}: Hour units within the solar day
\item \code{minute()}: Minute units within the solar hour
\item \code{second()}: Second units within the solar minute
\item \code{degree()}: Solar angle units within the day
\item \code{arcminute()}: Arcminute units within the solar degree
\item \code{arcsecond()}: Arcsecond units within the solar arcminute
\item \code{illumination()}: Illumination phases (night, astronomical/nautical/civil dawn, day, civil/nautical/astronomical dusk)
}
\subsection{AM/PM half-days}{

The \code{ampm} unit divides each solar day into two halves between solar noon and
solar midnight:\tabular{lll}{
   Half \tab Period \tab Description \cr
   AM \tab Solar midnight to noon \tab Morning half; before solar transit \cr
   PM \tab Solar noon to midnight \tab Afternoon half; after solar transit \cr
}

}

\subsection{Solar illumination phases}{

Phases describe the illumination state of the sky and correspond to standard
twilight definitions used in astronomy and navigation. Each phase is bounded
by a pair of solar altitude thresholds:\tabular{lll}{
   Phase \tab Solar altitude range \tab Description \cr
   Night \tab < -18° \tab Sky fully dark; from last dusk to first dawn (spans noon) \cr
   Astronomical dawn \tab -18° to -12° \tab Astronomical twilight before sunrise; faint objects obscured \cr
   Nautical dawn \tab -12° to -6° \tab Nautical twilight before sunrise; horizon visible at sea \cr
   Civil dawn \tab -6° to -0.833° \tab Civil twilight before sunrise; sky brightening in the east \cr
   Day \tab > -0.833° \tab Sun above the horizon; spans solar noon \cr
   Civil dusk \tab -0.833° to -6° \tab Civil twilight after sunset; sky fading in the west \cr
   Nautical dusk \tab -6° to -12° \tab Nautical twilight after sunset; horizon visible at sea \cr
   Astronomical dusk \tab -12° to -18° \tab Astronomical twilight after sunset; faint objects obscured \cr
}


The -0.833° threshold for sunrise and sunset accounts for the mean angular
radius of the solar disc (0.267°) plus the standard atmospheric refraction
at the horizon (0.566°). Noon and midnight are derived from the equation of
time rather than a fixed altitude. Locations that experience polar day or
polar night (civil days where sunrise does not occur) are not currently
supported, it is recommended to use an alternative reference location.
}
}
\examples{
# Find the current solar time in Melbourne
datetime(Sys.time(), calendar = cal_time_solar, lat = -37.8136, lon = 144.9631)

}
\seealso{
\code{\link{cal_time_civil}}, \code{\link{cal_time_lunar}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/timezone.R
\name{tz_offset}
\alias{tz_offset}
\title{Get timezone offset}
\usage{
tz_offset(x, ...)
}
\arguments{
\item{x}{A time class coercible to POSIXt with an associated time zone.}

\item{...}{Additional arguments passed to methods.}
}
\value{
A \code{mixtime} duration vector of offsets from UTC in the same chronon
(e.g. seconds for POSIXt, days for dates, etc.)
}
\description{
Returns the UTC offset for a given datetime in its specified timezone.
}
\examples{
tz_offset(as.POSIXct(Sys.time(), tz = Sys.timezone()))
tz_offset(as.POSIXct("2024-06-15 12:00:00", tz = "America/New_York"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01_chronon_divmod.R
\name{chronon_divmod}
\alias{chronon_divmod}
\title{Convert between chronons of different time granules}
\usage{
chronon_divmod(from, to, ...)
}
\arguments{
\item{from}{The time granule that \code{x} is measured in (e.g., \code{day(1L)}).}

\item{to}{The time granule to convert \code{x} into (e.g., \code{week(1L)}).}

\item{...}{Additional arguments for methods.}
}
\value{
An list of two elements:
\itemize{
\item \code{div}: integer vector of chronons measured in the \code{to} time granule.
\item \code{mod}: integer vector of the remainder (in \code{from} time granule) after
converting to the \code{to} time granule.
}
}
\description{
This function converts between chronons measured in different time granules.
It is used internally for converting between different continuous time types,
and is particularly useful for efficiently converting between irregular time
granules. The default method uses \code{chronon_cardinality()} to cast between
time granules, which is efficient for regular time granules.
}
\examples{
# Convert day 16 after epoch (1970-01-01) into weeks since epoch (and remainder days)
with(cal_isoweek, chronon_divmod(day(1L), week(1L), 16L))

# Convert week 4 after epoch (1970-W1) into days since epoch
with(cal_isoweek, chronon_divmod(week(1L), day(1L), 4L))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01_cyclical_labels.R
\name{cyclical_labels}
\alias{cyclical_labels}
\title{Friendly labels for cyclical relationships}
\usage{
cyclical_labels(granule, cycle, ...)
}
\arguments{
\item{granule}{A time granule object representing the granule (e.g., \code{month(1L)})}

\item{cycle}{A time granule object representing the cycle (e.g., \code{year(1L)})}

\item{...}{Additional arguments for methods.}
}
\value{
Character vector of labels for the time point within the cycle.
}
\description{
This S7 generic function provides the labels for cyclical relationships
between time granules. These functions should return locale specific labels.
}
\examples{
# Labels for months in a year
with(cal_gregorian, cyclical_labels(month(1L), year(1L), 1:12))

# Labels for days in a week
with(cal_isoweek, cyclical_labels(day(1L), week(1L), 1:7))

# Labels for weeks in a year, defaulted from time_unit_abbr()
with(cal_isoweek, cyclical_labels(week(1L), year(1L), 1:52))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01_chronon_cardinality.R
\name{chronon_cardinality}
\alias{chronon_cardinality}
\title{Cardinality between time granules}
\usage{
chronon_cardinality(x, y, ...)
}
\arguments{
\item{x}{The finer time granule (e.g. \code{cal_gregorian$month(1L)})}

\item{y}{The coarser time granule (e.g. \code{cal_gregorian$year(1L)})}

\item{...}{Additional arguments for methods.}
}
\value{
Numeric describing how many \code{x} time granules fit into \code{y} at time \code{at}.
}
\description{
This S7 generic function defines the calendrical relationships between two
chronons, and is one of the building block for defining calendars in mixtime.
It calculates how many \code{x} chronons fit into the \code{y} chronon. Some chronon
sizes are context-dependent (such as the number of days in a month),
and so an optional time point defined in terms of \code{y} chronons can be
provided with \code{at}.
}
\details{
The methods are dispatched based on the shortest path along defined methods.
This allows for defining only the direct relationships between adjacent
time units, and relying on graph traversal to find how to convert between
more distant units. For example the number of seconds in an hour can be
calculated from the number of seconds in a minute and then number of minutes
in an hour.

If a method is defined for converting between time units of different
calendar systems (e.g., Gregorian calendar days to Chinese calendar days),
then that method can be used to convert times at any granularity between the
two systems.
}
\examples{
# There are 12 months in a year
with(cal_gregorian, chronon_cardinality(month(1L), year(1L)))

# There are 7 days in a week
with(cal_isoweek, chronon_cardinality(day(1L), week(1L)))

# There are 3600 seconds in an hour
with(cal_gregorian, chronon_cardinality(second(1L), hour(1L)))

# There are 18 "2 months" in 3 years
with(cal_gregorian, chronon_cardinality(month(2L), year(3L)))

# There are 365 days in 2025 (a common year)
chronon_cardinality(
  cal_gregorian$day(1L), cal_gregorian$year(1L),
  at = year(as.Date("2025-01-01"))
)

# There are 366 days in 2024 (a leap year)
chronon_cardinality(
  cal_gregorian$day(1L), cal_gregorian$year(1L), 
  at = mixtime::year(as.Date("2024-01-01"))
)

# There are 29 days in February 2024 (a leap year)
chronon_cardinality(
  cal_gregorian$day(1L), cal_gregorian$month(1L), 
  at = yearmonth(as.Date("2024-02-01"))
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01_calendar.R
\name{new_calendar}
\alias{new_calendar}
\title{Create a new calendar}
\usage{
new_calendar(..., inherit = NULL, class = character())
}
\arguments{
\item{...}{Named time unit class definitions. Each argument should be a
time unit class (typically created with \code{S7::new_class()}) that inherits
from \code{mt_unit} or \code{mt_tz_unit}. The names define the calendar's fields
and are used to access unit constructors (e.g., \code{calendar$year()}).}

\item{inherit}{Optional calendar to inherit time units from. Units
defined in \code{...} will override inherited units with the same name.}

\item{class}{Character vector of additional classes for the calendar object.}
}
\value{
A calendar object of class \code{c(class, "mt_calendar")}, consisting of a
named list containing the specified time unit classes.
}
\description{
Define a new calendar as a collection of time units. Calendars are the
foundation for representing dates and times in terms of human-readable
components like years, months, days, hours, minutes, and seconds. Each
calendar is defined by specifying the time units it contains, which
determine how time values can be interpreted and manipulated.
}
\details{
Time units are typically S7 class definitions that inherit from \code{\link{mt_unit}}
for standard units, \code{\link{mt_tz_unit}} for timezone-aware units (civil time), or
\code{\link{mt_loc_unit}} for location-aware units (astronomical time). The calendar
object provides a namespace for accessing these unit constructors and
defines the relationships between them for calendar arithmetic.
}
\examples{
# Create a simple calendar with year and month units
# (inheriting from civil time units for day, hour, minute, second, ...)
cal_simple <- new_calendar(
  year = new_time_unit("tu_year", parent = mt_tz_unit),
  month = new_time_unit("tu_month", parent = mt_tz_unit),
  inherit = cal_time_civil,
  class = "cal_simple"
)

# Create time granules from the calendar
cal_simple$year(1L)
cal_simple$month(1L)

}
\seealso{
\code{\link[=linear_time]{linear_time()}}, \code{\link[=cyclical_time]{cyclical_time()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cal0_time_lunar.R
\name{calendar_time_lunar}
\alias{calendar_time_lunar}
\alias{cal_time_lunar_synodic}
\alias{cal_time_lunar}
\title{Lunar time unit classes}
\format{
A calendar containing synodic lunar time units.
}
\usage{
cal_time_lunar_synodic

cal_time_lunar
}
\value{
An S3 list of class \code{c("cal_time_lunar", "mt_calendar")} containing
the named time unit classes of the synodic lunar calendar. Each unit is
accessible via \code{$} notation and calling it with a step size produces a
time granule (e.g., 1 synodic month granule as
\code{cal_time_lunar$month(1L)}). Lunar month and phase boundaries are
location-independent.
}
\description{
Time unit constructors for the synodic lunar time system, where the boundary
of each month is at the new moon (lunar conjunction) and the boundary of
each phase is at a lunar octant (each eighth of the synodic cycle). Both
boundaries are location-independent, as the new moon is a geocentric
astronomical event. \code{cal_time_lunar} is an alias for
\code{cal_time_lunar_synodic}.
}
\details{
The following time units are available in the lunar calendar systems.
\itemize{
\item \code{month()}: Synodic month unit
\item \code{phase()}: Synodic phase unit
}
}
\examples{
# Find the time of a new moon in the Gregorian calendar
t <- linear_time(Sys.Date(), cal_time_lunar$month(1L))
datetime(t, tz = "Australia/Melbourne")


}
\seealso{
\code{\link{cal_time_civil}}, \code{\link{cal_time_solar}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{mt_duration-compare}
\alias{mt_duration-compare}
\alias{==,mt_duration-method}
\alias{!=,mt_duration-method}
\alias{<,mt_duration-method}
\alias{<=,mt_duration-method}
\alias{>,mt_duration-method}
\alias{>=,mt_duration-method}
\title{Comparison operators for durations (\code{mt_duration})}
\usage{
\S4method{==}{mt_duration}(e1, e2)
\S4method{!=}{mt_duration}(e1, e2)
\S4method{<}{mt_duration}(e1, e2)
\S4method{<=}{mt_duration}(e1, e2)
\S4method{>}{mt_duration}(e1, e2)
\S4method{>=}{mt_duration}(e1, e2)
}
\arguments{
\item{e1, e2}{\code{mt_duration} vectors (or values castable to one, such as plain
numeric vectors interpreted in the other operand's chronon).}
}
\value{
A logical vector.
}
\description{
A duration is a scalar magnitude of time measured in a given chronon (e.g.
\code{days(3)}), with no reference to a point in time. Comparing two durations is
therefore a plain magnitude comparison, once both operands have been
expressed in a common chronon:
\itemize{
\item \code{a == b} iff the two magnitudes are equal in their common chronon
\item \code{a < b}, \code{a <= b}, \code{a > b}, \code{a >= b} compare the magnitudes directly
}

Unlike \code{mt_linear} comparison there is no interval/span to consider, so this
\emph{is} a total order. When both operands already share a chronon the magnitudes
are compared as-is; otherwise both are scaled to their finest common chronon
(the same scaling used when combining durations arithmetically, see
\code{duration_combine()}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linear_time.R
\name{new_linear_time_fn}
\alias{new_linear_time_fn}
\title{Linear time function factory}
\usage{
new_linear_time_fn(chronon, default_calendar = cal_gregorian)
}
\arguments{
\item{chronon}{A bare call for a time granule object representing the chronon
(e.g., \code{day(1)})}

\item{default_calendar}{A default calendar used to find the time units for
conversion if they don't exist in the calendar of the input data (e.g.,
\code{cal_isoweek} for week chronons to work with gregorian calendar inputs).}
}
\value{
A function used to create linear time points with a specific chronon.
}
\description{
\code{new_linear_time_fn()} creates a linear time function for a specified
chronon. A chronon is the smallest indivisible time granule (e.g., days,
hours).
}
\examples{

# Linear time with 1 month granules as the chronon
ym <- new_linear_time_fn(month(1L))
ym(Sys.Date())

# Linear time with 1 day granules as the chronon
yd <- new_linear_time_fn(day(1L))
yd(Sys.Date())

# Linear time with 1 week granules as the chronon, using the ISO week calendar
yw <- new_linear_time_fn(week(1L), default_calendar = cal_isoweek)
yw(Sys.Date())

# Linear time with 1 hour granules as the chronon
ymd_h <- new_linear_time_fn(hour(1L))
ymd_h(Sys.time())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predicates.R
\name{time_is_complete_at}
\alias{time_is_complete_at}
\title{Test whether a granule is completed at a time point}
\usage{
time_is_complete_at(x, granule, ...)
}
\arguments{
\item{x}{A time object (typically a \code{mixtime} vector).}

\item{granule}{The time granule whose precision to test, given as a granule
generator (e.g. \code{cal_gregorian$month}) or a sized time unit (e.g.
\code{cal_gregorian$month(1L)}).}

\item{...}{Additional arguments for methods.}
}
\value{
A logical vector the same length as \code{x}.
}
\description{
\code{time_is_complete_at()} tests, for each element of a \code{mixtime} vector, whether
the coarser \code{granule} that element falls into is fully observed \emph{by the vector
as a whole} -- that is, whether every finer chronon making up that granule is
present somewhere in \code{x}.
}
\details{
Unlike \code{\link[=time_is_determinate_at]{time_is_determinate_at()}}, completeness is a collective property: an
element is \code{TRUE} only when the other elements needed to fill its granule are
also present. For example, in \code{year(1L)} the months of \verb{2020 Jan : 2020 Oct}
are all \code{FALSE} (November and December are missing, so 2020 is incomplete),
whereas in \verb{2020 Jan : 2021 Mar} the twelve months of 2020 are \code{TRUE} (they
complete 2020) while the three months of 2021 remain \code{FALSE}.

A granule equal to \code{x}'s own chronon is completed by each point on its own
(\code{TRUE}). A granule finer than \code{x} cannot be completed by coarser points
(\code{FALSE}). Missing (\code{NA}) and infinite times give \code{NA}.

Completeness is only defined within a single time granularity. Mixed-type
\code{mixtime} vectors (e.g. months alongside days) are not yet supported and
raise an error.
}
\examples{
# 2020 Jan : 2020 Oct does not complete the year -> all FALSE
time_is_complete_at(yearmonth(as.Date("2020-01-01")) + 0:9, cal_gregorian$year(1L))

# 2020 Jan : 2021 Mar completes 2020 (TRUE) but not 2021 (FALSE)
time_is_complete_at(yearmonth(as.Date("2020-01-01")) + 0:14, cal_gregorian$year(1L))

}
\seealso{
\code{\link[=time_is_determinate_at]{time_is_determinate_at()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predicates.R
\name{is_time}
\alias{is_time}
\alias{time_is_linear}
\alias{time_is_cyclical}
\alias{time_is_duration}
\alias{is_time_linear}
\alias{is_time_cyclical}
\alias{is_time_duration}
\title{Check the time type of values}
\usage{
time_is_linear(x, ...)

time_is_cyclical(x, ...)

time_is_duration(x, ...)

is_time_linear(x, ...)

is_time_cyclical(x, ...)

is_time_duration(x, ...)
}
\arguments{
\item{x}{A time object (typically a \code{mixtime} vector).}

\item{...}{Additional arguments for methods.}
}
\value{
A logical vector the same length as \code{x}.
}
\description{
Test whether elements of a \code{mixtime} vector are linear, cyclical, or durations.
}
\details{
These helpers return a logical vector the same length as \code{x} identifying the
type of time represented by each element.
}
\examples{
t <- c(yearmonth(0), month_of_year(0), months(0L))
time_is_linear(t)
time_is_cyclical(t)
time_is_duration(t)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mixtime.R
\name{mixtime}
\alias{mixtime}
\title{Create a mixtime vector}
\usage{
mixtime(
  data,
  chronon = time_chronon(data),
  cycle = time_cycle(data),
  discrete = TRUE
)
}
\arguments{
\item{data}{A vector of time values. This can be a character vector (e.g. "2024-01-01"),
a numeric vector (e.g. seconds since epoch), or a time class (e.g. Date, POSIXct, yearmonth, etc.).}

\item{chronon}{A time granule object representing the smallest indivisible time granule (chronon) for the mixtime. This is used to interpret the numeric values in \code{data} and to define the time resolution of the mixtime. If not provided, it will be inferred from \code{data}.}

\item{cycle}{An optional time granule object representing the cycle for cyclical time. This is used to define the repeating cycle for cyclical time representations (e.g. day-of-week, month-of-year). If not provided, the mixtime will be treated as linear time.}

\item{discrete}{A logical indicating whether the time values should be treated as discrete (integer) or continuous (fractional). This affects how numeric values are interpreted and how time arithmetic is performed. The default is \code{TRUE} (discrete).}
}
\value{
A \code{mixtime} object representing the time values in \code{data} according to the specified \code{chronon} and \code{cycle}.
}
\description{
A mixtime is a vector which describes a point in time. It uses a calendar
definition to translate a vector of numbers into a point in time.
}
\examples{
# Create a mixtime for today
mixtime(Sys.Date())

# Create a mixtime for the current date and time
mixtime(Sys.time())

# Convert time from tsibble classes to mixtime
mixtime(tsibble::yearmonth("2024 Jan"))

# Create a mixtime for the time of day (cyclical time)
mixtime(Sys.time(), cycle = cal_gregorian$day(1L))

# Specify a timezone for the chronon
mixtime(Sys.time(), chronon = cal_gregorian$second(1L, tz = Sys.timezone()))
mixtime(Sys.time(), chronon = cal_gregorian$second(1L, tz = "Pacific/Honolulu"))
mixtime(Sys.time(), chronon = cal_gregorian$second(1L, tz = "Australia/Melbourne"))

# Dates (and all granularities) can have timezones
mixtime(Sys.time(), chronon = cal_gregorian$day(1L, tz = Sys.timezone()))
mixtime(Sys.time(), chronon = cal_gregorian$day(1L, tz = "Pacific/Honolulu"))
mixtime(Sys.time(), chronon = cal_gregorian$day(1L, tz = "Australia/Melbourne"))

# Continuous time tracks progress within the chronon
mixtime(Sys.time(), chronon = cal_gregorian$day(1L, tz = Sys.timezone()), discrete = FALSE)

# Mixtime can combine different granularities and timezones in a vector
now <- Sys.time()
c(
  # Datetime (second chronon) in UTC
  mixtime(now),
  # Date (minute chronon) in local timezone
  mixtime(now, chronon = cal_gregorian$minute(1L, tz = Sys.timezone())),
  # Month (month chronon) in UTC
  mixtime(now, chronon = cal_gregorian$month(1L))
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01_chronon_epoch.R
\name{chronon_epoch}
\alias{chronon_epoch}
\title{Epoch offset for chronons}
\usage{
chronon_epoch(x, ...)
}
\arguments{
\item{x}{A chronon (time unit) object.}

\item{...}{Additional arguments for methods.}
}
\value{
A numeric value representing the epoch for the chronon.
}
\description{
Returns the epoch offset for a given chronon (time unit). The epoch defines
the starting point of the chronon's linear numbering, used when converting
from internal representations to common displays (e.g. applying an epoch of
1970-01-01).
}
\examples{
# The epoch for year linear time displays is 1970
chronon_epoch(cal_gregorian$year(1L))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01_linear_labels.R
\name{linear_labels}
\alias{linear_labels}
\title{Friendly labels for linear relationships}
\usage{
linear_labels(granule, ...)
}
\arguments{
\item{granule}{A time granule object representing the granule (e.g., \code{year(1L)})}

\item{...}{Additional arguments for methods.}
}
\value{
Character vector of labels for the time point.
}
\description{
This S7 generic function provides the labels for linear (non-repeating)
positions of a time granule. These functions should return locale specific
labels.
}
\examples{
# Labels for years on a linear axis
with(cal_gregorian, linear_labels(year(1L), 2020:2025))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01_time_unit.R
\name{new_time_unit}
\alias{new_time_unit}
\title{Create a new time unit class}
\usage{
new_time_unit(
  name,
  parent = mt_unit,
  package = topNamespaceName(parent.frame()),
  properties = list(),
  abstract = FALSE,
  constructor = NULL,
  validator = NULL
)
}
\arguments{
\item{name}{A string naming the new S7 class (e.g., \code{"tu_my_year"}).
By convention, time unit class names are prefixed with \code{tu_}.}

\item{parent}{The parent S7 class. Should be one of \code{\link{mt_unit}},
\code{\link{mt_tz_unit}}, or \code{\link{mt_loc_unit}}, or a class that itself inherits from
one of these. Defaults to \code{\link{mt_unit}}.}

\item{package}{A string giving the package name that owns the class.
Defaults to the name of the calling namespace, so typically does not
need to be set explicitly.}

\item{properties}{A named list of additional \code{\link[S7:new_property]{S7::new_property()}}
definitions beyond those inherited from \code{parent}. Each entry becomes
a slot on instances of the new class. Defaults to an empty list.}

\item{abstract}{Logical. If \code{TRUE} the class cannot be instantiated
directly; it serves only as a base for further subclassing.}

\item{constructor}{A function to use as the class constructor, or \code{NULL}
(default) to generate one automatically. The auto-generated constructor
accepts \code{...} (forwarded to the parent constructor) plus one argument per
entry in \code{properties}, with defaults taken from each property's
\code{default} field.}

\item{validator}{A function of one argument (\code{self}) that returns \code{NULL}
when the object is valid, or a character string describing the problem.
See \code{\link[S7:new_class]{S7::new_class()}} for details.}
}
\value{
An S7 class object, as returned by \code{\link[S7:new_class]{S7::new_class()}}.
}
\description{
Define a new S7 class representing a time unit for use in a mixtime calendar.
Time units are the building blocks of calendars: each unit represents a
specific temporal component (e.g., day, month, year) and carries a step size
\code{n}. Units are combined via \code{\link[=new_calendar]{new_calendar()}} to form a complete calendar
system.
}
\details{
Choose the parent class based on the type of time the unit represents:
\itemize{
\item \code{\link{mt_unit}} — abstract or calendar-only time (no timezone or location
context; e.g., Gregorian year or ISO week).
\item \code{\link{mt_tz_unit}} — civil time with a timezone (e.g., a clock hour that
needs \code{tz} to resolve wall-clock ambiguity).
\item \code{\link{mt_loc_unit}} — astronomical time with a geographic location (e.g., a
solar day tied to observer longitude/latitude).
}
}
\examples{
# An abstract unit with no properties
tu_my_seq <- new_time_unit("tu_my_seq", parent = mt_unit)

# A civil-time unit with an extra property
tu_my_quarter <- new_time_unit(
  "tu_my_quarter",
  parent = mt_tz_unit,
  properties = list(
    fiscal = S7::new_property(S7::class_logical, default = FALSE)
  )
)

}
\seealso{
\itemize{
\item \code{\link{mt_unit}}, \code{\link{mt_tz_unit}}, \code{\link{mt_loc_unit}} for the base classes to
inherit from.
\item \code{\link[=new_calendar]{new_calendar()}} for assembling time units into a calendar.
\item \code{\link[S7:new_class]{S7::new_class()}} for the underlying S7 class constructor.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linear_time.R
\name{linear_time_helpers}
\alias{linear_time_helpers}
\alias{year}
\alias{yearquarter}
\alias{yearmonth}
\alias{yearweek}
\alias{date}
\alias{datetime}
\title{Linear time helper functions}
\usage{
year(data, discrete = TRUE, calendar = time_calendar(data), ...)

yearquarter(data, discrete = TRUE, calendar = time_calendar(data), ...)

yearmonth(data, discrete = TRUE, calendar = time_calendar(data), ...)

yearweek(data, discrete = TRUE, calendar = time_calendar(data), ...)

date(data, discrete = TRUE, calendar = time_calendar(data), ...)

datetime(data, discrete = TRUE, calendar = time_calendar(data), ...)
}
\arguments{
\item{data}{A vector of time points (e.g. \link[base:Date]{base::Date}, \link[base:POSIXt]{base::POSIXt})}

\item{discrete}{If \code{TRUE}, the number of chronons since Unix epoch that
\code{data} falls into is returned as an integer. If \code{FALSE}, a fractional number
of chronons is returned (analagous to time using a continuous time model).}

\item{calendar}{A calendar used to evaluate the time units. Defaults to the
calendar of the input data. Common options include \link{cal_gregorian} and
\link{cal_isoweek}.}

\item{...}{Additional arguments for \code{\link[=linear_time]{linear_time()}}, such as \code{tz} for timezones.}
}
\value{
A \code{mixtime} time vector containing an \code{mt_linear} vector with chronons matching the function used.
}
\description{
Convenience functions for creating common linear time representations. These
functions work with different calendar systems and adapt based on the input
data's calendar.
}
\details{
These functions create linear time representations with different chronons
and granules:
\itemize{
\item \code{year()}: Represents time in whole years. The chronon is one year.
\item \code{yearquarter()}: Represents time in quarters, grouped by year. The chronon
is one quarter, with years as the granule.
\item \code{yearmonth()}: Represents time in months, grouped by year. The chronon is
one month, with years as the granule.
\item \code{yearweek()}: Represents time in weeks, grouped by year. The chronon is
one week, with years as the granule. Defaults to ISO 8601 week calendar.
}
}
\section{Calendar flexibility}{

These functions adapt to the calendar system of the input data. For example:
\itemize{
\item \code{year(date("2025-12-29"))} returns a Gregorian year
\item \code{year(yearweek(date("2025-12-29")))} returns an ISO week-based year
}

You can also explicitly specify a calendar using the \code{calendar} argument:

\preformatted{
year(date("2025-12-29"), calendar = cal_isoweek)
}
}

\section{Custom linear time representations}{

For more complex time structures, use \code{\link[=linear_time]{linear_time()}} or \code{\link[=new_linear_time_fn]{new_linear_time_fn()}}
to create custom representations with any combination of chronons and granules.
}

\examples{

# Gregorian year
year(Sys.Date())
year(Sys.Date(), discrete = FALSE)

# ISO week-based year
year(yearweek(Sys.Date()))

# Year-quarter
yearquarter(Sys.Date())
yearquarter(Sys.Date(), discrete = FALSE)

# Year-month
yearmonth(Sys.Date())
yearmonth(Sys.Date(), discrete = FALSE)

# Year-week (ISO 8601)
yearweek(Sys.Date())
yearweek(0:52)

}
\seealso{
\itemize{
\item \code{\link[=linear_time]{linear_time()}} for creating custom linear time representations
\item \code{\link[=new_linear_time_fn]{new_linear_time_fn()}} for creating reusable linear time functions
\item \link{cal_gregorian}, \link{cal_isoweek} for calendar systems
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cyclical_time.R
\name{new_cyclical_time_fn}
\alias{new_cyclical_time_fn}
\title{Cyclical time function factory}
\usage{
new_cyclical_time_fn(chronon, cycle, default_calendar = cal_gregorian)
}
\arguments{
\item{chronon}{A time granule object representing the chronon (e.g., \code{day(1L)})}

\item{cycle}{A time granule object representing the cycle (e.g., \code{week(1L)})}

\item{default_calendar}{A default calendar used to find the time units for
conversion if they don't exist in the calendar of the input data (e.g., \code{cal_isoweek})}
}
\value{
A function used to create cyclical time points with a specific chronon and cycle.
}
\description{
\code{new_cyclical_time_fn()} creates a cyclical time function for a specified
chronon and cycle. The cycle is the larger time granule that defines the time
period over which the chronon loops (e.g., a week). The chronon is the smaller
time granule that iterates within each cycle (e.g., a day). Combined, these two
granules form a cyclical time relationship (e.g., day of the week).
}
\examples{

day_of_week <- new_cyclical_time_fn(day(1L), week(1L), default_calendar = cal_isoweek)
day_of_week(Sys.Date())

month_of_year <- new_cyclical_time_fn(month(1L), year(1L))
month_of_year(Sys.Date())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duration.R
\name{new_duration_fn}
\alias{new_duration_fn}
\title{Duration function factory}
\usage{
new_duration_fn(chronon, default_calendar = cal_gregorian)
}
\arguments{
\item{chronon}{A bare call for a time unit object representing the chronon
(e.g., \code{month(1L)}, \code{day(1L)}).}

\item{default_calendar}{A default calendar used to resolve the time units
if they don't exist in the calendar of the input data (e.g.,
\code{cal_gregorian}).}
}
\value{
A function used to create duration vectors with a specific chronon.
The returned function accepts:
\describe{
\item{\code{data}}{A numeric vector of duration magnitudes.}
\item{\code{calendar}}{A calendar system used to evaluate \code{chronon}. Defaults
to \code{time_calendar(data)}.}
\item{\code{...}}{Additional arguments passed to the chronon (e.g., \code{tz} for
timezones).}
}
}
\description{
\code{new_duration_fn()} creates a duration function for a specified chronon. A
chronon is the smallest indivisible time unit (e.g., days, months) that
defines what the numeric magnitudes in the resulting duration vector
represent.
}
\examples{
# Create a months duration function
months <- new_duration_fn(month(1L), default_calendar = cal_gregorian)
months(1:6)

# Create a days duration function
days <- new_duration_fn(day(1L), default_calendar = cal_gregorian)
days(1:7)

}
\seealso{
\itemize{
\item \code{\link[=duration]{duration()}} for creating duration vectors directly
\item \link{cal_gregorian}, \link{cal_isoweek} for calendar systems
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01_labels.R
\name{label_scheme}
\alias{label_scheme}
\alias{linear_labels}
\alias{cyclical_labels}
\title{Describe a granule's label scheme}
\usage{
label_scheme(
  start = 0L,
  vocab = NULL,
  transform = NULL,
  width = NULL,
  locale = NULL
)

linear_labels(granule, ...)

cyclical_labels(granule, cycle, ...)
}
\arguments{
\item{start}{Author-fixed. Raw index 0 displays as \code{start}, shared by
numeric rendering and vocab indexing (e.g. \code{start = 1L} makes January,
raw index 0, display as \code{"1"} and index \code{vocab()[[1]]}). A caller can't
override this: get it wrong and dates are wrong, not just styled
differently.}

\item{vocab}{\code{NULL} (numeric only), or a function \verb{function(type = NULL, locale = NULL)} returning either the full named list of renderings for
one locale (\code{type = NULL}), or one rendering (e.g.
\code{vocab("abbreviated")}). \code{\link[=vocab_table]{vocab_table()}} builds this from a plain
lookup table; see "Locale specific labels" below for how to specify labels.}

\item{transform}{\code{NULL} (plain \code{start} shift), or \verb{list(encode = function(i, at = NULL) -> vocab index, decode = function(d, at = NULL) -> raw i)}, overriding the indexing used for named rendering only. Both
functions are authored scalar (\code{if}/\code{return}) and vectorized
internally. See "Irregular cycles" below.}

\item{width, locale}{Caller-overridable defaults for \code{cyc(granule, cycle, width = ...)}/\code{locale = ...} at format/parse time, not calendar facts.
\code{width} zero-pads numeric rendering (e.g. \code{width = 2L} for \code{"02"});
\code{locale} selects which \code{vocab()} entry to use by default.}

\item{granule}{A time granule instance (e.g. \code{month(1L)}).}

\item{...}{Unused; only present for S7 dispatch.}

\item{cycle}{For \code{cyclical_labels()}, the coarser granule instance this
cycle repeats within (e.g. \code{year(1L)}).}
}
\value{
A function \code{function(granule, cycle) list(start =, vocab =, transform =, width =, locale =)}, assignable directly as a
\code{\link[=linear_labels]{linear_labels()}}/\code{\link[=cyclical_labels]{cyclical_labels()}} method. \code{granule}/\code{cycle} are only
used for dispatch: a scheme describes a granule class, not the instance
a method happens to be called with.

A plain list, as constructed by \code{\link[=label_scheme]{label_scheme()}}.
}
\description{
Builds the function a \code{\link[=linear_labels]{linear_labels()}}/\code{\link[=cyclical_labels]{cyclical_labels()}} method is:
assign its result directly, e.g. \code{method(cyclical_labels, list(granule, cycle)) <- label_scheme(...)}. Three levels, from plain to most irregular:

\code{linear_labels()} and \code{cyclical_labels()} are S7 generics returning a
granule's label scheme: a plain list describing how its internal position
renders as text and parses back (\code{start}, \code{vocab}, \code{transform}, \code{width},
\code{locale}; see \code{\link[=label_scheme]{label_scheme()}}).
}
\details{
\enumerate{
\item Plain numeric. No \code{vocab}: labels are \code{i + start}, optionally
zero-padded to \code{width}. The default for most granules (minutes,
seconds).
\item Named, regular. \code{vocab} supplies the names (see \code{\link[=vocab_table]{vocab_table()}});
\code{start} positions raw index 0 in both the numeric and vocab-indexed
rendering. Covers most named calendar units (months, weekdays).
\item Named, irregular. \code{transform} overrides the plain \code{start} shift with
hand-written \code{encode}/\code{decode}, for cycles where raw index and name
aren't a constant offset apart (e.g. a leap month splitting one name
into two). See "Irregular cycles" below.
}

A scheme is only consulted for named (\code{label = TRUE}) rendering and its
inverse; numeric rendering is always the plain \code{i + start} shift (see
\code{\link[=linear_labels_format]{linear_labels_format()}}). Labels that aren't index-shaped at all (e.g.
"1BC") override the generics directly, documented in \code{\link[=linear_labels]{linear_labels()}}
instead.

\code{\link[=linear_labels_format]{linear_labels_format()}}/\code{\link[=linear_labels_parse]{linear_labels_parse()}} (and their cyclical
counterparts) do the actual rendering/parsing, with the default method
looking up the scheme and applying it. A granule with labels that are
incompatible with label schemes (e.g. 2BC, 1BC, 1, 2, ...) registers a method
directly on those generics instead, skipping the scheme system. See
"Overriding the generics directly" below.
}
\section{Locale specific labels}{

\code{vocab} is a plain closure, not a shared registry, so two calendar
packages can't collide on it. To add locale support to an existing
granule, redefine the \code{linear_labels()}/\code{cyclical_labels()} method and
call the previous one for anything you're not changing: the normal S7
rule that the last \verb{method(...) <-} wins. That needs a hand-written
wrapper rather than assigning \code{label_scheme()}'s result directly, since
overriding one field means fetching the old scheme and changing it:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{scheme <- cyclical_labels(cal_gregorian$month(1L), cal_gregorian$year(1L))
method(cyclical_labels, list(cal_gregorian$month, cal_gregorian$year)) <-
  function(granule, cycle) \{
    scheme$vocab <- function(type = NULL, locale = NULL) \{
      if (identical(locale \%||\% "en-GB", "fr-FR")) \{
        fr <- list(
          wide = c("janvier", "février", "mars", "avril", "mai", "juin",
                   "juillet", "août", "septembre", "octobre", "novembre", "décembre"),
          abbreviated = c("janv.", "févr.", "mars", "avr.", "mai", "juin",
                          "juil.", "août", "sept.", "oct.", "nov.", "déc.")
        )
        return(if (is.null(type)) fr else fr[[type]])
      \}
      scheme$vocab(type, locale)
    \}
    scheme
  \}
}\if{html}{\out{</div>}}

Locale tags follow BCP 47 (\code{"en-GB"}, \code{"fr-FR"}), like ICU, \code{stringi} and
\code{clock}. Type names follow CLDR's wide/abbreviated/narrow, so CLDR data
(e.g. \code{stringi::stri_datetime_symbols()}) works with \code{\link[=vocab_table]{vocab_table()}}
unchanged. Other \code{type} values are allowed too, e.g. \code{"emoji"} for lunar
phases, picked at render time via \code{\link[=cyclical_labels_format]{cyclical_labels_format()}}'s \code{type}
argument.
}

\section{Irregular cycles}{

\code{at} (in \code{transform}'s \code{encode}/\code{decode}, and in
\code{\link[=cyclical_labels_format]{cyclical_labels_format()}}/\code{\link[=cyclical_labels_parse]{cyclical_labels_parse()}}) is the coarser
granule instance's raw position, not epoch-shifted, with \code{n = 1}. It
follows the same convention as \code{\link[=chronon_cardinality]{chronon_cardinality()}}. It lets
\code{transform} pick the right name when that name depends on which cycle
instance a raw index falls in, e.g. the Hebrew calendar's leap year
splitting "Adar" into "Adar I"/"Adar II":

\if{html}{\out{<div class="sourceCode r">}}\preformatted{month_names <- c("Tishrei", "Cheshvan", "Kislev", "Tevet", "Shevat",
                  "Adar", "Adar I", "Adar II",
                  "Nisan", "Iyar", "Sivan", "Tammuz", "Av", "Elul")

method(cyclical_labels, list(cal_hebrew$month, cal_hebrew$year)) <- label_scheme(
  start = 1L,
  vocab = vocab_table(`en-GB` = list(wide = month_names, abbreviated = month_names)),
  transform = list(
    encode = function(i, at) \{
      leap <- is_hebrew_leap_year(at)
      if (i <= 4L) return(i + 1L)
      if (!leap) return(if (i == 5L) 6L else i + 3L)
      if (i == 5L) return(7L)
      if (i == 6L) return(8L)
      i + 2L
    \},
    decode = function(d, at) \{
      leap <- is_hebrew_leap_year(at)
      if (d <= 5L) return(d - 1L)
      if (!leap) \{
        if (d == 6L) return(5L)
        if (d \%in\% c(7L, 8L)) cli::cli_abort("'Adar I'/'Adar II' are not valid outside a leap year.")
        return(d - 3L)
      \}
      if (d == 6L) cli::cli_abort("'Adar' is ambiguous in a leap year, use 'Adar I' or 'Adar II'.")
      if (d == 7L) return(5L)
      if (d == 8L) return(6L)
      d - 2L
    \}
  )
)
}\if{html}{\out{</div>}}

The aborts in \code{decode()} matter: without them, "Adar II" typed in a
common year would fall through to the same offset as plain "Adar" and
silently parse to the wrong month. Only the label layer, which knows
both the text and \code{at}, can catch that. \code{transform}'s leap predicate must
also be the same one \code{\link[=chronon_cardinality]{chronon_cardinality()}} uses for this pair; put it
in one shared helper.
}

\section{Overriding the format and parsing methods}{

\code{linear_labels_format()}/\code{linear_labels_parse()} are ordinary S7
generics with a default method, not the only place formatting logic can
live. So escaping the scheme system for one granule is just registering
a method directly:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{method(linear_labels_format, cal_gregorian$year) <- function(granule, i, ...) \{
  ifelse(i <= 0L, paste0(-i + 1L, "BC"), i)
\}
method(linear_labels_parse, cal_gregorian$year) <- function(granule, ...) \{
  list(
    pattern = "\\\\d+(?:BC)?",
    decode = function(text, at = NULL) \{
      bc <- grepl("BC$", text)
      n <- as.integer(sub("BC$", "", text))
      ifelse(bc, 1L - n, n) - chronon_epoch(granule)
    \}
  )
\}
}\if{html}{\out{</div>}}

\code{decode()} needs \code{- chronon_epoch(granule)} because this method bypasses
the default, which normally does that step (see \code{\link[=linear_labels_format]{linear_labels_format()}}'s
"Epoch shift" section).

The \code{...} on \verb{_format()} is not just style: \code{format()} forwards whatever
\link{component_helpers}' \code{lin()}/\code{cyc()} were called with (\code{label},
\code{abbreviate}, \code{type}, \code{width}, \code{locale}, ...) as named arguments. A
method without \code{...} errors with "unused argument" as soon as a caller
passes something it doesn't declare.

Registering only \code{linear_labels_format()} is fine: that's the same "no
parsing yet" state a fresh granule starts in. But registering it without
\code{linear_labels_parse()} leaves parsing silently wrong instead of just
missing: text like \code{"1BC"} falls through to whatever the granule's
scheme says, usually plain numeric decoding, which doesn't understand
"BC" at all.
}

\examples{
# A scheme for months of the year, using R's localised month names.
S7::method(cyclical_labels, list(cal_gregorian$month, cal_gregorian$year)) <- label_scheme(
  start = 1L, width = 2L,
  vocab = vocab_table(`en-GB` = list(wide = month.name, abbreviated = month.abb))
)

}
\seealso{
\code{\link[=linear_labels]{linear_labels()}}/\code{\link[=cyclical_labels]{cyclical_labels()}}, the generics a scheme is
returned from; \code{\link[=vocab_table]{vocab_table()}} for the common hand-listed name table
case.

\code{\link[=label_scheme]{label_scheme()}} for the fields a method returns;
\code{\link[=linear_labels_format]{linear_labels_format()}}/\code{\link[=cyclical_labels_format]{cyclical_labels_format()}} for the generics
that interpret it; \code{\link[=vocab_table]{vocab_table()}} for the common hand-listed name
table case.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chronon_common.R
\name{chronon_common}
\alias{chronon_common}
\alias{chronon_common.mixtime}
\title{Find the common chronon of a time object}
\usage{
chronon_common(x, ...)

chronon_common.mixtime(x, .ptype = NULL, ...)
}
\arguments{
\item{x}{A time object (typically a \code{\link{mixtime}}).}

\item{...}{Additional arguments for methods.}

\item{.ptype}{If NULL, the default, the output returns the common chronon
across all chronons of \code{x}. Alternatively, a prototype chronon can be
supplied to \code{.ptype} to demand a specific chronon is used. If the supplied
\code{.ptype} cannot represent all input chronons without loss of information,
an error is raised.}
}
\value{
A time granule object representing the common chronon.
}
\description{
This utility function takes a set of chronons and identifies a common chronon
of the finest granularity that can represent all input chronons without loss
of information. This is useful for operations that require a shared time
granule, such as combining or comparing different time measured at different
precisions.

The result is obtained by finding the greatest lower bound (GLB) of the input
chronons using the ordered relationships defined by \code{chronon_cardinality()}
methods. The GLB represents the finest chronon that can represent all input
chronons without loss of information.
}
\examples{
# The common chronon between a year-month and a day is a day
chronon_common(c(yearmonth(Sys.Date()), date(Sys.Date())))

# The common chronon between a Gregorian month and an ISO week is a day
chronon_common(c(yearmonth(Sys.Date()), yearweek(Sys.Date())))

# The common chronon between a ISO week and an hour is an hour
chronon_common(c(yearweek(Sys.Date()), linear_time(Sys.time(), hour(1L))))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time_chronon.R
\name{time_chronon}
\alias{time_chronon}
\title{Obtain the chronon of a time object}
\usage{
time_chronon(x, ...)
}
\arguments{
\item{x}{A time object (e.g., \link[base:Date]{base::Date}, \link[base:POSIXct]{base::POSIXct}, \code{\link[=linear_time]{linear_time()}}, etc.)}

\item{...}{Additional arguments for methods.}
}
\value{
A time \code{\link[=duration]{duration()}} vector representing the chronon of each
value (e.g., \code{days(1L)}).
}
\description{
This S7 generic function extracts the chronon (the smallest time granule) from a
time object, such as continuous time or cyclical time representations.
}
\examples{

# The chronon of a Date object is 1 day
time_chronon(Sys.Date())

# The chronon of a POSIXct object is 1 second
time_chronon(Sys.time())

# The chronon of a continuous time year and month is 1 month
time_chronon(yearmonth(Sys.Date()))

# The common chronon of a mixed time object is the finest chronon
time_chronon(c(yearmonth(Sys.Date()), Sys.Date()))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/timezone.R
\name{tz_abbreviation}
\alias{tz_abbreviation}
\title{Get timezone abbreviation}
\usage{
tz_abbreviation(x, tz = tz_name(x))
}
\arguments{
\item{x}{A POSIXct datetime object or something coercible to POSIXct.
The timezone is extracted from this object.}

\item{tz}{A character vector of timezones to abbreviate at time point \code{x}.}
}
\value{
A character vector of timezone abbreviations.
}
\description{
Returns the timezone abbreviation (e.g., "EST", "PDT") for a given datetime
in its specified timezone.
}
\examples{
tz_abbreviation(Sys.time())
tz_abbreviation(as.POSIXct("2024-01-15 12:00:00", tz = "America/New_York"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cal0_time_civil.R
\name{calendar_time_civil}
\alias{calendar_time_civil}
\alias{cal_time_civil}
\title{Civil time unit classes}
\usage{
cal_time_civil
}
\value{
A time granule object for the civil time system.
}
\description{
Time unit constructors for the civil time system where the boundary of each
day is at midnight on the 24 hour clock. This calendar is intended to be
built on by other calendars (e.g. \verb{[cal_time_civil]} and \verb{[cal_isoweek]}) to
add common time components.  These units can be
used with \code{\link[=linear_time]{linear_time()}} to create custom time representations.
}
\details{
The following time units are available (\verb{cal_time_civil$}).
\itemize{
\item \code{day()}: Day unit
\item \code{hour()}: Hour unit
\item \code{minute()}: Minute unit
\item \code{second()}: Second unit
\item \code{millisecond()}: Millisecond unit
\item \code{microsecond()}: Microsecond unit
\item \code{nanosecond()}: Nanosecond unit
}
}
\examples{
# Create a custom time representation using civil time granules
hms <- new_cyclical_time_fn(
  chronon = second(1L),
  cycle = hour(1L)
)

}
\seealso{
\code{\link{cal_time_civil}}, \code{\link{cal_isoweek}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01_chronon_parse.R
\name{chronon_parse_linear}
\alias{chronon_parse_linear}
\alias{chronon_parse_cyclical}
\title{Default parsing format strings for chronons}
\usage{
chronon_parse_linear(x, cal = time_calendar(x), ...)

chronon_parse_cyclical(x, y, ...)
}
\arguments{
\item{x}{A time granule for the chronon.}

\item{cal}{The calendar of the chronon, used to disambiguate suitable format
strings for time units that are shared across calendars (e.g.
\code{cal_gregorian$day} and \code{cal_isoweek$day}).}

\item{...}{Additional arguments for methods.}

\item{y}{A time granule for the cycle}
}
\value{
A character vector of format templates, ordered from most to
least common, built with \code{\link[=parse_format]{parse_format()}}. The first element typically
matches \code{\link[=chronon_format_linear]{chronon_format_linear()}} (for \code{chronon_parse_linear()}) or
\code{\link[=chronon_format_cyclical]{chronon_format_cyclical()}} (for \code{chronon_parse_cyclical()}) for the
same chronon.
}
\description{
Provides candidate format strings for parsing text into a given chronon
(finest time granule), for use as the \code{format} argument of
\code{\link[=time_parse]{time_parse()}}. Dispatches the same way as
\code{\link[=chronon_format_linear]{chronon_format_linear()}}/\code{\link[=chronon_format_cyclical]{chronon_format_cyclical()}}, but returns every
common format instead of a single default, so \code{\link[=time_parse]{time_parse()}} can try each
in turn and keep whichever parses the most values. Methods should build
their return value with \code{\link[=parse_format]{parse_format()}}.
}
\examples{
chronon_parse_linear(cal_gregorian$year(1L))
chronon_parse_linear(cal_gregorian$month(1L))
chronon_parse_linear(cal_gregorian$day(1L))
chronon_parse_linear(cal_isoweek$day(1L))

chronon_parse_cyclical(cal_gregorian$month(1L), cal_gregorian$year(1L))
chronon_parse_cyclical(cal_isoweek$day(1L), cal_isoweek$week(1L))

}
\seealso{
\code{\link[=parse_format]{parse_format()}} for building candidate format strings (what
these methods are built from), \code{\link[=time_parse]{time_parse()}} for using the result as
candidate formats, \code{\link[=chronon_format_linear]{chronon_format_linear()}}/\code{\link[=chronon_format_cyclical]{chronon_format_cyclical()}}
for the single default format these are based on.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01_chronon_format.R
\name{chronon_format_attr}
\alias{chronon_format_attr}
\title{Default formatting strings for chronon attributes}
\usage{
chronon_format_attr(x, ...)
}
\arguments{
\item{x}{A chronon (time granule) object.}

\item{...}{Additional arguments for methods.}
}
\value{
A character string containing the default format suffix for the chronon.
}
\description{
Provides suffixes for default formatting strings for a given chronon (time granule).
This provides useful information such as timezones or locations in the string.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/timezone.R
\name{tz_name}
\alias{tz_name}
\title{Extract timezone from an object}
\usage{
tz_name(x, ...)
}
\arguments{
\item{x}{An object with timezone information.}

\item{...}{Additional arguments passed to methods.}
}
\value{
A character vector representing the timezone of each time point
(e.g., "America/New_York", "UTC").
}
\description{
Generic function to extract the timezone from objects that have timezone information.
}
\examples{
tz_name(Sys.time())
tz_name(as.POSIXct("2024-06-15 12:00:00", tz = "America/New_York"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cal_sym454.R
\name{calendar_sym454}
\alias{calendar_sym454}
\alias{cal_sym454}
\title{Symmetry454 time unit classes}
\format{
A civil-based calendar containing Symmetry454 time units.
}
\usage{
cal_sym454
}
\value{
An S3 list of class \code{c("cal_sym454", "mt_calendar")} containing
the named time unit classes of the Symmetry454 calendar. Each unit is
accessible via \code{$} notation and calling it with a step size produces a
time granule (e.g., 1 week granule as \code{cal_sym454$week(1L)}).
}
\description{
Time unit constructors for the Symmetry454 calendar system. These units can be
used with \code{\link[=linear_time]{linear_time()}} to create custom time representations.
}
\details{
The Symmetry454 calendar (Sym454) is a perennial solar calendar proposed by
Dr. Irv Bromberg. It preserves the traditional 12-month structure and 7-day
week, with months arranged in a symmetrical 4-5-4 week pattern per quarter.
Every month starts on Monday and has a whole number of weeks, meaning no
month ever contains a partial week.

The following time units are available in the Symmetry454 calendar (\verb{cal_sym454$}).
\itemize{
\item \code{year()}: Year unit
\item \code{month()}: Month unit
\item \code{week()}: Week unit
\item \code{day()}: Day unit
\item \code{hour()}: Hour unit
\item \code{minute()}: Minute unit
\item \code{second()}: Second unit
\item \code{millisecond()}: Millisecond unit
}
\subsection{Leap years}{

Rather than intercalary days, Symmetry454 uses a \strong{leap week} appended to
December once every 5 or 6 years, making December a 5-week month in leap
years. A year is a leap year if \code{(52 * year + 146) \%\% 293 < 52}.

This yields a mean year of 365 + 71/293 days (approx 365 days, 5 hours, 48
minutes, 56.5 seconds), intentionally slightly shorter than the mean
northward equinoctial year.
}
}
\examples{
# Create a custom time representation using Symmetry454 units
linear_time(
  Sys.time(),
  chronon = cal_sym454$week(1L)
)

}
\seealso{
\code{\link[=linear_time]{linear_time()}} for creating linear time points, and
\url{https://en.wikipedia.org/wiki/Symmetry454} for more calendar details.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/timezone.R
\name{tz_transitions}
\alias{tz_transitions}
\title{Get timezone transitions}
\usage{
tz_transitions(start, end)
}
\arguments{
\item{start}{A POSIXct datetime object or something coercible to POSIXct,
representing the start of the time range. The timezone is extracted from this object.}

\item{end}{A POSIXct datetime object or something coercible to POSIXct,
representing the end of the time range.}
}
\value{
A data frame with columns:
\itemize{
\item \code{time}: A \code{mixtime} linear time point (continuous, UTC seconds) giving
the instant of the transition.
\item \code{offset_before}, \code{offset_after}: \code{mixtime} durations (UTC seconds)
giving the UTC offset immediately before and after the transition.
}
}
\description{
Returns all timezone transitions (e.g., daylight saving time changes) that
occur between two datetimes. The timezone is taken from the start datetime.
}
\examples{
# Get all DST transitions in 2024 for New York
tz_transitions(
  as.POSIXct("2024-01-01", tz = "America/New_York"),
  as.POSIXct("2024-12-31", tz = "America/New_York")
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01_calendar.R
\name{time_calendar}
\alias{time_calendar}
\title{Obtain the calendar of a time object}
\usage{
time_calendar(x, ...)
}
\arguments{
\item{x}{A time object (e.g., \link[base:Date]{base::Date}, \link[base:POSIXct]{base::POSIXct}, \code{\link[=linear_time]{linear_time()}}, etc.)}

\item{...}{Additional arguments for methods.}
}
\value{
A calendar object (e.g., \code{cal_gregorian}, \code{cal_isoweek})
}
\description{
This S7 generic function extracts the calendar system from a time object.
The calendar defines the collection of time units (years, months, days, etc.)
used to interpret the time representation.
}
\examples{

# The calendar of a Date object is the Gregorian calendar
time_calendar(Sys.Date())

# The calendar of a POSIXct object is also Gregorian
time_calendar(Sys.time())

# The calendar of a yearweek object is the ISO week calendar
time_calendar(yearweek(Sys.Date()))

# A mixed time object returns a list of calendars
time_calendar(c(yearmonth(Sys.Date()), Sys.Date()))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format.R
\name{format.mt_time}
\alias{format.mt_time}
\alias{format,mt_time-method}
\title{Format mixtime vectors}
\usage{
\S4method{format}{mt_time}(x, ..., attr = TRUE)
}
\arguments{
\item{x}{A mixtime vector.}

\item{...}{Additional arguments for methods, including \code{format}: a
glue-style format string, defaulting to one derived automatically for \code{x}.}

\item{attr}{If \code{TRUE} (default), append attribute information (e.g.
timezone) to the default format.}
}
\value{
A character vector the same length as \code{x}.
}
\description{
Formats a mixtime vector as a character vector, using a glue-style format
string of \code{{lin(...)}}/\code{{cyc(...)}} tokens tailored to the vector's chronon
and cycle. If \code{format} is omitted, a sensible default is derived
automatically. See \code{vignette("time-format-strings")} for the format string
syntax.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/components.R
\name{component_helpers}
\alias{component_helpers}
\alias{lin}
\alias{cyc}
\title{Linear and cyclical component helpers}
\usage{
lin(granule)

cyc(granule, cycle)
}
\arguments{
\item{granule}{The time granule to address, given as a granule generator (e.g.
\code{year}) or a sized time unit (e.g. \code{year(1L)}). Resolved in the calendar of
the time vector being formatted or decomposed.}

\item{cycle}{The coarser granule defining the cycle a \code{cyc()} component repeats
within (e.g. \code{year} in \code{cyc(month, year)}).}
}
\value{
A component specification, consumed internally by the component-aware
context (e.g. \code{\link[=format]{format()}} or \code{\link[=time_components]{time_components()}}).
}
\description{
\code{lin()} and \code{cyc()} name the time components addressed by mixtime's
\emph{component-aware contexts} — a single vocabulary shared across \code{\link[=format]{format()}} (and
parsing) format strings and \code{\link[=time_components]{time_components()}} expressions. They are only
meaningful inside one of these contexts; calling them directly is an error.
}
\details{
A linear and a cyclical component of the same granule store the \emph{same} value:
the count of that chronon since the Unix epoch. The reduction to a within-cycle
position happens only when a cyclical vector is formatted. \code{cyc()} therefore
behaves like \code{lin()} on its finest granule but additionally records the cycle.
}
\seealso{
\code{\link[=format]{format()}} and \code{\link[=time_components]{time_components()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/round.R
\name{time_round}
\alias{time_round}
\alias{time_floor}
\alias{time_ceiling}
\alias{trunc_time}
\title{Round, floor and ceiling transformations for time objects}
\usage{
time_round(x, granule, ...)

time_ceiling(x, granule, ...)

time_floor(x, granule, ...)
}
\arguments{
\item{x}{A date/time object to be rounded. Accepted types include Date,
POSIXct, POSIXlt and other objects that inherit from POSIXt. The returned
object will be of the same class as the input.}

\item{granule}{A time granule (or object coercible to a time granule, e.g. "day").}

\item{...}{Additional arguments passed to specific implementations.}
}
\value{
An object of the same class as x with its time components adjusted
to the requested granule.
}
\description{
A family of helpers to round date/time objects to a specified time granule
such as second, minute, hour, or day. These functions preserve the input
time class, as rounded by the attributes of the \code{granule}.
}
\examples{
# Round POSIXct to the nearest minute (preserving tz)
t <- as.POSIXct("2020-01-01 12:34:56", tz = "UTC")
time_round(t, granule = cal_gregorian$minute(1L))

# Floor to the nearest hour
time_floor(t, granule = cal_gregorian$hour(1L))

# Ceiling a Date (treated as midnight-of-day rounding)
d <- as.Date("2020-01-01")
time_ceiling(d, granule = cal_gregorian$month(1L))

}
\seealso{
\link[base:round]{base::round}, \link[lubridate:round_date]{lubridate::round_date}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01_time_unit.R
\name{time_unit_full}
\alias{time_unit_full}
\alias{time_unit_plural}
\alias{time_unit_abbr}
\title{Time units as a string}
\usage{
time_unit_full(x, ...)

time_unit_plural(x, n = 2L)

time_unit_abbr(x, ...)
}
\arguments{
\item{x}{A time granule object (e.g., \code{cal_gregorian$month(1L)})}

\item{...}{Additional arguments for methods.}

\item{n}{Numeric quantity used to select singular or plural form.}
}
\value{
A string representing the time unit
}
\description{
These S7 generic functions provide the full and abbreviated names for time
units. \code{time_unit_full()} is used in messages and durations (e.g., "2 months").
\code{time_unit_abbr()} is used for tsibble index interval displays (e.g., "1M").
\code{time_unit_plural()} pluralises the full unit name for a given quantity using
cli-style pluralisation (e.g., \code{"year{?/s}"} becomes \code{"year"} or \code{"years"}).
}
\examples{
time_unit_full(cal_gregorian$year(1L))
time_unit_plural(cal_gregorian$year(1L), 1L)
time_unit_plural(cal_gregorian$year(1L), 2L)
time_unit_abbr(cal_gregorian$year(1L))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mixtime-package.R
\docType{package}
\name{mixtime-package}
\alias{mixtime-package}
\title{mixtime: Mixed Temporal Vectors and Operations}
\description{
\if{html}{\figure{logo.svg}{options: style='float: right' alt='logo' width='120'}}

Flexible time classes for time series analysis and forecasting with mixed temporal granularities. Supports linear and cyclical time representations in discrete and continuous forms, with timezone support, across multiple calendar systems including Gregorian and ISO week date calendars. Time points are stored numerically relative to a chronon; an atomic time granule defined by time units of a calendar. Calendrical arithmetic enables conversion between time granules (e.g. days to months) and calendar systems. Multi-unit arithmetic allows for temporal analysis with other granules of common calendars (e.g. fortnights are 2-week units). Time vectors of different granularities (e.g. monthly and quarterly) can be combined in a single vector, making 'mixtime' ideal for data that changes observation frequency over time or requires temporal reconciliation across scales. The package is extensible, allowing users to define custom calendars that build upon civil and astronomical time systems.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://pkg.mitchelloharawild.com/mixtime/}
  \item \url{https://github.com/mitchelloharawild/mixtime}
  \item Report bugs at \url{https://github.com/mitchelloharawild/mixtime/issues}
}

}
\author{
\strong{Maintainer}: Mitchell O'Hara-Wild \email{mail@mitchelloharawild.com} (\href{https://orcid.org/0000-0001-6729-7695}{ORCID})

Authors:
\itemize{
  \item Mitchell O'Hara-Wild \email{mail@mitchelloharawild.com} (\href{https://orcid.org/0000-0001-6729-7695}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duration.R
\name{duration}
\alias{duration}
\title{Duration vectors}
\usage{
duration(
  data,
  chronon = time_chronon(data),
  discrete = NULL,
  calendar = time_calendar(data)
)
}
\arguments{
\item{data}{A time vector of duration magnitudes, or an existing
\code{\link[=duration]{duration()}} vector to convert to \code{chronon} granules.}

\item{chronon}{A time granule expression representing the chronon, evaluated
in the context of \code{calendar}. Use unquoted expressions like \code{month(1L)} or
\code{day(1L)}. Chronons from a specific calendar can also be used (e.g.
\code{cal_gregorian$month(1L)}). Defaults to the time chronon of the input
\code{data} (\code{time_chronon(data)}).}

\item{discrete}{Logical. If \code{TRUE} (default), returns integer durations
always rounding down (discrete time model). If \code{FALSE}, returns fractional
durations (continuous time model).}

\item{calendar}{Calendar system used to evaluate \code{chronon}. Defaults to
\code{time_calendar(data)} for existing time objects. Common options include
\link{cal_gregorian} and \link{cal_isoweek}.}
}
\value{
A \code{mixtime} vector containing an \code{mt_duration} vector.
}
\description{
\code{duration()} creates a vector of durations with a specified chronon.
Durations represent a fixed span of time measured in a given time granule
(e.g., 3 months, 5 days), without reference to a specific point in time.
}
\examples{
# A duration of 3 months
duration(3L, cal_gregorian$month(1L))

# A vector of durations in days
duration(1:7, cal_gregorian$day(1L))

# Convert a duration of 4 days into weeks
duration(days(4), cal_isoweek$week(1L), discrete = FALSE)
duration(days(4), cal_isoweek$week(1L), discrete = TRUE)

}
\seealso{
\itemize{
\item \code{\link[=new_duration_fn]{new_duration_fn()}} for creating reusable duration functions
\item \link{cal_gregorian}, \link{cal_isoweek} for calendar systems
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/00_classes.R
\name{class_mixtime}
\alias{class_mixtime}
\title{Base S7 class for mixtime vector objects}
\usage{
class_mixtime(x = list(), i = seq_len(sum(lengths(x))))
}
\arguments{
\item{x}{A list of \code{"mt_time"} vectors, see \code{\link[=new_time]{new_time()}} for details.}

\item{i}{A vector of integers specifying the location of each element in \code{x}
as if they were combined in order. The values in \code{i} must be between 1 and
the total number of elements across all vectors in \code{x}, and can contain
duplicates. If not provided, it defaults to a sequence from 1 to the total
number of elements across all vectors in \code{x}.}
}
\value{
When used as a class definition (e.g., in \code{S7::method(generic, class_mixtime)}), an S7 class object representing the \code{mixtime} class,
inheriting from \link[vecvec:class_vecvec]{vecvec::class_vecvec}. When called as a constructor
(\code{class_mixtime(list(...))}), a mixtime vector of S7 class \code{mixtime}
(also inheriting the S3 class \code{"mixtime"}), containing the supplied list
of time vectors as a \link[vecvec:class_vecvec]{vecvec::class_vecvec} structure. End users should
prefer \code{\link[=mixtime]{mixtime()}} or \code{\link[=new_mixtime]{new_mixtime()}} for construction.
}
\description{
\code{class_mixtime} is the base S7 class for all mixtime vector objects,
inheriting from \link[vecvec:class_vecvec]{vecvec::class_vecvec}. While not intended to be used
directly, this S7 class is suitable to use when defining S7 methods for
mixtime vectors. S3 methods can be defined using the \code{mixtime::mixtime}
class.
}
\seealso{
\code{\link[=mixtime]{mixtime()}} for creating mixtime vectors, and \code{\link[=new_mixtime]{new_mixtime()}} for
the low-level constructor function of this S7 class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01_chronon_cardinality.R
\name{chronon_cardinality_fixed}
\alias{chronon_cardinality_fixed}
\title{Fixed cardinality between time granules}
\usage{
chronon_cardinality_fixed(x, y, ...)
}
\arguments{
\item{x}{The finer time granule (e.g. \code{cal_gregorian$month(1L)})}

\item{y}{The coarser time granule (e.g. \code{cal_gregorian$year(1L)})}

\item{...}{Additional arguments for methods.}
}
\value{
A single number describing how many unit \code{x} time granules fit
into a unit \code{y} time granule.
}
\description{
A restricted variant of \code{\link[=chronon_cardinality]{chronon_cardinality()}} for time granule pairs
whose relationship is a constant, context-independent number (e.g., 60
seconds in a minute, 24 hours in a day). Unlike \code{chronon_cardinality()},
methods for this generic do not receive (and must not need) an \code{at} time
point, and should return the number of unit (\code{n = 1L}) \code{x} granules that
fit within one unit \code{y} granule.
}
\details{
Defining a method for \code{chronon_cardinality_fixed()} automatically provides
a \code{chronon_cardinality()} method for the pair (scaled by the requested
granule sizes via the \code{list(mt_unit, mt_unit)} fallback method), and marks
the relationship as safe to use for \code{\link[=chronon_divmod]{chronon_divmod()}}'s graph traversal,
where variable (context-dependent) cardinalities cannot be used since no
\code{at} is available mid-traversal.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mixtime.R
\name{new_mixtime}
\alias{new_mixtime}
\title{Constructor for mixtime vectors}
\usage{
new_mixtime(x = mt_linear())
}
\arguments{
\item{x}{A mixtime time vector (created with \code{\link[=new_time]{new_time()}}) to wrap in a mixtime class.}
}
\value{
A \code{mixtime} object, which allows mixed-type time vectors to coexist in a single vector.
}
\description{
Creates a \code{mixtime} vector, which can contain time points of different
granularities (e.g. monthly and quarterly) in a single vector via \code{vecvec}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linear_time.R
\name{linear_time}
\alias{linear_time}
\title{Linear time points}
\usage{
linear_time(
  data,
  chronon = time_chronon(data),
  discrete = TRUE,
  calendar = time_calendar(data)
)
}
\arguments{
\item{data}{Input data to convert to linear time. Can be:
\itemize{
\item Numeric values (interpreted as chronons since Unix epoch)
\item Character strings (parsed as dates/times)
\item Date or POSIXct objects
\item Other time objects
}}

\item{chronon}{A time granule expression representing the chronon (smallest
indivisible time granule), evaluated in the context of \code{calendar}. Use
unquoted expressions like \code{month(1L)} or \code{hour(1L)}. Chronons from a
specific calendar can also be used (e.g. \code{cal_isoweek$week(1L)}).
Defaults to the time chronon of the input \code{data} (\code{time_chronon(data)}).}

\item{discrete}{Logical. If \code{TRUE} (default), returns integer chronons since
Unix epoch (discrete time model). If \code{FALSE}, returns fractional chronons
allowing representation of fractional time granules (continuous time model).}

\item{calendar}{Calendar system used to evaluate \code{chronon} and \code{granules}.
Defaults to \code{time_calendar(data)} for existing time objects. Common options
include \link{cal_gregorian} and \link{cal_isoweek}.}
}
\value{
A \code{mixtime} time vector containing an \code{mt_linear} vector.
}
\description{
\code{linear_time()} creates a vector of linear time points with a specified
chronon (smallest time granule). This function is useful for creating custom
time representations that aren't covered by the convenience functions like
\code{\link[=yearmonth]{yearmonth()}} or \code{\link[=yearweek]{yearweek()}}.
}
\examples{
# Hourly time
linear_time(
  Sys.time(),
  chronon = hour(1L)
)

# Monthly time
linear_time(
  Sys.Date(),
  chronon = month(1L)
)

# Discrete vs continuous time
linear_time(Sys.time(), chronon = day(1L), discrete = TRUE)
linear_time(Sys.time(), chronon = day(1L), discrete = FALSE)

# ISO week calendar with week-day structure
linear_time(
  Sys.Date(),
  chronon = day(1L),
  calendar = cal_isoweek
)

}
\seealso{
\itemize{
\item \code{\link[=new_linear_time_fn]{new_linear_time_fn()}} for creating reusable linear time functions
\item \code{\link[=yearmonth]{yearmonth()}}, \code{\link[=yearquarter]{yearquarter()}}, \code{\link[=year]{year()}} for Gregorian time representations
\item \code{\link[=yearweek]{yearweek()}} for ISO 8601 week-based time
\item \link{cal_gregorian}, \link{cal_isoweek} for calendar systems
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/components.R
\name{time_components}
\alias{time_components}
\title{Extract linear and cyclical time components}
\usage{
time_components(x, ..., calendar = time_calendar(x))
}
\arguments{
\item{x}{A \code{mixtime} (or an object coercible to one via \code{\link[=as_mixtime]{as_mixtime()}}, such
as a \code{Date} or \code{POSIXct}).}

\item{...}{Named expressions using \code{\link[=lin]{lin()}} and \code{\link[=cyc]{cyc()}} describing the
components to extract. The granule names (e.g. \code{year}, \code{month}, \code{day}) are
resolved in the calendar of \code{x}.}

\item{calendar}{Calendar system used to resolve granule names, overlaid on the
calendar of \code{x}. Defaults to \code{time_calendar(x)}. Supply e.g. \link{cal_isoweek}
to make ISO \code{week}-based components available.}
}
\value{
A data frame with one column per requested component. \code{lin()} columns
are linear (\code{mt_linear}) time vectors and \code{cyc()} columns are cyclical
(\code{mt_cyclical}) time vectors.
}
\description{
\code{time_components()} decomposes a time vector into its constituent parts using
\code{dplyr::mutate()}-like semantics. Each named expression is built from the
\code{\link[=lin]{lin()}} and \code{\link[=cyc]{cyc()}} helpers (the same vocabulary used in \code{\link[=format]{format()}} strings)
and produces a component time vector:
}
\details{
\itemize{
\item \verb{lin(<granule>)} extracts a \strong{linear} component (a non-repeating count,
e.g. the year), returning a linear time vector.
\item \verb{cyc(<granule>, <cycle>)} extracts a \strong{cyclical} component (a repeating
position within a larger cycle, e.g. the month within the year), returning a
cyclical time vector.
}

All requested components are computed together in a single decomposition of
the underlying time vector (via \code{chronon_parts()}), reusing the shared
recursive \code{chronon_divmod()} results rather than converting each component
independently.
}
\examples{
t <- yearmonth(as.Date("2026-02-14") + c(0, 40, 400))

# Extract the year (linear) and month-of-year (cyclical)
time_components(t, yr = lin(year), mth = cyc(month, year))

# Components can be named automatically from the expression
time_components(as.Date("2025-12-15") + 0:3, cyc(day, cal_isoweek$week))

}
\seealso{
\code{\link[=lin]{lin()}} and \code{\link[=cyc]{cyc()}} for the component helpers, \code{\link[=linear_time]{linear_time()}} and
\code{\link[=cyclical_time]{cyclical_time()}} for constructing individual component vectors, and
\code{\link[=format]{format()}} for the string counterpart of this interface.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cyclical_time.R
\name{cyclical_time}
\alias{cyclical_time}
\title{Cyclical time points}
\usage{
cyclical_time(
  data,
  chronon = time_chronon(data),
  cycle = time_cycle(data),
  discrete = TRUE,
  calendar = time_calendar(data)
)
}
\arguments{
\item{data}{Input data to convert to cyclical time. Can be:
\itemize{
\item Numeric values (interpreted as chronons, 1-indexed)
\item Character strings (parsed as dates/times)
\item Date or POSIXct objects
\item Other time objects
}}

\item{chronon}{A time granule representing the chronon (finest
indivisible time granule), evaluated in the context of \code{calendar}. Use
unquoted expressions like \code{day(1L)} or \code{month(1L)}. Chronons from a
specific calendar can also be used (e.g. \code{cal_isoweek$day(1L)}).}

\item{cycle}{A time granule representing the cycle (coarser time granule
that defines the period), evaluated in the context of \code{calendar}. Use
unquoted expressions like \code{week(1L)} or \code{year(1L)}.}

\item{discrete}{Logical. If \code{TRUE} (default), returns integer positions within
the cycle (discrete time model). If \code{FALSE}, returns fractional positions
allowing representation of fractional time chronons (continuous time model).}

\item{calendar}{Calendar system used to evaluate \code{chronon} and \code{cycle}.
Defaults to \code{time_calendar(data)} for existing time objects. Common options
include \link{cal_gregorian} and \link{cal_isoweek}.}
}
\value{
A \code{mixtime} time vector containing an \code{mt_cyclical} vector.
}
\description{
\code{cyclical_time()} creates a vector of cyclical time points representing
positions within repeating cycles. This function is useful for creating custom
cyclical time representations that aren't covered by the convenience functions
like \code{\link[=day_of_week]{day_of_week()}} or \code{\link[=month_of_year]{month_of_year()}}.
}
\examples{
# Day of week (1-7, Monday = 1)
cyclical_time(
  Sys.Date(),
  chronon = day(1L),
  cycle = week(1L),
  calendar = cal_isoweek
)

# Month of year (1-12)
cyclical_time(
  Sys.Date(),
  chronon = month(1L),
  cycle = year(1L)
)

# Continuous time (discrete = FALSE) for fractional month of year
cyclical_time(
  Sys.Date(),
  chronon = month(1L),
  cycle = year(1L),
  discrete = FALSE
)

# Day of month with Gregorian calendar
cyclical_time(
  Sys.Date(),
  chronon = day(1L),
  cycle = month(1L),
  calendar = cal_gregorian
)

# Hours, minutes, and seconds
cyclical_time(
  Sys.time(),
  chronon = second(1L),
  cycle = day(1L)
)

}
\seealso{
\itemize{
\item \code{\link[=new_cyclical_time_fn]{new_cyclical_time_fn()}} for creating reusable cyclical time functions
\item \code{\link[=day_of_week]{day_of_week()}}, \code{\link[=day_of_month]{day_of_month()}}, \code{\link[=day_of_year]{day_of_year()}} for common cyclical representations
\item \code{\link[=month_of_year]{month_of_year()}}, \code{\link[=week_of_year]{week_of_year()}} for other cyclical time helpers
\item \link{cal_gregorian}, \link{cal_isoweek} for calendar systems
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01_chronon_parse.R
\name{parse_format}
\alias{parse_format}
\title{Candidate format strings for parsing time}
\usage{
parse_format(..., regex = FALSE)
}
\arguments{
\item{...}{Format strings, as for the \code{format} argument of \code{\link[=time_parse]{time_parse()}}.}

\item{regex}{Whether the literal (non-token) text in \code{...} should be
matched as regular expression syntax rather than escaped literally
(e.g. \code{"[/-]"} to accept either \code{/} or \code{-} as a separator). This is a
niche option for irregular text - most formats leave it at the
default, \code{FALSE}. See the \code{regex} argument of \code{\link[=time_parse]{time_parse()}} for
details.}
}
\value{
A character vector of format strings suitable for \code{\link[=time_parse]{time_parse()}},
with a \code{"regex"} attribute for the \code{regex} argument the parser.
}
\description{
Combines one or more format strings into a character vector for use as
the \code{format} argument of \code{\link[=time_parse]{time_parse()}}. When multiple format strings are
given, \code{\link[=time_parse]{time_parse()}} tries each in turn and keeps whichever parses the
most values, so \code{parse_format()} is the usual way to build up a set of
candidates to try (it's also how \code{\link[=chronon_parse_linear]{chronon_parse_linear()}}/
\code{\link[=chronon_parse_cyclical]{chronon_parse_cyclical()}} methods build theirs). It can optionally mark
its candidates as using regex syntax rather than literal text; see the
\code{regex} argument below.
}
\examples{
parse_format("{lin(year)}-{cyc(month, year)}-{cyc(day, month)}")

# Multiple candidates are tried in turn, keeping whichever parses most values
parse_format(
  "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}",
  "{lin(year)}/{cyc(month, year)}/{cyc(day, month)}"
)

# regex = TRUE treats the surrounding text as regular expression syntax
parse_format(
  "{lin(year)}[/-]{cyc(month, year)}[/-]{cyc(day, month)}",
  regex = TRUE
)
}
\seealso{
\code{\link[=time_parse]{time_parse()}} for using the result as \code{format},
\code{\link[=chronon_parse_linear]{chronon_parse_linear()}}/\code{\link[=chronon_parse_cyclical]{chronon_parse_cyclical()}} for calendar-specific
candidates built this way.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cal_isoweek.R
\name{calendar_isoweek}
\alias{calendar_isoweek}
\alias{cal_isoweek}
\title{ISO 8601 time unit classes}
\format{
A civil-based calendar containing ISO 8601 time units.
}
\usage{
cal_isoweek
}
\value{
An S3 list of class \code{c("cal_isoweek", "mt_calendar")} containing
the named time unit classes of the ISO 8601 week calendar. Each unit is
accessible via \code{$} notation and calling it with a step size produces a
time granule (e.g., 1 week granule as \code{cal_isoweek$week(1L)}).
}
\description{
Time unit constructors for the ISO 8601 calendar system. These units can be
used with \code{\link[=linear_time]{linear_time()}} to create custom time representations.
}
\details{
The following time units are available in the ISO week date calendar:
\itemize{
\item \code{year()}: ISO year unit (years start on the week containing the first Thursday)
\item \code{week()}: Week unit (7-day periods)
\item \code{day()}: Day unit
\item \code{hour()}: Hour unit
\item \code{minute()}: Minute unit
\item \code{second()}: Second unit
\item \code{millisecond()}: Millisecond unit
}

ISO 8601 weeks always start on Monday and the first week of a year is the week
containing the first Thursday of that year. This means that some days in early
January may belong to the last week of the previous ISO year, and some days in
late December may belong to the first week of the next ISO year.
}
\seealso{
\code{\link[=linear_time]{linear_time()}} for creating custom time representations,
\code{\link[=yearweek]{yearweek()}} for a pre-defined ISO 8601 year-week representation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse.R
\name{time_parse}
\alias{time_parse}
\title{Parse text into a time point}
\usage{
time_parse(
  x,
  chronon = NULL,
  cycle = NULL,
  format = NULL,
  regex = FALSE,
  na = c("", "NA"),
  calendar = NULL,
  locale = NULL,
  discrete = TRUE
)
}
\arguments{
\item{x}{A character vector to parse.}

\item{chronon}{Target time granule for the result, and (with \code{cycle}) the
source of \code{format} candidates when \code{format} is \code{NULL}. Its attributes
(e.g. \code{tz}) fill in whatever \code{format} leaves unset, and the result is
converted onto it if \code{format} reaches a different chronon.}

\item{cycle}{Target cycle granule, pairing with \code{chronon} for a cyclical
result. Requires \code{chronon}.}

\item{format}{A glue-style format string of \code{\link[=lin]{lin()}}/\code{\link[=cyc]{cyc()}} tokens, e.g.
\code{"{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"} (see
\code{vignette("time-format-strings")}), or several to try: whichever parses
the most values of \code{x} is used for the whole vector (ties keep the
earliest-listed format), and its unparsed values become \code{NA} (with a
warning). Aborts if no format matches the shape of even one value.
\code{time_parse(format(x, fmt), format = fmt)} round-trips back to \code{x}.
\code{NULL} (the default) derives candidates from \code{chronon}/\code{cycle} via
\code{\link[=chronon_parse_linear]{chronon_parse_linear()}}/\code{\link[=chronon_parse_cyclical]{chronon_parse_cyclical()}}; requires \code{chronon}.}

\item{regex}{If \code{FALSE} (the default), literal text surrounding tokens is
matched exactly. If \code{TRUE}, it's instead used verbatim as a regular
expression, e.g. \code{"[/-]"} to accept either \code{/} or \code{-} as a separator;
\code{(...)} groups you write are treated as non-capturing, since capturing
groups are reserved for the tokens. Ignored when \code{format} is derived
from \code{chronon}, which carries its own regex mode.}

\item{na}{Strings to treat as missing (\code{NA}), checked before matching
\code{format}. Not counted in the parsing-failure warning, unlike a value
that fails to match \code{format}.}

\item{calendar}{Calendar used to resolve granule names in \code{format}, and
to disambiguate \code{chronon}'s \code{\link[=chronon_parse_linear]{chronon_parse_linear()}} candidates when
\code{format} is \code{NULL}. \code{NULL} (the default) uses \code{time_calendar(cycle)} or
\code{time_calendar(chronon)}, whichever is supplied, else \link{cal_gregorian}.}

\item{locale}{Default locale for named (\code{label = TRUE}) tokens that don't
specify their own. \code{NULL} defers to each token's own scheme.}

\item{discrete}{Whether the result is discrete (integer chronon counts)
or continuous (fractional). See \code{\link[=linear_time]{linear_time()}}.}
}
\value{
A \code{mixtime} time vector, the same length as \code{x}. Linear if
\code{format} includes a \code{{lin(...)}} token (or \code{cycle} is \code{NULL}),
cyclical otherwise.
}
\description{
\code{time_parse()} is the inverse of \code{\link[=format]{format()}}: given text and the same
\code{{lin(...)}}/\code{{cyc(...)}} template \code{format} uses, it reconstructs the time
points that would have produced that text, via \code{\link[=time_compose]{time_compose()}}.
}
\details{
A \code{format} with a \code{{lin(...)}} token parses to linear time; one of only
\code{{cyc(...)}} tokens parses to cyclical time, e.g. \code{time_parse("Feb", format = "{cyc(month, year, label = TRUE)}")} recovers the same kind of
value as \code{\link[=month_of_year]{month_of_year()}}.

Granule-specific extraction and decoding labels for each token is done by
\code{\link[=linear_labels_parse]{linear_labels_parse()}}/\code{\link[=cyclical_labels_parse]{cyclical_labels_parse()}}.
}
\examples{
time_parse("2024-02-15", format = "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}")
time_parse(
  "15 Feb 2024",
  format = "{cyc(day, month)} {cyc(month, year, label = TRUE)} {lin(year)}"
)

# One bad value becomes NA (with a warning) instead of aborting the batch
time_parse(
  c("2024-02-15", "not a date"),
  format = "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}"
)

# No {lin(...)} token: parses to cyclical time
time_parse("Feb", format = "{cyc(month, year, label = TRUE)}")

# Several formats: whichever parses the most values is used for the whole
# vector; here none of the "Y-M-D" format's values match, so the "D/M/Y"
# format (which matches both) is used instead
time_parse(
  c("15/02/2024", "20/03/2024"),
  format = c(
    "{lin(year)}-{cyc(month, year)}-{cyc(day, month)}",
    "{cyc(day, month)}/{cyc(month, year)}/{lin(year)}"
  )
)

# regex = TRUE: match "/" or "-" as the separator, 
# and tolerate a trailing comment after the date.
time_parse(
  c("2024-02-15", "2024/02/15 (approx)"),
  format = "{lin(year)}[/-]{cyc(month, year)}[/-]{cyc(day, month)}( .*)?",
  regex = TRUE
)

# Default format strings from the target chronon, and results with `tz`.
time_parse("2024-02-15 09:00:00", chronon = cal_gregorian$second(1L, tz = "America/Los_Angeles"))

}
\seealso{
\code{\link[=format]{format()}} for the inverse direction, \code{\link[=time_compose]{time_compose()}} for
composing a time point from already-decoded components,
\code{\link[=label_scheme]{label_scheme()}} for declaring how a granule's labels parse,
\code{\link[=chronon_parse_linear]{chronon_parse_linear()}}/\code{\link[=chronon_parse_cyclical]{chronon_parse_cyclical()}} for the candidate
formats derived from \code{chronon}/\code{cycle}, and \code{vignette("time-format-strings")}
for the format string syntax.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/location.R
\name{mixtime_location}
\alias{mixtime_location}
\alias{loc_latitude}
\alias{loc_longitude}
\alias{loc_altitude}
\title{Extract locations from an object}
\usage{
loc_latitude(x, ...)

loc_longitude(x, ...)

loc_altitude(x, ...)
}
\arguments{
\item{x}{An object with location information.}

\item{...}{Additional arguments passed to methods.}
}
\value{
A numeric value representing the location (e.g., longitude, latitude, etc.).
}
\description{
Generic function to extract the location from objects that have location information.
}
\examples{
t <- linear_time(
  1:3, 
  cal_time_solar$day(1L, lat = -37.8136, lon = 144.9631)
)

loc_longitude(t)
loc_latitude(t)
loc_altitude(t)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01_labels.R
\name{linear_labels_format}
\alias{linear_labels_format}
\alias{cyclical_labels_format}
\alias{linear_labels_parse}
\alias{cyclical_labels_parse}
\title{Render and parse a granule's labels}
\usage{
linear_labels_format(granule, ...)

cyclical_labels_format(granule, cycle, ...)

linear_labels_parse(granule, ...)

cyclical_labels_parse(granule, cycle, ...)
}
\arguments{
\item{granule}{A time granule object representing the granule (e.g.
\code{month(1L)}).}

\item{...}{Passed on to the method. The default method (see below) takes:
\itemize{
\item \code{i}: Integer vector: the position along the linear axis, or within
the cycle for \code{cyclical_labels_format()}.
\item \code{at}: The linear position of the cycle granule, letting a method
produce labels specific to that cycle instance (mainly for irregular
cycles, e.g. a leap month). Same convention as \code{at} in
\code{\link[=chronon_cardinality]{chronon_cardinality()}}.
\item \code{label}: If \code{TRUE}, return named labels (e.g. "February"). If
\code{FALSE}, return the numeric position as character.
\item \code{abbreviate}: If \code{TRUE}, return abbreviated labels (e.g. "Feb"). If
\code{FALSE}, return full labels (e.g. "February").
\item \code{type}: Overrides \code{abbreviate}'s abbreviated/wide choice of
\code{\link[=vocab_table]{vocab_table()}} type outright, e.g. \code{type = "emoji"}. \code{NULL} (default)
defers to \code{abbreviate}.
\item \code{width},\code{locale}: Call-time overrides of the scheme's \code{width}/\code{locale}
defaults (see \code{\link[=label_scheme]{label_scheme()}}); \code{NULL} defers to the scheme.
}}

\item{cycle}{A time granule object representing the cycle (e.g.
\code{year(1L)}).}
}
\value{
Character vector of labels for the time point.
}
\description{
The generics turning a granule's internal position into display text
(\code{linear_labels_format()} for a non-repeating position, e.g. the year;
\code{cyclical_labels_format()} for a position within a larger cycle, e.g. the
month within the year) and back (\code{linear_labels_parse()}/
\code{cyclical_labels_parse()}).
}
\details{
Each has one default method (on the base \code{mt_unit} class) that looks up
the granule's scheme via \code{\link[=linear_labels]{linear_labels()}}/\code{\link[=cyclical_labels]{cyclical_labels()}} and
applies it (see \code{\link[=label_scheme]{label_scheme()}} for the fields). Calendar authors
usually declare a scheme instead of writing a method here directly.
Writing one directly overrides the generics for labels that aren't
index-shaped at all, see \code{\link[=linear_labels]{linear_labels()}}'s "Overriding the generics
directly" section.
}
\section{Epoch shift}{

\code{linear_labels_format()}'s \code{i} arrives already epoch-shifted for display
(\code{chronon_parts()} adds \code{\link[=chronon_epoch]{chronon_epoch()}} before calling it, e.g. a
Gregorian year's \code{i} is \code{2024}, not \code{54}). The default
\code{linear_labels_parse()} method shifts \code{decode()}'s output back, by
subtracting \code{\link[=chronon_epoch]{chronon_epoch()}} after the scheme's own decode runs. A
hand-written override method skips this and must do the same
subtraction itself if the granule has a non-zero epoch; see the example
in \code{\link[=linear_labels]{linear_labels()}}.
}

\examples{
# Labels for years on a linear axis
with(cal_gregorian, linear_labels_format(year(1L), 2020:2025))

# Labels for months in a year
with(cal_gregorian, cyclical_labels_format(month(1L), year(1L), 0:11))

}
\seealso{
\code{\link[=linear_labels]{linear_labels()}}/\code{\link[=cyclical_labels]{cyclical_labels()}} for the authoring
interface these generics' default methods interpret.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mixtime.R
\name{is_mixtime}
\alias{is_mixtime}
\title{Check if an object is a mixtime}
\usage{
is_mixtime(x)
}
\arguments{
\item{x}{An object to test.}
}
\value{
A scalar logical: \code{TRUE} if \code{x} is a \code{mixtime} vector, \code{FALSE}
otherwise.
}
\description{
Tests whether \code{x} inherits from the \code{mixtime} class.
}
\examples{
is_mixtime(Sys.Date())
is_mixtime(mixtime(Sys.Date()))

}
\seealso{
\code{\link[=as_mixtime]{as_mixtime()}} to coerce objects to \code{mixtime},
\code{\link[=mixtime]{mixtime()}} to construct a \code{mixtime}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{all.equal.mixtime::mt_time}
\alias{all.equal.mixtime::mt_time}
\title{Tolerant comparison of mixtime time values}
\usage{
\method{all.equal}{`mixtime::mt_time`}(target, current, tolerance = sqrt(.Machine$double.eps), ...)
}
\arguments{
\item{target, current}{\code{mt_time} vectors of the same length to compare (as
passed down by \link[vecvec:class_vecvec]{vecvec::class_vecvec}'s \code{all.equal()}).}

\item{tolerance}{Numeric tolerance, as an absolute number of \code{target}'s
chronon units. Defaults to \code{sqrt(.Machine$double.eps)}, as for
\code{\link[base:all.equal]{base::all.equal()}}.}

\item{...}{Ignored.}
}
\value{
\code{TRUE} if \code{target} and \code{current} are equal within \code{tolerance},
otherwise a string describing the discrepancy.
}
\description{
\link[vecvec:class_vecvec]{vecvec::class_vecvec}'s \code{all.equal()} method (which \code{mixtime} inherits,
since a mixtime vector is a \code{vecvec}) handles length/\code{NA} mismatches and
groups elements by their underlying storage slot, then compares each
group's \emph{raw} values with \code{all.equal()}. That raw comparison has no notion
of chronon: two elements that are \code{==} (e.g. \code{days(1)} and \code{hours(24)}, or
a \code{yearmonth} and an equivalent-instant \code{yearweek}) can end up stored with
different chronons or magnitudes, and so are wrongly reported as unequal.

This method fixes that at the per-slot level: elements already \code{==} (which
is chronon-aware, see \link{mt_linear-compare}/\link{mt_duration-compare}) count as
equal outright; for the rest, the discrepancy is measured as the duration
between them in their common chronon. A time point has no "typical
magnitude" to scale a \emph{relative} tolerance against the way a plain number
does, so - unlike \code{\link[base:all.equal.numeric]{base::all.equal.numeric()}} - \code{tolerance} is always
absolute, in (possibly fractional) chronon units.
}
\examples{
all.equal(yearmonth(0), yearmonth(0))
all.equal(yearmonth(0), yearmonth(1))
all.equal(days(1), hours(24))

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{mt_cyclical-compare}
\alias{mt_cyclical-compare}
\alias{==,mt_cyclical-method}
\alias{!=,mt_cyclical-method}
\alias{<,mt_cyclical-method}
\alias{<=,mt_cyclical-method}
\alias{>,mt_cyclical-method}
\alias{>=,mt_cyclical-method}
\title{Comparison operators for cyclical time (\code{mt_cyclical})}
\usage{
\S4method{==}{mt_cyclical}(e1, e2)
\S4method{!=}{mt_cyclical}(e1, e2)
\S4method{<}{mt_cyclical}(e1, e2)
\S4method{<=}{mt_cyclical}(e1, e2)
\S4method{>}{mt_cyclical}(e1, e2)
\S4method{>=}{mt_cyclical}(e1, e2)
}
\arguments{
\item{e1, e2}{\code{mt_cyclical} vectors sharing a cycle (or values castable to one,
such as plain numeric vectors sharing the other operand's chronon).}
}
\value{
A logical vector.
}
\description{
A cyclical time value stores an absolute chronon count but \emph{means} a position
within its cycle (e.g. \code{day_of_week()} means a weekday, not a particular
Wednesday). Comparison therefore reduces both operands to their position
within the cycle - the same reduction \code{\link[=format]{format()}} displays - and compares
those:
\itemize{
\item \code{day_of_week(date("2020-01-15")) == day_of_week(date("2020-01-22"))} is
\code{TRUE}, since both are a Wednesday.
\item \code{day_of_year(date("2020-01-15")) == day_of_year(date("2021-01-15"))} is
\code{TRUE}, since both are the 15th day of their year.
}

Both operands must share a \code{cycle}: a cycle is a modulus rather than a unit,
so there is no meaningful common cycle between (say) a weekday and a
day-of-year, and comparing them is an error. Differing \emph{chronons} within a
shared cycle are reconciled as they are for \link[=mt_linear-compare]{mt_linear}:
both positions are expressed in the finest common chronon, and a discrete
value spans the closed interval of its chronon, so
\itemize{
\item \code{a == b} iff \code{start(a) == start(b)} and \code{end(a) == end(b)}
\item \code{a < b} iff \code{end(a) < start(b)}, \code{a > b} iff \code{start(a) > end(b)}
\item \code{a <= b} iff \code{end(a) <= end(b)}, \code{a >= b} iff \code{start(a) >= start(b)}
}

As for \code{mt_linear}, this is not a total order when chronons differ. Ordering
follows the position within the cycle (so \code{Mon < Wed}); the cycle's wrap-around
is not treated as circular.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cal_gregorian.R
\name{calendar_gregorian}
\alias{calendar_gregorian}
\alias{cal_gregorian}
\title{Gregorian time unit classes}
\format{
A civil-based calendar containing Gregorian time units.
}
\usage{
cal_gregorian
}
\value{
An S3 list of class \code{c("cal_gregorian", "mt_calendar")} containing
the named time unit classes of the Gregorian calendar. Each unit is
accessible via \code{$} notation and calling it with a step size produces a
time granule (e.g., 1 month granule as \code{cal_gregorian$month(1L)}).
}
\description{
Time unit constructors for the Gregorian calendar system. These units can be
used with \code{\link[=linear_time]{linear_time()}} to create custom time representations.
}
\details{
The following time units are available in the Gregorian calendar (\verb{cal_gregorian$}).
\itemize{
\item \code{year()}: Year unit
\item \code{quarter()}: Quarter (3-month period) unit
\item \code{month()}: Month unit
\item \code{day()}: Day unit
\item \code{hour()}: Hour unit
\item \code{minute()}: Minute unit
\item \code{second()}: Second unit
\item \code{millisecond()}: Millisecond unit
}

These units form a hierarchy where conversions between adjacent units follow
the Gregorian calendar rules. For units that don't have a fixed relationship
(e.g., months to days), the conversion requires a time context.
}
\examples{
# Create a custom time representation using Gregorian units
linear_time(
  Sys.time(),
  chronon = hour(1L)
)

}
\seealso{
\code{\link[=linear_time]{linear_time()}} for creating linear time points.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01_labels.R
\name{vocab_table}
\alias{vocab_table}
\title{Build a \code{vocab} function from a plain name table}
\usage{
vocab_table(..., default_locale = "en-GB")
}
\arguments{
\item{...}{Named entries, one per locale (e.g. \code{ `en-GB` = list(wide = month.name, abbreviated = month.abb)}). Locale tags follow BCP 47.}

\item{default_locale}{The locale returned when \code{locale} is unsupplied (or
explicitly \code{NULL}, meaning "caller didn't ask", since
\code{time_parse()}/\code{format()} always pass some value for \code{locale}).}
}
\value{
A function \verb{function(type = NULL, locale = NULL)}, suitable as
the \code{vocab} field of a \code{\link[=label_scheme]{label_scheme()}}.
}
\description{
The common case for a \code{\link[=label_scheme]{label_scheme()}} \code{vocab} argument: a hand-listed
table of names, one entry per locale, each a named list of renderings. By
convention these follow CLDR's wide/abbreviated/narrow, but \code{type} can be
anything. \code{vocab_table()} wraps it in the \verb{function(type = NULL, locale = NULL)} shape \code{vocab} requires.
}
\details{
For vocab backed by an external i18n source instead of a hand-listed
table, write that \verb{function(type, locale)} directly instead. Same shape,
no list needed.
}
\examples{
month_vocab <- vocab_table(`en-GB` = list(wide = month.name, abbreviated = month.abb))
month_vocab("abbreviated")
month_vocab(locale = "en-GB")

}
\seealso{
\code{\link[=label_scheme]{label_scheme()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/00_classes.R, R/cal0_time_astro.R,
%   R/cal0_time_civil.R
\name{mt_unit}
\alias{mt_unit}
\alias{mt_loc_unit}
\alias{mt_tz_unit}
\title{Base S7 class for creating new time units}
\usage{
mt_unit(n = 1L)

mt_loc_unit(n = 1L, lat = naive_loc, lon = naive_loc, alt = naive(0))

mt_tz_unit(n = 1L, tz = naive_tz)
}
\arguments{
\item{n}{The step size of time granule. For example, \code{n = 2L} is 2 time
units, and \code{cal_isoweek$week(2L)} would represent 2 weeks (a fortnight).}

\item{lat}{Numeric. Latitude in decimal degrees. Range: -90 to 90.
Default: naive location (astronomical calculations require \code{lat}).}

\item{lon}{Numeric. Longitude in decimal degrees. Range: -180 to 180.
Default: naive location (astronomical calculations require \code{lon}).}

\item{alt}{Numeric. Altitude in meters above sea level.
Default: 0 (sea level).}

\item{tz}{The timezone name for the unit (valid units can be found with \verb{[tzdb::tzdb_names()]})}
}
\value{
A time granule object of class \code{mt_unit}
}
\description{
This class is the primitive class for time units, and should
be extended from when creating new time units. A new class
is typically created with S7 using:
\code{S7::new_class("tu_***", parent = mt_tz_unit)}
}
\details{
Time units are the building blocks of calendars in mixtime. Each unit
represents a specific temporal component (e.g., day, month, year) and
can be combined using \code{\link[=new_calendar]{new_calendar()}} to create a calendar system.

When creating custom calendars, define time unit classes that inherit
from either \code{mt_unit} (for standard units) or \code{mt_tz_unit} (for
timezone-aware units), then pass them as named arguments to
\code{\link[=new_calendar]{new_calendar()}}. The calendar will use these names to create
constructor functions accessible via \code{$} notation (e.g., \code{calendar$day(1L)}).
}
\section{Calendar Algebra Methods}{


Time units enable calendar arithmetic through key generic methods
that should be implemented for custom time units:
\itemize{
\item \code{chronon_cardinality_fixed(x, y)} - Returns the number of unit \code{x}
granules that fit within one unit \code{y} granule, for relationships that are
a constant, context-independent number (e.g., 7 days per week, 24 hours
per day). Prefer this over \code{chronon_cardinality()} whenever the
relationship does not depend on \code{at}, since it is also used to determine
which relationships are safe to use for \code{\link[=chronon_divmod]{chronon_divmod()}}'s graph
traversal.
\item \code{chronon_cardinality(x, y, at)} - Returns the number of \code{x} granule
that fit within one \code{y} granule. This is variable based on \code{at} (e.g.,
28-31 days per month). Only implement this directly (rather than
\code{chronon_cardinality_fixed()}) when the relationship genuinely depends on
\code{at}, and pair it with a direct \code{chronon_divmod()} method so that the
relationship remains reachable during graph traversal.
\item \code{chronon_divmod(x, from, to)} - Converts time point \code{x} from granules of
\code{from} to granules of \code{to}, returning a list with \code{div} (the quotient)
and \code{mod}. This enables conversions between granules that have
variable cardinality (e.g., the date 2020-03-23 to the month 2020-03).
All conversions should be based on chronons since epoch (1970-01-01),
in the UTC time zone.
}

These methods work together to enable mixtime to perform calendar-aware
arithmetic, understanding that months have variable lengths and handling
timezone-aware conversions.
}

\examples{
# Create a timezone-aware unit class

# Use these units to create a calendar
my_calendar <- new_calendar(
  day = S7::new_class("tu_my_day", parent = mt_unit),
  month = S7::new_class("tu_my_month", parent = mt_tz_unit),
  class = "my_calendar"
)

# Access unit constructors from the calendar
my_calendar$day(1L)
my_calendar$month(3L, tz = "America/New_York")

}
\seealso{
\code{\link[=new_calendar]{new_calendar()}} for creating calendars from time units
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duration.R
\name{duration_helpers}
\alias{duration_helpers}
\alias{years}
\alias{quarters}
\alias{months}
\alias{weeks}
\alias{days}
\alias{hours}
\alias{minutes}
\alias{seconds}
\alias{milliseconds}
\title{Duration helper functions}
\usage{
years(data, calendar = time_calendar(data), ...)

quarters(data, calendar = time_calendar(data), ...)

months(data, calendar = time_calendar(data), ...)

weeks(data, calendar = time_calendar(data), ...)

days(data, calendar = time_calendar(data), ...)

hours(data, calendar = time_calendar(data), ...)

minutes(data, calendar = time_calendar(data), ...)

seconds(data, calendar = time_calendar(data), ...)

milliseconds(data, calendar = time_calendar(data), ...)
}
\arguments{
\item{data}{A time vector of duration magnitudes, or an existing
\code{\link[=duration]{duration()}} vector to convert to \code{chronon} granules.}

\item{calendar}{Calendar system used to evaluate \code{chronon}. Defaults to
\code{time_calendar(data)} for existing time objects. Common options include
\link{cal_gregorian} and \link{cal_isoweek}.}

\item{...}{Additional arguments passed to the chronon (e.g. \code{tz} for
timezones).}
}
\value{
A \code{mixtime} vector containing an \code{mt_duration} vector.
}
\description{
Convenience functions for creating duration vectors of common time units.
Each function wraps \code{\link[=new_duration_fn]{new_duration_fn()}} for its respective chronon.
}
\examples{
years(3L)
quarters(2L)
months(6L)
weeks(4L)
days(7L)
hours(12L)
minutes(30L)
seconds(45L)
milliseconds(500L)

}
\seealso{
\itemize{
\item \code{\link[=new_duration_fn]{new_duration_fn()}} for creating custom duration functions
\item \code{\link[=duration]{duration()}} for creating duration vectors directly
\item \link{cal_gregorian}, \link{cal_isoweek} for calendar systems
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time_cycle.R
\name{time_cycle}
\alias{time_cycle}
\title{Obtain the cycle of a time object}
\usage{
time_cycle(x, ...)
}
\arguments{
\item{x}{A time object (e.g., \link[base:Date]{base::Date}, \link[base:POSIXct]{base::POSIXct}, \code{\link[=linear_time]{linear_time()}}, etc.)}

\item{...}{Additional arguments for methods.}
}
\value{
A time \code{\link[=duration]{duration()}} object representing the cycle of each value
(e.g. \code{weeks(1L)}), or \code{NA} if the object has no cyclical component.
}
\description{
This S7 generic function extracts the cycle (the cyclical time granule) from
a time object, such as cyclical time representations.
}
\examples{

# Non-cyclical objects return NA
time_cycle(Sys.Date())

# The cycle of a cyclical time object
time_cycle(month_of_year(Sys.Date()))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cyclical_time.R
\name{cyclical_time_helpers}
\alias{cyclical_time_helpers}
\alias{month_of_year}
\alias{day_of_year}
\alias{day_of_month}
\alias{time_of_day}
\alias{day_of_week}
\alias{week_of_year}
\title{Cyclical time helpers}
\usage{
month_of_year(data, discrete = TRUE, calendar = time_calendar(data), ...)

day_of_year(data, discrete = TRUE, calendar = time_calendar(data), ...)

day_of_month(data, discrete = TRUE, calendar = time_calendar(data), ...)

time_of_day(data, discrete = TRUE, calendar = time_calendar(data), ...)

day_of_week(data, discrete = TRUE, calendar = time_calendar(data), ...)

week_of_year(data, discrete = TRUE, calendar = time_calendar(data), ...)
}
\arguments{
\item{data}{Another object to be coerced into the specified cyclical time.}

\item{discrete}{If \code{TRUE}, the position within the cycle that \code{data}
falls into is returned as an integer. If \code{FALSE}, a fractional
position is returned (analagous to time using a continuous time model).}

\item{calendar}{A calendar object specifying the calendar system to use.}

\item{...}{Additional arguments for \code{\link[=cyclical_time]{cyclical_time()}}, such as \code{tz} for timezones.}
}
\value{
A \code{mixtime} time vector containing an \code{mt_cyclical} vector with chronon and cycle matching the function used.
}
\description{
Helper functions for creating cyclical time representations. These functions
create time objects that repeat within a larger time cycle, useful for identifying
seasonal patterns or positions within a calendar period.
}
\section{Cyclical time representations}{

\itemize{
\item \code{day_of_week()}: Represents the day position within a week (1-7) using
the ISO 8601 standard where weeks start on Monday.
\item \code{day_of_month()}: Represents the day position within a month (1-28, 1-29,
1-30, or 1-31 depending on the month). The chronon is one day, cycling
within a month.
\item \code{day_of_year()}: Represents the day position within a year (1-365 or 1-366
for leap years). The chronon is one day, cycling within a year.
\item \code{week_of_year()}: Represents the week position within a year (1-52 or 1-53)
using the ISO 8601 week numbering system.
\item \code{month_of_year()}: Represents the month position within a year (1-12).
The chronon is one month, cycling within a year.
}
}

\section{Custom cyclical time representations}{

You can create custom cyclical time representations using \code{\link[=cyclical_time]{cyclical_time()}}
with any of the supported time units (see \link{calendar_gregorian} and \link{calendar_isoweek}).

For example, to create a representation for day of the month:

\preformatted{
day_of_month <- new_cyclical_time_fn(
  chronon = day(1L), cycle = month(1L),
  default_calendar = cal_gregorian
)
}
}

\examples{

month_of_year(Sys.Date())
day_of_year(Sys.Date())
day_of_week(Sys.Date())
day_of_week(as.Date("2025-12-15") + 0:6)

}
\seealso{
\code{\link[=cyclical_time]{cyclical_time()}} for creating cyclical time vectors,
\code{\link[=new_cyclical_time_fn]{new_cyclical_time_fn()}} for creating cyclical time helper functions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compose.R
\name{time_compose}
\alias{time_compose}
\title{Compose a linear time vector from linear and cyclical components}
\usage{
time_compose(..., discrete = TRUE, calendar = cal_gregorian)
}
\arguments{
\item{...}{Components used to build the time point. Each element is either:
\itemize{
\item a two-sided formula, \verb{lin(<granule>) ~ <value>} or
\verb{cyc(<granule>, <cycle>) ~ <value>} (see \code{\link[=lin]{lin()}}/\code{\link[=cyc]{cyc()}}), or
\item an existing linear or cyclical \code{mixtime} vector.
}}

\item{discrete}{Logical. If \code{TRUE} (default), returns integer chronons since
Unix epoch (discrete time model). If \code{FALSE}, returns fractional chronons
allowing representation of fractional time granules (continuous time model).}

\item{calendar}{Calendar used to resolve bare granule names in \code{lin()}/
\code{cyc()} formulas. Defaults to \link{cal_gregorian}.}
}
\value{
A \code{mixtime} time vector, at the finest chronon reached by the
chain (or the root's own chronon, if only one component is given).
Linear with a \code{lin()} anchor, cyclical otherwise.
}
\description{
\code{time_compose()} is the inverse of \code{\link[=time_components]{time_components()}}: given a set of
\code{\link[=lin]{lin()}}/\code{\link[=cyc]{cyc()}} components it reconstructs the corresponding time points.
Each component is either a two-sided formula pairing a spec with its value,
or an already-tagged linear/cyclical time vector (e.g. produced by
\code{\link[=linear_time]{linear_time()}}, \code{\link[=cyclical_time]{cyclical_time()}}, or a \code{\link[=time_components]{time_components()}} column).
}
\details{
A \code{lin()} component (the \strong{anchor}), when supplied, fixes the absolute
position at some granule (e.g. the year). Every other component must be
\code{cyc()}, chaining without gaps or branches from the anchor down to the
target chronon: each cycle must equal another component's chronon exactly.

With no \code{lin()} anchor, every component must be \code{cyc()}, chained the same
way but rooted at whichever component's cycle isn't itself another
component's chronon. The result is cyclical time tagged with that root's
cycle: \code{cyc(month, year) ~ 3} alone matches \code{month_of_year()} for any
March; chaining \code{cyc(day, month) ~ 15} onto it collapses to one (day,
year) pair, day-of-year 74, matching \code{day_of_year()} for 15 March.

Values of linear and cyclical components are specified on the right-hand-side
of the formula. A \code{lin()} value is the real-world count (e.g. the literal
year 1980); a \code{cyc()} value is the 1-indexed position within the cycle
(e.g. \code{cyc(month, year) ~ 3} is the 3rd month, March), matching everyday
counting rather than the raw 0-indexed position \code{\link[=time_components]{time_components()}} uses
internally.
}
\examples{
# cyc() values are 1-indexed positions: month 3 is March, day 15 is the 15th
time_compose(lin(year) ~ 1980, cyc(month, year) ~ 3, cyc(day, month) ~ 15)

# A lin() anchor alone is a valid (coarser) time point
time_compose(lin(year) ~ 1980)

# No lin() anchor: composes cyclical time
time_compose(cyc(month, year) ~ 3)

# Chaining collapses to one (chronon, cycle) pair: day 15 of month 3
# becomes day-of-year 74
time_compose(cyc(day, month) ~ 15, cyc(month, year) ~ 3)

# Round-tripping through time_components()
parts <- time_components(as.Date("2024-03-15"), yr = lin(year), mth = cyc(month, year))
with(parts, time_compose(yr, mth))

# Multi-unit (self-referencing) cycles: the 3rd month (1-indexed) of the
# 4th 3-month block since epoch (block 3 = months 9-11 -> December 1970)
time_compose(lin(month(3L)) ~ 3, cyc(month(1L), month(3L)) ~ 3)

}
\seealso{
\code{\link[=time_components]{time_components()}} for the inverse operation, \code{\link[=lin]{lin()}}/\code{\link[=cyc]{cyc()}} for
the component vocabulary shared with \code{\link[=time_components]{time_components()}} and \code{\link[=format]{format()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predicates.R
\name{time_is_determinate_at}
\alias{time_is_determinate_at}
\title{Test whether time is determinate at a granule's precision}
\usage{
time_is_determinate_at(x, granule, ...)
}
\arguments{
\item{x}{A time object (typically a \code{mixtime} vector).}

\item{granule}{The time granule whose precision to test, given as a granule
generator (e.g. \code{cal_gregorian$month}) or a sized time unit (e.g.
\code{cal_gregorian$month(1L)}).}

\item{...}{Additional arguments for methods.}
}
\value{
A logical vector the same length as \code{x}.
}
\description{
\code{time_is_determinate_at()} tests, for each element of a \code{mixtime} vector, whether
the time point is well-defined at the precision of \code{granule}.
}
\details{
Discrete (integer) time cannot resolve a granule finer than its own chronon (a
\code{year()} has no determinate month), so those elements are \code{FALSE}. Continuous
(fractional) time tracks progress within its chronon and so resolves finer
granules exactly (0\% through 2020 is 0\% through January), giving \code{TRUE}.
Coarser-or-equal granules are always determinate. Missing (\code{NA}) and infinite
times give \code{NA}.
}
\examples{
# Discrete: a year has no determinate month
time_is_determinate_at(year(2020L), cal_gregorian$month(1L))

# Continuous: 0\% through 2020 is 0\% through January
time_is_determinate_at(year(2020), cal_gregorian$month(1L))

# A coarser granule is always determinate
time_is_determinate_at(yearmonth(as.Date("2020-02-01")), cal_gregorian$year(1L))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence.R
\name{seq.mixtime::mixtime}
\alias{seq.mixtime::mixtime}
\alias{seq.mixtime::mt_time}
\title{Generate sequences of mixtime values}
\usage{
\method{seq}{`mixtime::mixtime`}(...)

\method{seq}{`mixtime::mt_time`}(
  from,
  to,
  by,
  length.out = NULL,
  along.with = NULL,
  on_invalid = c("nearest", "overflow"),
  ...
)
}
\arguments{
\item{...}{Additional arguments passed to the underlying sequence method.}

\item{from}{Starting value of the sequence.}

\item{to}{End value of the sequence (if provided).}

\item{by}{Increment of the sequence. Can be:
\itemize{
\item A numeric for the number of time chronons
\item A character string specifying the interval (e.g., "1 day", "2 weeks",
"1 month", "1 year")
\item A time granule object created with time unit functions (e.g., \code{cal_gregorian$year(1L)},
\code{cal_gregorian$month(1L)}, \code{cal_gregorian$day(1L)})
\item A time \code{\link[=duration]{duration()}} object (e.g., \code{years(1L)}, \code{months(1L)}, \code{days(1L)})
}}

\item{length.out}{Desired length of the sequence (alternative to \code{to}).}

\item{along.with}{Take the length from this argument (alternative to \code{length.out}).}

\item{on_invalid}{How to handle time points that overflow the cycle when
using a \code{by} argument with different time granule than the sequence chronon.
Options are:
\itemize{
\item \code{"nearest"} (default): Adjust overflowing time points to the nearest
valid time point within the cycle
\item \code{"overflow"}: Allow time points to overflow into the next cycle
}

This is relevant when the starting time point has an offset that doesn't
exist in all cycles. For example, starting on day 31 with \code{by = "1 month"}
will overflow in months with fewer than 31 days (e.g., February). With
\code{"nearest"}, these will be adjusted to the last day of the month (e.g.,
Feb 28/29). With \code{"overflow"}, the extra days carry into the next month.

If not explicitly specified and overflow occurs, a warning is issued with
the default \code{"nearest"} behavior applied.}
}
\value{
A mixtime vector containing the sequence.
}
\description{
Create regular sequences of time values. This method handles both linear
time sequences (dates, date-times) and cyclical time sequences (day of week,
month of year).
}
\details{
For linear time types (Date, POSIXct, yearmonth, etc.), sequences progress
forward or backward in time. For cyclical time types (month_of_year,
day_of_week, etc.), sequences wrap around cyclically.
}
\examples{
# Linear time sequences with integer by
seq(yearmonth("2020 Jan"), yearmonth("2020 Dec"))
seq(yearquarter("2020 Q1"), length.out = 5, by = 3)

# Linear time sequences with string intervals
seq(date("2020-01-01"), date("2020-12-31"), by = "1 month")
seq(yearmonth("2020 Jan"), yearmonth("2025 Jan"), by = "1 year")
seq(date("2020-01-01"), length.out = 10, by = "2 weeks")

# Linear time sequences incrementing by time granules
seq(yearmonth("2020 Jan"), yearmonth(("2020 Dec")), by = cal_gregorian$month(2L))
seq(date("2020-01-01"), length.out = 5, by = cal_gregorian$year(1L))
seq(date("2020-01-01"), date("2020-01-31"), by = cal_gregorian$day(7L))

# Handling invalid dates with on_invalid
seq(date("2020-01-31"), length.out = 3, by = "1 month")  # warns, uses nearest
seq(date("2020-01-31"), length.out = 3, by = "1 month", on_invalid = "nearest")
seq(date("2020-01-31"), length.out = 3, by = "1 month", on_invalid = "overflow")

# Cyclical time sequences
seq(month_of_year(0L), month_of_year(11L))
seq(day_of_week(0L), day_of_week(6L), by = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce.R
\name{as_mixtime}
\alias{as_mixtime}
\title{Convert a time class into a mixtime}
\usage{
as_mixtime(x, ...)
}
\arguments{
\item{x}{A time value to convert to a \code{mixtime}. Any time class with a defined \code{time_chronon()} method can be converted (e.g. \code{Date}, \code{POSIXct}, \code{yearmonth}, etc.).}

\item{...}{Additional arguments passed to the underlying \code{\link[vctrs:vec_cast]{vec_cast()}} method.}
}
\value{
A \code{mixtime} object corresponding to \code{x}.
}
\description{
Coerces a time object (e.g. \code{Date}, \code{POSIXct}, \code{yearmonth}) to a \code{mixtime}
vector using \code{\link[vctrs:vec_cast]{vctrs::vec_cast()}}. The chronon and cycle are inferred from
\code{x} via \code{\link[=time_chronon]{time_chronon()}} and \code{\link[=time_cycle]{time_cycle()}}.
}
\examples{
as_mixtime(Sys.Date())
as_mixtime(Sys.time())

}
\seealso{
\code{\link[=mixtime]{mixtime()}} for constructing a \code{mixtime} directly from data,
\code{\link[=is_mixtime]{is_mixtime()}} for testing if an object is a \code{mixtime}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01_chronon_format.R
\name{chronon_format_linear}
\alias{chronon_format_linear}
\alias{chronon_format_cyclical}
\alias{chronon_format_duration}
\title{Default formatting strings for chronons}
\usage{
chronon_format_linear(x, cal = time_calendar(x), ...)

chronon_format_cyclical(x, y, ...)

chronon_format_duration(x, ...)
}
\arguments{
\item{x}{A time granule for the chronon.}

\item{cal}{The calendar of the chronon, used to disambiguate suitable format
strings for time units that are shared across calendars (e.g.
\code{cal_gregorian$day} and \code{cal_isoweek$day}).}

\item{...}{Additional arguments for methods.}

\item{y}{A time granule for the cycle}
}
\value{
A character string containing the default format template for the chronon.
}
\description{
Provides default linear time formatting strings for a given chronon (finest
time granule). The format strings use placeholders like \code{{lin(year(1L))}},
\verb{\{cyc(month(1L), year(1L)\}} and \verb{\{cyc(day(1L), month(1L)\}}, which are
evaluated in the context of the data's \code{\link[=time_calendar]{time_calendar()}}.
}
\examples{
chronon_format_linear(cal_gregorian$year(1L))
chronon_format_linear(cal_gregorian$month(1L))
chronon_format_linear(cal_gregorian$day(1L))
chronon_format_linear(cal_isoweek$day(1L))

chronon_format_cyclical(cal_gregorian$month(1L), cal_gregorian$year(1L))
chronon_format_cyclical(cal_gregorian$day(1L), cal_gregorian$month(1L))
chronon_format_cyclical(cal_isoweek$day(1L), cal_isoweek$week(1L))
chronon_format_cyclical(cal_isoweek$week(1L), cal_isoweek$year(1L))

chronon_format_duration(cal_gregorian$year(1L))
chronon_format_duration(cal_gregorian$month(1L))
chronon_format_duration(cal_gregorian$day(1L))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time.R
\name{new_time}
\alias{new_time}
\title{Constructor for mixtime time vectors}
\usage{
new_time(x = integer(), chronon = mt_unit(1L), cycle = NULL, class = NULL)
}
\arguments{
\item{x}{A numeric vector of time points, integers for discrete time or
doubles for continuous time.}

\item{chronon}{A time granule object representing the smallest indivisible time
granule (chronon) for the time vector (e.g. \code{cal_gregorian$day(1L)}).}

\item{cycle}{An optional time granule object representing the cycle for cyclical
time (e.g. \code{cal_gregorian$week(1L)} for day-of-week). If not provided, the
time vector will be treated as linear time.}

\item{class}{An optional character vector of additional S3 classes to assign
to the resulting time vector. This allows for further subclassing of
\code{mt_time} for specific time types (e.g. linear, cyclical, durations, etc.).}
}
\value{
A \code{mt_time} vector representing the time points in \code{x} according to
the specified \code{chronon} and \code{cycle}.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{new_time()} was the low-level constructor for \code{mt_time} vectors. It has been
deprecated in favour of calling the concrete time class constructors directly:
\code{\link[=mt_linear]{mt_linear()}} for linear time, \code{\link[=mt_duration]{mt_duration()}} for durations, and
\code{\link[=mt_cyclical]{mt_cyclical()}} for cyclical time.

Creates a \code{mixtime} time vector at a specific time point, with a specified
chronon and optional cycle. The \code{chronon} defines the smallest indivisible
time granule for the time vector, while the \code{cycle} allows for cyclical time
representations (e.g. day-of-week, month-of-year).
}
\examples{
# Create a continuous mixtime time vector for today
new_time(
  as.double(Sys.Date()),
  chronon = cal_gregorian$day(1L, tz = Sys.timezone()),
  class = "mt_linear"
)

# Create a discrete mixtime time vector for the current date and time
new_time(
  as.integer(Sys.time()),
  chronon = cal_gregorian$second(1L, tz = Sys.timezone()),
  class = "mt_linear"
)

# Create a discrete mixtime time vector for the time of day (cyclical time)
new_time(
  as.integer(Sys.time()), 
  chronon = cal_gregorian$second(1L, tz = Sys.timezone()), 
  cycle = cal_gregorian$day(1L, tz = Sys.timezone()),
  class = "mt_cyclical"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{mt_linear-compare}
\alias{mt_linear-compare}
\alias{==,mt_linear-method}
\alias{!=,mt_linear-method}
\alias{<,mt_linear-method}
\alias{<=,mt_linear-method}
\alias{>,mt_linear-method}
\alias{>=,mt_linear-method}
\title{Comparison operators for linear time (\code{mt_linear})}
\usage{
\S4method{==}{mt_linear}(e1, e2)
\S4method{!=}{mt_linear}(e1, e2)
\S4method{<}{mt_linear}(e1, e2)
\S4method{<=}{mt_linear}(e1, e2)
\S4method{>}{mt_linear}(e1, e2)
\S4method{>=}{mt_linear}(e1, e2)
}
\arguments{
\item{e1, e2}{\code{mt_linear} vectors (or values castable to one, such as
plain numeric vectors sharing the other operand's chronon).}
}
\value{
A logical vector.
}
\description{
Discrete linear time values represent a \emph{closed interval} spanning their
chronon (e.g. \code{year(2020)} spans every instant from the start of 2020 to
the end of 2020), while continuous linear time values represent a single
instant. Comparing two \code{mt_linear} vectors therefore compares the
start/end instants of the (possibly zero-width) interval each value
represents:
\itemize{
\item \code{a == b} iff \code{start(a) == start(b)} and \code{end(a) == end(b)}
\item \code{a < b} iff \code{end(a) < start(b)}
\item \code{a > b} iff \code{start(a) > end(b)}
\item \code{a <= b} iff \code{end(a) <= end(b)} (right-bound comparison)
\item \code{a >= b} iff \code{start(a) >= start(b)} (left-bound comparison)
}

This is \strong{not} a total order: \code{<=}/\code{>=} are endpoint comparisons, not
shorthand for \verb{(< or ==)}/\verb{(> or ==)}, so it is possible for none of
\code{==}, \code{<}, \code{>} to hold between two values.

If both operands are continuous (fractional chronons), or share an
identical chronon, the comparison simplifies to a direct numeric
comparison, since there is no interval to consider.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/00_utils.R
\name{circsum}
\alias{circsum}
\title{Compute circular rolling sums}
\usage{
circsum(x, size, step = size)
}
\arguments{
\item{x}{A numeric vector to compute circular sums over.}

\item{size}{Integer; the window size (number of consecutive elements to
sum). A negative size anchors the window at its end, with the window
counting backwards.}

\item{step}{Integer; the step size (the increment in starting index for
each sum). A negative step walks the windows backwards around the circle.}
}
\value{
A numeric vector containing the sum of each contiguous subsequence
around the circle. The length of the resulting vector is the number of
combinations until the pattern between \code{x} and \code{step} repeats
}
\description{
Calculates rolling sums of length \code{k} for all contiguous subsequences
around a circular vector. Returns sums for each valid k-element window
that wraps around the vector as if arranged in a circle.
}
\examples{
# Simple circular sum with window of 2
circsum(c(1, 2, 3, 4), 2)
# Returns: 3 7 (1+2, 3+4)

# Window of 3 elements
circsum(c(1, 2, 3, 4, 5), 3)
# Returns: 6 10 9 8 12 (1+2+3, 4+5+1, 2+3+4, 5+1+2, 3+4+5)

# Negative step walks the same windows backwards
circsum(c(1, 2, 3, 4, 5), 1, -1)
# Returns: 1 5 4 3 2

# Negative size anchors each window at its end instead of its start - a
# trailing rather than leading rolling sum
circsum(c(1, 2, 3, 4, 5), -2, 1)
# Returns: 6 3 5 7 9 (5+1, 1+2, 2+3, 3+4, 4+5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/00_classes.R
\name{mt_time-class}
\alias{mt_time-class}
\alias{mt_time}
\alias{mt_linear}
\alias{mt_duration}
\alias{mt_cyclical}
\title{Time vector classes}
\usage{
mt_linear(.data = integer(), chronon = mt_unit(1L))

mt_duration(.data = integer(), chronon = mt_unit(1L))

mt_cyclical(.data = integer(), chronon = mt_unit(1L), cycle = mt_unit(1L))
}
\arguments{
\item{.data}{A numeric vector of chronon counts (integer or double).}

\item{chronon, cycle}{Time granules (\link{mt_unit} objects) giving the unit of the
chronon counts and, for \code{mt_cyclical}, the length of the cycle.}
}
\value{
An S7 class object (used for method dispatch), or a time vector when
called as a constructor.
}
\description{
The \code{mt_time} family are the S7 vector classes that store time points as a
numeric count of chronons. \code{mt_time} is the (internal) base class carrying
the \code{chronon} property; the common modes of time are:
\itemize{
\item \code{mt_linear} - linear time points, typically produced with \code{\link[=linear_time]{linear_time()}}.
\item \code{mt_cyclical} - cyclical time points with additional \code{cycle} granule,
typically produced with \code{\link[=cyclical_time]{cyclical_time()}}.
\item \code{mt_duration} - time durations, typically produced with \code{\link[=duration]{duration()}}.
}

The underlying data can be either \strong{integer} (discrete time) or \strong{double}
(continuous time). The \code{chronon} (and, for \code{mt_cyclical}, the \code{cycle}) are
time granules, the result from a \link{mt_unit} object.
}
\seealso{
\code{\link[=mt_linear]{mt_linear()}}, \code{\link[=mt_duration]{mt_duration()}}, and \code{\link[=mt_cyclical]{mt_cyclical()}} to construct
time vectors, and \link{mt_unit} for the granule type stored in \code{chronon}/\code{cycle}.
}
